
import * as BABYLON from "./render3d/context";

export interface AnimationGroupChangeCfg {
    [changeName: string]: AnimationGroupChangeInfo;
}
export interface AnimationGroupChangeInfo {
    from: string;
    to: string;
    isLoop: boolean;
}
const cfgMap: Map<string, AnimationGroupChangeCfg> = new Map();

export class AnimationGroupChange {
    public static animChangeTo(modelName: string, changeName: string, animationGroupMap: Map<string, BABYLON.AnimationGroupRS>, curAnim: BABYLON.AnimationGroupRS, startCall: Function, endCall: Function) {
        let nextAnim: BABYLON.AnimationGroupRS;
        const cfg = cfgMap.get(modelName);

        if (cfg) {
            nextAnim = this.changeTo(changeName, animationGroupMap, cfg, curAnim, startCall, endCall);
        }

        return nextAnim;
    }
    public static changeTo(changeName: string, animationGroupMap: Map<string, BABYLON.AnimationGroupRS>, cfg: AnimationGroupChangeCfg, curAnim: BABYLON.AnimationGroupRS, startCall: Function, endCall: Function) {
        let nextAnim: BABYLON.AnimationGroupRS;
        const info = cfg[changeName];

        if (info) {
            nextAnim = animationGroupMap.get(info.to);
            // curAnim.loopAnimation = false;
            if (curAnim.isPlaying) {
                curAnim.onAnimationGroupEndObservable.addOnce(() => {
                    try {
                        startCall && startCall();
                    } catch (err) {

                    }
                });
            } else {
                if (nextAnim) {
                    nextAnim.onAnimationGroupEndObservable.addOnce(() => {
                        try {
                            endCall && endCall();
                        } catch (err) {

                        }
                    });
                    nextAnim.start(info.isLoop);
                }
            }
        }

        return nextAnim;
    }
    public static registAnimationGroupChnageCfg(modelName: string, cfg: AnimationGroupChangeCfg) {
        cfgMap.set(modelName, cfg);
    }
    public static getAnimationGroupChnageCfg(modelName: string) {
        return cfgMap.get(modelName);
    }
}