/**
 * Composed of a frame, and an action function
 */
export class AnimationEvent {
    /**
     * Specifies if the animation event is done
     */
    public isDone: boolean = false;

    /**
     * Initializes the animation event
     * @param frame The frame for which the event is triggered
     * @param action The event to perform when triggered
     * @param onlyOnce Specifies if the event should be triggered only once
     */
    constructor(
        /** The frame for which the event is triggered **/
        public frame: number,
        /** The event to perform when triggered **/
        public action: (currentFrame: number) => void,
        /** Specifies if the event should be triggered only once**/
        public onlyOnce?: boolean) {
    }
    // PI_BEGIN - 20200814 - 动画事件
    /**
     * 帧事件监听记录列表
     */
    public _pi_listenList: (() => boolean | undefined)[]  = [];
    /**
     * 添加监听
     * @param func 监听函数
     */
    public _pi_addListen(func: () => boolean) {
        this._pi_listenList.push(func);
    }
    /**
     * 帧事件名称
     */
    public _pi_eventKey: string;
    // PI_END

    /** @hidden */
    public _clone(): AnimationEvent {
        // PI_BEGIN - 20200814 - 动画事件
        // return new AnimationEvent(this.frame, this.action, this.onlyOnce);
        // ///////////////////////////////
        const animEvent = new AnimationEvent(this.frame, this.action, this.onlyOnce);
        animEvent._pi_eventKey = this._pi_eventKey;
        return animEvent;
        // PI_END
    }
}
