import * as Ctx from "../../render/index";
import { ETempNodeAnimationDataKey, LoaderRecord } from "../base";
import { AnimationChannelTargetPath, AnimationSamplerInterpolation } from "../glTF2Interface";
import { EAnimationPropertyID, GLTFInterpolationType, IBakedKeyframes, IBakedKeyframesNew, INewAnimationChannel, INewAnimationSampler, InterpolationType } from "../interface";

export class LoadExtraAnimation {
    public static load(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {

        let propertyID: EAnimationPropertyID;
        if (channel.target.path) {
            switch (channel.target.path) {
                case (AnimationChannelTargetPath.TRANSLATION): {
                    propertyID = EAnimationPropertyID.Position;
                    break;
                }
                case (AnimationChannelTargetPath.SCALE): {
                    propertyID = EAnimationPropertyID.Scaling;
                    break;
                }
                case (AnimationChannelTargetPath.ROTATION): {
                    propertyID = EAnimationPropertyID.RotationQuaternion;
                    break;
                }
            }
        }
        if (channel.extras?.property != undefined) {
            propertyID = channel.extras.property;
        }

        switch (propertyID) {
            case (EAnimationPropertyID.Alpha): {
                this.Alpha(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.AlphaCutOff): {
                this.AlphaCutoff(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.cellId): {
                console.error(`Not Support Animation Property ${propertyID}CellId`);
                break;
            }
            case (EAnimationPropertyID.MainColor): {
                this.MainColor(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.intensity): {
                console.error(`Not Support Animation Property ${propertyID} intensity`);
                break;
            }
            case (EAnimationPropertyID.lightdiffuse): {
                console.error(`Not Support Animation Property ${propertyID} Light Diffuse`);
                break;
            }
            case (EAnimationPropertyID.MaskCutoff): {
                this.Maskcutoff(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.ort): {
                console.error(`Not Support Animation Property ${propertyID} Camera Ort`);
                break;
            }
            case (EAnimationPropertyID.per): {
                console.error(`Not Support Animation Property ${propertyID} Camera Per`);
                break;
            }
            case (EAnimationPropertyID.IsActive): {
                this.IsActive(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.RotationQuaternion): {
                this.RotationQuaternion(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.Euler): {
                this.Euler(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.Scaling): {
                this.Scaling(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.Position): {
                this.Position(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.MainTexUOffset): {
                this.MainTexUOffset(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.MainTexVOffset): {
                this.MainTexVOffset(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.MainTexUScale): {
                this.MainTexUScale(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.MainTexVScale): {
                this.MainTexVScale(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.MaskTexUOffset): {
                this.MaskTexUOffset(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.MaskTexUScale): {
                this.MaskTexUScale(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.MaskTexVOffset): {
                this.MaskTexVOffset(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.MaskTexVScale): {
                this.MaskTexVScale(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.OpacityTexUOffset): {
                this.OpacityTexUOffset(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.OpacityTexUScale): {
                this.OpacityTexUScale(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.OpacityTexVOffset): {
                this.OpacityTexVOffset(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.OpacityTexVScale): {
                this.OpacityTexVScale(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.MainTexTilloff): {
                this.MainTexTilloff(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.MaskTexTilloff): {
                this.MaskTexTilloff(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            case (EAnimationPropertyID.OpacityTexTilloff): {
                this.OpacityTexTilloff(loader, indexGroup, indexChannel, channel, sampler, animationGroup);
                break;
            }
            default: {
                console.error(`Not Support Animation Property ${propertyID}`);
                break;
            }
        }
    }
    public static Position(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Node, Ctx.FrameCurvePositionStep, Ctx.FrameCurvePosition, Ctx.FrameCurvePositionPICubicSpline, Ctx.FrameCurvePositionCubicSpline);
    }
    public static Euler(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Node, Ctx.FrameCurveEulerStep, Ctx.FrameCurveEuler, Ctx.FrameCurveEulerPICubicSpline, Ctx.FrameCurveEulerCubicSpline);
    }
    public static RotationQuaternion(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Node, Ctx.FrameCurveRotationQuaternionStep, Ctx.FrameCurveRotationQuaternion, Ctx.FrameCurveRotationQuaternionPICubicSpline, Ctx.FrameCurveRotationQuaternionCubicSpline);
    }
    public static Scaling(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Node, Ctx.FrameCurveScalingStep, Ctx.FrameCurveScaling, Ctx.FrameCurveScalingPICubicSpline, Ctx.FrameCurveScalingCubicSpline);
    }
    public static Alpha(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveAlphaStep, Ctx.FrameCurveAlpha, Ctx.FrameCurveAlphaPICubicSpline, Ctx.FrameCurveAlphaCubicSpline);
    }
    public static AlphaCutoff(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveAlphaCutoffStep, Ctx.FrameCurveAlphaCutoff, Ctx.FrameCurveAlphaCutoffPICubicSpline, Ctx.FrameCurveAlphaCutoffCubicSpline);
    }
    public static MainTexUScale(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveMainTexUScaleStep, Ctx.FrameCurveMainTexUScale, Ctx.FrameCurveMainTexUScalePICubicSpline, Ctx.FrameCurveMainTexUScaleCubicSpline);
    }
    public static MainTexVScale(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveMainTexVScaleStep, Ctx.FrameCurveMainTexVScale, Ctx.FrameCurveMainTexVScalePICubicSpline, Ctx.FrameCurveMainTexVScaleCubicSpline);
    }
    public static MainTexUOffset(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveMainTexUOffsetStep, Ctx.FrameCurveMainTexUOffset, Ctx.FrameCurveMainTexUOffsetPICubicSpline, Ctx.FrameCurveMainTexUOffsetCubicSpline);
    }
    public static MainTexVOffset(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveMainTexVOffsetStep, Ctx.FrameCurveMainTexVOffset, Ctx.FrameCurveMainTexVOffsetPICubicSpline, Ctx.FrameCurveMainTexVOffsetCubicSpline);
    }
    public static MainColor(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveMainColorStep, Ctx.FrameCurveMainColor, Ctx.FrameCurveMainColorPICubicSpline, Ctx.FrameCurveMainColorCubicSpline);
    }
    public static IsActive(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Node, Ctx.FrameCurveIsActiveStep, Ctx.FrameCurveIsActive, Ctx.FrameCurveIsActivePICubicSpline, Ctx.FrameCurveIsActiveCubicSpline);
    }
    public static Maskcutoff(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveMaskCutoffStep, Ctx.FrameCurveMaskCutoff, Ctx.FrameCurveMaskCutoffPICubicSpline, Ctx.FrameCurveMaskCutoffCubicSpline);
    }
    public static MaskTexUScale(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveMaskTexUScaleStep, Ctx.FrameCurveMaskTexUScale, Ctx.FrameCurveMaskTexUScalePICubicSpline, Ctx.FrameCurveMaskTexUScaleCubicSpline);
    }
    public static MaskTexVScale(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveMaskTexVScaleStep, Ctx.FrameCurveMaskTexVScale, Ctx.FrameCurveMaskTexVScalePICubicSpline, Ctx.FrameCurveMaskTexVScaleCubicSpline);
    }
    public static MaskTexUOffset(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveMaskTexUOffsetStep, Ctx.FrameCurveMaskTexUOffset, Ctx.FrameCurveMaskTexUOffsetPICubicSpline, Ctx.FrameCurveMaskTexUOffsetCubicSpline);
    }
    public static MaskTexVOffset(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveMaskTexVOffsetStep, Ctx.FrameCurveMaskTexVOffset, Ctx.FrameCurveMaskTexVOffsetPICubicSpline, Ctx.FrameCurveMaskTexVOffsetCubicSpline);
    }
    public static OpacityTexUScale(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveOpacityTexUScaleStep, Ctx.FrameCurveOpacityTexUScale, Ctx.FrameCurveOpacityTexUScalePICubicSpline, Ctx.FrameCurveOpacityTexUScaleCubicSpline);
    }
    public static OpacityTexVScale(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveOpacityTexVScaleStep, Ctx.FrameCurveOpacityTexVScale, Ctx.FrameCurveOpacityTexVScalePICubicSpline, Ctx.FrameCurveOpacityTexVScaleCubicSpline);
    }
    public static OpacityTexUOffset(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveOpacityTexUOffsetStep, Ctx.FrameCurveOpacityTexUOffset, Ctx.FrameCurveOpacityTexUOffsetPICubicSpline, Ctx.FrameCurveOpacityTexUOffsetCubicSpline);
    }
    public static OpacityTexVOffset(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveOpacityTexVOffsetStep, Ctx.FrameCurveOpacityTexVOffset, Ctx.FrameCurveOpacityTexVOffsetPICubicSpline, Ctx.FrameCurveOpacityTexVOffsetCubicSpline);
    }
    public static MainTexTilloff(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveMainTexTilloffStep, Ctx.FrameCurveMainTexTilloff, Ctx.FrameCurveMainTexTilloffPICubicSpline, Ctx.FrameCurveMainTexTilloffCubicSpline);
    }
    public static MaskTexTilloff(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveMaskTexTilloffStep, Ctx.FrameCurveMaskTexTilloff, Ctx.FrameCurveMaskTexTilloffPICubicSpline, Ctx.FrameCurveMaskTexTilloffCubicSpline);
    }
    public static OpacityTexTilloff(loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS) {
        LoadExtraAnimation.analy(loader, indexGroup, indexChannel, channel, sampler, animationGroup, ETempNodeAnimationDataKey.Material, Ctx.FrameCurveOpacityTexTilloffStep, Ctx.FrameCurveOpacityTexTilloff, Ctx.FrameCurveOpacityTexTilloffPICubicSpline, Ctx.FrameCurveOpacityTexTilloffCubicSpline);
    }
    private static analy<T0 extends Ctx.FrameCurveFramesStep, T1 extends Ctx.FrameCurveFrames, T2 extends Ctx.FrameCurveFramesPIGLTFCubicSpline, T3 extends Ctx.FrameCurveFramesGLTFCubicSpline>(
        loader: LoaderRecord, indexGroup: number, indexChannel: number, channel: INewAnimationChannel, sampler: INewAnimationSampler, animationGroup: Ctx.AnimationGroupRS, targetKey: ETempNodeAnimationDataKey,
        C0: new (key: string, designFPS: number) => T0, C1: new (key: string, designFPS: number) => T1, C2: new (key: string, designFPS: number) => T2, C3: new (key: string, designFPS: number) => T3
    ) {
        indexGroup += 1;
        indexChannel += 1;
        let animTarget = loader.nodeAnimationAbout.get(channel.target.node);
        let target = animTarget ? animTarget[targetKey] : undefined;
        if (loader.gltfres) {
            let key = `${loader.rootUrl}#${indexGroup}#${indexChannel}`;
            let curve = new C0(key, loader.animationFPS);
            curve.set_id_by_gltf(indexGroup, indexChannel, loader.gltfres);
            // console.error("Animation: Key - ", Ctx.FrameCurve.check(curve.id, curve.property), curve.property);
            let animation = new Ctx.AnimationRS<T0>(key, curve);
            if (target instanceof Array) {
                target.forEach((item) => {
                    animationGroup.addTargetedAnimation(animation, curve.targetAnimation(animationGroup.entity, item.entity));
                });
            } else {
                animationGroup.addTargetedAnimation(animation, curve.targetAnimation(animationGroup.entity, target.entity));
            }
        } else {
            throw new Error("Not Impl !!")
            if (target) {
                let times = loader.f32_accessors.get(sampler.input);
                let values = loader.f32_accessors.get(sampler.output);
                if (times && values) {
                    let key = `${loader.rootUrl}#${indexGroup}#${indexChannel}`;
                    let interpolation = GLTFInterpolationType.LINEAR;
                    if (sampler.interpolation) {
                        switch (sampler.interpolation) {
                            case (AnimationSamplerInterpolation.STEP): {
                                interpolation = GLTFInterpolationType.STEP;
                                break;
                            }
                            case (AnimationSamplerInterpolation.CUBICSPLINE): {
                                interpolation = GLTFInterpolationType.CUBICSPLINE;
                                break;
                            }
                        }
                    } else if (sampler.extras?.interpolation != undefined) {
                        interpolation = sampler.extras.interpolation;
                    }
    
                    let animation: Ctx.AnimationRS<Ctx.FrameCurve>;
                    switch (interpolation) {
                        case (GLTFInterpolationType.STEP): {
                            let curve = new C0(key, loader.animationFPS);
                            animation = new Ctx.AnimationRS<T0>(key, curve);
                            if (curve.format(times, values)) {
                            }
                            break;
                        }
                        case (GLTFInterpolationType.PICUBICSPLINE): {
                            let curve = new C2(key, loader.animationFPS);
                            animation = new Ctx.AnimationRS<T2>(key, curve);
                            if (curve.format(times, values)) {
                            }
                            break;
                        }
                        case (GLTFInterpolationType.CUBICSPLINE): {
                            let curve = new C3(key, loader.animationFPS);
                            animation = new Ctx.AnimationRS<T3>(key, curve);
                            if (curve.format(times, values)) {
                            }
                            break;
                        }
                        default: {
                            let curve = new C1(key, loader.animationFPS);
                            animation = new Ctx.AnimationRS<T1>(key, curve);
                            if (curve.format(times, values)) {
                            }
                            break;
                        }
                    }
                    
                    if (animation) {
                        // if (target instanceof Array) {
                        //     target.forEach((item) => {
                        //         animationGroup.addTargetedAnimation(animation, item);
                        //     });
                        // } else {
                        //     animationGroup.addTargetedAnimation(animation, target);
                        // }
                    }
                }
            }
        }
    }
    public static IndiceRenderRange(loader: LoaderRecord, indexNode: number, indexMesh: number, data: IBakedKeyframes, mesh: Ctx.Mesh, root: Ctx.Entity): Ctx.AnimationGroupRS {
        let key = `${loader.rootUrl}#bake#${indexNode}#${indexMesh}`;
        let len = data.keyframes.length;
        let fps = Math.round(data.keyframes[len - 1].frames / data.duration * 0.1) * 10;

        let curve = new Ctx.FrameCurveIndiceRenderRangeStep(key, fps);
        let animation = new Ctx.AnimationRS<Ctx.FrameCurveIndiceRenderRangeStep>(key, curve);
        
        if (!Ctx.FrameCurve.check(curve.id, curve.property)) {
            data.keyframes.forEach((val) => {
                let temp = val.data[0];
                curve.addKey(Math.round(val.frames), temp.start, temp.start + temp.count);
            });
            curve.build();
        }
        let animationGroup = new Ctx.AnimationGroupRS(mesh.name, loader.scene, loader.animationFPS);
        animationGroup.addTargetedAnimation(
            animation,
            <Ctx.ITargetAnimationProperty> {
                property: animation.property,
                target: mesh.entity,
                curve: animation.id,
                group: animationGroup.entity
            }
        );

        return animationGroup;
    }
    public static IndiceRenderRangeNew(loader: LoaderRecord, indexNode: number, indexMesh: number, data: IBakedKeyframesNew, mesh: Ctx.Mesh, root: Ctx.Entity): Ctx.AnimationGroupRS {
        let key = `${loader.rootUrl}#bake#${indexNode}#${indexMesh}`;
        let len = data.keyframes.length;
        let fps = Math.round(data.keyframes[len - 3] / data.duration * 0.1) * 10;

        let curve = new Ctx.FrameCurveIndiceRenderRangeStep(key, fps);
        let animation = new Ctx.AnimationRS<Ctx.FrameCurveIndiceRenderRangeStep>(key, curve);
        
        if (!Ctx.FrameCurve.check(curve.id, curve.property)) {
            let len = data.keyframes.length / 3;
            let vals = data.keyframes;
            for (let i = 0; i < len; i++) {
                curve.addKey(Math.round(vals[i * 3 + 0]), vals[i * 3 + 1], vals[i * 3 + 2]);
            }
            curve.build();
        }
        let animationGroup = new Ctx.AnimationGroupRS(mesh.name, loader.scene, loader.animationFPS);
        animationGroup.addTargetedAnimation(
            animation, 
            <Ctx.ITargetAnimationProperty> {
                property: animation.property,
                target: mesh.entity,
                curve: animation.id,
                group: animationGroup.entity
            }
        );

        return animationGroup;
    }
}