import { IDistortionUVMaterial } from "../interface";
import { LoadExtraMaterial } from "./material";
import { Material, Color3, DistortionUVMaterial } from "../../render/index";
import { IRenderState, LoaderRecord } from "../base";
import { StandardEffectMaterial } from "../../render/materials/standard_effect_material";

export class MaterialLoaderDistortionUV {
    public static init() {
        LoadExtraMaterial.regist(
            "distortionUV",
            MaterialLoaderDistortionUV.load
        );
    }
    public static load(loaders: LoaderRecord, info: IDistortionUVMaterial, indexNode: number, callback: (mat: Material, state: IRenderState) => void): Material {

            // StandardEffectMaterial.regist();
            // let mat = new StandardEffectMaterial("Mat", loaders.scene);
            DistortionUVMaterial.regist();
            let mat = new DistortionUVMaterial("Mat", loaders.scene, undefined, loaders.scene.queryPassTag(info.renderQueue || 3000));

            if (info.diffuseTexture) {
                LoadExtraMaterial.diffuseTexture(mat, loaders, indexNode, info);
            }

            if (info.maskTexture) {
                LoadExtraMaterial.maskTexture(mat, loaders, indexNode, info);
            }

            let numbers = info.diffuseColor;
            if (numbers) {
                mat.emissiveColor = new Color3(numbers[0], numbers[1], numbers[2]);
            }
            let float = info.distortionX;
            if (float != undefined) {
                mat.distortionX = float;
            }
            float = info.distortionY;
            if (float != undefined) {
                mat.distortionY = float;
            }
            float = info.distortionSX;
            if (float != undefined) {
                mat.distortionSpeedX = float;
            }
            float = info.distortionSY;
            if (float != undefined) {
                mat.distortionSpeedY = float;
            }
            float = info.maskFlowMode;
            if (float != undefined) {
                mat.maskFlowMode = float;
            }
            float = info.vertexColorFactor;
            if (float != undefined) {
                mat.vertexColorFactor = float;
            }
            float = info.opacityFromMaskBlue;
            if (float != undefined) {
                mat.opacityFromMaskBlue = !!float;
            }

            let state = LoadExtraMaterial.state(mat, info);

            callback(mat, state);

            return mat;
    }
}