import { IMainOpacityEmissiveFresnelMaterial, IMainOpacityMaterial } from "../interface";
import { LoadExtraMaterial } from "./material";
import { Material, Color3, MainOpacityEmissiveFresnelMaterial } from "../../render/index";
import { IRenderState, LoaderRecord } from "../base";

export class MaterialLoaderMainOpacityEmissiveFresnel {
    public static init() {
        LoadExtraMaterial.regist(
            "main_opacity_emissive_fresnel",
            MaterialLoaderMainOpacityEmissiveFresnel.load
        );
    }
    public static load(loaders: LoaderRecord, info: IMainOpacityMaterial, indexNode: number, callback: (mat: Material, state: IRenderState) => void): Material {

            MainOpacityEmissiveFresnelMaterial.regist();
            let mat = new MainOpacityEmissiveFresnelMaterial("Mat", loaders.scene, undefined, loaders.scene.queryPassTag(info.renderQueue || 3000));

            MaterialLoaderMainOpacityEmissiveFresnel._loadProperty(loaders, mat, info, indexNode, callback);

            return mat;
    }
    public static _loadProperty(loaders: LoaderRecord, mat: MainOpacityEmissiveFresnelMaterial, info: IMainOpacityEmissiveFresnelMaterial, indexNode: number, callback: (mat: Material, state: IRenderState) => void) {
        if (info.diffuseTexture) {
            LoadExtraMaterial.diffuseTexture(mat, loaders, indexNode, info);
        }

        // if (info.emissionTexture) {
        //     LoadExtraMaterial.emissiveTexture(mat, loaders, indexNode, info);
        // }

        if (info.opacityTexture) {
            LoadExtraMaterial.opacityTexture(mat, loaders, indexNode, info);
        }

        let numbers = info.diffuseColor;
        // if (numbers) {s
        //     mat.diffuseColor = new Color3(numbers[0], numbers[1], numbers[2]);
        // }
        numbers = info.emissionColor;
        if (numbers) {
            mat.diffuseColor = new Color3(numbers[0], numbers[1], numbers[2]);
        }

        mat.emissiveFresnelParameters.bias = info.EFBias;
        mat.emissiveFresnelParameters.power = info.EFPower;
        mat.emissiveFresnelParameters.leftColor = new Color3(...info.EFLeft);
        mat.emissiveFresnelParameters.rightColor = new Color3(...info.EFRight);

        let state = LoadExtraMaterial.state(mat, info);

        callback(mat, state);
    }
}