
import { ISampler } from "gltf2/glTF2Interface";
import { ETempNodeAnimationDataKey, IRenderState, LoaderRecord } from "../../gltf2/base";
import { UriLoader } from "../../gltf2/uri";
import { Texture, Material, DefaultMaterial, MainOpacityMaterial, Constants } from "../../render/index";
import { IMaterialData, IMaterialDataNew, IPiTextureInfo } from "../interface";

export class LoadExtraMaterial {
    public static loadAsync(loaders: LoaderRecord, index: number, indexNode: number, callback: (mat: Material, state: IRenderState) => void): Promise<void> {
        let matInfo = loaders.gltf.materials[index];
        if (matInfo) {
            if (matInfo.extras) {
                let flag = false;
                for (let key in matInfo.extras ) {
                    // console.error("Material ! ", key);
                    let call = this.loadCalls.get(key);
                    if (call) {
                        if (call(loaders, matInfo.extras[key], indexNode, callback) != undefined) {
                            flag = true;
                        }
                    } else if (matInfo.extras[key].type) {
                        let key2 = matInfo.extras[key].type;
                        // console.error("Material ! ", key2);
                        let call = this.loadCalls.get(key2);
                        if (call) {
                            if (call(loaders, matInfo.extras[key], indexNode, callback) != undefined) {
                                flag = true;
                            }
                        }
                    }
                }

                if (!flag) {
                    console.error("Cannt Load Material ! ", matInfo.extras);
                    let mat = new DefaultMaterial(matInfo.name + index, loaders.scene);
                    callback(mat, { depthTest: true, depthWrite: false, alphaMode: Material.ALPHA_COMBINE, cullMode: Constants.NONE, renderQueue: 0 });
    
                    return Promise.resolve(null);
                }
            }
            else {
                console.error("Not Found Material In GLTF !");
                let mat = new MainOpacityMaterial(matInfo.name + index, loaders.scene);
                callback(mat, { depthTest: true, depthWrite: false, alphaMode: Material.ALPHA_COMBINE, cullMode: Constants.NONE, renderQueue: 0 });

                return Promise.resolve(null);
            }
        } else {
            throw new Error(`Material Index Error ! ${index}`);
        }
    }
    public static loadCalls: Map<string, (loaders: LoaderRecord, data: IMaterialData, indexNode: number, callback: (mat: Material, state: IRenderState) => void) => Material> = new Map();
    public static regist(key: string, call: (loaders: LoaderRecord, data: IMaterialData, indexNode: number, callback: (mat: Material, state: IRenderState) => void) => Material) {
        this.loadCalls.set(key, call);
    }

    public static diffuseTexture(mat: { diffuseTexture: Texture }, loaders: LoaderRecord, indexNode: number, info: { diffuseTexture: IPiTextureInfo, diffuseLevel?: number, diffuseOU?: number, diffuseOV?: number }) {
        let indexTexture = info.diffuseTexture.index;
        let texInfo = loaders.gltf.textures[indexTexture];
        let indexImage = texInfo.source;
        let imageInfo = loaders.gltf.images[indexImage];
        let url = imageInfo.uri;
        url = UriLoader.formatUrl(loaders.rootDir, url);
        loaders.textures.add(url);


        let texture = new Texture(null, loaders.scene, true, false, Texture.LINEAR_LINEAR);

        LoadExtraMaterial.sampler(loaders.gltf.samplers[texInfo.sampler], texture);

        texture.updateURL(url);
        let float = info.diffuseLevel;
        if (float != undefined) {
            texture.level = float;
        }
        float = info.diffuseOU;
        if (float != undefined) {
            texture.offsetSpeedU = 1000 / float;
        }
        float = info.diffuseOV;
        if (float != undefined) {
            texture.offsetSpeedV = 1000 / float;
        }
        LoadExtraMaterial.textureTilloff(info.diffuseTexture, texture);

        mat.diffuseTexture = texture;

        let animationAbout = loaders.nodeAnimationAbout.get(indexNode);
        animationAbout[ETempNodeAnimationDataKey.MainTex].push(texture);
    }
    public static emissiveTexture(mat: { emissiveTexture: Texture }, loaders: LoaderRecord, indexNode: number, info: { emissionTexture?: IPiTextureInfo, emissiveMapLevel?: number, emissionOU?: number, emissionOV?: number }) {
        let indexTexture = info.emissionTexture.index;
        let texInfo = loaders.gltf.textures[indexTexture];
        let indexImage = texInfo.source;
        let url = loaders.gltf.images[indexImage].uri;
        url = UriLoader.formatUrl(loaders.rootDir, url);
        loaders.textures.add(url);

        let texture = new Texture(null, loaders.scene, true, false, Texture.LINEAR_LINEAR);

        LoadExtraMaterial.sampler(loaders.gltf.samplers[texInfo.sampler], texture);

        texture.updateURL(url);
        let float = info.emissiveMapLevel;
        if (float != undefined) {
            texture.level = float;
        }
        float = info.emissionOU;
        if (float != undefined) {
            texture.offsetSpeedU = 1000 / float;
        }
        float = info.emissionOV;
        if (float != undefined) {
            texture.offsetSpeedV = 1000 / float;
        }
        LoadExtraMaterial.textureTilloff(info.emissionTexture, texture);

        mat.emissiveTexture = texture;

        let animationAbout = loaders.nodeAnimationAbout.get(indexNode);
        animationAbout[ETempNodeAnimationDataKey.EmissiveTex].push(texture);
    }
    
    public static opacityTexture(mat: { opacityTexture: Texture }, loaders: LoaderRecord, indexNode: number, info: { opacityTexture?: IPiTextureInfo, opacityLevel?: number, opacityOU?: number, opacityOV?: number }) {
        let indexTexture = info.opacityTexture.index;
        let texInfo = loaders.gltf.textures[indexTexture];
        let indexImage = texInfo.source;
        let url = loaders.gltf.images[indexImage].uri;
        url = UriLoader.formatUrl(loaders.rootDir, url);
        loaders.textures.add(url);

        let texture = new Texture(null, loaders.scene, true, false, Texture.LINEAR_LINEAR);

        LoadExtraMaterial.sampler(loaders.gltf.samplers[texInfo.sampler], texture);

        texture.updateURL(url);
        let float = info.opacityLevel;
        if (float != undefined) {
            texture.level = float;
        }
        float = info.opacityOU;
        if (float != undefined) {
            texture.offsetSpeedU = 1000 / float;
        }
        float = info.opacityOV;
        if (float != undefined) {
            texture.offsetSpeedV = 1000 / float;
        }
        LoadExtraMaterial.textureTilloff(info.opacityTexture, texture);

        mat.opacityTexture = texture;

        let animationAbout = loaders.nodeAnimationAbout.get(indexNode);
        animationAbout[ETempNodeAnimationDataKey.OpacityTex].push(texture);
    }
    
    public static maskTexture(mat: { maskTexture: Texture }, loaders: LoaderRecord, indexNode: number, info: { maskTexture?: IPiTextureInfo, emissiveMapLevel?: number, emissionOU?: number, emissionOV?: number }) {
        let indexTexture = info.maskTexture.index;
        let texInfo = loaders.gltf.textures[indexTexture];
        let indexImage = texInfo.source;
        let url = loaders.gltf.images[indexImage].uri;
        url = UriLoader.formatUrl(loaders.rootDir, url);
        loaders.textures.add(url);

        let texture = new Texture(null, loaders.scene, true, false, Texture.LINEAR_LINEAR);

        LoadExtraMaterial.sampler(loaders.gltf.samplers[texInfo.sampler], texture);

        texture.updateURL(url);
        // let float = info.emissiveMapLevel;
        // if (float != undefined) {
        //     texture.level = float;
        // }
        // float = info.emissionOU;
        // if (float != undefined) {
        //     texture.offsetSpeedU = 1000 / float;
        // }
        // float = info.emissionOV;
        // if (float != undefined) {
        //     texture.offsetSpeedV = 1000 / float;
        // }
        LoadExtraMaterial.textureTilloff(info.maskTexture, texture);

        mat.maskTexture = texture;

        let animationAbout = loaders.nodeAnimationAbout.get(indexNode);
        animationAbout[ETempNodeAnimationDataKey.MaskTex].push(texture);
    }

    public static sampler(samplerInfo: ISampler, texture: Texture) {
        if (samplerInfo && texture) {
            if (samplerInfo.wrapS !== undefined) texture.wrapU = samplerInfo.wrapS;
            if (samplerInfo.wrapT !== undefined) texture.wrapV = samplerInfo.wrapT;
            if (samplerInfo.minFilter !== undefined) texture.min_filter = <any>samplerInfo.minFilter;
            if (samplerInfo.magFilter !== undefined) texture.mag_filter = samplerInfo.magFilter;
        }
    }

    public static textureTilloff(info: { scale?: [number, number], offset?: [number, number] }, texture: Texture) {
        let numbers = info.scale;
        if (numbers) {
            texture.uScale = numbers[0];
            texture.vScale = numbers[1];
        }
        numbers = info.offset;
        if (numbers) {
            texture.uOffset = numbers[0];
            texture.vOffset = numbers[1];
        }
    }

    public static state(mat: Material, info: IMaterialData): IRenderState  {
        let float = info.alpha;
        if (float != undefined) {
            mat.alpha = float;
        }
        float = info.alphaCutOff;
        if (float != undefined) {
            mat.alphaCutOff = float;
        }

        let cullMode: Constants.CullMode = Constants.BACK;
        if (info.cull == "off") {
            cullMode = Constants.NONE;
        } else if (info.cull == "front") {
            cullMode = Constants.FRONT;
        }

        let depthTest = true;
        let depthWrite = false;
        if (info.zWrite) {
            depthWrite = true;
        }

        let renderQueue = info.renderQueue == undefined ? 3000 : info.renderQueue;
        let alphaMode = info.alphaMode == undefined ? (renderQueue >= 3000 ? 2 : 0) : info.alphaMode;

        return {
            depthTest, depthWrite, alphaMode, cullMode, renderQueue
        }
    }
}