import { ITwoOpacityMixMaterial } from "../interface";
import { LoadExtraMaterial } from "./material";
import { Material, Color3, TwoOpacityMixMaterial } from "../../render/index";
import { IRenderState, LoaderRecord } from "../base";
import { StandardEffectMaterial } from "../../render/materials/standard_effect_material";

export class MaterialLoaderTwoOpacityMix {
    public static init() {
        LoadExtraMaterial.regist(
            "two_opacity_mix",
            MaterialLoaderTwoOpacityMix.load
        );
    }
    public static load(loaders: LoaderRecord, info: ITwoOpacityMixMaterial, indexNode, callback: (mat: Material, state: IRenderState) => void): Material {
            // StandardEffectMaterial.regist();
            // let mat = new StandardEffectMaterial("Mat", loaders.scene);
            // mat.mixOpacityActive = true;
            TwoOpacityMixMaterial.regist();
            let mat = new TwoOpacityMixMaterial("Mat", loaders.scene, undefined, loaders.scene.queryPassTag(info.renderQueue || 3000));

            if (info.diffuseTexture) {
                LoadExtraMaterial.diffuseTexture(mat, loaders, indexNode, info);
            }

            if (info.opacityTexture) {
                LoadExtraMaterial.opacityTexture(mat, loaders, indexNode, info);
            }

            mat.mixOpacityUScale            = info.MixOUS;
            mat.mixOpacityVScale            = info.MixOVS;
            mat.mixOpacityUOffset           = info.MixOUO;
            mat.mixOpacityVOffset           = info.MixOVO;
            mat.mixOpacityUSpeed            = info.MixOUSp;
            mat.mixOpacityVSpeed            = info.MixOVSp;
            mat.mixOpacityIntensity         = info.MixOIns;
            mat.mixOpacityControlUScale     = info.MixCUS;
            mat.mixOpacityControlVScale     = info.MixCVS;
            mat.mixOpacityControlUOffset    = info.MixCUO;
            mat.mixOpacityControlVOffset    = info.MixCVO;
            mat.mixOpacityControlIntensity  = info.MixCIns;
            mat.mixOpacityControlValue      = info.MixCVal;

            let numbers = info.diffuseColor;
            // if (numbers) {
            //     mat.diffuseColor = new Color3(numbers[0], numbers[1], numbers[2]);
            // }
            numbers = info.emissionColor;
            if (numbers) {
                mat.diffuseColor = new Color3(numbers[0], numbers[1], numbers[2]);
            }
            
            let state = LoadExtraMaterial.state(mat, info);

            callback(mat, state);

            return mat;
    }
}