import * as GLTF2 from "./glTF2Interface";
import { LoaderRecord } from "./base";
import { BufferViewLoader } from "./bufferView";

export class AccessorLoader {
    public static loadAsyncUint16(loaders: LoaderRecord, index: number, accessor: GLTF2.IAccessor): Promise<Uint16Array> {
        let loader = loaders.u16_accessor_loaders.get(index);
        if (!loader) {
            let indexBufferView = accessor.bufferView;
            loader = BufferViewLoader.loadAsync(loaders, indexBufferView).then((data) => {
                let temp = new Uint16Array(data.buffer, data.byteOffset + (accessor.byteOffset || 0), accessor.count * AccessorLoader._GetNumComponents(accessor.type));
                loaders.u16_accessors.set(index, temp)
                return temp;
            });
            loaders.u16_accessor_loaders.set(index, loader);
        }
        return loader;
    }
    public static loadAsyncUint32(loaders: LoaderRecord, index: number, accessor: GLTF2.IAccessor): Promise<Uint32Array>  {
        let loader = loaders.u32_accessor_loaders.get(index);
        if (!loader) {
            let indexBufferView = accessor.bufferView;
            loader = BufferViewLoader.loadAsync(loaders, indexBufferView).then((data) => {
                let temp = new Uint32Array(data.buffer, data.byteOffset + (accessor.byteOffset || 0), accessor.count * AccessorLoader._GetNumComponents(accessor.type));
                loaders.u32_accessors.set(index, temp)
                return temp;
            });
            loaders.u32_accessor_loaders.set(index, loader);
        }
        return loader;
    }
    public static loadAsyncFloat32(loaders: LoaderRecord, index: number, accessor: GLTF2.IAccessor): Promise<Float32Array>  {
        let loader = loaders.f32_accessor_loaders.get(index);
        if (!loader) {
            let indexBufferView = accessor.bufferView;
            loader = BufferViewLoader.loadAsync(loaders, indexBufferView).then((data) => {
                let temp = new Float32Array(data.buffer, data.byteOffset + (accessor.byteOffset || 0), accessor.count * AccessorLoader._GetNumComponents(accessor.type));
                loaders.f32_accessors.set(index, temp);
                return temp;
            });
            loaders.f32_accessor_loaders.set(index, loader);
        }
        return loader;
    }
    
    public static _GetNumComponents(type: string): number {
        switch (type) {
            case "SCALAR": return 1;
            case "VEC2": return 2;
            case "VEC3": return 3;
            case "VEC4": return 4;
            case "MAT2": return 4;
            case "MAT3": return 9;
            case "MAT4": return 16;
        }

        throw new Error(`Invalid type (${type})`);
    }
}