import { ResTab } from "../interface/index";
import { IGLTFNew } from "./interface";
import * as BABYLON from "../render/index";

export interface IRenderState {
    renderQueue: number;
    cullMode: BABYLON.Constants.CullMode;
    depthTest: boolean;
    depthWrite: boolean;
    alphaMode: number;
}

export enum ETempNodeAnimationDataKey {
    Node = "Node",
    Mesh = "Mesh",
    Material = "Material",
    MainTex = "MainTex",
    OpacityTex = "OpacityTex",
    MaskTex = "MaskTex",
    EmissiveTex = "EmissiveTex",
    Camera = "Camera",
    Light = "Light",
}

export interface ITempNodeAnimationAbout {
    [ETempNodeAnimationDataKey.Node]: BABYLON.Node,
    [ETempNodeAnimationDataKey.Mesh]?: BABYLON.Mesh[],
    [ETempNodeAnimationDataKey.Material]?: BABYLON.Material[],
    [ETempNodeAnimationDataKey.MainTex]?: BABYLON.Texture[],
    [ETempNodeAnimationDataKey.OpacityTex]?: BABYLON.Texture[],
    [ETempNodeAnimationDataKey.EmissiveTex]?: BABYLON.Texture[],
    [ETempNodeAnimationDataKey.MaskTex]?: BABYLON.Texture[],
    [ETempNodeAnimationDataKey.Camera]?: BABYLON.Camera,
    [ETempNodeAnimationDataKey.Light]?: BABYLON.Light,
}

export class LoaderRecord {
    gltfres: BABYLON.IGLTFRes;
    rootNode: BABYLON.TransformNode;
    scene: BABYLON.Scene;
    gltf: IGLTFNew;
    rootDir: string;
    rootUrl: string;
    resTab: ResTab;
    uris: Map<any, Promise<Uint8Array>> = new Map();
    bufferLoaders: Map<any, Promise<Uint8Array>> = new Map();
    bufferViewLoaders: Map<any, Promise<Uint8Array>> = new Map();
    u16_accessor_loaders: Map<any, Promise<Uint16Array>> = new Map();
    u32_accessor_loaders: Map<any, Promise<Uint32Array>> = new Map();
    f32_accessor_loaders: Map<any, Promise<Float32Array>> = new Map();
    u16_accessors: Map<number, Uint16Array> = new Map();
    u32_accessors: Map<number, Uint32Array> = new Map();
    f32_accessors: Map<number, Float32Array> = new Map();
    nodeAnimationAbout: Map<number, ITempNodeAnimationAbout> = new Map();
    animationFPS: number;
    materials: Map<number, { material: BABYLON.Material, state: IRenderState}> = new Map();
    textures: Set<string> = new Set();
    skeletons: BABYLON.Skeleton[] = [];
    public promise() {
        let result = [];

        if (this.uris.size > 0) {
            this.uris.forEach((item) => { result.push(item); });
        } else {
            result.push(Promise.resolve(null));
        }

        return Promise.all(result)
        // .then(() => {
        //     console.error("uris Loaded !!");
        // })
        ;
    }
    public accessors() {
        let result = [];

        this.bufferLoaders.forEach((item) => { result.push(item); });
        this.bufferViewLoaders.forEach((item) => { result.push(item); });
        this.u16_accessor_loaders.forEach((item) => { result.push(item); });
        this.u32_accessor_loaders.forEach((item) => { result.push(item); });
        this.f32_accessor_loaders.forEach((item) => { result.push(item); });
        
        return Promise.all(result);
    }
    public textureLoaders() {
        let result = [];

        this.textures.forEach((item) => {
            result.push(BABYLON.BaseTexture.CreateLoad(item, false, false));
        });
        
        return Promise.all(result)
        // .then(() => {
        //     console.error(`Texture All Loaded !! ${this.rootUrl}`);
        // })
        ;
    }
    public dispose() {
        if (this.gltfres) {
            this.gltfres.free && this.gltfres.free();
            // console.log("Dispose GLTF");
        }
        if (this.rootNode) {
            this.rootNode.dispose();
        }
        if (this.skeletons) {
            this.skeletons.forEach((item) => {
                item.dispose();
            });
        }
        this.rootNode = undefined;
        this.gltfres = undefined;
        this.scene = undefined;
        this.gltf = undefined;
        this.rootUrl = undefined;
        this.resTab = undefined;
        this.uris = undefined;
        this.bufferLoaders = undefined;
        this.bufferViewLoaders = undefined;
        this.u16_accessor_loaders = undefined;
        this.u32_accessor_loaders = undefined;
        this.f32_accessor_loaders = undefined;
    }
}