export interface ITexture {
    url: string,
    texCoord?: number,
    scale?: [number, number],
    offset?: [number, number],
    filter?: [number, number],
    wrap?: [number, number],
}
export interface IMaterial {
    type: string,
    alpha?: number;
    alphaCutOff?: number;
    alphaMode?: number;
    zWrite?: 1 | undefined;
    cull?: string;
    renderQueue?: number;
}
export interface IMainOpacityMaterial extends IMaterial {
    diffuseTexture: ITexture;
    diffuseLevel?: number;
    diffuseOV?: number;
    diffuseOU?: number;

    opacityTexture?: ITexture;
    opacityLevel?: number;
    opacityOV?: number;
    opacityOU?: number;

    emissionTexture?: ITexture;
    emissiveMapLevel?: number;
    emissionOU?: number;
    emissionOV?: number;
    
    diffuseColor: [number, number, number, number];
    emissionColor: [number, number, number, number];
}

export interface IMainOpacityEmissiveFresnelMaterial extends IMainOpacityMaterial {
    // Fresnel
    EFBias?: number;
    EFPower?: number;
    EFLeft?: [number, number, number];
    EFRight?: [number, number, number];
}

export interface ITwoOpacityMixMaterial extends IMaterial {

    diffuseTexture: ITexture;
    diffuseLevel?: number;
    diffuseOV?: number;
    diffuseOU?: number;

    opacityTexture: ITexture;
    opacityLevel?: number;
    opacityOV?: number;
    opacityOU?: number;

    diffuseColor: [number, number, number, number];
    emissionColor: [number, number, number, number];

    // _ClipEdgeColor
    coEColor?: [number, number, number, number];
    // _ClipEdgeThickness
    coET?: number;
    // _ClipEdgeLevel
    coE?: number;
    // _ClipEdgeAlphaDiff
    coEAD?: number;

    // mixOpacityUScale
    MixOUS: number;
    // mixOpacityVScale
    MixOVS: number;
    // mixOpacityUOffset
    MixOUO: number;
    // mixOpacityVOffset
    MixOVO: number;
    // mixOpacityUSpeed
    MixOUSp: number;
    // mixOpacityVSpeed
    MixOVSp: number;
    // mixOpacityLevel
    MixOIns: number;
    // mixOpacityControlUScale
    MixCUS: number;
    // mixOpacityControlVScale
    MixCVS: number;
    // mixOpacityControlUOffset
    MixCUO: number;
    // mixOpacityControlVOffset
    MixCVO: number;
    // mixOpacityControlLevel
    MixCIns: number;
    // mixOpacityControlValue
    MixCVal: number;
}

export interface IDistortionUVMaterial extends IMaterial {
    alpha?: number;
    alphaCutOff?: number;
    alphaMode?: number;
    zWrite?: 1 | undefined;
    cull?: string;
    renderQueue?: number;
    diffuseTexture: ITexture;
    diffuseLevel?: number;
    maskTexture: ITexture;
    opacityTexture: ITexture;
    opacityLevel?: number;
    distortionX: number;
    distortionY: number;
    distortionSX: number;
    distortionSY: number;
    maskFlowMode: 0 | 1;
    vertexColorFactor: number;
    opacityFromMaskBlue: number;
    diffuseColor: [number, number, number, number];
}

export interface INode {

}
export interface IGLTFTemplate {

}