import { IEntity } from "./entity";
import { IIsometryNode } from "./node";
import { Constants } from "./constants";
import { Option } from "./types";
import { IMatrix, IVector3 } from "./math";

export interface IOrthographicCamera {
    size: number;
    piOrthSize: number;
    piSizeForWidth: boolean;
    orthoRight: number;
    orthoLeft: number;
    orthoTop: number;
    orthoBottom: number;
}

export interface IPerspectiveCamera {
    fov: number;
    fovMode: Constants.CameraFovMode;
}

export interface ITargetCamera {
    target: IVector3;
}

export interface ICamera extends IIsometryNode {
    active: boolean;
    aspect: Option<number>;
    minZ: number;
    maxZ: number;
    getViewMatrix(): IMatrix;
    getProjectMatrix(): IMatrix;
    getTransformationMatrix(): IMatrix;
    forceInclude(mesh: IEntity, doadd?: boolean): void;
    attachControl(canvas: HTMLCanvasElement, flag: boolean): void;
}

export interface IFreeCamera extends ICamera, IPerspectiveCamera, ITargetCamera, IOrthographicCamera {
    mode: Constants.FREECAMERA_MODE;
}