import { INodeD3 } from "pi_babylon/src/render";

export type EventTypes = 'pointerup' | 'pointerdown' | 'pointerclick' | "pointerclick" | 'pointermove';
export interface Event {
    /**
     * 事件响应 起始节点
     */
    source: Element;
    /**
     * 事件类型
     */
    // tslint:disable-next-line:no-reserved-keywords
    type: string;
    /**
     * 事件响应事件戳
     */
    timeStamp: number;

    /**
     * 当前事件响应节点
     */
    current: Element;
    /**
     * 是否为后处理 - down - up , moveenter - moveleave 等成对事件 失去焦点后处理 up - leave
     */
    isPostprocess: boolean;
    /**
     * 是否仅冒泡，不执行
     */
    isOnlyBubble: boolean;

    /**
     * 事件路径上的节点
     */
    path: Element[];

    [other: string]: any,
}

/**
 * 节点使用的事件数据
 */
export interface EventData extends Event {
    /**
     * 事件 坐标
     */
    x: number;
    /**
     * 事件 坐标
     */
    y: number;
    /**
     * 事件 上一次响应时坐标 - (move， 滚动)
     */
    lastX: number;
    /**
     * 事件 上一次响应时坐标 - (move， 滚动)
     */
    lastY: number;
    /**
     * X 滑动量
     */
    deltaX: number;
    /**
     * Y 滑动量
     */
    deltaY: number;
    /**
     * 事件类型
     */
    pointerType: EventTypes;
    /**
     * 是否为主事件
     */
    isPrimary: boolean;
    /**
     * 事件 ID
     */
    pointerID: number;
    /**
     * 多点事件数据堆 - 坐标堆
     * * key: 事件ID
     * * value: { x, y } 事件坐标
     */
    pointers: { x: number; y: number; id: number, sourceEvent?: any }[];
    /**
     * 是否中止事件传播
     */
    stopPropagation: boolean | Function;
    /**
     * 是否阻止事件默认处理
     */
    preventDefault: boolean | Function;
    /**
     * 是否阻止事件默认处理
     */
    __preventDefault: boolean;
    /**
     * 事件响应事件戳
     */
    timeNow: number;
    /**
     * 事件是否传递到 GUI 外去 <比如另一个3D场景>
     */
    isSendNextLayer: boolean;

    /**
     * 原生事件
     */
    sourceEvent?: PointerEvent;
}

/**
 * 事件监听函数模板
 */
export type EventListener = (e: EventData) => void;
export type KeyboardEventListener = (e: KeyboardEvent) => void;
