import { Constants } from "./constants";
import { Option } from "./types";

export interface IGeometryLink {
    geometry: any;
}

export interface IVertexAttribute {
    kind: Constants.VertexAttr;
    format: Constants.VertexFormat;
}

export interface IVertexBuffer {
    name: string;
    range: Option<[number, number]>;
    attrs: IVertexAttribute[];
}

export interface IIndiceBuffer {
    name: string;
    range: Option<[number, number]>;
    as_u16: boolean;
}

export enum EInstanceKind {
    TSR = "TSR",
    Color = "Color",
    Tilloff = "Tilloff",
}

export interface InstanceAttribute {
    key: string,
    code: string,
    foruniformKey?: string,
    vtype: Constants.INSTANCE_ATTR,
}

export interface GeometryDesc {
    vertices: IVertexBuffer[];
    // instance: EInstanceKind[];
    indices: Option<IIndiceBuffer>;
}

export interface IGeometry {
    setVerticesBuffer(vb: IVertexBuffer): void;
    setIndices(ib: IIndiceBuffer): void;
    applyToMesh(mesh: IGeometryLink, meta?: any): void;
}

export interface ICustomGeometry {
    update(pointCount: number): void;
    vertices(kind: typeof Constants.VERTEX_ATTR_POSITION | typeof Constants.VERTEX_ATTR_COLOR4 | typeof Constants.VERTEX_ATTR_UV): [Float32Array, number, number];
    indices(): Uint16Array;
    dispose(): void;
}
