import { IQuaternion, IVector3 } from "./math";
import { IDisposable } from "./base";

export interface INodePostion {
    position: IVector3,
}

export interface INodeEuler {
    rotation: IVector3,
}

export interface INodeScaling {
    scaling: IVector3,
}

export interface INodeRotation {
    rotationQuaternion: IQuaternion,
}

export interface INode {
    _scene: any; // IScene,
    parent: INode,
    // enable: boolean,
    setEnabled(val: boolean): void;
    isEnabled(): boolean;
    isDisposed(): boolean;
    
    /** @hidden */
    _getDescendants(results: INode[], directDescendantsOnly?: boolean, predicate?: (node: INode) => boolean): void;

    /**
     * Will return all nodes that have this node as ascendant
     * @param directDescendantsOnly defines if true only direct descendants of 'this' will be considered, if false direct and also indirect (children of children, an so on in a recursive manner) descendants of 'this' will be considered
     * @param predicate defines an optional predicate that will be called on every evaluated child, the predicate must return true for a given child to be part of the result, otherwise it will be ignored
     * @return all children nodes of all types
     */
    getDescendants(directDescendantsOnly?: boolean, predicate?: (node: INode) => boolean): INode[];
}

export interface IIsometryNode extends IDisposable, INode, INodeEuler, INodeRotation, INodePostion {
    lookAt(target: IVector3): void;
}
