export interface IPostprocessParamType {
    // etype: string;
}

export const KeyBloomCombine = "BloomCombine";
export interface IBloomCombine {
    etype: typeof KeyBloomCombine;
    /**
     * Bloom 强度
     */
    intensity: number;
}

export const KeyBloomThreshold = "BloomThreshold";
export interface IBloomThreshold {
    etype: typeof KeyBloomThreshold;
    thresholdKnee: number;
    threshold: number;
}

export const KeyBloom = "Bloom";
export interface IBloom extends IPostprocessParamType {
    etype: typeof KeyBloom;
    thresholdKnee: number;
    threshold: number;
    /**
     * Bloom 强度
     */
    intensity: number;
    /**
     * 模糊半径
     */
    blurRadius: number;
    /**
     * 效果迭代次数
     */
    useBlurIteration: number;
}

export const KeyColorEffect = "ColorEffect";
export interface IColorEffect extends IPostprocessParamType {
    etype: typeof KeyColorEffect;
    /**
     * 色相 - 对应Photoshop的[-180, 180]
     */
    hue: number;
    /**
     * 饱和度 对应Photoshop的[-100, 100]
     */
    saturate: number;
    /**
     * 明度 对应Photoshop的[-100, 100]
     */
    brightness: number;

    /**
     * 色彩平衡 - R 对应Photoshop的[-100, 100]
     */
    colorBalanceR: number;
    /**
     * 色彩平衡 - G 对应Photoshop的[-100, 100]
     */
    colorBalanceG: number;
    /**
     * 色彩平衡 - B 对应Photoshop的[-100, 100]
     */
    colorBalanceB: number;

    /**
     * 色阶 - 输入阈值的区间左边界 - 0-255
     */
    colorScaleRInShadow: number;
    /**
     * 色阶 - 中间调 - 9.99~0.1
     */
    colorScaleRInMid: number;
    /**
     * 色阶 - 输入阈值的区间右边界 - 0-255
     */
    colorScaleRInHighlight: number;
    /**
     * 色阶 - 输出阈值区间左边界 - 0-255
     */
    colorScaleROutShadow: number;
    /**
     * 色阶 - 输出阈值区间右边界 - 0-255
     */
    colorScaleROutHighlight: number;
}

export const KeyCopy = "Copy";
export interface ICopy extends IPostprocessParamType {
    etype: typeof KeyCopy;
    alpha: number;
    intensity: number;
}

export const KeyDirectBlur = "DirectBlur";
export interface IDirectBlur extends IPostprocessParamType {
    etype: typeof KeyDirectBlur;
    /**
     * 模糊半径 - 像素值
     */
    blurRadius: number;
    /**
     * direction
     */
    dx: number;
    /**
     * direction
     */
    dy: number;
}

export const KeyDualBlur = "DualBlur";
export interface IDualBlur extends IPostprocessParamType {
    etype: typeof KeyDualBlur;
    /**
     * 模糊半径
     */
    blurRadius: number;
    /**
     * 模糊强度
     */
    intensity: number;
}

export const KeyDualBlurDown = "DualBlurDown";
export interface IDualBlurDown extends IPostprocessParamType {
    etype: typeof KeyDualBlurDown;
    /**
     * 模糊半径
     */
    blurRadius: number;
    /**
     * 模糊强度
     */
    intensity: number;
}

export const KeyDualBlurUp = "DualBlurUp";
export interface IDualBlurUp extends IPostprocessParamType {
    etype: typeof KeyDualBlurUp;
    /**
     * 模糊半径
     */
    blurRadius: number;
    /**
     * 模糊强度
     */
    intensity: number;
}


export const KeyHueDegree = "HueDegree";
export interface IHueDegree extends IPostprocessParamType {
    etype: typeof KeyHueDegree;
    /**
     * 颜色变换系数
     */
    degree: number;
}

export const KeyRadialBlur = "RadialBllur";
/**
 * 效果以屏幕中心为原点
 */
export interface IRadialBlur {
    etype: typeof KeyRadialBlur;
    /**
     * 模糊半径 - 像素值
     */
    blurRadius: number;
    /**
     * 径向中心坐标
     */
    centerX: number;
    /**
     * 径向中心坐标
     */
    centerY: number;
    /**
     * 效果迭代次数 - 基础为 6
     */
    iteration: number;
}

export const KeyRadiusWave = "RadiusWave";
export interface IRadiusWave extends IPostprocessParamType {
    etype: typeof KeyRadiusWave;
    /**
     *  是否处理为圆形
     */
    asCircle: boolean;
    /**
     * 模糊半径 - 像素值
     */
    blurRadius: number;
    /**
     * 波纹径向起点 0-2
     */
    waveStartRadius: number;
    /**
     * 波纹径向终点 0-2
     */
    waveEndRadius: number;
    /**
     * 波纹数量
     */
    waveCounter: number;
    /**
     * 波纹高度 - 扭曲值域
     */
    waveWeight: number;
}

export const KeyVignette = "Vignette";
/**
 * 该效果坐标系 以屏幕中心为原点, 在屏幕径向边缘产生效果
 */
export interface IVignette extends IPostprocessParamType {
    etype: typeof KeyVignette;
    /**
     * 效果开始的位置
     */
    begin: number;
    /**
     * 效果达到最强时的位置
     */
    end: number;
    /**
     * 效果由开始到最强的过渡控制
     */
    scale: number;
    /**
     * 效果颜色
     */
    vignetteColor: [number, number, number];
}

export interface IPostprocessTexture {
    width: number;
    height: number;
    _webGLTexture: WebGLTexture;
}

export interface IPostprocessTarget {
    x: number;
    y: number;
    width: number;
    height: number;
    framebuffer: WebGLFramebuffer;
}

export type KeyPostprocess = typeof KeyBloom | typeof KeyBloomCombine | typeof KeyBloomThreshold | typeof KeyRadiusWave | typeof KeyRadialBlur
    | typeof KeyVignette | typeof KeyCopy | typeof KeyHueDegree | typeof KeyDirectBlur | typeof KeyColorEffect
    | typeof KeyDualBlur | typeof KeyDualBlurDown | typeof KeyDualBlurUp
;
export type IPostprocessParam = IBloom | IBloomCombine | IBloomThreshold | IRadiusWave | IRadialBlur
    | IVignette | ICopy | IHueDegree | IDirectBlur | IColorEffect
    | IDualBlur | IDualBlurDown | IDualBlurUp
;

export const SortKeys: KeyPostprocess[] = [KeyColorEffect, KeyHueDegree, KeyVignette, KeyRadialBlur, KeyDirectBlur, KeyBloom, KeyDualBlur, KeyRadiusWave, KeyCopy];

export function resortPostprocessParams(postprocess: IPostprocessParam[], actived: Map<KeyPostprocess, any>) {
    let result = [];
    SortKeys.forEach((key) => {
        let item = postprocess.find((v) => v.etype == key);
        if (item && actived.get(key) != undefined) {
            result.push(item);
        }
    });

    return result;
}