import { ICustomRenderTarget, IDisposable, ITempFrameBuffer, TAutoRenderTarget } from "./base";
import { Constants } from "./constants";

export interface IRenderPipeline {
    renderers: IRenderer[];
    queryPass(queue: number): Constants.PassTag;
}

export interface IRenderer extends IDisposable {
    pass: Constants.PassTag;
    autoClearColor(flag: boolean): void;
    autoClearDepth(flag: boolean): void;
    autoClearStencil(flag: boolean): void;
    /**
     * 
     * @param r 0 ~ 255
     * @param g 0 ~ 255
     * @param b 0 ~ 255
     * @param a 0 ~ 255
     */
    clearColor(r: number, g: number, b: number, a: number): void;
    /**
     * 
     * @param depth 0 ~ 1.
     */
    clearDepth(depth: number): void;
    clearStencil(value: number): void;
    viewport(x: number, y: number, width: number, height: number): void;
    setRenderTarget(target: ICustomRenderTarget | TAutoRenderTarget | ITempFrameBuffer): void;

}