import { BABYLON } from "../render3d/babylon";
import { IDisposable } from "./base";
import { IMeshRenderQueue } from "./mesh";
import { EAtlasRenderMode, ITexturePackAtlas, ITexturePackFrame } from "./texturepacker";
import { ITransformNode } from "./transform";
export { ITexturePackFrame, ITexturePackAtlas } from "./texturepacker";


export interface ISpriteAtlasCells {
    image: string;
    hCount: number;
    vCount: number;
    hsize: number;
    vsize: number;
    samplerMode?: number,
    alphaMode?: number,
    noMipmap?: boolean,
    renderMode?: EAtlasRenderMode;
}
export interface ISpriteAtlasRects {
    image: string;
    rects: {
        [index: string | number]: [number, number, number, number]
    },
    samplerMode?: number,
    alphaMode?: number,
    noMipmap?: boolean,
    renderMode?: EAtlasRenderMode;
}
export type ISpriteAtlas = ISpriteAtlasCells | ISpriteAtlasRects | ITexturePackAtlas;
export type ISpriteInfo = [string, number];

export interface ISprite extends IMeshRenderQueue, ITransformNode {
    color: BABYLON.Color3;
    alpha: number;
    readonly spriteId: number;
    spriteFrame: string | number;
    vertexPosition(data: Float32Array, offset: number, strip?: number, localoff?: number): number;
    vertexColor(data: Float32Array, offset: number, strip?: number, localoff?: number): number;
    vertexUV(data: Float32Array, offset: number, strip?: number, localoff?: number): number;
    dispose(doNotRecurse?: boolean, disposeMaterialAndTextures?: boolean): void;
    sortParam(): any;
    _onFrameUpdate(): boolean;
    setAnim(isLoop: boolean): void;
    duration: number;
}

export interface ISpriteManager extends IDisposable {
    material: any;
    readonly atlas: ISpriteAtlas;
    readonly atlasImage: string;
    getFrame(item: ISprite): [number, number, number, number];
    getTexturePackFrame(item: ISprite): ITexturePackFrame;
    extendAtlasFrame(key: string | number, rect: [number, number, number, number]): void;
    addSprite(item: ISprite): void;
    delSprite(item: ISprite): void;
    dirty(): void;
    sort(list: { param: any }[]): void;
}

export interface ISpriteScene {
    registSpritAtlas(atlas: ISpriteAtlas): void;
    getSpritManager(atlas: string): ISpriteManager;
    delSpritManager(managr: ISpriteManager): void;
}