import { ExtensionTools } from "../interface/index";
import { BABYLON } from "../render3d/babylon";
import { p3d } from "../render/types";

const KEY_OVERLAY = "KEY_OVERLAY";

export class OverlayColorMaterial extends BABYLON.MainTextureMaterial {
    public static readonly ShaderName = KEY_OVERLAY;
    constructor(materialName: string, babylonScene: BABYLON.Scene, shaderName = OverlayColorMaterial.ShaderName, _pass?: BABYLON.Constants.PassTag) {
        OverlayColorMaterial.regist();
        super(
            materialName,
            babylonScene,
            shaderName,
            _pass
        );
    }
    public static regist() {
        if (p3d.p3d_check_shader(BABYLON.Engine.native(), BABYLON.Engine.param(), OverlayColorMaterial.ShaderName)) {
            return;
        }
        let uniforms = p3d.MaterialUniformDefines.create();
        let includes = p3d.NodematerialIncludes.create();
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockUVOffsetSpeed);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
                // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(BABYLON.UniformMainTextureSimple.BLOCK_NAME));
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTextureUVOffsetSpeed);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTexture);
        // p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(BABYLON.UniformOpacityTexture.BLOCK_NAME));
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTextureUVOffsetSpeed);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTexture);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTextureUVOffsetSpeed);

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let binds_defines_base = BABYLON.Constants.BIND_MODEL | BABYLON.Constants.BIND_VIEWER | BABYLON.Constants.BIND_EFFECT_TEXTURE;

        p3d.p3d_regist_material(
            BABYLON.Engine.native(), BABYLON.Engine.param(),
            OverlayColorMaterial.ShaderName,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            mat4 finalWorld = PI_ObjectToWorld;
            vec4 position =  vec4(A_POSITION, 1.);
            vec4 worldPos =  finalWorld * position;
            gl_Position = PI_MATRIX_VP * worldPos;
            v_uv    = A_UV;
            v_color = A_COLOR4;
            `,
            `
        float alpha = 1.;
        vec4 src = Get_MainTex(v_uv, vec2(0., 0.), vec4(1., 1., 0., 0.));
        gl_FragColor = vec4(src.rgb * v_color.a + v_color.rgb, alpha * src.w);
        // gl_FragColor.rgb = vec3(uvAtlas(v_uv, _MainTex_Atlas[int(2)], _MainTex_Address[int(2)]), _MainTex_Address[int(2)].w * 1.);
        // gl_FragColor.a = 1.0;
            `,
            includes,
            "",
            varyings,
            binds_defines_base
        );

        varyings.free && varyings.free();
        uniforms.free && uniforms.free();
        includes.free && includes.free();
    }
}
