
import { Constants, ExtensionTools } from "../interface/index";
import { BABYLON } from "../render3d/babylon";
import { p3d } from "../render/types";
import { SpriteManager } from "../sprite/sprite_manager";

const KEY_TEXTUREPROGRESS_CIRCLE = "KEY_TEXTUREPROGRESS_CIRCLE";

export class TextureProgressCircleMaterial extends BABYLON.MainTextureMaterial {
    public static readonly ShaderName = KEY_TEXTUREPROGRESS_CIRCLE;
    constructor(materialName: string, babylonScene: BABYLON.Scene, shaderName = TextureProgressCircleMaterial.ShaderName, _pass?: BABYLON.Constants.PassTag) {
        TextureProgressCircleMaterial.regist();
        super(
            materialName,
            babylonScene,
            shaderName,
            _pass
        );
    }
    public static regist() {
        if (p3d.p3d_check_shader(BABYLON.Engine.native(), BABYLON.Engine.param(), TextureProgressCircleMaterial.ShaderName)) {
            return;
        }
        let uniforms = p3d.MaterialUniformDefines.create();
        let includes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
                // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(BABYLON.UniformMainTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTextureUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(BABYLON.UniformOpacityTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTextureUVOffsetSpeed);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTexture);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTextureUVOffsetSpeed);

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
            ["vec2", "vCenter"],
            ["float", "vAlpha"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let binds_defines_base = BABYLON.Constants.BIND_MODEL | BABYLON.Constants.BIND_VIEWER | BABYLON.Constants.BIND_EFFECT_VALUE | BABYLON.Constants.BIND_EFFECT_TEXTURE;

        p3d.p3d_regist_material(
            BABYLON.Engine.native(), BABYLON.Engine.param(),
            TextureProgressCircleMaterial.ShaderName,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            mat4 finalWorld = PI_ObjectToWorld;
            vec4 position =  vec4(A_POSITION, 1.);
            vec4 worldPos =  finalWorld * position;
            gl_Position = PI_MATRIX_VP * worldPos;
            v_uv    = A_UV;
            v_color = A_COLOR4;
            vCenter = vec2(22 >> gl_VertexIndex & 1, 54 >> gl_VertexIndex & 1) - vec2(0.5, 0.5);
            vAlpha = (1.0 - A_COLOR4.a) * 1.002 - 0.001;
            `,
            `
        float alpha = opacity(matParam);
        vec4 src = mainTexture(v_uv, applyUVOffsetSpeed(matParam.uMainUVOS), matParam);

        vec2 base_v = vec2(0.0, 1.0);
        vec2 diff   = vCenter;
        vec2 dist   = normalize(diff);
        float d     = dot(vec2(0.0, 1.0), dist);
        
        float angle = acos(d) * mix(-1., 1., step(0., diff.x));
        angle       = angle + mix(6.283185307179586, 0., step(0., diff.x));
        angle       /= 6.283185307179586;

        float intensity = smoothstep(vAlpha, vAlpha + 0.001, 1.0 - angle);

        // gl_FragColor = vec4(src.rgb * v_color.a + v_color.rgb, alpha * src.w);
        gl_FragColor = vec4(src.rgb * v_color.a, src.a * intensity);
            `,
            includes,
            "",
            varyings,
            binds_defines_base
        );

        varyings.free && varyings.free();
        uniforms.free && uniforms.free();
        includes.free && includes.free();
    }
    public static RegistForSprite() {
        SpriteManager.MaterialFactory.set(5, function(scene: BABYLON.Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag) {
            let _material = new TextureProgressCircleMaterial(`${atlasImage}`, scene, undefined, pass);
            let texture = new BABYLON.Texture(null, scene, noMipmap, false, samplerMode || BABYLON.Texture.LINEAR_LINEAR);
            texture.updateURL(atlasImage);
            texture.hasAlpha = true;
            texture.wrapU = BABYLON.Constants.CLAMP_ADDRESSMODE;
            texture.wrapV = BABYLON.Constants.CLAMP_ADDRESSMODE;
            _material.mainTexture = texture;
            // _material.emissiveColor = new BABYLON.Color3(1, 1, 1);
            // _material.disableDepthWrite = true;
            // _material.backFaceCulling = false;
            // _material.alphaMode = alphaMode;
            // _material.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
            return _material;
        });
    }
}
