
import * as BABYLON from "../../render/index";
// import * as BABYLON from "../../pi_babylon/render3d/babylon";
import { computeRadians, EShapeEmitterArcMode, EShapeEmitterDirectionMode, IShapeEmitterType } from "./ishape_emitter_type";
/**
 * 锥体发射器
 */
export class ConeShapeEmitter implements IShapeEmitterType {
    public static MAX_Z = 999999999;
    public directionMode: EShapeEmitterDirectionMode = EShapeEmitterDirectionMode.Unity;
    private _startRadius: number;
    private _angle: number = Math.PI;
    private _height: number = 1;
    private _baseHeight: number;
     /**
      * 半径域
      */
     public radiusRange: number;
     /**
      * 弧形范围
      */
     public arcValue: number = Math.PI * 2;
     /**
      * 弧形范围发射模式
      */
     public arcMode: EShapeEmitterArcMode = EShapeEmitterArcMode.Random;
     /**
      * 弧形周围可产生粒子的离散间隔 - 小于0.01 时, 不做间隔计算
      */
     public arcSpread: number = 0;
     /**
      * 弧形范围发射速度
      */
     public arcSpeed: number = 1;
     /**
      * 弧形范围精度
      */
     public arcSpreadLimit: number = 0.001;

    /**
     * 高度范围
     */
    public heightRange = 0;

    /**
     * Gets or sets a value indicating if all the particles should be emitted from the spawn point only (the base of the cone)
     */
    public emitFromSpawnPointOnly = false;

    /**
     * 半径
     */
    public get radius(): number {
        return this._startRadius;
    }
    /**
     * 半径
     */
    public set radius(value: number) {
        this._startRadius = Math.max(0.0001, value);
        this._buildShape();
    }

    /**
     * 锥体顶角弧度
     */
    public get angle(): number {
        return this._angle;
    }

    /**
     * 锥体顶角弧度
     */
    public set angle(value: number) {
        this._angle = value;
        this._buildShape();
    }

    /**
     * 锥体高度 - 当 heightRange > 0
     */
    public get height(): number {
        return this._height;
    }

    /**
     * 锥体高度 - 当 heightRange > 0
     */
    public set height(value: number) {
        this._height = Math.max(0.0, value);
        this._buildShape();
    }

    private _buildShape() {
        if (this._angle !== 0) {
            this._baseHeight = this._startRadius / Math.tan(this._angle / 2);
        }
        else {
            this._baseHeight = ConeShapeEmitter.MAX_Z;
        }
    }

    public rotation: BABYLON.Vector3 = new BABYLON.Vector3(0, 0, 0);
    public position: BABYLON.Vector3 = new BABYLON.Vector3(0, 0, 0);
    public scaling: BABYLON.Vector3 = new BABYLON.Vector3(1, 1, 1);
    public localMatrix: BABYLON.Matrix = BABYLON.Matrix.Identity();

    alignDirection: boolean = false;
    randomizeDirection: number = 0;
    spherizeDirection: number = 0;
    randomizePosition: number = 0;
    /**
     * Creates a new instance ConeParticleEmitter
     * @param radius the radius of the emission cone (1 by default)
     * @param angle the cone base angle (PI by default)
     * @param directionRandomizer defines how much to randomize the particle direction [0-1] (default is 0)
     */
    constructor(radius = 1, angle = Math.PI,
        /** defines how much to randomize the particle direction [0-1] (default is 0) */
        public directionRandomizer = 0) {
        this.angle = angle;
        this.radius = radius;
    }
    public dispose() {
    }

    public rotateByQuaternionAroundPoint = (position: BABYLON.Vector3, rotateQuat: BABYLON.Quaternion, centerOffset: BABYLON.Vector3, orbtialDiff: BABYLON.Vector3) => {
        position.z = 0;
        position.rotateByQuaternionAroundPointToRef(rotateQuat, centerOffset, orbtialDiff);
    }
    /**
     * Called by the particle System when the direction is computed for the created particle.
     * @param worldMatrix is the world matrix of the particle system
     * @param directionToUpdate is the direction vector to update with the result
     * @param particle is the particle we are computed the direction for
     * @param isLocal defines if the direction should be set in local space
     */
    public startDirectionFunction(worldMatrix: BABYLON.Matrix, directionToUpdate: BABYLON.Vector3, position: BABYLON.Vector3, localPosition: BABYLON.Vector3, isLocal: boolean): void {
        if (isLocal) {
            BABYLON.TmpVectors.Vector3[0].copyFrom(localPosition!).addInPlaceFromFloats(0, 0, this._baseHeight).normalize();
        }
        else {
            BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(position.x, position.y, position.z + this._baseHeight, worldMatrix, BABYLON.TmpVectors.Vector3[0]);
            BABYLON.TmpVectors.Vector3[0].subtractToRef(worldMatrix.getTranslation(), BABYLON.TmpVectors.Vector3[0]).normalize();
        }

        var direction = BABYLON.TmpVectors.Vector3[0];

        localPosition.normalizeToRef(BABYLON.TmpVectors.Vector3[1]);
        direction.x = direction.x * (1.0 - this.spherizeDirection) + BABYLON.TmpVectors.Vector3[1].x * this.spherizeDirection;
        direction.y = direction.y * (1.0 - this.spherizeDirection) + BABYLON.TmpVectors.Vector3[1].y * this.spherizeDirection;
        direction.z = direction.z * (1.0 - this.spherizeDirection) + BABYLON.TmpVectors.Vector3[1].z * this.spherizeDirection;
        direction.normalize();

        direction.x += BABYLON.Scalar.RandomRange(0, this.randomizeDirection);
        direction.y += BABYLON.Scalar.RandomRange(0, this.randomizeDirection);
        direction.z += BABYLON.Scalar.RandomRange(0, this.randomizeDirection);

        directionToUpdate.copyFrom(direction);
        directionToUpdate.normalize();
    }

    /**
     * Called by the particle System when the position is computed for the created particle.
     * @param worldMatrix is the world matrix of the particle system
     * @param positionToUpdate is the position vector to update with the result
     * @param particle is the particle we are computed the position for
     * @param isLocal defines if the position should be set in local space
     */
    startPositionFunction(worldMatrix: BABYLON.Matrix, positionToUpdate: BABYLON.Vector3, emissionLoop: number, emissionProgress: number, emissionIndex: number, emissionTotal: number, isLocal: boolean): void {
        let s = computeRadians(emissionLoop, emissionProgress, emissionIndex, emissionTotal, Math.PI * 2, this.arcValue, this.arcSpread, this.arcSpeed, this.arcMode);

        var h: number;

        if (!this.emitFromSpawnPointOnly) {
            h = BABYLON.Scalar.RandomRange(0, this.heightRange);
        } else {
            h = 0.0;
        }
        h = h * h;
        h = Math.max(0.00001, h);

        let t = BABYLON.Scalar.RandomRange(0, this.radiusRange);
        var radius = this._startRadius - this._startRadius * t * t;
        if (this._baseHeight > 0) {
            radius = radius * (h * this._height + this._baseHeight) / this._baseHeight;
        }

        var randX = 0;
        var randZ = 0;
        var randY = 0;

        if (this.directionMode == EShapeEmitterDirectionMode.Unity) {
            randX = radius * Math.sin(-s);
            randY = radius * Math.cos(-s);
            randZ = h * this._height;
        }
        else {
            randX = radius * Math.sin(s);
            randZ = radius * Math.cos(s);
            randY = h * this._height;
        }

        randX += BABYLON.Scalar.RandomRange(-this.randomizePosition, this.randomizePosition);
        randZ += BABYLON.Scalar.RandomRange(-this.randomizePosition, this.randomizePosition);
        randY += BABYLON.Scalar.RandomRange(-this.randomizePosition, this.randomizePosition);

        if (isLocal) {
            positionToUpdate.x = randX;
            positionToUpdate.y = randY;
            positionToUpdate.z = randZ;
        }
        else {
            BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(randX, randY, randZ, worldMatrix, positionToUpdate);
        }
    }

    /**
     * Clones the current emitter and returns a copy of it
     * @returns the new emitter
     */
    public clone(): ConeShapeEmitter {
        throw new Error("Method not implemented.");
    }

    /**
     * Returns the string "ConeParticleEmitter"
     * @returns a string containing the class name
     */
    public getClassName(): string {
        return "ConeParticleEmitter";
    }

    /**
     * Serializes the particle system to a JSON object.
     * @returns the JSON object
     */
    public serialize(): any {
        var serializationObject: any = {};

        serializationObject.type = this.getClassName();
        serializationObject.radius = this._startRadius;
        serializationObject.angle = this._angle;
        serializationObject.directionRandomizer = this.directionRandomizer;
        serializationObject.radiusRange = this.radiusRange;
        serializationObject.heightRange = this.heightRange;
        serializationObject.emitFromSpawnPointOnly = this.emitFromSpawnPointOnly;

        return serializationObject;
    }

    /**
     * Parse properties from a JSON object
     * @param serializationObject defines the JSON object
     */
    public parse(serializationObject: any): void {
        this.radius = serializationObject.radius;
        this.angle = serializationObject.angle;
        this.directionRandomizer = serializationObject.directionRandomizer;

        this.radiusRange = serializationObject.radiusRange !== undefined ? serializationObject.radiusRange : 1;
        this.heightRange = serializationObject.radiusRange !== undefined ? serializationObject.heightRange : 1;
        this.emitFromSpawnPointOnly = serializationObject.emitFromSpawnPointOnly !== undefined ? serializationObject.emitFromSpawnPointOnly : false;
    }
}
