
import * as BABYLON from "../../render/index";
// import * as BABYLON from "../../pi_babylon/render3d/babylon";
import { EShapeEmitterDirectionMode, IShapeEmitterType } from "./ishape_emitter_type";

/**
 * 点发射器
 */
export class PointShapeEmitter implements IShapeEmitterType {
    public static MAX_Z = 999999999;
    /**
     * 创建模式
     */
    public directionMode: EShapeEmitterDirectionMode = EShapeEmitterDirectionMode.Unity;
    /**
     * 第一发射方向
     */
    public direction1 = new BABYLON.Vector3(0, 0, 1.0);
    /**
     * 第二发射方向
     */
    public direction2 = new BABYLON.Vector3(0, 0, 1.0);

    public rotation: BABYLON.Vector3 = new BABYLON.Vector3(0, 0, 0);
    public position: BABYLON.Vector3 = new BABYLON.Vector3(0, 0, 0);
    public scaling: BABYLON.Vector3 = new BABYLON.Vector3(1, 1, 1);
    public localMatrix: BABYLON.Matrix = BABYLON.Matrix.Identity();

    alignDirection: boolean = false;
    randomizeDirection: number = 0;
    spherizeDirection: number = 0;
    randomizePosition: number = 0;
    public rotateByQuaternionAroundPoint = (position: BABYLON.Vector3, rotateQuat: BABYLON.Quaternion, centerOffset: BABYLON.Vector3, orbtialDiff: BABYLON.Vector3) => {
        position.rotateByQuaternionAroundPointToRef(rotateQuat, centerOffset, orbtialDiff);
    }
    /**
     * 
     */
    public startDirectionFunction(worldMatrix: BABYLON.Matrix, directionToUpdate: BABYLON.Vector3, position: BABYLON.Vector3, localPosition: BABYLON.Vector3, isLocal: boolean): void {
        var randX = BABYLON.Scalar.RandomRange(this.direction1.x, this.direction2.x);
        var randY = BABYLON.Scalar.RandomRange(this.direction1.y, this.direction2.y);
        var randZ = BABYLON.Scalar.RandomRange(this.direction1.z, this.direction2.z);

        if (isLocal) {
            directionToUpdate.copyFromFloats(randX, randY, randZ);
        }
        else {
            BABYLON.Vector3.TransformNormalFromFloatsToRef(randX, randY, randZ, worldMatrix, directionToUpdate);
        }
    }

    /**
     * Called by the particle System when the position is computed for the created particle.
     * @param worldMatrix is the world matrix of the particle system
     * @param positionToUpdate is the position vector to update with the result
     * @param particle is the particle we are computed the position for
     * @param isLocal defines if the position should be set in local space
     */
    startPositionFunction(worldMatrix: BABYLON.Matrix, positionToUpdate: BABYLON.Vector3, emissionLoop: number, emissionProgress: number, emissionIndex: number, emissionTotal: number, isLocal: boolean): void {
        if (isLocal) {
            positionToUpdate.copyFromFloats(0, 0, 0);
        }
        else {
            BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(0, 0, 0, worldMatrix, positionToUpdate);
        }
    }

    /**
     * Clones the current emitter and returns a copy of it
     * @returns the new emitter
     */
    public clone(): PointShapeEmitter {
        throw new Error("Method not implemented.");
    }

    /**
     * @returns a string containing the class name
     */
    public getClassName(): string {
        return "PointParticleEmitter";
    }

    /**
     * Serializes the particle system to a JSON object.
     * @returns the JSON object
     */
    public serialize(): any {
        var serializationObject: any = {};

        serializationObject.type = this.getClassName();
        serializationObject.direction1 = this.direction1.asArray();
        serializationObject.direction2 = this.direction2.asArray();

        return serializationObject;
    }

    /**
     * Parse properties from a JSON object
     * @param serializationObject defines the JSON object
     */
    public parse(serializationObject: any): void {
        BABYLON.Vector3.FromArrayToRef(serializationObject.direction1, 0, this.direction1);
        BABYLON.Vector3.FromArrayToRef(serializationObject.direction2, 0, this.direction2);
    }
    public dispose() {
        // this.direction1 = undefined;
        // this.direction2 = undefined;
    }
}
