// export function RotationYawPitchRollToRef(yaw: number, pitch: number, roll: number, result: {x: number, y: number, z: number, w: number}) {
//     // Produces a quaternion from Euler angles in the z-y-x orientation (Tait-Bryan angles)
//     var halfRoll    = Math.round(roll   * 0.5 / RADIANS_ITER_E) % RADIANS_ITER_COUNT;
//     var halfPitch   = Math.round(pitch  * 0.5 / RADIANS_ITER_E) % RADIANS_ITER_COUNT;
//     var halfYaw     = Math.round(yaw    * 0.5 / RADIANS_ITER_E) % RADIANS_ITER_COUNT;
    
//     halfRoll = halfRoll < 0 ? RADIANS_ITER_COUNT + halfRoll : halfRoll;
//     halfPitch = halfPitch < 0 ? RADIANS_ITER_COUNT + halfPitch : halfPitch;
//     halfYaw = halfYaw < 0 ? RADIANS_ITER_COUNT + halfYaw : halfYaw;

//     var sinRoll     = _sin(halfRoll);
//     var cosRoll     = _cos(halfRoll);
//     var sinPitch    = _sin(halfPitch);
//     var cosPitch    = _cos(halfPitch);
//     var sinYaw      = _sin(halfYaw);
//     var cosYaw      = _cos(halfYaw);

//     result.x        = (cosYaw * sinPitch * cosRoll) + (sinYaw * cosPitch * sinRoll);
//     result.y        = (sinYaw * cosPitch * cosRoll) - (cosYaw * sinPitch * sinRoll);
//     result.z        = (cosYaw * cosPitch * sinRoll) - (sinYaw * sinPitch * cosRoll);
//     result.w        = (cosYaw * cosPitch * cosRoll) + (sinYaw * sinPitch * sinRoll);
// }

export function RotationYawPitchRollToRef(yaw: number, pitch: number, roll: number, result: {x: number, y: number, z: number, w: number}) {
    // Produces a quaternion from Euler angles in the z-y-x orientation (Tait-Bryan angles)
    var halfRoll    = roll   * 0.5;
    var halfPitch   = pitch  * 0.5;
    var halfYaw     = yaw    * 0.5;

    var sinRoll     = Math.sin(halfRoll);
    var cosRoll     = Math.cos(halfRoll);
    var sinPitch    = Math.sin(halfPitch);
    var cosPitch    = Math.cos(halfPitch);
    var sinYaw      = Math.sin(halfYaw);
    var cosYaw      = Math.cos(halfYaw);

    result.x        = (cosYaw * sinPitch * cosRoll) + (sinYaw * cosPitch * sinRoll);
    result.y        = (sinYaw * cosPitch * cosRoll) - (cosYaw * sinPitch * sinRoll);
    result.z        = (cosYaw * cosPitch * sinRoll) - (sinYaw * sinPitch * cosRoll);
    result.w        = (cosYaw * cosPitch * cosRoll) + (sinYaw * sinPitch * sinRoll);
};

export function sin(radians: number) {
    let index = Math.round(radians / RADIANS_ITER_E) % RADIANS_ITER_COUNT;
    index = index < 0 ? RADIANS_ITER_COUNT + index: index;
    return SIN_TABLE[index];
}

export function cos(radians: number) {
    let index = Math.round(radians / RADIANS_ITER_E) % RADIANS_ITER_COUNT;
    index = index < 0 ? RADIANS_ITER_COUNT + index: index;
    return COS_TABLE[index];
}


function _sin(index: number) {
    return SIN_TABLE[index];
}

function _cos(index: number) {
    return COS_TABLE[index];
}

const RADIANS_ITER_COUNT = 360;
const RADIANS_ITER_E = Math.PI * 2 / 360;
const SIN_TABLE: number[] = [];
const COS_TABLE: number[] = [];

for (let i = 0; i < RADIANS_ITER_COUNT; i++) {
    SIN_TABLE[i] = Math.sin(i / 360 * Math.PI * 2);
    COS_TABLE[i] = Math.cos(i / 360 * Math.PI * 2);
}

export function directionToQuaternion(direction: {x: number, y: number, z: number}, result: {x: number, y: number, z: number, w: number}) {
    const xAxis = direction.x;
    const yAxis = direction.y;
    const zAxis = direction.z;

    let yaw = -Math.atan2(zAxis, xAxis) + Math.PI / 2;
    let len = Math.sqrt(xAxis * xAxis + zAxis * zAxis);
    let pitch = -Math.atan2(yAxis, len);

    RotationYawPitchRollToRef(yaw, pitch, 0, result);

}