import * as BABYLON from "../render/index";
import { ParticleSystemTool } from "./particle_system_tool";
import { Entity, Engine } from "../render/engines/engine";
import { IParticleSystemD3, p3d } from "../render/types";
import { ExtensionTools } from "../base";

/**
 * 对接 Unity ParticleSystem Mesh Mode
 */
export class MeshParticleSystem extends BABYLON.Mesh implements IParticleSystemD3 {

    public static sqrt3 = Math.sqrt(3);
    public static registCompute = (call: () => void) => {
        ExtensionTools.setPermanentBefore(call);
    }
    public static registRecycle = (call: () => void) => {
        ExtensionTools.setPermanent(call);
    }
    public static unregistCompute = (call: () => void) => {
        ExtensionTools.clearPermanent(call);
    }
    public static unregistRecycle = (call: () => void) => {
        ExtensionTools.clearPermanent(call);
    }

    public psTool: ParticleSystemTool;

    private source: BABYLON.Mesh;

    private _alwaysSelectAsActive: boolean = true;
    public get alwaysSelectAsActive() {
        return false;
    }
    public set alwaysSelectAsActive(v: boolean) {
        this._alwaysSelectAsActive = v;
        if (this.source) {
            // this.source.alwaysSelectAsActiveMesh = v;
        }
    }

    private _trailmesh: Entity;
    private _trailgeometry: Entity;

    constructor(name: string, scene?: BABYLON.Scene) {
        let instancestates: BABYLON.InstanceAttribute[] = [
            { key: "InsColor4", vtype: BABYLON.Constants.INSTANCE_ATTR_VEC4, code: "A_COLOR4 = InsColor4;" },
            { key: "InsTilloff", vtype: BABYLON.Constants.INSTANCE_ATTR_VEC4, code: "A_UV = A_UV * InsTilloff.xy + InsTilloff.zw;" },
        ];

        super(name, scene, instancestates, undefined, true);

        let idtrailmesh = p3d.p3d_entity(Engine.native());
        let idtrailgeo = p3d.p3d_entity(Engine.native());
        this._trailmesh = new Entity(idtrailmesh);
        this._trailgeometry = new Entity(idtrailgeo);

        // // if (this._scene.getEngine().getCaps().instancedArrays) {
        //     this.psTool = new ParticleSystemTool(scene);
        //     this.psTool.getParentWorldMatrix = () => {
        //         return this.parent ? this.parent.computeWorldMatrix() : undefined;
        //     };

        //     this.psTool.getCameraRotationMatrixInvert = () => {
        //         (this._scene).activeCamera.computeWorldMatrix();
        //         return (this._scene).activeCamera.getViewMatrix().getRotationMatrix().invert();
        //     };
        //     this.psTool.getCameraPosition = () => {
        //         (this._scene).activeCamera.computeWorldMatrix();
        //         return (this._scene).activeCamera.globalPosition;
        //     };

        //     this.psTool.getCameraMatrix = () => {
        //         (this._scene).activeCamera.computeWorldMatrix();
        //         return (this._scene).activeCamera.getViewMatrix();
        //     };

        //     this.psTool.getWorldMatrix = () => {
        //         return this.computeWorldMatrix();
        //     };

        //     this.psTool.getLocalMatrix = () => {
        //         return this._localMatrix;
        //     };
        // // } else {
        // //     console.error(`Can not use instancedArrays`);
        // // }
    }
    get speed() {
        return this._speed;
    }

    private _speed: number = 1.0;
    public set speedRatio(val: number) {
        if (this._speed != val) {
            this._speed = val;
            // p3d.p3d_particle_system_timescale(Engine.native(), Engine.param(), this.entity, this._speed);
            this._scene.cmdsParticleSpeed.set(this._entity, this);
        }
    }
    public get speedRatio() {
        return this._speed;
    }
    public isPlaying(): boolean {
        return this._isMPPlaying;
    }

    public formatByGLTF(gltf: p3d.GLTFRes, index: number) {

        p3d.p3d_particle_system_with_gltf(
            Engine.commands(), this._scene.entity,
            this.entity,
            this._trailmesh.entity, this._trailgeometry.entity, gltf, index,
            ExtensionTools.getAtom("InsColor4"), ExtensionTools.getAtom("InsTilloff"), 
        );

        this.start();
    }

    public setSourceMesh(mesh: BABYLON.Mesh) {
        // this.source = mesh;
        // if (mesh) {
        //     if (!this.psTool) {
        //         mesh.setEnabled(false);
        //         return;
        //     }
        //     // mesh.renderAsMeshParticle = true;
        //     // mesh.useVertexColors = true;
        //     if (mesh.alphaIndex >= 3000) {
        //         // mesh.hasVertexAlpha  = true;
        //     }
        //     // mesh.alwaysSelectAsActiveMesh = this._alwaysSelectAsActive;
        //     mesh.parent = this;
        //     // mesh.freezeWorldMatrix(BABYLON.Matrix.Identity());
        // }
        this.build();
    }

    public dispose(doNotRecurse?: boolean, disposeMaterialAndTextures?: boolean) {
        if (this._disposed) {
            return;
        }
        // this.stop();

        this._trailmesh.dispose();
        this._trailgeometry.dispose();

        // this.psTool.dispose();

        this.psTool = undefined;

        // if (this.source) {
        //     // this.source.material?.dispose(true, true, true);
        //     this.source.dispose(false, true);
        // }

        super.dispose();
    }

    private build() {

        // this.psTool.build();

        // if (this.source) {
        //     // this.source.thinInstanceSetBuffer("matrix", this.psTool.mpMatrixList, 16, false);
        //     // this.source.thinInstanceSetBuffer("color", this.psTool.mpColorData, 4, false);
        //     // this.source.thinInstanceSetBuffer(ATTRIBUTE_PS_UV_SHEET, this.psTool.mpUVSheetData, 4, false);
        //     // var colorBuffer = new BABYLON.VertexBuffer(this._scene.getEngine(), this.psTool.mpColorData, BABYLON.VertexBuffer.ColorKind, true, false, 4, true);
        //     // this.source.setVerticesBuffer(colorBuffer);
        //     // p3d.p3d_mesh_instance_world_matrixs(Engine.native(), Engine.param(), this.source.entity, this.psTool.mpMatrixList, 0, this.psTool.mpColorData);
        // }
    }

    setEnabled(val: boolean): void {
        super.setEnabled(val);
    }

    private _isMPPlaying: boolean = true;
    public start() {
        this._isMPPlaying = true;
        this._scene.cmdsParticlePlay.set(this._entity, this);
        // p3d.p3d_particle_system_start(Engine.native(), Engine.param(), this.entity);
        // if (!this._isMPPlaying && this.psTool) {
        //     this.psTool.start();

        //     MeshParticleSystem.registCompute(this._computeCall);
        //     // MeshParticleSystem.registRecycle(this._recycleCall);
        //     MeshParticleSystem.registCompute(this._updateCall);
        //     // this._scene.onBeforeRenderObservable.add(this._updateCall);

        //     this._isMPPlaying = true;
        // }
    }

    private _computeCall = () => {
        // if (this._disposed) {
        //     return;
        // }
        // // this.computeParents();
        // this.psTool.varCompute(this._scene.deltaTime || 0) ; // this._scene.deltaTime || 0);
        // this._mpDirty = true;

        // this._recycleCall();
    }

    private _mpDirty = true;
    private _updateCall = () => {
        // if (this._disposed) {
        //     return;
        // }
        // if (this._mpDirty) {
        //     // this.computeParents();
        //     const count = this.psTool.mpUpdate();
        //     let enable = count > 0 || this.psTool.maxLifetimeOverage > 0;

        //     if (this.source) {
        //         // this.source.thinInstanceBufferUpdated("matrix");
        //         // this.source.updateVerticesData(BABYLON.VertexBuffer.ColorKind, this.psTool.mpColorData);
        //         // this.source.thinInstanceBufferUpdated("color");
        //         // this.source.thinInstanceBufferUpdated(ATTRIBUTE_PS_UV_SHEET);
        //         // this.source.thinInstanceCount = count;

        //         if (this.source._geometryID) {
        //             if (count > 0) {
        //                 p3d.p3d_mesh_instance_world_matrixs(Engine.native(), Engine.param(), this.source._geometryID, this.psTool.mpMatrixList, 0, count * 16);
        //                 p3d.p3d_mesh_instance_colors(Engine.native(), Engine.param(), this.source._geometryID, this.psTool.mpColorData, 0, count * 4);
        //                 p3d.p3d_mesh_instance_tilloffs(Engine.native(), Engine.param(), this.source._geometryID, this.psTool.mpUVData, 0, count * 4);
        //             } else if (enable) {
        //                 this.psTool.mpColorData[0] = 0;
        //                 this.psTool.mpColorData[1] = 0;
        //                 this.psTool.mpColorData[2] = 0;
        //                 this.psTool.mpColorData[3] = 0;
        //                 p3d.p3d_mesh_instance_world_matrixs(Engine.native(), Engine.param(), this.source._geometryID, this.psTool.mpMatrixList, 0, 1 * 16);
        //                 p3d.p3d_mesh_instance_colors(Engine.native(), Engine.param(), this.source._geometryID, this.psTool.mpColorData, 0, 1 * 4);
        //                 p3d.p3d_mesh_instance_tilloffs(Engine.native(), Engine.param(), this.source._geometryID, this.psTool.mpUVData, 0, 1 * 4);
        //             }
        //         }

        //         // this.source.setEnabled(count > 0 && this.isEnabled());
        //     }

        //     p3d.p3d_node_enable(Engine.native(), Engine.param(), this.source.entity, enable);

        //     this._mpDirty = false;
        // }
    }

    // private computeParents() {
    //     let max = 32;
    //     let list: BABYLON.TransformNode[] = [this];
    //     let node = <BABYLON.TransformNode>this;
    //     while (node.parent != null) {
    //         list.push(<BABYLON.TransformNode>node.parent);
    //         node = <any>node.parent;
    //         if (max <= 0) {
    //             break;
    //         }
    //     }
    //     let len = list.length;
    //     for (let i = len - 1; i >= 0; i--) {
    //         list[i].computeWorldMatrix();
    //     }
    // }

    private _recycleCall = () => {
        // this.psTool.recycle();
    }

    public stop() {
        // // p3d.p3d_()
        // if (this._isMPPlaying && this.psTool) {
        //     this.psTool.stop();

        //     MeshParticleSystem.unregistCompute(this._computeCall);
        //     // MeshParticleSystem.unregistRecycle(this._recycleCall);
        //     // this._scene.onBeforeRenderObservable.removeCallback(this._updateCall);
        //     MeshParticleSystem.unregistCompute(this._updateCall);

        //     this._isMPPlaying = false;
        // }
        this._isMPPlaying = false;
        this._scene.cmdsParticlePlay.set(this._entity, this);
        // p3d.p3d_particle_system_stop(Engine.native(), Engine.param(), this.entity);
    }

    public reset() {
        this.stop();
    }
}