
import * as BABYLON from "../../render/index";
// import * as BABYLON from "../../pi_babylon/render3d/babylon";
import { Color4Gradient, FloatInterpolation, ICurve } from "../interpolation";
import { IParticle } from "../particle_base";
import { EInterpolationCurveMode, EInterpolationGradienMode, FourGradientInfo, OneParamInfo, ThreeParamInfo, TInterpolateColor, TInterpolateConstant, TInterpolateCurve, TInterpolateGradient, TInterpolateRandom, TInterpolateTwoColors, TInterpolateTwoConstants, TInterpolateTwoCurves, TInterpolateTwoGradients } from "../iparticle_system_config";

export const TempVector3A = new BABYLON.Vector3(0, 0, 0);
export const TempVector3B = new BABYLON.Vector3(0, 0, 0);
export const TempVector3C = new BABYLON.Vector3(0, 0, 0);
export const TempVector3D = new BABYLON.Vector3(0, 0, 0);
export const TempVector3E = new BABYLON.Vector3(0, 0, 0);

export const TempColor4A = new BABYLON.Color4(1, 1, 1, 1);
export const TempColor4B = new BABYLON.Color4(1, 1, 1, 1);
export const TempColor4C = new BABYLON.Color4(1, 1, 1, 1);

export const TempQuaternionA = new BABYLON.Quaternion(0, 0, 0);
export const TempQuaternionB = new BABYLON.Quaternion(0, 0, 0);
export const TempQuaternionC = new BABYLON.Quaternion(0, 0, 0);

export const TempMatrixA = BABYLON.Matrix.Identity();
export const TempMatrixB = BABYLON.Matrix.Identity();
export const TempMatrixC = BABYLON.Matrix.Identity();

export interface IParticleModifier {
    modify(particle: IParticle, amount: number, deltaSeconds: number): void;
}

export function transformVectorAsLocalSpace(source: BABYLON.Vector3, transformInvert: BABYLON.Matrix, result: BABYLON.Vector3) {
    // no thing
}

export function transformVectorAsWorldSpace(source: BABYLON.Vector3, transformInvert: BABYLON.Matrix, result: BABYLON.Vector3) {
    BABYLON.Vector3.TransformCoordinatesToRef(source, transformInvert, result);
}

export function interpolateThree(
    forceInterpolationX: FloatInterpolation,
    forceInterpolationY: FloatInterpolation,
    forceInterpolationZ: FloatInterpolation,
    amount: number,
    random: [number, number, number, number],
    result: BABYLON.Vector3
) {
    let x = forceInterpolationX.interpolate(amount, random[0]);
    let y = forceInterpolationY.interpolate(amount, random[1]);
    let z = forceInterpolationZ.interpolate(amount, random[2]);
    result.copyFromFloats(x, y, z);
}

export function interpolateTranslationNoAxis(
    forceInterpolationX: FloatInterpolation,
    forceInterpolationY: FloatInterpolation,
    forceInterpolationZ: FloatInterpolation,
    amount: number,
    random: [number, number, number, number],
    result: BABYLON.Vector3
) {
    let x = forceInterpolationX.interpolate(amount, random[0]);
    result.copyFromFloats(x, x, x);
}

export function interpolateRotationNoAxis(
    forceInterpolationX: FloatInterpolation,
    forceInterpolationY: FloatInterpolation,
    forceInterpolationZ: FloatInterpolation,
    amount: number,
    random: [number, number, number, number],
    result: BABYLON.Vector3
) {
    let z = forceInterpolationX.interpolate(amount, random[0]);
    result.copyFromFloats(0, 0, z);
}

export function interpolateScalingNoAxis(
    forceInterpolationX: FloatInterpolation,
    forceInterpolationY: FloatInterpolation,
    forceInterpolationZ: FloatInterpolation,
    amount: number,
    random: [number, number, number, number],
    result: BABYLON.Vector3
) {
    let x = forceInterpolationX.interpolate(amount, random[0]);
    result.copyFromFloats(x, x, x);
}

export function interpolateFour(
    forceInterpolationX: FloatInterpolation,
    forceInterpolationY: FloatInterpolation,
    forceInterpolationZ: FloatInterpolation,
    amount: number,
    random: [number, number, number, number],
    result: BABYLON.Color4
) {
    let r = forceInterpolationX.interpolate(amount, random[0]);
    let g = forceInterpolationY.interpolate(amount, random[1]);
    let b = forceInterpolationZ.interpolate(amount, random[2]);
    let a = forceInterpolationZ.interpolate(amount, random[3]);
    result.copyFromFloats(r, g, b, a);
}

export interface Vector3Interpolate {
    x: FloatInterpolation;
    y: FloatInterpolation;
    z: FloatInterpolation;
}

export function applyVector3Interpolation(config: OneParamInfo | ThreeParamInfo, target: Vector3Interpolate) {
    applyFloatInterpolationFromMulti(config, 0, target.x);
    applyFloatInterpolationFromMulti(config, 1, target.y);
    applyFloatInterpolationFromMulti(config, 2, target.z);
}

export function applyFloatInterpolationFromMulti(config: OneParamInfo | ThreeParamInfo, index: number, target: FloatInterpolation) {
    let temp = target;

    switch (config[1]) {
        case TInterpolateConstant: {
            let tempMode = EInterpolationCurveMode.Constant;
            let tempData0: number;
            let tempData1: number;

            tempMode = EInterpolationCurveMode.Constant;
            if (config[0] == 1) {
                tempData0 = config[2];
                tempData1 = config[2];
            } else {
                tempData0 = config[2][index];
                tempData1 = config[2][index];
            }
            temp.mode = tempMode;
            temp.constant0 = tempData0;
            temp.constant1 = tempData1;
            break;
        }
        case TInterpolateTwoConstants: {
            let tempMode = EInterpolationCurveMode.TwoConstants;
            let tempData0: number;
            let tempData1: number;

            tempMode = EInterpolationCurveMode.TwoConstants;
            if (config[0] == 1) {
                tempData0 = config[2];
                tempData1 = config[3];
            } else {
                tempData0 = config[2][index];
                tempData1 = config[3][index];
            }
            temp.mode = tempMode;
            temp.constant0 = tempData0;
            temp.constant1 = tempData1;
            break;
        }
        case TInterpolateCurve: {
            let tempMode = EInterpolationCurveMode.Curve;
            let tempData0: ICurve;
            let tempData1: ICurve;

            tempMode = EInterpolationCurveMode.Curve;
            if (config[0] == 1) {
                tempData0 = config[2];
                tempData1 = config[2];
            } else {
                tempData0 = config[2][index];
                tempData1 = config[2][index];
            }
            temp.mode = tempMode;
            temp.minCurve = tempData0;
            temp.maxCurve = tempData1;
            break;
        }
        case TInterpolateTwoCurves: {
            let tempMode = EInterpolationCurveMode.Curve;
            let tempData0: ICurve;
            let tempData1: ICurve;

            tempMode = EInterpolationCurveMode.Curve;
            if (config[0] == 1) {
                tempData0 = config[2];
                tempData1 = config[3];
            } else {
                tempData0 = config[2][index];
                tempData1 = config[3][index];
            }
            temp.mode = tempMode;
            temp.minCurve = tempData0;
            temp.maxCurve = tempData1;
            break;
        }
    }
}

export class TranslationInterpolate implements Vector3Interpolate {
    private _isAxis: boolean = false;
    public set isAxis(value: boolean) {
        if (this._isAxis !== value) {
            this._isAxis = value;
            if (value) {
                this.interpolate = interpolateThree;
            } else {
                this.interpolate = interpolateTranslationNoAxis;
            }
        }
    }
    public get isAxis() {
        return this._isAxis;
    }

    constructor(
        public x: FloatInterpolation,
        public y: FloatInterpolation,
        public z: FloatInterpolation,
    ) {
        this.transformForce = transformVectorAsLocalSpace;
        this.interpolate = interpolateTranslationNoAxis;
    }

    public compute(amount: number, random: [number, number, number, number], transformInvert: BABYLON.Matrix, result: BABYLON.Vector3) {
        this.interpolate(
            this.x,
            this.y,
            this.z,
            amount,
            random,
            result
        );
        this.transformForce(result, transformInvert, result);
    }

    protected transformForce: (source: BABYLON.Vector3, transformInvert: BABYLON.Matrix, result: BABYLON.Vector3) => void;
    protected interpolate: (
        forceInterpolationX: FloatInterpolation,
        forceInterpolationY: FloatInterpolation,
        forceInterpolationZ: FloatInterpolation,
        amount: number,
        random: [number, number, number, number],
        result: BABYLON.Vector3
    ) => void;

    public static format(config: OneParamInfo | ThreeParamInfo, target: TranslationInterpolate) {
        target.isAxis = config[0] !== 1;
        applyVector3Interpolation(config, target);
    }
}

export class RotationInterpolate implements Vector3Interpolate {
    private _isAxis: boolean = false;
    public set isAxis(value: boolean) {
        if (this._isAxis !== value) {
            this._isAxis = value;
            if (value) {
                this.interpolate = interpolateThree;
            } else {
                this.interpolate = interpolateRotationNoAxis;
            }
        }
    }
    public get isAxis() {
        return this._isAxis;
    }

    constructor(
        public x: FloatInterpolation,
        public y: FloatInterpolation,
        public z: FloatInterpolation,
    ) {
        this.interpolate = interpolateRotationNoAxis;
    }

    protected compute(amount: number, random: [number, number, number, number], result: BABYLON.Vector3) {
        this.interpolate(
            this.x,
            this.y,
            this.z,
            amount,
            random,
            result
        );
    }

    protected interpolate: (
        x: FloatInterpolation,
        y: FloatInterpolation,
        z: FloatInterpolation,
        amount: number,
        random: [number, number, number, number],
        result: BABYLON.Vector3
    ) => void;

    public static format(config: OneParamInfo | ThreeParamInfo, target: RotationInterpolate) {
        target.isAxis = config[0] !== 1;
        applyVector3Interpolation(config, target);
    }
}

export class ScalingInterpolate implements Vector3Interpolate {
    private _isAxis: boolean = false;
    public set isAxis(value: boolean) {
        if (this._isAxis !== value) {
            this._isAxis = value;
            if (value) {
                this.interpolate = interpolateThree;
            } else {
                this.interpolate = interpolateScalingNoAxis;
            }
        }
    }
    public get isAxis() {
        return this._isAxis;
    }

    constructor(
        public x: FloatInterpolation,
        public y: FloatInterpolation,
        public z: FloatInterpolation,
    ) {
        this.interpolate = interpolateScalingNoAxis;
    }

    protected compute(amount: number, random: [number, number, number, number], result: BABYLON.Vector3) {
        this.interpolate(
            this.x,
            this.y,
            this.z,
            amount,
            random,
            result
        );
    }

    protected interpolate: (
        forceInterpolationX: FloatInterpolation,
        forceInterpolationY: FloatInterpolation,
        forceInterpolationZ: FloatInterpolation,
        amount: number,
        random: [number, number, number, number],
        result: BABYLON.Vector3
    ) => void;

    public static format(config: OneParamInfo | ThreeParamInfo, target: ScalingInterpolate) {
        target.isAxis = config[0] !== 1;
        applyVector3Interpolation(config, target);
    }
}

export class Color4Interpolate {

    private static tempResult: [number, number, number, number] = [1, 1, 1, 1];
    constructor(
        public gradient: Color4Gradient,
    ) {
        //
    }

    protected compute(amount: number, result: BABYLON.Color4, startAmount: number) {
        this.gradient.interpolate(amount, Color4Interpolate.tempResult, startAmount);
        result.copyFromFloats(
            Color4Interpolate.tempResult[0],
            Color4Interpolate.tempResult[1],
            Color4Interpolate.tempResult[2],
            Color4Interpolate.tempResult[3]
        );
    }

    public static format(config: FourGradientInfo, target: Color4Interpolate) {
        let interpolation = target.gradient;

        switch (config[1]) {
            case TInterpolateRandom: {
                interpolation.mode = EInterpolationGradienMode.Random;
                break;
            }
            case TInterpolateColor: {
                interpolation.mode = EInterpolationGradienMode.Color;
                interpolation.constant0 = config[2];
                interpolation.constant1 = config[2];
                break;
            }
            case TInterpolateTwoColors: {
                interpolation.mode = EInterpolationGradienMode.TwoColors;
                interpolation.constant0 = config[2];
                interpolation.constant1 = config[3];
                break;
            }
            case TInterpolateGradient: {
                interpolation.mode = EInterpolationGradienMode.Gradient;
                interpolation.minGradients = config[2];
                interpolation.maxGradients = config[2];
                break;
            }
            case TInterpolateTwoGradients: {
                interpolation.mode = EInterpolationGradienMode.TwoGradients;
                interpolation.minGradients = config[2];
                interpolation.maxGradients = config[3];
                break;
            }
        }
    }
}