
import * as BABYLON from "../../render/index";
// import * as BABYLON from "../../pi_babylon/render3d/babylon";
import { IParticle } from "../particle_base";
import { IParticleModifier, TempQuaternionA, TempVector3A, TempVector3B, TempVector3C, TempVector3D, TempVector3E, TranslationInterpolate } from "./base";
import { FloatInterpolation } from "../interpolation";
import { EInterpolationCurveMode } from "../iparticle_system_config";

export class LocalPosition implements IParticleModifier {
    public orbitalRotateSpeed: TranslationInterpolate = new TranslationInterpolate(new FloatInterpolation(), new FloatInterpolation(), new FloatInterpolation());
    public orbitalOffset: TranslationInterpolate = new TranslationInterpolate(new FloatInterpolation(), new FloatInterpolation(), new FloatInterpolation());
    public radial: FloatInterpolation = new FloatInterpolation();
    public speedModifier: FloatInterpolation = new FloatInterpolation();

    constructor() {
        this.speedModifier.mode = EInterpolationCurveMode.Constant;
        this.speedModifier.constant0 = 1;
    }
    public rotateByQuaternionAroundPoint = (position: BABYLON.Vector3, rotateQuat: BABYLON.Quaternion, centerOffset: BABYLON.Vector3, orbtialDiff: BABYLON.Vector3) => {
        position.rotateByQuaternionAroundPointToRef(rotateQuat, centerOffset, orbtialDiff);
    }

    modify(particle: IParticle, amount: number, deltaSeconds: number) {

        let orbitalRotate = TempVector3B;
        this.orbitalRotateSpeed.compute(amount, particle.baseRandom, particle.startWorldMatrixInvert, orbitalRotate);

        let speedModifier = this.speedModifier.interpolate(amount, particle.baseRandom[0]);

        let radial = this.radial.interpolate(amount, particle.baseRandom[1]);

        let centerOffset = TempVector3C;
        this.orbitalOffset.compute(amount, particle.baseRandom, particle.startWorldMatrixInvert, centerOffset);

        // centerOffset.scale(orbitalRadial / tempRadial);

        // // 由半径和角速度计算向量速度
        // // let orbtialVelocity = orbitalRotate.scale(orbitalRadial);

        // let zAxis = TempVector3D;
        // zAxis.copyFrom(orbitalRotate);
        // // zAxis.scaleInPlace(-1);
        // zAxis.normalizeToRef(zAxis);

        // // let xAxis = TempVector3D;
        // let xAxis = centerOffset;
        // xAxis.normalizeToRef(xAxis);
        // // xAxis.copyFrom(centerOffset);

        // /// yAxis 即环绕运动方向
        // let yAxis = TempVector3E;
        // BABYLON.Vector3.CrossToRef(zAxis, xAxis, yAxis);
        // yAxis.normalize();

        // yAxis.scaleInPlace(orbitalRadial * Math.PI);

        // // 总速度
        // yAxis.addInPlace(particle.direction);

        // let localResult = TempVector3A;
        // yAxis.scaleToRef(deltaSeconds, localResult);
        // localResult.scaleInPlace(speedModifier);
        // particle.position.addInPlace(localResult);

        // 旋转量
        orbitalRotate.scaleInPlace(deltaSeconds * speedModifier);

        // 旋转的目标位置
        let orbtialDiff = TempVector3D;
        particle.position.subtractToRef(centerOffset, orbtialDiff);

        if (orbtialDiff.lengthSquared() > 0) {
            let rotate = TempQuaternionA;
            BABYLON.Quaternion.FromEulerVectorToRef(orbitalRotate, rotate);

            orbitalRotate.copyFrom(particle.position);
            this.rotateByQuaternionAroundPoint(orbitalRotate, rotate, centerOffset, orbtialDiff);

            // 基础径向向量
            orbtialDiff.subtractToRef(centerOffset, orbtialDiff);
            let len = orbtialDiff.length();
            let scale = (len + radial * deltaSeconds) / len;
            orbtialDiff.scaleInPlace(scale);

            // 新的位置
            orbtialDiff.addToRef(centerOffset, orbtialDiff);

            // 最终径向位移
            orbtialDiff.subtractToRef(orbitalRotate, orbtialDiff);
        } else {
            orbtialDiff.copyFromFloats(0, 0, 0);
        }

        let localResult = TempVector3A;
        particle.direction.scaleToRef(deltaSeconds, localResult);
        localResult.scaleInPlace(speedModifier);

        particle.readldirection.copyFrom(orbtialDiff);
        particle.readldirection.addInPlace(localResult);

        particle.position.addInPlace(particle.readldirection);

        if (deltaSeconds != 0) {
            particle.readldirection.scaleInPlace(1 / Math.abs(deltaSeconds));
            particle.directionLength = particle.readldirection.length();
        }
    }
}