
import * as BABYLON from "../../render/index";
// import * as BABYLON from "../../pi_babylon/render3d/babylon";
import { directionToQuaternion } from "../math";
import { IParticle } from "../particle_base";
import { TempVector3A, TempVector3B, TempVector3C } from "./base";

export type TRenderWorldMatrixCompute = (
    parentWorldMatrix: BABYLON.Matrix,
    realStartMatrix: BABYLON.Matrix,
    resultMatrix: BABYLON.Matrix,
    cameraRotationMatrix: BABYLON.Matrix,
    camraGlobalPos: BABYLON.Vector3,
    direction: BABYLON.Vector3,
    directionLength: number,
    position: BABYLON.Vector3,
    scaling: BABYLON.Vector3,
    rotation: BABYLON.Vector3,
    stretchedLengthScale: number,
    stretchedVelocityScale: number
) => void;

const TempVector3Translation = new BABYLON.Vector3();
const TempVector3Scaling = new BABYLON.Vector3();
const TempVector3Rotation = new BABYLON.Vector3();
const TempQuaternion_0 = new BABYLON.Quaternion();
const TempQuaternion_1 = new BABYLON.Quaternion();
const TempQuaternion_Y = BABYLON.Quaternion.RotationYawPitchRoll(Math.PI / 2, 0, 0);
const TempMatrix_View = new BABYLON.Matrix();
const TempMatrix_1 = new BABYLON.Matrix();
const TempVector3Zero = new BABYLON.Vector3(0, 0, 0);
const TempVector3One = new BABYLON.Vector3(1, 1, 1);

const _StretchedBillboardRot = new BABYLON.Vector3(0, 0, -1.5707963267948966);
(<any>window)._StretchedBillboardRot = _StretchedBillboardRot;

export class StretchedBillboard {
    public static compute(
        parentWorldMatrix: BABYLON.Matrix,
        realStartMatrix: BABYLON.Matrix,
        resultMatrix: BABYLON.Matrix,
        cameraRotationMatrix: BABYLON.Matrix,
        camraGlobalPos: BABYLON.Vector3,
        direction: BABYLON.Vector3,
        directionLength: number,
        position: BABYLON.Vector3,
        scaling: BABYLON.Vector3,
        rotation: BABYLON.Vector3,
        stretchedLengthScale: number,
        stretchedVelocityScale: number
) {
        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(position.x, position.y, position.z, parentWorldMatrix, TempVector3Translation);
        realStartMatrix.setTranslationFromFloats(0, 0, 0);

        realStartMatrix.decompose(TempVector3Scaling, undefined, undefined);

        let speedDirection = TempVector3A;
        let xAxis = TempVector3B;
        let viewDirection = TempVector3C;
        viewDirection.copyFromFloats(0, 0, 1);
        BABYLON.Vector3.TransformCoordinatesToRef(viewDirection, cameraRotationMatrix, viewDirection);
        direction.scaleToRef(1, speedDirection);
        BABYLON.Vector3.TransformCoordinatesToRef(speedDirection, realStartMatrix, speedDirection);
        let speedDirectionLength = directionLength;
        // camraGlobalPos.subtractToRef(TempVector3Translation, viewDirection);

        viewDirection.normalize();
        speedDirection.normalize();
        let zAxis = viewDirection;
        let yAxis = speedDirection;
        BABYLON.Vector3.CrossToRef(speedDirection, zAxis, xAxis);
        const xSquareLength = xAxis.lengthSquared();
        if (xSquareLength === 0) {
            xAxis.x = 1.0;
        } else {
            xAxis.normalizeFromLength(Math.sqrt(xSquareLength));
        }
        BABYLON.Vector3.CrossToRef(xAxis, yAxis, zAxis);
        yAxis.normalize();

        BABYLON.Matrix.FromXYZAxesToRef(xAxis, yAxis, zAxis, realStartMatrix);
        realStartMatrix.setTranslation(TempVector3Translation);

        TempVector3Scaling.multiplyInPlace(scaling);
        TempVector3Scaling.y *= stretchedLengthScale;
        TempVector3Scaling.y += stretchedVelocityScale * speedDirectionLength;
        speedDirection.copyFromFloats(0, -(TempVector3Scaling.y) / 2, 0);

        TempQuaternion_0.copyFromFloats(0, 0, 0, 1);
        // BABYLON.Quaternion.FromEulerAnglesToRef(_StretchedBillboardRot.x, _StretchedBillboardRot.y, _StretchedBillboardRot.z, TempQuaternion_0);
        BABYLON.Matrix.ComposeToRef(TempVector3Scaling, TempQuaternion_0, speedDirection, resultMatrix);

        resultMatrix.multiplyToRef(realStartMatrix, resultMatrix);

        BABYLON.Quaternion.FromEulerAnglesToRef(_StretchedBillboardRot.x, _StretchedBillboardRot.y, _StretchedBillboardRot.z, TempQuaternion_0);
        // // BABYLON.Matrix.ComposeToRef(TempVector3One, TempQuaternion_0, TempVector3Zero, TempMatrix_1);
        TempQuaternion_0.toRotationMatrix(TempMatrix_1);

        TempMatrix_1.multiplyToRef(resultMatrix, resultMatrix);
    }
}

export class HorizontalBillboard {
    public static compute(
        parentWorldMatrix: BABYLON.Matrix,
        realStartMatrix: BABYLON.Matrix,
        resultMatrix: BABYLON.Matrix,
        cameraRotationMatrix: BABYLON.Matrix,
        camraGlobalPos: BABYLON.Vector3,
        direction: BABYLON.Vector3,
        directionLength: number,
        position: BABYLON.Vector3,
        scaling: BABYLON.Vector3,
        rotation: BABYLON.Vector3,
        stretchedLengthScale: number,
        stretchedVelocityScale: number
) {
        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(position.x, position.y, position.z, parentWorldMatrix, TempVector3Translation);

        realStartMatrix.setTranslationFromFloats(0, 0, 0);

        BABYLON.Quaternion.RotationYawPitchRollToRef(rotation.y, rotation.x, rotation.z, TempQuaternion_0);
        BABYLON.Matrix.ComposeToRef(scaling, TempQuaternion_0, TempVector3Zero, TempMatrix_1);
        TempMatrix_1.multiplyToRef(realStartMatrix, realStartMatrix);

        realStartMatrix.decompose(TempVector3Scaling, undefined, undefined);
        BABYLON.Quaternion.FromEulerAnglesToRef(-Math.PI / 2, 0, 0, TempQuaternion_0);
        BABYLON.Matrix.ComposeToRef(TempVector3Scaling, TempQuaternion_0, TempVector3Translation, resultMatrix);
    }
}

export class VerticalBillboard {
    public static compute(
        parentWorldMatrix: BABYLON.Matrix,
        realStartMatrix: BABYLON.Matrix,
        resultMatrix: BABYLON.Matrix,
        cameraRotationMatrix: BABYLON.Matrix,
        camraGlobalPos: BABYLON.Vector3,
        direction: BABYLON.Vector3,
        directionLength: number,
        position: BABYLON.Vector3,
        scaling: BABYLON.Vector3,
        rotation: BABYLON.Vector3,
        stretchedLengthScale: number,
        stretchedVelocityScale: number
) {
        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(position.x, position.y, position.z, parentWorldMatrix, TempVector3Translation);

        realStartMatrix.setTranslationFromFloats(0, 0, 0);

        BABYLON.Quaternion.RotationYawPitchRollToRef(rotation.y, rotation.x, rotation.z, TempQuaternion_0);
        BABYLON.Matrix.ComposeToRef(scaling, TempQuaternion_0, TempVector3Zero, TempMatrix_1);
        TempMatrix_1.multiplyToRef(realStartMatrix, realStartMatrix);

        realStartMatrix.decompose(TempVector3Scaling, undefined, undefined);
        TempVector3Translation.subtractToRef(camraGlobalPos, TempVector3A);
        TempVector3A.y = 0;
        TempVector3A.z = 0;
        if (TempVector3A.x == 0 && TempVector3A.z == 0) {
            TempVector3A.y = 1;
        }
        BABYLON.Vector3.NormalizeToRef(TempVector3A, TempVector3A);
        directionToQuaternion(TempVector3A, TempQuaternion_0);
        BABYLON.Matrix.ComposeToRef(TempVector3Scaling, TempQuaternion_0, TempVector3Translation, resultMatrix);
    }
}

export class RenderAlignmentView {
    public static compute(
        parentWorldMatrix: BABYLON.Matrix,
        realStartMatrix: BABYLON.Matrix,
        resultMatrix: BABYLON.Matrix,
        cameraRotationMatrix: BABYLON.Matrix,
        camraGlobalPos: BABYLON.Vector3,
        direction: BABYLON.Vector3,
        directionLength: number,
        position: BABYLON.Vector3,
        scaling: BABYLON.Vector3,
        rotation: BABYLON.Vector3,
        stretchedLengthScale: number,
        stretchedVelocityScale: number
) {
        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(position.x, position.y, position.z, parentWorldMatrix, TempVector3Translation);

        realStartMatrix.setTranslationFromFloats(0, 0, 0);

        // 自身旋转
        BABYLON.Quaternion.RotationYawPitchRollToRef(rotation.y, rotation.x, rotation.z, TempQuaternion_1);
        BABYLON.Matrix.ComposeToRef(scaling, TempQuaternion_1, TempVector3Zero, TempMatrix_1);

        TempMatrix_1.multiplyToRef(realStartMatrix, realStartMatrix);

        realStartMatrix.decompose(TempVector3Scaling, undefined, undefined);
        // TempQuaternion_0.toEulerAnglesToRef(TempVector3Rotation);
        // let tempRotateZ = TempVector3Rotation.z;

        // 相机方向
        BABYLON.Quaternion.FromRotationMatrixToRef(cameraRotationMatrix, TempQuaternion_0);
        // 叠加自身旋转
        TempQuaternion_0.multiplyInPlace(TempQuaternion_1);
        TempQuaternion_0.normalize();

        BABYLON.Matrix.ComposeToRef(TempVector3Scaling, TempQuaternion_0, TempVector3Translation, resultMatrix);
    }
}

export class RenderAlignmentWorld {
    public static compute(
        parentWorldMatrix: BABYLON.Matrix,
        realStartMatrix: BABYLON.Matrix,
        resultMatrix: BABYLON.Matrix,
        cameraRotationMatrix: BABYLON.Matrix,
        camraGlobalPos: BABYLON.Vector3,
        direction: BABYLON.Vector3,
        directionLength: number,
        position: BABYLON.Vector3,
        scaling: BABYLON.Vector3,
        rotation: BABYLON.Vector3,
        stretchedLengthScale: number,
        stretchedVelocityScale: number
) {
        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(position.x, position.y, position.z, parentWorldMatrix, TempVector3Translation);

        realStartMatrix.setTranslationFromFloats(0, 0, 0);

        // Local - Scaling - Rotation
        BABYLON.Quaternion.RotationYawPitchRollToRef(rotation.y, rotation.x, rotation.z, TempQuaternion_0);
        BABYLON.Matrix.ComposeToRef(scaling, TempQuaternion_0, TempVector3Zero, TempMatrix_1);

        TempMatrix_1.multiplyToRef(realStartMatrix, realStartMatrix);

        realStartMatrix.decompose(TempVector3Scaling, undefined, undefined);
        TempQuaternion_0.copyFromFloats(0, 0, 0, 1);
        BABYLON.Matrix.ComposeToRef(TempVector3Scaling, TempQuaternion_0, TempVector3Translation, resultMatrix);
    }
}

export class RenderAlignmentFacing {
    public static useRightHandedSystem = false;
    public static compute(
        parentWorldMatrix: BABYLON.Matrix,
        realStartMatrix: BABYLON.Matrix,
        resultMatrix: BABYLON.Matrix,
        cameraRotationMatrix: BABYLON.Matrix,
        camraGlobalPos: BABYLON.Vector3,
        direction: BABYLON.Vector3,
        directionLength: number,
        position: BABYLON.Vector3,
        scaling: BABYLON.Vector3,
        rotation: BABYLON.Vector3,
        stretchedLengthScale: number,
        stretchedVelocityScale: number
) {
        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(position.x, position.y, position.z, parentWorldMatrix, TempVector3Translation);
        realStartMatrix.setTranslationFromFloats(0, 0, 0);

        // 自身旋转
        BABYLON.Quaternion.RotationYawPitchRollToRef(rotation.y, rotation.x, rotation.z, TempQuaternion_1);
        BABYLON.Matrix.ComposeToRef(scaling, TempQuaternion_1, TempVector3Zero, TempMatrix_1);

        TempMatrix_1.multiplyToRef(realStartMatrix, realStartMatrix);

        realStartMatrix.decompose(TempVector3Scaling, undefined, undefined);

        // 相机方向
        // TempVector3Translation.subtractFromFloatsToRef(camraGlobalPos.x, camraGlobalPos.y, camraGlobalPos.z, TempVector3A);
        // if (RenderAlignmentFacing.useRightHandedSystem) {
        //     BABYLON.Matrix.LookAtRHToRef(BABYLON.Vector3.ZeroReadOnly, TempVector3A, BABYLON.Vector3.UpReadOnly, TempMatrix_View);
        // } else {
        //     BABYLON.Matrix.LookAtLHToRef(BABYLON.Vector3.ZeroReadOnly, TempVector3A, BABYLON.Vector3.UpReadOnly, TempMatrix_View);
        // }
        if (RenderAlignmentFacing.useRightHandedSystem) {
            BABYLON.Matrix.LookAtRHToRef(camraGlobalPos, TempVector3Translation, BABYLON.Vector3.UpReadOnly, TempMatrix_View);
        } else {
            BABYLON.Matrix.LookAtLHToRef(camraGlobalPos, TempVector3Translation, BABYLON.Vector3.UpReadOnly, TempMatrix_View);
        }
        TempMatrix_View.getRotationMatrixToRef(TempMatrix_View);

        TempQuaternion_0.fromRotationMatrix(cameraRotationMatrix);

        // 叠加自身旋转
        TempQuaternion_0.multiplyInPlace(TempQuaternion_1);
        TempQuaternion_0.normalize();

        BABYLON.Matrix.ComposeToRef(TempVector3Scaling, TempQuaternion_0, TempVector3Translation, resultMatrix);
    }
}

export class RenderAlignmentVelocity {
    public static compute(
        parentWorldMatrix: BABYLON.Matrix,
        realStartMatrix: BABYLON.Matrix,
        resultMatrix: BABYLON.Matrix,
        cameraRotationMatrix: BABYLON.Matrix,
        camraGlobalPos: BABYLON.Vector3,
        direction: BABYLON.Vector3,
        directionLength: number,
        position: BABYLON.Vector3,
        scaling: BABYLON.Vector3,
        rotation: BABYLON.Vector3,
        stretchedLengthScale: number,
        stretchedVelocityScale: number
) {
        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(position.x, position.y, position.z, parentWorldMatrix, TempVector3Translation);
        realStartMatrix.setTranslationFromFloats(0, 0, 0);

        BABYLON.Quaternion.RotationYawPitchRollToRef(rotation.y, rotation.x, rotation.z, TempQuaternion_1);
        BABYLON.Matrix.ComposeToRef(scaling, TempQuaternion_1, TempVector3Zero, TempMatrix_1);

        direction.scaleToRef(1, TempVector3A);
        BABYLON.Vector3.TransformCoordinatesToRef(TempVector3A, realStartMatrix, TempVector3A);
        BABYLON.Vector3.NormalizeToRef(TempVector3A, TempVector3A);
        // 如何旋转以`看向`速度方向
        BABYLON.Matrix.LookAtLHToRef(BABYLON.Vector3.ZeroReadOnly, TempVector3A, BABYLON.Vector3.UpReadOnly, TempMatrix_View);
        TempMatrix_View.getRotationMatrixToRef(TempMatrix_View);
        TempQuaternion_0.fromRotationMatrix(TempMatrix_View);
        // 叠加自身旋转
        TempQuaternion_0.multiplyInPlace(TempQuaternion_1);
        TempQuaternion_0.normalize();

        TempMatrix_1.multiplyToRef(realStartMatrix, realStartMatrix);
        realStartMatrix.decompose(TempVector3Scaling, undefined, undefined);
        BABYLON.Matrix.ComposeToRef(TempVector3Scaling, TempQuaternion_0, TempVector3Translation, resultMatrix);
    }
}

export class RenderAlignmentLocal {
    public static compute(
        parentWorldMatrix: BABYLON.Matrix,
        realStartMatrix: BABYLON.Matrix,
        resultMatrix: BABYLON.Matrix,
        cameraRotationMatrix: BABYLON.Matrix,
        camraGlobalPos: BABYLON.Vector3,
        direction: BABYLON.Vector3,
        directionLength: number,
        position: BABYLON.Vector3,
        scaling: BABYLON.Vector3,
        rotation: BABYLON.Vector3,
        stretchedLengthScale: number,
        stretchedVelocityScale: number
    ) {
        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(position.x, position.y, position.z, parentWorldMatrix, TempVector3Translation);
        realStartMatrix.setTranslationFromFloats(0, 0, 0);

        BABYLON.Quaternion.RotationYawPitchRollToRef(rotation.y, rotation.x, rotation.z, TempQuaternion_0);
        BABYLON.Matrix.ComposeToRef(scaling, TempQuaternion_0, TempVector3Zero, TempMatrix_1);
        TempMatrix_1.multiplyToRef(realStartMatrix, resultMatrix);

        resultMatrix.setTranslation(TempVector3Translation);
    }
}