import { IParticle } from "../particle_base";
import { FloatInterpolation, UpdateRandom } from "../interpolation";
import { IParticleModifier } from "./base";

export enum RowMode {
    Custom = 0,
    Random = 1,
}

export enum TimeMode {
    Liftime = 0,
    Speed = 1,
}

export enum AnimationMode {
    WholeSheet = 0,
    SingleRow = 1,
}

export const ATTRIBUTE_PS_UV_SHEET = 'uv_sheet';

export class TextureSheet implements IParticleModifier {
    public rowMode: RowMode = RowMode.Random;
    public customRow: number = 0;
    public timeMode: TimeMode = TimeMode.Liftime;
    public animMode: AnimationMode = AnimationMode.WholeSheet;
    private _tilesX: number = 1;
    private _tilesY: number = 1;
    // private _tilesXY: number = 1;
    private _uScale: number = 1;
    private _vScale: number = 1;
    public set tilesX(v: number) {
        this._tilesX = v;
        this._uScale = 1 / v;
        // this._tilesXY = this._tilesX * this._tilesY;
    }
    public get tilesX() {
        return this._tilesX;
    }
    public set tilesY(v: number) {
        this._tilesY = v;
        this._vScale = 1 / v;
        // this._tilesXY = this._tilesX * this._tilesY;
    }
    public get tilesY() {
        return this._tilesY;
    }
    public cycles: number = 1;
    public active: boolean = true;

    constructor(
        public frameOverTime: FloatInterpolation,
        public startFrame: FloatInterpolation
    ) {
        this.modifyCall = TextureSheet.modifyForStart;
    }

    public set runAsStart(value: boolean) {
        if (value) {
            this.modifyCall = TextureSheet.modifyForStart;
        } else {
            this.modifyCall = TextureSheet.modifyForOverLifetime;
        }
    }

    modify(particle: IParticle, amount: number, deltaSeconds: number) {
        UpdateRandom();

        this.modifyCall(particle, amount, deltaSeconds, this);
    }

    private modifyCall: (particle: IParticle, amount: number, deltaSeconds: number, modifier: TextureSheet) => void;

    public static modifyForStart(particle: IParticle, amount: number, deltaSeconds: number, modifier: TextureSheet) {
        particle.textureStartFrame  = modifier.startFrame.interpolate(0, particle.baseRandom[0]);

        if (modifier.rowMode == RowMode.Custom) {
            particle.textureRow         = modifier.customRow;
        } else {
            particle.textureRow         = Math.round(Math.random() * modifier._tilesY) % modifier._tilesY;
        }
    }

    public static modifyForOverLifetime(particle: IParticle, amount: number, deltaSeconds: number, modifier: TextureSheet) {
        if (modifier.active && modifier.frameOverTime) {
            const interpolation = modifier.frameOverTime.interpolate((amount * modifier.cycles) % 1, particle.baseRandom[1]);
            let start = particle.textureStartFrame;
            let cellId = 0;
            let cellX = 0;
            let cellY = 0;

            if (modifier.animMode == AnimationMode.SingleRow) {
                cellY = particle.textureRow;
                cellX = Math.floor(((start  + interpolation) * modifier._tilesX) % modifier._tilesX);
            }
            else {
                cellId = (start  + interpolation) * modifier._tilesX * modifier._tilesY;
                cellX = Math.floor(cellId % modifier._tilesX);
                cellY = Math.floor(cellId / modifier._tilesX);
            }

            particle.uv[2] = cellX / modifier.tilesX,
            particle.uv[3] = (modifier._tilesY - cellY - 1) / modifier.tilesY;
        }
        else {
            particle.uv[2] = 0,
            particle.uv[3] = (modifier._tilesY - 0 - 1) / modifier.tilesY;
        }

        particle.uv[0] = 1 / modifier.tilesX;
        particle.uv[1] = 1 / modifier.tilesY;
    }
}