
import { IParticle } from "../particle_base";
import { IParticleModifier, TempVector3A, transformVectorAsLocalSpace, transformVectorAsWorldSpace, TranslationInterpolate } from "./base";
import { FloatInterpolation } from "../interpolation";

export class VelocityOverLifetime extends TranslationInterpolate implements IParticleModifier {
    private _isLocalSpace: boolean = true;
    public set isLocalSpace(value: boolean) {
        if (this._isLocalSpace !== value) {
            this._isLocalSpace = value;
            if (value) {
                this.transformForce = transformVectorAsLocalSpace;
            } else {
                this.transformForce = transformVectorAsWorldSpace;
            }
        }
    }
    public get isLocalSpace() {
        return this._isLocalSpace;
    }

    constructor(x: FloatInterpolation, y: FloatInterpolation, z: FloatInterpolation) {
        super(x, y, z);
        this.isAxis = true;
    }

    modify(particle: IParticle, amount: number, deltaSeconds: number) {
        let localResult = TempVector3A;

        this.compute(amount, particle.baseRandom, particle.startWorldMatrixInvert, localResult);

        let x = localResult.x;
        let y = localResult.y;
        let z = localResult.z;
        localResult.subtractInPlace(particle.velocity);
        particle.velocity.copyFromFloats(x, y, z);

        particle.direction.addInPlace(localResult);
    }
}