import { BABYLON } from "../render3d/babylon";
import { MixGroungMaterial } from "./material";
import { CustomScene } from "../index";

export class MixGround {
    private plane: BABYLON.Mesh;
    private root: BABYLON.TransformNode;
    public material: MixGroungMaterial;
    constructor(
        public scene: CustomScene,
        hsize: number,
        vsize: number,
    ) {
        MixGroungMaterial.regist();
        this.material = new MixGroungMaterial(scene, "MixGround", scene.renderpipeline.opaqueRenderer.pass);

        let mesh = BABYLON.PlaneBuilder.CreatePlane("MixGround", { size: 1 }, scene.impl);
        mesh.scaling = new BABYLON.Vector3(hsize, vsize, 1);

        this.plane = mesh;
        // this.plane = BABYLON.PlaneBuilder.CreatePlane("MixGround", { size: 1 }, scene.impl);

        this.plane.setMaterial(this.material);
        this.material.sceneHSize = hsize;
        this.material.sceneVSize = vsize;
        // debugger;

        this.root = new BABYLON.TransformNode("MixGround", this.scene.impl);
        this.root.rotation.copyFromFloats(Math.PI / 2, 0, 0);
        this.plane.parent = this.root;
        window["rr"] = this.root;
    }

    public set position(val: BABYLON.Vector3) {
        this.root.position.copyFrom(val);
    }
    public set scaling(val: BABYLON.Vector3) {
        this.plane.scaling = val;
    }
    public get position() {
        return this.root.position;
    }
    public get scaling() {
        return this.plane.scaling;
    }
    public set alphaIndex(val: number) {
        this.plane.alphaIndex = val;
    }

    public dispose() {
        this.material?.dispose(true, true);
        this.plane?.dispose(true, true);
    }

    public static rotateSin = Math.sin(0 / 180 * Math.PI);
    public static rotateCos = Math.cos(0 / 180 * Math.PI);
    /**
     * 行列坐标转换为场景坐标
     * 以 Map 中心 为旋转中心 顺时针旋转 45 度, 再纵向缩放为 0.8 
     * @param row 
     * @param col 
     * @param result 
     */
    public static rowcolToPosition(row: number, col: number, result: [number, number]) {
        let x = col;
        let y = row;

        result[0] = MixGround.rotateCos * x - MixGround.rotateSin * y;
        result[1] = MixGround.rotateSin * x + MixGround.rotateCos * y;

        // result[0] += 1 / 2;
        // result[1] += 1 / 2;

        // result[0] += this.tileMap.width / 2;
        result[1] *= MixGround.scaleY;
    }
    public static scaleY = 0.999;
}