
import { SceneStruct } from "../scene_struct";
import { Texture } from "../render/materials/Textures/texture";
import { IMixGroundMaterial, IMixGroundTextureTilloff } from "./interface";
import { EffectMaterial } from "../render/materials/material";
import { Engine } from "../render/engines/engine";
import { Constants } from "../interface/index";
import { UniformTexture, p3d } from "../render/index";
import { ExtensionTools } from "../base";

export class GridInfo {
    static TOP = 1;
    static RIGHT = 2;
    static BOTTOM = 4;
    static LEFT = 8;
    static RIGHT_TOP = 16;
    static RIGHT_BOTTOM = 32;
    static LEFT_BOTTOM = 64;
    static LEFT_TOP = 128;
}

export class MixGroungMaterial extends EffectMaterial implements IMixGroundMaterial {
    static KEY = "MixGroungMaterial";
    static POSITION = "position";
    static UV = "uv";
    static ASSET_TEX0 = "ASSET_TEX0";
    static ASSET_TEX1 = "ASSET_TEX1";
    static ASSET_TEX2 = "ASSET_TEX2";
    static ASSET_TEX3 = "ASSET_TEX3";
    static ASSET_TEX4 = "ASSET_TEX4";
    static ASSET_TILLOFF_0 = "ASSET_TILLOFF_0";
    static ASSET_TILLOFF_1 = "ASSET_TILLOFF_1";
    static ASSET_TILLOFF_2 = "ASSET_TILLOFF_2";
    static ASSET_TILLOFF_3 = "ASSET_TILLOFF_3";
    static ASSET_TILLOFF_4 = "ASSET_TILLOFF_4";
    static MIX_TEX = "CHANNEL_MIX_TEX";
    static MIX_TILLOFF = "MIX_TILLOFF";
    static GRID_TEX = "GRID_TEX";
    static WORLD_VIEW_PROJECTION = "worldViewProjection";
    static VIEW_PROJECTION = "viewProjection";
    static VIEW = "view";
    static WORLD = "world";
    static LIGHT_TEX = "LIGHT_TEX";
    static SCENE_PARAM = "SCENE_PARAM";
    static GRID_PARAM = "GRID_PARAM";
    static LIGHT_PARAM = "LIGHT_PARAM";

    protected _assetTex0: Texture;
    protected _assetTex0Dirty: boolean;
    set assetTex0(val: Texture) {
        this._assetTex0 = val;
        this._assetTex0Dirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    get assetTex0() {
        return this._assetTex0;
    }
    protected _assetTex1: Texture;
    protected _assetTex1Dirty: boolean;
    set assetTex1(val: Texture) {
        this._assetTex1 = val;
        this._assetTex1Dirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    get assetTex1() {
        return this._assetTex1;
    }
    protected _assetTex2: Texture;
    protected _assetTex2Dirty: boolean;
    set assetTex2(val: Texture) {
        this._assetTex2 = val;
        this._assetTex2Dirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    get assetTex2() {
        return this._assetTex2;
    }
    protected _assetTex3: Texture;
    protected _assetTex3Dirty: boolean;
    set assetTex3(val: Texture) {
        this._assetTex3 = val;
        this._assetTex3Dirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    get assetTex3() {
        return this._assetTex3;
    }
    protected _assetTex4: Texture;
    protected _assetTex4Dirty: boolean;
    set assetTex4(val: Texture) {
        this._assetTex4 = val;
        this._assetTex4Dirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    get assetTex4() {
        return this._assetTex4;
    }
    protected _mixTex: Texture;
    protected _mixTexDirty: boolean;
    set mixTex(val: Texture) {
        this._mixTex = val;
        this._mixTexDirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    get mixTex() {
        return this._mixTex;
    }
    protected _lightTex: Texture;
    protected _lightTexDirty: boolean;
    set lightTex(val: Texture) {
        this._lightTex = val;
        this._lightTexDirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    get lightTex() {
        return this._lightTex;
    }
    protected _gridTex: Texture;
    protected _gridTexDirty: boolean;
    set gridTex(val: Texture) {
        this._gridTex = val;
        this._gridTexDirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    get gridTex() {
        return this._gridTex;
    }
    protected _assetTilloff0: IMixGroundTextureTilloff;
    set assetTilloff0(val: IMixGroundTextureTilloff) {
        this._assetTilloff0 = val;
        p3d.p3d_material_uniform_vec4(Engine.commands(), this._entity, ExtensionTools.getAtom(MixGroungMaterial.ASSET_TILLOFF_0), val.uScale, val.vScale, val.uOffset, val.vOffset);
    }
    get assetTilloff0() {
        return this._assetTilloff0;
    }
    protected _assetTilloff1: IMixGroundTextureTilloff;
    set assetTilloff1(val: IMixGroundTextureTilloff) {
        this._assetTilloff1 = val;
        p3d.p3d_material_uniform_vec4(Engine.commands(), this._entity, ExtensionTools.getAtom(MixGroungMaterial.ASSET_TILLOFF_1), val.uScale, val.vScale, val.uOffset, val.vOffset);
    }
    get assetTilloff1() {
        return this._assetTilloff1;
    }
    protected _assetTilloff2: IMixGroundTextureTilloff;
    set assetTilloff2(val: IMixGroundTextureTilloff) {
        this._assetTilloff2 = val;
        p3d.p3d_material_uniform_vec4(Engine.commands(), this._entity, ExtensionTools.getAtom(MixGroungMaterial.ASSET_TILLOFF_2), val.uScale, val.vScale, val.uOffset, val.vOffset);
    }
    get assetTilloff2() {
        return this._assetTilloff2;
    }
    protected _assetTilloff3: IMixGroundTextureTilloff;
    set assetTilloff3(val: IMixGroundTextureTilloff) {
        this._assetTilloff3 = val;
        p3d.p3d_material_uniform_vec4(Engine.commands(), this._entity, ExtensionTools.getAtom(MixGroungMaterial.ASSET_TILLOFF_3), val.uScale, val.vScale, val.uOffset, val.vOffset);
    }
    get assetTilloff3() {
        return this._assetTilloff3;
    }
    protected _assetTilloff4: IMixGroundTextureTilloff;
    set assetTilloff4(val: IMixGroundTextureTilloff) {
        this._assetTilloff4 = val;
        p3d.p3d_material_uniform_vec4(Engine.commands(), this._entity, ExtensionTools.getAtom(MixGroungMaterial.ASSET_TILLOFF_4), val.uScale, val.vScale, val.uOffset, val.vOffset);
    }
    get assetTilloff4() {
        return this._assetTilloff4;
    }
    protected _mixTilloff: IMixGroundTextureTilloff;
    set mixTilloff(val: IMixGroundTextureTilloff) {
        this._mixTilloff = val;
        p3d.p3d_material_uniform_vec4(Engine.commands(), this._entity, ExtensionTools.getAtom(MixGroungMaterial.MIX_TILLOFF), val.uScale, val.vScale, val.uOffset, val.vOffset);
    }
    get mixTilloff() {
        return this._mixTilloff;
    }
    protected _sceneHSize: number;
    set sceneHSize(val: number) {
        this._sceneHSize = val;
    }
    get sceneHSize() {
        return this._sceneHSize;
    }
    protected _sceneVSize: number;
    set sceneVSize(val: number) {
        this._sceneVSize = val;
    }
    get sceneVSize() {
        return this._sceneVSize;
    }
    protected _gridEdgeLineBaseSize: number = 0.3;
    set gridEdgeLineBaseSize(val: number) {
        this._gridEdgeLineBaseSize = val;
        this.updateGridParam();
    }
    get gridEdgeLineBaseSize() {
        return this._gridEdgeLineBaseSize;
    }
    protected _gridEdgeLineDeltaSize: number = 0.05;
    set gridEdgeLineDeltaSize(val: number) {
        this._gridEdgeLineDeltaSize = val;
        this.updateGridParam();
    }
    get gridEdgeLineDeltaSize() {
        return this._gridEdgeLineDeltaSize;
    }
    protected _gridEdgeLineStrength: number = 0.8;
    set gridEdgeLineStrength(val: number) {
        this._gridEdgeLineStrength = val;
        this.updateGridParam();
    }
    get gridEdgeLineStrength() {
        return this._gridEdgeLineStrength;
    }
    protected _gridEdgeLightStrength: number = 0.8;
    set gridEdgeLightStrength(val: number) {
        this._gridEdgeLightStrength = val;
        this.updateGridParam();
    }
    get gridEdgeLightStrength() {
        return this._gridEdgeLightStrength;
    }

    protected _gridEdgeLightRangeSize: number = 0.05;
    set gridEdgeLightRangeSize(val: number) {
        this._gridEdgeLightRangeSize = val;
        this.updateSceneParam();
    }
    get gridEdgeLightRangeSize() {
        return this._gridEdgeLightRangeSize;
    }

    protected updateSceneParam = () => {
        p3d.p3d_material_uniform_vec4(Engine.commands(), this._entity, ExtensionTools.getAtom(MixGroungMaterial.SCENE_PARAM),
            this._sceneHSize,
            this._sceneVSize,
            (Date.now() - this._time) / 1000,
            this._gridEdgeLightRangeSize,
        );
    }
    protected updateGridParam() {
        p3d.p3d_material_uniform_vec4(Engine.commands(), this._entity, ExtensionTools.getAtom(MixGroungMaterial.GRID_PARAM),
            this._gridEdgeLineBaseSize,
            this._gridEdgeLineDeltaSize,
            this._gridEdgeLineStrength,
            this._gridEdgeLightStrength,
        );
    }

    protected _time: number = Date.now();

    name?: string;
    constructor(
        scene: SceneStruct,
        materialName: string,
        _pass: Constants.PassTag
    ) {
        super(materialName, scene.impl, MixGroungMaterial.KEY, _pass);
        scene.impl.onBeforeRenderObservable.add(this.updateSceneParam);
    }

    _onFrameUpdate(): void {
        if (this._isDisposed) { return; }

        if (this._assetTex0 && this._assetTex0Dirty) {
            UniformTexture.texture(this.entity, MixGroungMaterial.ASSET_TEX0, this._assetTex0);
        }
        if (this._assetTex1 && this._assetTex1Dirty) {
            UniformTexture.texture(this.entity, MixGroungMaterial.ASSET_TEX1, this._assetTex1);
        }
        if (this._assetTex2 && this._assetTex2Dirty) {
            UniformTexture.texture(this.entity, MixGroungMaterial.ASSET_TEX2, this._assetTex2);
        }
        if (this._assetTex3 && this._assetTex3Dirty) {
            UniformTexture.texture(this.entity, MixGroungMaterial.ASSET_TEX3, this._assetTex3);
        }
        if (this._assetTex4 && this._assetTex4Dirty) {
            UniformTexture.texture(this.entity, MixGroungMaterial.ASSET_TEX4, this._assetTex4);
        }
        if (this._mixTex && this._mixTexDirty) {
            UniformTexture.texture(this.entity, MixGroungMaterial.MIX_TEX, this._mixTex);
        }
        if (this._lightTex && this._lightTexDirty) {
            UniformTexture.texture(this.entity, MixGroungMaterial.LIGHT_TEX, this._lightTex);
        }
        if (this._gridTex && this._gridTexDirty) {
            UniformTexture.texture(this.entity, MixGroungMaterial.GRID_TEX, this._gridTex);
        }

        this._assetTex0Dirty = false;
        this._assetTex1Dirty = false;
        this._assetTex2Dirty = false;
        this._assetTex3Dirty = false;
        this._assetTex4Dirty = false;
        this._mixTexDirty = false;
        this._lightTexDirty = false;
        this._gridTexDirty = false;
    }

    public static regist() {
        if (Engine.CheckShader(MixGroungMaterial.KEY)) {
            return;
        }
        let uniforms: p3d.MaterialUniformDefines = p3d.MaterialUniformDefines.create();
        let includes: p3d.NodematerialIncludes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
        p3d.p3d_shader_uniform_tex(uniforms, MixGroungMaterial.ASSET_TEX0, true, Constants.SHADER_STAGE_FRAGMENT, Constants.DEFAULT_TEXTURE_WHITE, Constants.TextureViewDimensionD2);
        p3d.p3d_shader_uniform_tex(uniforms, MixGroungMaterial.ASSET_TEX1, true, Constants.SHADER_STAGE_FRAGMENT, Constants.DEFAULT_TEXTURE_WHITE, Constants.TextureViewDimensionD2);
        p3d.p3d_shader_uniform_tex(uniforms, MixGroungMaterial.ASSET_TEX2, true, Constants.SHADER_STAGE_FRAGMENT, Constants.DEFAULT_TEXTURE_WHITE, Constants.TextureViewDimensionD2);
        p3d.p3d_shader_uniform_tex(uniforms, MixGroungMaterial.ASSET_TEX3, true, Constants.SHADER_STAGE_FRAGMENT, Constants.DEFAULT_TEXTURE_WHITE, Constants.TextureViewDimensionD2);
        p3d.p3d_shader_uniform_tex(uniforms, MixGroungMaterial.ASSET_TEX4, true, Constants.SHADER_STAGE_FRAGMENT, Constants.DEFAULT_TEXTURE_WHITE, Constants.TextureViewDimensionD2);
        p3d.p3d_shader_uniform_tex(uniforms, MixGroungMaterial.MIX_TEX, true, Constants.SHADER_STAGE_FRAGMENT, Constants.DEFAULT_TEXTURE_BLACK, Constants.TextureViewDimensionD2);
        p3d.p3d_shader_uniform_tex(uniforms, MixGroungMaterial.GRID_TEX, true, Constants.SHADER_STAGE_FRAGMENT, Constants.DEFAULT_TEXTURE_WHITE, Constants.TextureViewDimensionD2);
        p3d.p3d_shader_uniform_tex(uniforms, MixGroungMaterial.LIGHT_TEX, true, Constants.SHADER_STAGE_FRAGMENT, Constants.DEFAULT_TEXTURE_WHITE, Constants.TextureViewDimensionD2);
        p3d.p3d_shader_uniform_vec4(uniforms, MixGroungMaterial.MIX_TILLOFF, 1, 1, 0, 0, false);
        p3d.p3d_shader_uniform_vec4(uniforms, MixGroungMaterial.ASSET_TILLOFF_0, 1, 1, 0, 0, false);
        p3d.p3d_shader_uniform_vec4(uniforms, MixGroungMaterial.ASSET_TILLOFF_1, 1, 1, 0, 0, false);
        p3d.p3d_shader_uniform_vec4(uniforms, MixGroungMaterial.ASSET_TILLOFF_2, 1, 1, 0, 0, false);
        p3d.p3d_shader_uniform_vec4(uniforms, MixGroungMaterial.ASSET_TILLOFF_3, 1, 1, 0, 0, false);
        p3d.p3d_shader_uniform_vec4(uniforms, MixGroungMaterial.ASSET_TILLOFF_4, 1, 1, 0, 0, false);
        p3d.p3d_shader_uniform_vec4(uniforms, MixGroungMaterial.LIGHT_PARAM, 1, 1, 1, 1, false);
        p3d.p3d_shader_uniform_vec4(uniforms, MixGroungMaterial.GRID_PARAM, 0.3, 0.05, 0.1, 0.8, false);
        p3d.p3d_shader_uniform_vec4(uniforms, MixGroungMaterial.SCENE_PARAM, 1, 1, 0, 0.1, false);

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec2", "vUV"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });

        let binds_defines_base = Constants.BIND_MODEL | Constants.BIND_VIEWER | Constants.BIND_EFFECT_VALUE;

        Engine.RegistShader(
            MixGroungMaterial.KEY,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            #define PI 3.141592653589793
            `,
            vertexShader(),
            fragmentShader(),
            includes,
            ``,
            varyings,
            binds_defines_base
        );
    }
}

function vertexShader() {
    return `
    mat4 finalWorld = PI_ObjectToWorld;
    vec4 position =  vec4(A_POSITION, 1.);
    vec4 worldPos =  finalWorld * position;
    // vec4 worldPos =  position;
    gl_Position = PI_MATRIX_VP * worldPos;
    vUV         = A_UV;
`;
}
function fragmentShader() {
    return `
    vec2 mainUV = vUV;
    vec4 diffuseColor = vec4(1., 1., 1., 1.);

    vec4 mixColor = texture(sampler2D(${MixGroungMaterial.MIX_TEX}, sampler${MixGroungMaterial.MIX_TEX}), mainUV * ${MixGroungMaterial.MIX_TILLOFF}.zw + ${MixGroungMaterial.MIX_TILLOFF}.xy);

    vec4 asset0 = texture(sampler2D(${MixGroungMaterial.ASSET_TEX0}, sampler${MixGroungMaterial.ASSET_TEX0}), mainUV * (${MixGroungMaterial.ASSET_TILLOFF_0}.zw ) + ${MixGroungMaterial.ASSET_TILLOFF_0}.xy);
    vec4 asset1 = texture(sampler2D(${MixGroungMaterial.ASSET_TEX1}, sampler${MixGroungMaterial.ASSET_TEX1}), mainUV * (${MixGroungMaterial.ASSET_TILLOFF_1}.zw ) + ${MixGroungMaterial.ASSET_TILLOFF_1}.xy);
    vec4 asset2 = texture(sampler2D(${MixGroungMaterial.ASSET_TEX2}, sampler${MixGroungMaterial.ASSET_TEX2}), mainUV * (${MixGroungMaterial.ASSET_TILLOFF_2}.zw ) + ${MixGroungMaterial.ASSET_TILLOFF_2}.xy);
    vec4 asset3 = texture(sampler2D(${MixGroungMaterial.ASSET_TEX3}, sampler${MixGroungMaterial.ASSET_TEX3}), mainUV * (${MixGroungMaterial.ASSET_TILLOFF_3}.zw ) + ${MixGroungMaterial.ASSET_TILLOFF_3}.xy);
    vec4 asset4 = texture(sampler2D(${MixGroungMaterial.ASSET_TEX4}, sampler${MixGroungMaterial.ASSET_TEX4}), mainUV * (${MixGroungMaterial.ASSET_TILLOFF_4}.zw ) + ${MixGroungMaterial.ASSET_TILLOFF_4}.xy);

    vec4 lightColor = texture(sampler2D(${MixGroungMaterial.LIGHT_TEX}, sampler${MixGroungMaterial.LIGHT_TEX}), mainUV) * ${MixGroungMaterial.LIGHT_PARAM}.w;

    diffuseColor.rgb = asset0.rgb;
    diffuseColor.rgb = mix(diffuseColor.rgb, asset1.rgb, mixColor.r);
    diffuseColor.rgb = mix(diffuseColor.rgb, asset2.rgb, mixColor.g);
    diffuseColor.rgb = mix(diffuseColor.rgb, asset3.rgb, mixColor.b);
    diffuseColor.rgb = mix(diffuseColor.rgb, asset4.rgb, 1.0 - mixColor.a);

    float lightStrength = 1.; //GetLuminance(diffuseColor.rgb);
    // lightStrength = pow(lightStrength, 0.5);
    // grid
    vec2 grid = fract(mainUV * ${MixGroungMaterial.SCENE_PARAM}.xy);
    vec2 gridRowCol = floor(mainUV * ${MixGroungMaterial.SCENE_PARAM}.xy);

    float random = GetLuminance(diffuseColor.rgb) + cos(54981. * gridRowCol.x * gridRowCol.y);

    float lineDiff = ${MixGroungMaterial.SCENE_PARAM}.w;
    // 网格线
    float gridLine = smoothstep(0.5 - lineDiff, 0.5, abs(grid.x - 0.5)) + smoothstep(0.5 - lineDiff, 0.5, abs(grid.y - 0.5));
    // 网格线 由内向外增强
    gridLine = smoothstep(0.5, 1., gridLine);

    vec4 gridData = texture(sampler2D(${MixGroungMaterial.GRID_TEX}, sampler${MixGroungMaterial.GRID_TEX}), mainUV);
    float gridLum = gridData.r + gridData.g + gridData.b;
    // 为不同颜色实现时间差异,以实现呼吸差异
    float gridFlag = gridData.a * 255.;

    float deltaLightStrength = 0.4 + smoothstep(-0.5, 1.5, random) * 0.6;
    // 网格线 左&下 变暗, 右&上 变亮 模拟右上向左下的光照下的瓷砖边界凹槽
    float gridEdgeLightStrength = 1.0 - gridLine * (mix(${MixGroungMaterial.GRID_PARAM}.w * deltaLightStrength, 0.-${MixGroungMaterial.GRID_PARAM}.w * deltaLightStrength, smoothstep(0.9, 1.1, grid.x + grid.y)));
    lightStrength = lightStrength * gridEdgeLightStrength;

    // 这里控制网格线是否显示
    diffuseColor.rgb = diffuseColor.rgb * lightStrength;

    float randomEdge = gridData.r * 1. + gridData.g * 2. + gridData.b * 4.;
    float edge = ${MixGroungMaterial.GRID_PARAM}.x + ${MixGroungMaterial.GRID_PARAM}.y * cos((${MixGroungMaterial.SCENE_PARAM}.z + randomEdge) * PI * 2. * 0.2);
    float leftV = (1.0 - smoothstep(0.05, edge, grid.x));
    float bottomV = (1.0 - smoothstep(0.05, edge, grid.y));
    float rightV = smoothstep(1.0 - edge, 1.0 - 0.05, grid.x);
    float topV = smoothstep(1.0 - edge, 1.0 - 0.05, grid.y);

    float gridLeftTop = step(${GridInfo.LEFT_TOP}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.LEFT_TOP}. * gridLeftTop;
    float gridLeftBottom = step(${GridInfo.LEFT_BOTTOM}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.LEFT_BOTTOM}. * gridLeftBottom;
    float gridRightBottom = step(${GridInfo.RIGHT_BOTTOM}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.RIGHT_BOTTOM}. * gridRightBottom;
    float gridRightTop = step(${GridInfo.RIGHT_TOP}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.RIGHT_TOP}. * gridRightTop;
    float gridLeft = step(${GridInfo.LEFT}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.LEFT}. * gridLeft;
    float gridBottom = step(${GridInfo.BOTTOM}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.BOTTOM}. * gridBottom;
    float gridRight = step(${GridInfo.RIGHT}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.RIGHT}. * gridRight;
    float gridTop = step(${GridInfo.TOP}., gridFlag);
    float gridEdge = max( mix(0., bottomV, gridTop),
        max( mix(0., rightV, gridRight),
            max( mix(0., leftV, gridLeft),
                max( mix(0., topV, gridBottom),
                    max( mix(0., leftV * bottomV, gridLeftTop),
                        max( mix(0., leftV * topV, gridLeftBottom),
                            max( mix(0., rightV * topV, gridRightBottom),
                                mix(0., rightV * bottomV, gridRightTop)
                            )
                        )
                    )
                )
            )
        )
    );
    gridEdge = gridEdge * step(0.01, gridLum);
    // gridEdge = smoothstep(0.0, 0.8, gridEdge) - step(0.95, gridEdge);
    gridEdge = gridEdge * ${MixGroungMaterial.GRID_PARAM}.z;
    // gridEdge = smoothstep(0.0, ${MixGroungMaterial.GRID_PARAM}.z, gridEdge) - step(${MixGroungMaterial.GRID_PARAM}.z, gridEdge);

    // diffuseColor.rgb = gridData.rgb; //
    diffuseColor.rgb = mix(diffuseColor.rgb, gridData.rgb, gridEdge);

    // diffuseColor.rgb = diffuseColor.rgb * lightColor.rgb * smoothstep(0., 1.0 + deltaLightStrength, 1.0 + deltaLightStrength); // * (1.0 + gridEdge * 2.);
    diffuseColor.rgb = diffuseColor.rgb * lightColor.rgb; // * (1.0 + gridEdge * 2.);

    gl_FragColor = diffuseColor;
`;
}