
import { NodeTools, LoaderTool } from "./scene_tool";
import { IModelChildOpt, IModelAttachOpt, IModelOpt, IImageSolt, IMaterialSolt, IMaterialReplace, IModelObj, IPrefabAnime, IMeshMaterial, IMeshRenderQueue, INodeCollider, Res, ResTab, IModel, IAnimationGroup } from "./interface/index";

import { MeshParticleSystem } from "./mesh_particle/mesh_particle_system";
import { BABYLON } from "./render3d/babylon";
import { ExtensionTools } from "./base";
import { PiCollider } from "./scene_pick";
import { CustomScene } from "./index";
import { N3Pool } from "./render/engines/datapool";


declare let module;
let logger;
ExtensionTools.RegistStartUp(() => { logger = ExtensionTools.Logger(module.name, "ModelObj"); });

export class Model implements IModel {
    private _extendData: any;
    set extendData(val: any) {
        this._extendData = val;
    }
    get extendData() {
        return this._extendData;
    }
    rootImpl: BABYLON.ITransformNodeD3;
    isDisposed: boolean;
    protected _isDisposeOver: boolean = false;
    public isLoaded: boolean;
    protected _position: [number, number, number];
    protected _scale: [number, number, number];
    protected _rotate: [number, number, number];
    protected _lookAt: [number, number, number];
    protected _color: [number, number, number];
    protected _alpha: number;
    public get position(): [number, number, number] {
        return this._position;
    }
    public get scale(): [number, number, number] {
        return this._scale;
    }
    public get rotate(): [number, number, number] {
        return this._rotate;;
    }
    public get lookAt(): [number, number, number] {
        return this._lookAt;
    }
    public get color(): [number, number, number] {
        return this._color;
    }
    public get alpha(): number {
        return this._alpha;
    }
    public animOpt: IPrefabAnime;

    protected _animationMap?: Map<string, BABYLON.AnimationGroup> = new Map();
    public get animationMap() {
        return this._animationMap;
    }

    constructor(
        protected _name: string,
        public readonly scene: CustomScene,
    ) {

    }
    setAnimeFrame(animName: string, progress: number): IModel {
        // throw new Error("Method not implemented.");
        return this;
    }
    protected _loadPromise?: Promise<any>

    public get name() {
        return this._name;
    }

    public get loadPromise() {
        return this._loadPromise;
    }

    protected _isEnabled: boolean;
    public isEnabled(): boolean {
        if (this._isEnabled === undefined) {
            return true;
        } else {
            return this._isEnabled;
        };
    }
    protected _rayID: number;
    public get rayID() {
        return this._rayID;
    }
    public set rayID(val: any) {
        this._rayID = val;
        this._applyRayID();
    }
    protected _applyRayID() {
        if (this.isLoaded && !this.isDisposed) {
            this.rootImpl.rayID = this.rayID;
            let nodes = [this.rootImpl];
            while (nodes.length > 0) {
                let temp = [];
                nodes.forEach((node) => {
                    node.rayID = this.rayID;
                    node.getChildren().forEach((child) => {
                        temp.push(child);
                    });
                });
                nodes = temp;
            }
        }
    }
    protected _state: any;
    public modifyState(state: any) {
        this._state = state;
        this._modifyState();
    }
    protected _modifyState() {
        if (this.isLoaded && this.isDisposed == false) {
            //
        }
    }

    public setPostion(data: number[]) {
        this.setPosition(...<[number, number, number]>data);
    }
    public setPosition(x: number | number[], y?: number, z?: number) {
        N3Pool.recycle(this._position);
        if (x instanceof Array) {
            this._position = N3Pool.request(x[0], x[1], x[2]);
        } else {
            this._position = N3Pool.request(x, y, z);
        }
        this._changePosition();
    }
    public setScaling(x: number | number[], y?: number, z?: number) {
        N3Pool.recycle(this._scale);
        if (x instanceof Array) {
            this._scale = N3Pool.request(x[0], x[1], x[2]);
        } else {
            this._scale = N3Pool.request(x, y, z);
        }
        this._changeScale();
    }
    public setScale(x: number | number[], y?: number, z?: number) {
        this.setScaling(x, y, z);
    }
    public setRotation(x: number | number[], y?: number, z?: number) {
        N3Pool.recycle(this._rotate);
        if (x instanceof Array) {
            this._rotate = N3Pool.request(x[0], x[1], x[2]);
        } else {
            this._rotate = N3Pool.request(x, y, z);
        }
        this._changeRotate();
    }
    public setRotate(x: number | number[], y?: number, z?: number) {
        this.setRotation(x, y, z);
    }
    public setLookAt(x: number | number[], y?: number, z?: number) {
        N3Pool.recycle(this._lookAt);
        if (x instanceof Array) {
            this._lookAt = N3Pool.request(x[0], x[1], x[2]);
        } else {
            this._lookAt = N3Pool.request(x, y, z);
        }
        this._changeLookAt();
    }

    public setEnabled(isEnabled: boolean) {
        this._isEnabled = isEnabled;
        this.rootImpl?.setEnabled(isEnabled);
    }
    setColor(r: number, g: number, b: number): void {
        N3Pool.recycle(this._color);
        this._color = N3Pool.request(r, g, b);
    }
    setAlpha(val: number): void {
        this._alpha = val;
    }
    dispose(disposeMaterialAndTextures?: boolean): void {
        this._pointerUps.length = 0;
        this._pointerDowns.length = 0;
        this.rootImpl?.dispose();
    }
    addAnimationEventListen(key: string, func: () => boolean): void {
        // throw new Error("Method not implemented.");
    }
    pickAboutItems(): INodeCollider[] {
        // throw new Error("Method not implemented.");
        return [];
    }
    setAnim(animOpt: IPrefabAnime): void {
        // throw new Error("Method not implemented.");
    }
    setAnimeSpeed(param: number): void {
        // throw new Error("Method not implemented.");
    }
    setParent(parent: IModel, indexOfParent?: number): void {
        this.rootImpl.parent = <BABYLON.TransformNode>parent?.rootImpl;
    }
    reuse() {

    }

    protected _animations?: BABYLON.AnimationGroup[];
    public get animations() {
        return this._animations;
    }

    protected _renderingGroupId: number;
    protected _alphaIndex: number;
    protected _lastuseAlphaIndex: number;
    set alphaIndex(param: number) {
        this._alphaIndex = param;
    }
    public get alphaIndex() {
        // debugger;
        return this._alphaIndex;
    }
    public get renderingGroupId() {
        return this._renderingGroupId;
    }
    public set renderingGroupId(param: number) {
        // debugger;
        this._renderingGroupId = param;
    }

    protected __cached: boolean;
    public get _cached() {
        return this.__cached;
    }
    public set _cached(val: boolean) {
        if (this.__cached == undefined) {
            this.__cached = val;
        }
    }

    public unuse(): void {
        this.setEnabled(false);
        this._modifyUseState();
        this._pointerUps.length = 0;
        this._pointerDowns.length = 0;
    }
    protected _modifyUseState(): void {
        if (this._isEnabled == false) {
            this.animationMap?.forEach((group) => {
                group.onAnimationGroupEndObservable.clear();
                group.onAnimationGroupPlayObservable.clear();
                group.onAnimationGroupLoopObservable.clear();
                // group.onAnimationGroupFrameEventObservable.clear();
                group.stop();
            });
        }
    }

    public renderQueueItems(): IMeshRenderQueue[] {
        let result: IMeshRenderQueue[] = [];
        return result;
    }

    public renderMaterialItems(): IMeshMaterial[] {
        let result: IMeshMaterial[] = [];
        return result;
    }

    /**
     * 停止当前的动画 - （模型）一次只播一个动画
     */
    public stopAnim() {

    }
    /**
     * 暂停当前的动画 - （模型）一次只播一个动画
     */
    public pauseAnim() {

    }

    protected _changePosition() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this._position) {
            NodeTools.positNode(<BABYLON.Transform>this.rootImpl, this._position);
        }
    }
    protected _changeScale() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this._scale) {
            NodeTools.scaleNode(<BABYLON.Transform>this.rootImpl, this._scale);
        }
    }
    protected _changeRotate() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this._rotate) {
            NodeTools.rotateNode(<BABYLON.Transform>this.rootImpl, this._rotate);
        }
    }
    protected _changeLookAt() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this._lookAt) {
            NodeTools.nodeLookAt(<BABYLON.Transform>this.rootImpl, this._lookAt);
        }
    }
    
    getChild(index?: number): BABYLON.Transform {
        if (index !== undefined) {
            return <BABYLON.Transform>this.rootImpl.getChildren()[index]
        }
        return <BABYLON.Transform>this.rootImpl;
    }
    protected _pointerUps: Function[] = [];
    protected _pointerDowns: Function[] = [];
    addEventListener(etype: "pointerup" | "pointerdown", func: Function) {
        if (this.isDisposed) return;
        this._isPickable = true;
        if (etype == "pointerup") {
            let idx = this._pointerUps.indexOf(func);
            if (idx < 0) {
                this._pointerUps.push(func);
            }
        } else {
            let idx = this._pointerDowns.indexOf(func);
            if (idx < 0) {
                this._pointerDowns.push(func);
            }
        }
    }
    removeEventListener(etype: "pointerup" | "pointerdown", func: Function) {
        if (this.isDisposed) return;
        if (etype == "pointerup") {
            let idx = this._pointerUps.indexOf(func);
            if (idx >= 0) {
                this._pointerUps.splice(idx, 1);
            }
        } else {
            let idx = this._pointerDowns.indexOf(func);
            if (idx < 0) {
                this._pointerDowns.splice(idx, 1);
            }
        }
    }
    colliderSphereTreshold(val: number) {
        if (this._collider == undefined) {
            this._collider = new PiCollider(new BABYLON.Vector3(-0.5, -0.5, -0.5), new BABYLON.Vector3( 0.5,  0.5,  0.5));
        }
        this._collider.sphereTreshold = val;
        this._updateCollider();
    }
    protected _collider: PiCollider;
    protected _isPickable: boolean;
    protected _onPointerAttached: boolean = false;
    protected _getColliderNode() {
        let node: BABYLON.Node = <any>this.rootImpl;
        return node;
    }
    collider(min: BABYLON.Vector3, max: BABYLON.Vector3) {
        if (this.isDisposed) return;
        this._isPickable = true;
        if (this._collider == undefined) {
            this._collider = new PiCollider(min, max);
        } else {
            this._collider.minimum.copyFrom(min);
            this._collider.maximum.copyFrom(max);
        }
        this._updateCollider();
    }
    protected _updateCollider() {
        if (this.isDisposed) return;
        if (this._isPickable) {
            let node: BABYLON.Node = this._getColliderNode();
            if (this._onPointerAttached == false) {
                node.onPointerObservable.add(this._onPointer);
                this._onPointerAttached = true;
            }
            node?.setCollider(this._collider)
        }
    }
    protected _onPointer = (info: BABYLON.PointerInfo, eventState: BABYLON.EventState) => {
        if (this.isDisposed) return;
        if (info.type == BABYLON.PointerEventTypes.POINTERDOWN) {
            let result = this._onPointerDown();
            eventState.skipNextObservers = result;
        } else if (info.type == BABYLON.PointerEventTypes.POINTERUP) {
            let result = this._onPointerUp();
            eventState.skipNextObservers = result;
        }
    }
    protected _onPointerUp = (): boolean => {
        if (this.isDisposed) return false;
        let result = true;
        this._pointerUps.forEach(function(func) {
            try {
                if (func() === false) {
                    result = false;
                }
            } catch (error) {
                
            }
        });
        return result;
    }
    protected _onPointerDown = (): boolean => {
        if (this.isDisposed) return false;
        let result = true;
        this._pointerDowns.forEach(function(func) {
            try {
                if (func() === false) {
                    result = false;
                }
            } catch (error) {
                
            }
        });
        return result;
    }
}


export class ModelObj extends Model {
    public static ListenCounter = 0;
    public isDisposed: boolean = false;
    protected _isDisposeOver: boolean = false;
    /**
     * 导入的动画是否默认播放
     */
    public animDefault: boolean;
    public particleAutoStart: boolean;
    public gltfModifyConfig: {
        imageSolts?: string[] | IImageSolt[];
        materialSolts?: string[] | IMaterialSolt[];
        materialReplace?: IMaterialReplace[];
    } = {};
    /**
     * 切换动画时是否平滑过渡
     */
    public changeAnimEnableBlending: boolean = false;
    
    /**
     * 切换动画时过渡 需要的时间 单位:秒
     */
    public changeAnimBlendUseSeconds: number = (1 / 0.01) * (1000 / 60) / 1000;

    /**
     * 切换动画时过渡速度 默认 0.01 - 每帧进度
     * @deprecated
     */
    public set changeAnimBlendSpeed(val: number) {
        this.changeAnimBlendUseSeconds = (1 / val) * (1000 / 60) / 1000;
    }
    /**
     * @deprecated
     */
    public workWithAnimationGroupChange: boolean = false;
    public needDepthPrePass: boolean = false;
    /**
     * 是否具有骨骼动画优化
     */
    public skeletonAnimOptimize: boolean = false;
    /**
     * 动画切换前是否重置各属性为无动画状态
     */
    public initialWhileAnimChange: boolean = false;

    /**
     * 父节点
     */
    // public parent: ModelObj;
    // protected _parent: ModelObj;
    // public get parent() {
    //     return this._parent;
    // }
    // public readonly path: string;
    protected _path: string;
    public get path() {
        return this._path;
    }
    // public readonly fileName: string;
    protected _fileName: string;
    public get fileName() {
        return this._fileName;
    }
    // public readonly modelName: string | string[];
    protected _modelName: string | string[];
    public get modelName() {
        return this._modelName;
    }
    public insertedCall: (model: ModelObj) => any;
    /**
     * 可拓展的实例化物件创建方法
     */
    // public createInstancedMesh: (instanceID: any, scene: CustomScene) => BABYLON.InstancedMesh;

    private _meshMap: Map<string, BABYLON.Mesh> = new Map();
    public get meshMap() {
        return this._meshMap;
        // throw new Error("Method not implemented.");
    }

    protected _skeletonMap: Map<string, BABYLON.Skeleton> = new Map();
    public get skeletonMap() {
        return this._skeletonMap;
        // throw new Error("Method not implemented.");
    }
    // public readonly particleSysMap: Map<string, MeshParticleSystem> = new Map();
    public get particleSysList() {
        return this._particleSystems;
    }
    public get attachOptList() {
        throw new Error("Method not implemented.");
    }
    public get childOptList() {
        throw new Error("Method not implemented.");
    }
    public get particleSysMap() {
        return this._particleSystems;
    }
    /**
     * 模型当前动画
     */
    protected _animation: BABYLON.AnimationGroupRS;
    public get animation() {
        return this._animation;
    }
    /**
     * 模型当前动画 对应 骨头动画
     */
    // protected animationForBone: BABYLON.AnimationGroup;
    /**
     * 当前动画是否循环播放
     */
    protected isAnimationLoop: boolean = false;
    public get loadPromise() {
        return this._loadPromise;
    }
    protected _animEventMap: Map<string, () => boolean> = new Map();
    protected get animEventMap() {
        // throw new Error("Method not implemented.");
        return this._animEventMap;
    }
    /**
     * 加载时准备的克隆数据
     */
    // protected cloneData: ICloneData;
    protected get cloneData() {
        throw new Error("Method not implemented.");
    }
    /**
     * 骨骼缓存文件路径
     */
    // protected sklaiPath: string;
    protected get sklaiPath() {
        throw new Error("Method not implemented.");
    }
    // protected sklaiRes: Res<any, ArrayBuffer>;
    protected get sklaiRes() {
        throw new Error("Method not implemented.");
    }
    /**
     * 克隆标识
     */
    // protected isClone: boolean = false;
    protected get isClone() {
        throw new Error("Method not implemented.");
    }
    /**
     * 模型创建时的配置
     */
    protected option: IModelOpt;
    protected cloneInfo: { ref: number } = { ref: 0 };
    public get isReady() {
        // throw new Error("Method not implemented.");
        return this.rootImpl; // && this.rootImpl.isReady;
    }
    constructor(meshName: string, scene: CustomScene, opt: IModelOpt = {}) {
        super(meshName, scene);

        // this.scene = scene;

        this.scene._modelMap.set(this, this);

        this._name = meshName;
        // this._parent = opt.parent;
        this._rayID = opt.rayID;
        this._path = opt.path || '';
        this._fileName = opt.fileName;
        this._modelName = opt.modelName;
        this.insertedCall = opt.insertedCall;
        this.animDefault = true; // !!opt.animDefault;
        this.particleAutoStart = !!opt.particleAutoStart;
        this._isEnabled = false;

        this.gltfModifyConfig.imageSolts = opt.imageSolts;
        this.gltfModifyConfig.materialSolts = opt.materialSolts;
        this.gltfModifyConfig.materialReplace = opt.materialReplace;

        this.needDepthPrePass = !!opt.needDepthPrePass;
        this.workWithAnimationGroupChange = opt.workWithAnimationGroupChange;
        this.skeletonAnimOptimize = !!opt.skeletonAnimOptimize;
        this.initialWhileAnimChange = !!opt.initialWhileAnimChange;
        this.option = opt;

        if (this.fileName) {
            const arr = [];
            arr.push(
                LoaderTool.insertMeshAsync(this, opt.absolutePath, opt.absoluteFileName).then((res) => {
                    if (this.isDisposed) { throw new Error(`Disposed !!`); }
                    this._gltf = res.gltf;
                    this.loaded(res._root_, res.nodes, res.meshes, res.particles, res.materials, res.animationGroups, res.skeletons);
                    return this;
                }).catch((e) => {
                    console.error(e);
                })
            );
            // arr.push(this.loadSklai());

            this._loadPromise = Promise.all(arr).then((value: [void, [string, Res<any, ArrayBuffer>] | undefined]) => {
                if (this.isDisposed) {
                    this.dispose();
                } else if (this.insertedCall) {
                    this.insertedCall(this);
                }
                return this;
            });

            // this._loadPromise = new Promise((resolve, reject) => {

            // });
        }
    }
    public static Creation(meshName: string, scene: CustomScene, opt: IModelOpt = {}) {
        return new ModelObj(meshName, scene, opt);
    }
    protected _transformNodes?: BABYLON.ITransformNode[];
    protected _meshes?: BABYLON.AbstractMesh[];
    protected _materials?: BABYLON.Material[];
    protected _particleSystems?: BABYLON.MeshParticleSystem[];
    protected _gltf?: BABYLON.IGLTFRes;
    public get transformNodes() {
        return this._transformNodes;
    }
    public get meshes() {
        return this._meshes;
    }
    public get materials() {
        return this._materials;
    }
    public get animations() {
        return this._animations;
    }
    public get particleSystems() {
        return this._particleSystems;
    }
    public get _cached() {
        return this.__cached;
    }
    public set _cached(val: boolean) {
        if (this.__cached === undefined) {
            this.__cached = val;
        }
    }
    protected frameListen: (data: BABYLON.IAnimationEventDataFrame) => void;
    /**
     * 添加动画事件监听
     * @deprecated use animationGroup.onAnimationGroupFrameEventObserva
     * @param key 动画事件关键字
     * @param func 监听 - 返回值为 true 表示该监听仅生效一次
     */
    public addAnimationEventListen(key: string, func: () => boolean) {
        // throw new Error("Method not implemented.");
        this._animEventMap.set(key, func);
    }
    private _animationEventCall = (data: any) => {
        let call = this._animEventMap?.get(data);
        if (call) {
            if (call()) {
                this._animEventMap.delete(data);
            }
        }
    }
    /**
     * 返回 Promise
     * @param opt
     */
    public loadAsync(opt: IModelOpt) {
        if (!this.fileName) {
            // this._parent = opt.parent;
            this._rayID = opt.rayID;
            this._path = opt.path || '';
            this._fileName = opt.fileName;
            this._modelName = opt.modelName;
            // this.insertedCall = opt.insertedCall;
            this.animDefault = true; // !!opt.animDefault;
            this.particleAutoStart = !!opt.particleAutoStart;
            // this._isEnabled = true;

            this.gltfModifyConfig.imageSolts = opt.imageSolts;
            this.gltfModifyConfig.materialSolts = opt.materialSolts;
            this.gltfModifyConfig.materialReplace = opt.materialReplace;

            this.needDepthPrePass = !!opt.needDepthPrePass;
            this.workWithAnimationGroupChange = opt.workWithAnimationGroupChange;
            this.skeletonAnimOptimize = !!opt.skeletonAnimOptimize;
            this.initialWhileAnimChange = !!opt.initialWhileAnimChange;
            this.option = opt;

            const arr = [];
            arr.push(
                LoaderTool.insertMeshAsync(this, opt.absolutePath, opt.absoluteFileName).then((res) => {
                    this.loaded(res._root_, res.nodes, res.meshes, res.particles, res.materials, res.animationGroups, res.skeletons);
                    return this;
                }).catch((e) => {
                    console.error(e);
                })
            );
            // arr.push(this.loadSklai());

            this._loadPromise = Promise.all(arr).then((value: [void, [string, Res<any, ArrayBuffer>] | undefined]) => {
                if (this.isDisposed) {
                    this.dispose();
                } else if (this.insertedCall) {
                    this.insertedCall(this);
                }
                return this;
            }).catch(e => {
                console.error(e);
                this.dispose();
                return this;
            });
        }
        else {
            this._loadPromise = Promise.reject(`已在创建时进行加载`);
        }

        return this._loadPromise;
    }
    protected loadSklai(): Promise<[string, Res<any, ArrayBuffer>] | undefined> {
        // if (!this.skeletonAnimOptimize) {
            return Promise.resolve(undefined);
        // } else {
        //     return LoaderTool.loadSklai(this, this.option.sklaiPath, this.option.sklaiName);
        // }
    }
    protected analySkin() {
    }
    public dispose = (disposeMaterialAndTextures: boolean = true) => {
        this.isDisposed = true;
        this._dispose();
        
        super.dispose();
    }
    protected _dispose() {
        this.scene._modelMap.delete(this);

        if (this._gltf) {
            this._gltf && this._gltf.free && this._gltf.free();
            // console.error("Dispose GLTF");
        }
        
        this.rootImpl?.dispose();
        this._animations?.forEach((item) => {
            item.dispose();
        });
        this._materials?.forEach((item) => {
            item.dispose();
        });

        N3Pool.recycle(this._position   );
        N3Pool.recycle(this._scale      );
        N3Pool.recycle(this._rotate     );
        N3Pool.recycle(this._lookAt     );
        N3Pool.recycle(this._color      );    

        this._gltf          = undefined;
        this._meshes        = undefined;
        this._materials     = undefined;
        this._animations    = undefined;
        this.rootImpl       = undefined;
        this._transformNodes = undefined;
        this._particleSystems = undefined;

        this.option         = undefined;
        this.animOpt        = undefined;
        this.parent         = undefined;
        this._animation     = undefined;
        this._lookAt        = undefined;
        this._rotate        = undefined;
        this._position      = undefined;
        this._scale         = undefined;
        this._color         = undefined;
        this.insertedCall   = undefined;
        this.gltfModifyConfig = undefined;
    }
    public reuse() {
        if (this.isDisposed) return;
        if (this.animOpt == undefined && this.animDefault && this._animations && this._animations.length > 0) {
            this.animOpt = {
                animName: this.animations[0].name,
                isLoop: true,
            }
            this._changeAnim();
        }
    }
    public unuse(): void {
        this._isUnuse = true;
        this.setEnabled(false);
        this._modifyUseState();
    }
    protected _isUnuse: boolean = false;
    protected _modifyUseState(): void {
        if (this._isUnuse == true) {
            this.animationMap?.forEach((group) => {
                // group.onAnimationGroupEndObservable.clear();
                // group.onAnimationGroupPlayObservable.clear();
                // group.onAnimationGroupLoopObservable.clear();
                // group.onAnimationGroupFrameEventObservable.clear();
                group.stop();
            });
            this._animEventMap.clear();
        }
    }
    // tslint:disable-next-line:max-line-length
    protected loaded = (root: BABYLON.ITransformNodeD3, nodes: BABYLON.ITransformNodeD3[], meshes: BABYLON.AbstractMesh[], particleSystems: MeshParticleSystem[], materials: BABYLON.Material[], animationGroups: BABYLON.AnimationGroupRS[], skins: BABYLON.Skeleton[]) => {
        this.scene.isDirty = true;

        this.rootImpl = root;
        this._particleSystems = particleSystems;
        this._meshes = meshes;
        this._animations = <any>animationGroups;
        this._materials = materials;
        this._transformNodes = nodes;

        if (this.isDisposed == true) {
            return;
        }
        
        if (meshes !== undefined) {
            meshes.forEach((item) => {
                let mesh = <BABYLON.Mesh>item;
                this.meshMap.set(mesh.id, mesh);
                // (<any>mesh).rayID = this.rayID;
                if (mesh.material && this.needDepthPrePass != undefined) {
                    // mesh.material.needDepthPrePass = this.needDepthPrePass;
                }
            });
        }

        if (animationGroups !== undefined) {
            // this._animation = animationGroups[0];
            animationGroups.forEach((animGroup) => {
                this._animationMap.set(animGroup.name, <any>animGroup);
                animGroup.onAnimationGroupFrameEventObservable.add((eventdata) => {
                    this._animationEventCall(eventdata.data);
                });
            });
        }

        if (this.animOpt == undefined && this.animDefault && animationGroups[0]) {
            this.animOpt = {
                animName: animationGroups[0].name,
                isLoop: true,
            };
        }

        skins?.forEach((skin) => {
            this._skeletonMap.set(skin.name, skin);
        });

        this.isLoaded = true;
        this._changePosition();
        this._changeScale();
        this._changeRotate();
        this._changeLookAt();
        this._changeAlpha();
        this._changeEnable();
        this._changeColor();
        this._updateCollider();

        this._updateAttach();
        this._updateAttachOpt();
        this._updateChildOpt();
        this._updateParent();
        // this._changeParentNode();
        this._changeAlphaIndex();
        this._changeRenderingGroupId();
        this._changeAnim();
        this._changeAnimeSpeed();
        this._modifyUseState();
        
        this._applyRayID();
    
        return;
    }
    protected animEndCall = () => {
        throw new Error("Method not implemented.");
    }
    /**
     * 动画执行结束的回调
     */
    protected animEndCB() {
        throw new Error("Method not implemented.");
    }
    /**
     * 停止模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 start 配对
     */
    public stopAllAnim() {
        throw new Error("Method not implemented.");
    }
    /**
     * 暂停模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 play 配对
     */
    public pauseAllAnim() {
        throw new Error("Method not implemented.");
    }
    /**
     * 继续播放模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 pause 配对
     * @param loop 是否循环
     */
    public playAllAnim(loop: boolean = true) {
        throw new Error("Method not implemented.");
    }
    /**
     * 重新播放模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 pause 配对,从第0帧开始
     */
    public restartAllAnim() {
        throw new Error("Method not implemented.");
    }
    /**
     * 重新播放模型所有粒子
     * @description 与 pause 配对,从第0帧开始
     */
    public restartAllParticle() {
        throw new Error("Method not implemented.");
    }
    /**
     * 启动模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 stop 配对
     * @param loop 是否循环
     * @param speed 播放速度
     */
    public startAllAnim(loop: boolean = true, speed: number = 1) {
        // throw new Error("Method not implemented.");
        if (this.animOpt == undefined) {
            this.animDefault = true;
            if (this.animations?.length > 0) {
                this.animOpt = {
                    animName: this.animations[0].name,
                    isLoop: loop,
                }
            }
        }
        this.setAnimeSpeed(speed);
        this._changeAnim();
    }
    /**
     * 停止模型所有粒子
     */
    public stopAllParticle() {
        throw new Error("Method not implemented.");
    }
    /**
     * 启动模型所有粒子
     * @param delay 延时启动时间
     */
    public startAllParticle(delay: number = 0) {
        throw new Error("Method not implemented.");
    }
    /**
     * 附加到指定骨骼
     * @param opt
     */
    public setAttach(opt: IModelAttachOpt) {
        throw new Error("Method not implemented.");
    }
    /**
     * 添加子模型
     * @param opt
     */
    public setChildModelOpt(opt: IModelChildOpt) {
        throw new Error("Method not implemented.");
    }

    public setAlpha(data: number) {
        this._alpha = data;
        this._changeAlpha();
    }
    public setColor(r: number, g: number, b: number) {
        super.setColor(r, g, b);
        this._changeColor();
    }
    // private _parent: IPrefab;
    // private _indexOfParent: number;
    // public setParent(parent: IPrefab, indexOfParent?: number): void {
    //     this._parent = parent;
    //     this._indexOfParent = indexOfParent;
    //     this._changeParentNode();
    // }
    /**
     * @deprecated
     * @deprecated use setEnabled
     * @param data 
     */
    public setVisible(data: boolean) {
        throw new Error("Method not implemented. Please Use `setEnabled`.");
    }
    
    private _enableRecord: [number, boolean][] =[];
    private _enableTime = Date.now();
    public setEnabled(isEnabled: boolean) {
        let time = Date.now();
        this._enableRecord.push([this._enableTime, isEnabled]);
        this._enableTime = time;
        this._isEnabled = isEnabled;
        this._changeEnable();
    }
    /**
     * 启动指定某个动画 - （模型）一次只播一个动画
     * @param animOpt
     */
    public setAnim(animOpt: IPrefabAnime) {
        this.animOpt = animOpt;
        this._changeAnim();
    }
    set alphaIndex(param: number) {
        this._alphaIndex = param;
        this._changeAlphaIndex();
    }
    public get alphaIndex() {
        // debugger;
        return this._alphaIndex;
    }
    public get renderingGroupId() {
        return this._renderingGroupId;
    }
    public set renderingGroupId(param: number) {
        // debugger;
        this._renderingGroupId = param;
        this._changeRenderingGroupId();
    }

    public renderQueueItems(): IMeshRenderQueue[] {
        let result: IMeshRenderQueue[] = [];
        this._meshes?.forEach((item) => {
            result.push(item);
        });
        this._particleSystems?.forEach((item) => {
            result.push(item);
        });

        return result;
    }

    public renderMaterialItems(): IMeshMaterial[] {
        let result: IMeshMaterial[] = [];
        this._meshes?.forEach((item) => {
            result.push(<IMeshMaterial>item);
        });
        this._particleSystems?.forEach((item) => {
            result.push(<IMeshMaterial>item);
        });

        return result;
    }

    public pickAboutItems(): INodeCollider[] {
        let result: INodeCollider[] = [];
        this._meshes?.forEach((item) => {
            result.push(<any>item);
        });

        return result;
    }

    protected _animeSpeed = 1.;
    setAnimeSpeed(param: number): void {
        this._animeSpeed = param;
        this._changeAnimeSpeed();
    }
    /**
     * 停止当前的动画 - （模型）一次只播一个动画
     */
    public stopAnim() {
        if (this.workWithAnimationGroupChange) {
            return;
        }
        if (this.isLoaded && !this.isDisposed) {
            this.stopAnimation();
        } else {
            this.animOpt = undefined;
        }
    }
    /**
     * 暂停当前的动画 - （模型）一次只播一个动画
     */
    public pauseAnim() {
        if (this.isLoaded && !this.isDisposed) {
            this.pauseAnimation();
        } else {
            this.animOpt = undefined;
        }
    }
    // protected _changeParentNode() {
    //     if (!this.isLoaded || this.isDisposed) { return }
    //     this.rootImpl.parent = this._parent?.getChild(this._indexOfParent);
    // }
    protected _changeAlpha() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this._alpha !== undefined) {
            this.materials?.forEach((material) => {
                material.alpha = this._alpha;
            });
        }
    }
    protected _changeColor(): void {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this._color) {
            this.materials?.forEach((material) => {
                material.mainColor(this._color[0], this._color[1], this._color[2]);
            });
        }
    }
    /**
     * @deprecated
     */
    protected _changeVisible() {
        this._changeEnable();
    }

    protected _changeEnable() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this._isEnabled !== undefined) {
            if (this.rootImpl?.isEnabled() != this._isEnabled && this._isEnabled == true) {
                this._changeAnim();
            }
            this.rootImpl.setEnabled(this._isEnabled);
        }
    }
    protected _updateAttachOpt() {
        // let useCount = 0;
        // let tempIndex = 0;
        // let newListLen = 0;
        // const listLen = this.attachOptList.length;

        // for (let len = listLen - 1; len >= 0; len--) {
        //     const opt = this.attachOptList[len];

        //     if (opt.mesh !== undefined) {
        //         const skeleton = this.skeletonMap.get(opt.skeletonName);

        //         if (skeleton !== undefined) {
        //             const boneIndex = skeleton.getBoneIndexByName(opt.boneName);
        //             if (boneIndex >= 0) {
        //                 // opt.mesh.attachToBone(skeleton.bones[boneIndex], this.rootImpl);
        //                 NodeTools.nodeTransform(opt.mesh, opt.transform);
        //             }
        //         }

        //         this.attachOptList[len] = undefined;
        //         useCount++;
        //     }
        // }

        // newListLen = listLen - useCount;
        // for (let i = 0, k = listLen - useCount; i < k; i++) {
        //     while (this.attachOptList[tempIndex] === undefined && tempIndex < listLen) {
        //         tempIndex++;
        //     }

        //     if (this.attachOptList[tempIndex] !== undefined) {
        //         this.attachOptList[i] = this.attachOptList[tempIndex];
        //     }
        // }

        // this.attachOptList.length = newListLen;
    }
    protected _updateChildOpt() {
        // this.childOptList.forEach((opt) => {
        //     if (opt.isFinished !== true) {
        //         if (opt.model !== undefined) {
        //             if (opt.model.rootImpl !== undefined) {
        //                 opt.isFinished = true;
        //                 opt.model.rootImpl.parent = this.rootImpl;
        //                 NodeTools.nodeTransform(opt.model.rootImpl, opt.transform);
        //             }
        //         } else if (opt.mesh !== undefined) {
        //             opt.isFinished = true;
        //             opt.mesh.parent = this.rootImpl;
        //             NodeTools.nodeTransform(opt.mesh, opt.transform);
        //         }
        //     }
        // });
    }
    protected _updateAttach() {
        //
    }
    protected _changeAnim() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this.animOpt !== undefined) {
            this.stopAnimation();

            let anime = this.animationMap.get(this.animOpt.animName);
            if (anime) {
                if (this.animOpt.startCall) {
                    anime.onAnimationGroupPlayObservable.addOnce(this.animOpt.startCall);
                }
                if (this.animOpt.endCall) {
                    anime.onAnimationGroupEndObservable.addOnce(this.animOpt.endCall);
                }
                if (this.animOpt.loopCall) {
                    anime.onAnimationGroupLoopObservable.addOnce(this.animOpt.loopCall);
                }
                if (this.animOpt.frameCall) {
                    anime.onAnimationGroupFrameEventObservable.addOnce(this.animOpt.frameCall);
                }
                anime.speedRatio = this._animeSpeed;
                anime.start(
                    this.animOpt.isLoop,
                );
                this._animation = <any>anime;
            }
        } else {
            this.stopAnimation();
        }
    }
    protected _changeAlphaIndex() {
        if (this.isLoaded && this._meshes && this._alphaIndex != undefined) {
            let min = Number.MAX_SAFE_INTEGER;
            this._meshes.forEach((item) => {
                min = Math.min(min, item.alphaIndex);
            });
            this._particleSystems.forEach((item) => {
                min = Math.min(min, item.alphaIndex);
            });
            this._meshes.forEach((item) => {
                item.alphaIndex = item.alphaIndex - min + this._alphaIndex;
            });
            this._particleSystems.forEach((item) => {
                item.alphaIndex = item.alphaIndex - min + this._alphaIndex;
            });
        }
    }
    protected _changeRenderingGroupId() {
        if (this.isLoaded && this._meshes && this._renderingGroupId != undefined) {
            this._meshes.forEach((item) => {
                item.renderingGroupId = this._renderingGroupId;
            });
            this._particleSystems.forEach((item) => {
                item.renderingGroupId = this._renderingGroupId;
            });
        }
    }
    protected _changeAnimeSpeed() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this.animation) {
            this.animation.speedRatio = this._animeSpeed;
        }
        this._particleSystems?.forEach((item) => {
            item.speedRatio = this._animeSpeed;
        });
    }
    protected parent: Model = undefined;
    protected parentIndex: number;
    setParent(parent: IModel, parentIndex?: number): void {
        this.parent = <Model>parent;
        this.parentIndex = parentIndex;
        this._updateParent();
    }
    getChild(index?: number): BABYLON.TransformNode {
        if (!this.isLoaded || this.isDisposed) { return }
        if (index != undefined) {
            return <BABYLON.TransformNode>this.transformNodes[index];
        } else {
            return <BABYLON.TransformNode>this.rootImpl;
        }
    }
    protected _updateParent() {
        if (this.isLoaded && !this.isDisposed && this.rootImpl) {
            if (this.parent && !this.isDisposed) {
                if (this.parent.isLoaded) {
                    this.rootImpl.parent = this.parent.getChild(this.parentIndex);
                } else {
                    let parent = this.parent;
                    this.parent.loadPromise.then((model) =>{
                        if (parent == model) {
                            this.rootImpl.parent = this.parent?.getChild(this.parentIndex);
                        }
                    });
                }
            } else {
                this.rootImpl.parent = null;
            }
        }
    }
    /**
     * 更改模型动画
     * @param animName 目标动画名称
     * @param isLoop 是否循环播放
     * @param stopFlag 动画停止配置
     * @param endCall 动画结束回调
     */
    // protected changeAnimation = () => {
    //     this._animation = this._changeAnimation(this.animOpt, false);
    //     this.isAnimationLoop = this.animOpt.isLoop;
    // }
    // protected _changeAnimation(animOpt: IModelAnimOpt, forBoneAttach: boolean = false) {

    //     const speed = animOpt.speed === undefined ? 1 : animOpt.speed;
    //     const animName = animOpt.animName + (forBoneAttach ? BABYLON.Skeleton.ANIME_NAME_SUFFIX_WHILE_BAKED : "");
    //     const isLoop = !!animOpt.isLoop;

    //     const animGroup = this.animationMap.get(animName);


    //     if (animGroup !== undefined) {
    //         if (this.changeAnimEnableBlending) {
    //             animGroup.enableBlend(false, this.changeAnimBlendUseSeconds);
    //         }

    //         if (animGroup.isStarted && isLoop) {
    //             logger.warn(`${animName} - Animation is played!`);
    //             animGroup.onAnimationGroupEndObservable.clear();
    //             animGroup.stop();
    //         }

    //         animGroup.speed = speed;
    //         animGroup.start(isLoop, null);
    //     } else {
    //         if (!forBoneAttach) {
    //             logger.warn(`${animName} - Animation Do Not Exist!`);
    //         }
    //     }

    //     return animGroup;
    // }
    protected stopAnimation = () => {
        this._stopAnimation(this._animation);
        this._animation = undefined;
    }
    protected _stopAnimation = (animation: BABYLON.AnimationGroupRS) => {
        if (animation && this.changeAnimEnableBlending && animation.isStarted) {
            animation.enableBlend(true, this.changeAnimBlendUseSeconds);
        } else {
            if (animation) {
                animation.onAnimationGroupPlayObservable.clear();
                animation.onAnimationGroupEndObservable.clear();
                animation.onAnimationGroupLoopObservable.clear();
                animation.onAnimationGroupFrameEventObservable.clear();
                animation.stop();
            }
        }
    }
    protected pauseAnimation = () => {
        this._animation && this._animation.pause();
    }
    protected initAnimationEventMap() {
        throw new Error("Method not implemented.");
    }
    // -------------------------------------------------------------------- 克隆
    /**
     * 以当前模型 进行克隆
     * @deprecated
     * @param newName 新的名称
     * @param rayID rayID
     */
    public clone(newName: string, rayID: number = -1) {
        throw new Error("Method not implemented.");
    }
    /**
     * @deprecated
     * @param newName 
     * @param rayID 
     */
    protected _clone(newName: string, rayID: number = -1) {
        throw new Error("Method not implemented.");
    }
    /**
     * @deprecated
     * @param oldParent 
     * @param newParent 
     */
    protected _instanceChild(oldParent: BABYLON.TransformNode, newParent: BABYLON.TransformNode) {
        throw new Error("Method not implemented.");
    }
    /**
     * 克隆动画 - 动画需要作用的目标对象
     * @deprecated
     * @param newNodes 节点
     * @param newSkeletons 骨骼
     * @param newMaterials 材质
     * @param newTextures 纹理
     */
    protected _cloneAnimation(newNodes: { metadata: any }[], newSkeletons: { metadata: any }[], newMaterials: { metadata: any }[], newTextures: { metadata: any }[], bones: { metadata: any }[]) {
    
        throw new Error("Method not implemented.");
    }
    /**
     * @deprecated
     * @param newMeshes 
     */
    protected _cloneSkeleton(newMeshes: BABYLON.AbstractMesh[]) {
        throw new Error("Method not implemented.");
    }
    /**
     * @deprecated
     * @param newNodes 
     */
    protected _cloneParicle(newNodes: BABYLON.TransformNode[]) {
        throw new Error("Method not implemented.");
    }
    /**
     * @deprecated
     */
    protected animationsInitRecordMap: Map<any, Map<string, any>> = new Map();
    /**
     * @deprecated
     */
    protected animationsInitRecorde() {
        throw new Error("Method not implemented.");
    }
    /**
     * @deprecated
     * @param target 
     * @param path 
     */
    protected animationInitRecorde(target: any, path: string) {
        throw new Error("Method not implemented.");
    }
    /**
     * @deprecated
     */
    protected animationsInitUse() {
        throw new Error("Method not implemented.");
    }
    /**
     * @deprecated
     * @param target 
     * @param path 
     * @param data 
     */
    protected animationInitUse(target: any, path: string, data: any) {
        throw new Error("Method not implemented.");
    }
    protected _updateCollider() {
        if (this.isDisposed) return;
        if (this.isLoaded && this._isPickable) {
            super._updateCollider();
        }
    }
    public setAnimeFrame(animName: string, progress: number): IModel {
        if (this.isLoaded && !this.isDisposed) {
            let anime = <IAnimationGroup>this.animationMap.get(animName);
            if (anime != this.animation) {
                this.stopAnimation();
                this._animation = <any>anime;
            }
            if (anime) {
                anime.goto(progress);
            }
        }
        return this;
    }
}
