import { IBloom, IBloomCombine, IBloomThreshold, KeyBloom, KeyBloomCombine, KeyBloomThreshold } from "./bloom";
import { IColorEffect, KeyColorEffect } from "./color_effect";
import { ICopy, KeyCopy } from "./copy";
import { IDirectBlur, KeyDirectBlur } from "./direct_blur";
import { IDualBlur, IDualBlurDown, IDualBlurUp, KeyDualBlur, KeyDualBlurDown, KeyDualBlurUp } from "./dual_blur";
import { IHueDegree, KeyHueDegree } from "./hue_degree";
import { IRadialBlur, KeyRadialBlur } from "./radial_blur";
import { IRadiusWave, KeyRadiusWave } from "./radius_wave";
import { IVignette, KeyVignette } from "./vignette";

export interface IPostprocessTexture {
    width: number;
    height: number;
    _webGLTexture: WebGLTexture;
}

export interface IPostprocessTarget {
    x: number;
    y: number;
    width: number;
    height: number;
    framebuffer: WebGLFramebuffer;
}

export type KeyPostprocess = typeof KeyBloom | typeof KeyBloomCombine | typeof KeyBloomThreshold | typeof KeyRadiusWave | typeof KeyRadialBlur
    | typeof KeyVignette | typeof KeyCopy | typeof KeyHueDegree | typeof KeyDirectBlur | typeof KeyColorEffect
    | typeof KeyDualBlur | typeof KeyDualBlurDown | typeof KeyDualBlurUp
;
export type IPostprocessParam = IBloom | IBloomCombine | IBloomThreshold | IRadiusWave | IRadialBlur
    | IVignette | ICopy | IHueDegree | IDirectBlur | IColorEffect
    | IDualBlur | IDualBlurDown | IDualBlurUp
;

export const SortKeys: KeyPostprocess[] = [KeyColorEffect, KeyHueDegree, KeyVignette, KeyRadialBlur, KeyDirectBlur, KeyBloom, KeyDualBlur, KeyRadiusWave, KeyCopy];

export function resortPostprocessParams(postprocess: IPostprocessParam[], actived: Map<KeyPostprocess, any>) {
    let result = [];
    SortKeys.forEach((key) => {
        let item = postprocess.find((v) => v.etype == key);
        if (item && actived.get(key) != undefined) {
            result.push(item);
        }
    });

    return result;
}