import { FrameCurve } from "./curves/base";
import { p3d } from "../types";

/**
 * 动画数据
 * * 指定 曲线名称, 作用的属性
 * * 通过 keys 获取曲线 插入关键帧
 * @example
 * ```typescript
 * let animation = new Animation<FrameCurveAlpha>("anime01", EAnimePropertyID.Alpha, 30);
 * let curve = animation.keys;
 * curve.addKey(0, 0.0);
 * curve.addKey(30, 1.0);
 * ```
 */
export class AnimationRS<T extends FrameCurve> {
    private _curve: T;
    private _property: p3d.EAnimePropertyID;
    public get keys(): T {
        return <T>this._curve;
    }
    constructor(private _key: string, curve: T, loopMode?: number, enableBlending?: boolean) {
        this._curve = curve;
        this._property = curve.property;
    }
    public get key() {
        return this._curve.key;
    }
    public get id() {
        return this._curve.id;
    }
    public get property() {
        return this._property;
    }
}