
import { AnimationRS } from "./animation";
import { FrameCurve } from "./curves/index";
import { DefaultKeyRes, Engine, Entity } from "../engines/engine";
import { Scene } from "../scene";
import { Observable, IAnimationGroup, ITargetAnimation
    , ITargetAnimationAttribute, ITargetAnimationProperty, ITargetAnimationUniform
} from "../../interface/index";
import { p3d, IFrameUpdate, IAnimationEventDataEnd, IAnimationEventDataFrame, IAnimationEventDataLoop, IAnimationEventDataStart, IAnimationGroup3D } from "../../render/types";
import { ExtensionTools } from "../../base";

type FrameData = any;
type FrameDataIndex = number;
type RustFrameData = FrameDataIndex;

export class AnimationGroupRS extends Entity implements IAnimationGroup3D, IFrameUpdate {
    private _name: string;
    private _scene: Scene;
    private _attach: Entity;
    public get atach(): Entity {
        return this._attach;
    }
    public get isPlaying() {
        return this.isStarted && !this._isPaused;
    }
    private _animations: AnimationRS<FrameCurve>[] = [];
    constructor(name: string, scene: Scene, private fps: number = 30) {
        let entity = p3d.p3d_animation_group(Engine.native(), Engine.commands(), scene.entity);
        // scene.cmdsAnimationAttach.push([scene.entity, attach.entity]);
        super(entity);
        this._attach = scene;
        this._name = name;
        this._scene = scene;
        this._scene.animationGroupsRS.set(this.entity, this);
        scene.animationGroupMap.set(this.entity, this);
    }
    gotoFrame(frame: number) {
        throw new Error("Method not implemented.");
    }
    goto(amount: number): IAnimationGroup {
        p3d.p3d_anime_group_goto(Engine.commands(), this.entity, amount);
        return this;
    }

    public get name() {
        return this._name;
    }

    _onStart(data: IAnimationEventDataStart) {
        this._onAnimationGroupPlayObservable.notifyObservers(data, -1, this, this);
        return this;
    }
    _onEnd(data: IAnimationEventDataEnd) {
        this._onAnimationGroupEndObservable.notifyObservers(data, -1, this, this);
        return this;
    }
    _onLoop(data: IAnimationEventDataLoop) {
        this._onAnimationGroupLoopObservable.notifyObservers(data, -1, this, this);
        return this;
    }
    _onFrame(data: IAnimationEventDataFrame) {
        let framedata = this._frameListenData[data.data];
        data.data = framedata;
        this._onAnimationGroupFrameEventObservable.notifyObservers(data, -1, this, this);
        return this;
    }
    private _runtimeinfos: ITargetAnimation[] = [];
    public addTargetedAnimation<T extends FrameCurve>(animation: AnimationRS<T>, target: ITargetAnimation) {
        if (this.entity != 0) {
            this._animations.push(animation);
            // this._runtimeinfos.push(target);
            // Engine.addCmdTargetAnimation(animation.id, animation.property, this._scene.entity, this._entity, target.entity);
            // p3d.p3d_target_animation(Engine.native(), Engine.param(), animation.id, animation.property, this._scene.entity, this._entity, target.entity);
            
            let item = target;
            if ((<ITargetAnimationProperty>item).target) {
                p3d.p3d_property_target_animation(Engine.commands(),  item.curve, (<ITargetAnimationProperty>item).property, item.group, (<ITargetAnimationProperty>item).target);
            } else if ((<ITargetAnimationUniform>item).uniformProperty) {
                p3d.p3d_uniform_target_animation(
                    Engine.commands(), (<ITargetAnimationUniform>item).uniformTarget, item.group,
                    ExtensionTools.getAtom((<ITargetAnimationUniform>item).uniformProperty), item.curve
                );
                // console.warn(`uniform Anime: ${(<ITargetAnimationUniform>item).uniformProperty} `);
            } else if ((<ITargetAnimationAttribute>item).attributeProperty) {
                p3d.p3d_attribute_target_animation(
                    Engine.commands(), (<ITargetAnimationAttribute>item).attributeTarget, item.group,
                    ExtensionTools.getAtom((<ITargetAnimationAttribute>item).attributeProperty), item.curve
                );
            }
        }
    }
    private _isStarted: boolean;
    public get isStarted() {
        return this._isStarted;
    }

    private _speed: number = 1;
    public get speedRatio() {
        return this._speed;
    }
    public set speedRatio(val: number) {
        this._speed = val;
    }

    public delay_ms: number = 0;
    public fillmode: p3d.EFillMode = p3d.EFillMode.Forwards;

    private _loopMode: p3d.ELoopMode = p3d.ELoopMode.Not;
    private _loopCount: number = null;
    private _amountMode: number = p3d.EAmountMode.None;
    private _from: number = 0;
    private _to: number = 1;
    public start(isLoop: boolean = false, loopCount: number = undefined, amountMode: p3d.EAmountMode = p3d.EAmountMode.None, from: number = 0, to: number = 1, ) {
        if (this.entity != 0) {
            this._isStarted = true;
            this._isPaused = false;
            this._loopMode = isLoop ? p3d.ELoopMode.Positive : p3d.ELoopMode.Not;
            this._loopCount = loopCount;
            this._amountMode = amountMode;
            this._from = from;
            this._to = to;

            // this._runtimeinfos.forEach((item) => {
            //     if ((<ITargetAnimationProperty>item).target) {
            //         p3d.p3d_property_target_animation(Engine.commands(),  item.curve, (<ITargetAnimationProperty>item).property, item.group, (<ITargetAnimationProperty>item).target);
            //     } else if ((<ITargetAnimationUniform>item).uniformProperty) {
            //         p3d.p3d_uniform_target_animation(
            //             Engine.commands(), (<ITargetAnimationUniform>item).uniformTarget, item.group,
            //             getAtom((<ITargetAnimationUniform>item).uniformProperty), item.curve
            //         );
            //         console.warn(`uniform Anime: ${(<ITargetAnimationUniform>item).uniformProperty} `);
            //     } else if ((<ITargetAnimationAttribute>item).attributeProperty) {
            //         p3d.p3d_uniform_target_animation(
            //             Engine.commands(), (<ITargetAnimationAttribute>item).attributeTarget, item.group,
            //             getAtom((<ITargetAnimationAttribute>item).attributeProperty), item.curve
            //         );
            //     }
            // });
            // this._runtimeinfos.length = 0;

            p3d.p3d_anime_group_start(
                Engine.commands(), this.entity, this._speed,
                this._loopMode, this._loopCount, this._from, this._to, this.fps, this._amountMode, 0, 0, 0, 0, this.delay_ms, this.fillmode
            );
        }
    }
    
    public restart() {
        if (this.entity != 0) {
            this._isPaused = false;
            p3d.p3d_anime_group_start(
                Engine.commands(), this.entity, this._speed,
                this._loopMode, this._loopCount, this._from, this._to, this.fps, this._amountMode, 0, 0, 0, 0, this.delay_ms, this.fillmode
            );
        }
    }

    protected _isPaused: boolean = true;
    public pause() {
        if (this._entity != 0) {
            p3d.p3d_anime_group_pause(
                Engine.commands(), this._entity
            );
            this._isPaused = true;
        }
    }
    public stop() {
        if (this._entity != 0) {
            this._isStarted = false;
            p3d.p3d_anime_group_stop(
                Engine.commands(), this._entity
            );
        }
    }
    public dispose(): void {
        if (this._entity != 0) {
            this._scene.animationGroupsRS.delete(this._entity);
            p3d.p3d_animation_group_delete(Engine.commands(), this._entity);
        }
        super.dispose();
    }

    private _frameListenData: FrameData[] = [];

    public addFrameEvent(percent: number, data: any) {
        if (this._disposed) {
            return;
        }
        percent = Math.min(1., Math.max(0., percent));
        let idx = this._frameListenData.length;
        this._frameListenData.push(data);
        p3d.p3d_animation_group_add_frame_event(Engine.commands(), this._entity, percent, idx);
    }

    _onFrameUpdate(): void {
        if (this._enableBlend) {
            let factor = (this._scene._animationTime - this._blendStart) / 1000 * this._blendingSpeed;
            factor = Math.min(1., Math.max(0., factor));
            if (this._blendFadeOut) {
                factor = 1.0 - factor;
            }

            this._updateBlendingFactor(factor);
        }
    }

    /**
     * 卸载动画对目标的作用
     * @returns 
     */
    public _unintallEffect() {
        if (this._disposed) {
            return;
        }
        p3d.p3d_animation_group_target_reset(Engine.commands(), this._entity);
    }

    private _blendingSpeed: number = 0.01;
    private _enableBlend: boolean = false;
    private _blendStart: number = 0;
    /**
     * 混合时作为退出的动画
     */
    private _blendFadeOut: boolean = true;

    private _blendingFactor: number = 1.;
    private _updateBlendingFactor(factor: number) {
        if (this._blendingFactor != factor) {
            this._blendingFactor = factor;
            p3d.p3d_animation_group_weight(Engine.commands(), this._entity, factor);
        }
    }

    public enableBlend(blendFadeOut: boolean, usesecond: number = 1) {
        this._blendFadeOut = blendFadeOut;
        this._blendingSpeed = 1 / usesecond;
        this._enableBlend = true;
        this._blendStart = this._scene._animationTime;
    }

    /**
     * This observable will notify when all animations are playing.
     */
    private _flagStart = false;
    private _flagEnd = false;
    private _flagLoop = false;
    private _flagFrame = false;
    private _onAnimationGroupPlayObservable = new Observable<IAnimationEventDataStart>();
    public get onAnimationGroupPlayObservable() {
        if (this._flagStart == false) {
            p3d.p3d_animation_group_listen(Engine.commands(), this._entity, p3d.EAnimationGroupListen.Start);
            this._flagStart = true;
        }
        return this._onAnimationGroupPlayObservable;
    }

    /**
     * Observer raised when all animations have looped
     */
    private _onAnimationGroupLoopObservable = new Observable<IAnimationEventDataLoop>();
    public get onAnimationGroupLoopObservable() {
        if (this._flagLoop == false) {
            p3d.p3d_animation_group_listen(Engine.commands(), this._entity, p3d.EAnimationGroupListen.Loop);
            this._flagLoop = true;
        }
        return this._onAnimationGroupLoopObservable;
    }

    /**
     * This observable will notify when all animations have ended.
     */
    private _onAnimationGroupEndObservable = new Observable<IAnimationEventDataEnd>();
    public get onAnimationGroupEndObservable() {
        if (this._flagEnd == false) {
            p3d.p3d_animation_group_listen(Engine.commands(), this._entity, p3d.EAnimationGroupListen.End);
            this._flagEnd = true;
        }
        return this._onAnimationGroupEndObservable;
    }

    /**
     * This observable will notify when all animations have paused.
     */
    private _onAnimationGroupFrameEventObservable = new Observable<IAnimationEventDataFrame>();
    public get onAnimationGroupFrameEventObservable() {
        if (this._flagFrame == false) {
            p3d.p3d_animation_group_listen(Engine.commands(), this._entity, p3d.EAnimationGroupListen.Frame);
            this._flagFrame = true;
        }
        return this._onAnimationGroupFrameEventObservable;
    }

}