import { Entity, Engine } from "../engines/engine";
import { Scene } from "../scene";
import { Node, Transform } from "../node";
import { Matrix, Quaternion, Space, Vector3 } from "../../maths/index";
import { p3d } from "../types";
import { IBone, IMatrix } from "../../interface/index";

export class Bone extends Transform implements IBone  {
    public static _rotationAxisCache: Quaternion = new Quaternion();
    public static _PoseCache: Float32Array = new Float32Array(16);
    metadata: any = {};

    constructor(name: string, scene: Scene) {
        let entity = p3d.p3d_bone(Engine.native(), Engine.commands(), scene.entity);
        super(name, scene, entity)
    }
    _baseMatrix: Matrix = Matrix.Identity();
    _absoluteTransform: Matrix = Matrix.Identity();
    _invertedAbsoluteTransform: Matrix = Matrix.Identity();
    setBindPose(matrix: Matrix): void {
        this._baseMatrix.copyFrom(matrix);
        if (this.parent) {
            this._baseMatrix.multiplyToRef((<Bone>this.parent)._absoluteTransform, this._absoluteTransform);
        } else {
            this._absoluteTransform.copyFrom(this._baseMatrix);
        }
        this._absoluteTransform.invertToRef(this._invertedAbsoluteTransform);

        matrix = this._baseMatrix.transpose();
        for (let i = 0; i < 16; i++) {
            Bone._PoseCache[i] = matrix.m[i];
        }
        // console.log(matrix);
        p3d.p3d_bone_pose(Engine.commands(), this.entity, Bone._PoseCache);
    }
    setBindPoseFromArray(matrix: ArrayLike<number>): void {
        for (let i = 0; i < 16; i++) {
            Bone._PoseCache[i] = matrix[i];
        }
        p3d.p3d_bone_pose(Engine.commands(), this.entity, Bone._PoseCache);
    }

    protected __linkedTransformNode: Node;
    public set _linkedTransformNode(val: Node) {
        this.__linkedTransformNode = val;
        p3d.p3d_bone_link(Engine.commands(), this.entity, val.entity);
    }
    public get _linkedTransformNode() {
        return this.__linkedTransformNode;
    }

    forceDirty() {
        this._scene?.cmdsForceDirty.set(this._entity, this);
    }

    dispose(): void {
        super.dispose();
    }
}