
import { Engine, Entity } from "../engines/engine";
import { IAbstractMesh, IAnimationAttach, IEntity, ISkeleton } from "../../interface/index";
import { Bone } from "./bone";
import { Matrix } from "../../maths/math.vector";
import { IEntityD3, p3d } from "../types";

export class Skeleton extends Entity implements IAnimationAttach, ISkeleton {
    public static ANIME_NAME_SUFFIX_WHILE_BAKED = "_BONES_ATTACH";
    public static _Bones: Float64Array = new Float64Array(256);

    private _name: string;
    constructor(name: string, root: Entity, bones: Bone[]) {
        bones.forEach((bone, i) => {
            Skeleton._Bones[i] = bone.entity;
        });
        let entity = p3d.p3d_skeleton(Engine.native(), Engine.commands(), 4, root.entity, Skeleton._Bones, bones.length, 1);
        super(entity);
        this._name = name;
        this.name = name;
        this.bones = bones;
    }

    _totalBones: Bone[] = [];

    getPoseMatrix(): Matrix {
        throw new Error("Method not implemented.");
    }
    overrideMesh: IAbstractMesh;
    name?: string;

    public applyToMesh(mesh: IEntityD3) {
        if (this._isDisposed) return;
        p3d.p3d_skin_use(Engine.commands(), mesh.entity, this.entity)        
    }

    public bones = new Array<Bone>();

    public get id() {
        return this._name;
    }

    public getBoneIndexByName(name: string): number {
        return -1;
    }

    dispose(): void {
        if (this._disposed) return;

        this.bones.forEach((item) => {
            item.dispose();
        });
        this.bones = undefined;

        super.dispose();
    }
}