
import { Scene } from "../scene";
import { Vector3 } from "../../maths/math.vector";
import { Camera } from "./camera";
import { Engine } from "../engines/engine";
import { p3d, IFreeCameraD3 } from "../types";
import { Constants, } from "../../interface/index";

export class FreeCamera extends Camera implements IFreeCameraD3 {
    constructor(name: string, position: Vector3, scene: Scene) {
        super(name, scene);
        this.position = position;
    }
    private _fov: number = 0.75;
    public get fov() {
        return this._fov;
    }
    public set fov(val: number) {
        if (this._fov != val) {
            this._fov = val;
            p3d.p3d_camera_fov(Engine.commands(), this.camera.entity, this._fov);
        }
    }
    
    private _size: number = 4;
    public get size() {
        return this._size;
    }
    public set size(val: number) {
        if (this._size != val) {
            this._size = val;
            p3d.p3d_camera_size(Engine.commands(), this.camera.entity, this._size);
        }
    }
    public get piOrthSize() {
        return this._size;
    }
    public set piOrthSize(val: number) {
        if (this._size != val) {
            this._size = val;
            p3d.p3d_camera_size(Engine.commands(), this.camera.entity, this._size);
        }
    }

    public set piSizeForWidth(val: boolean) {
        this.fovMode = val ? Constants.FOVMODE_HORIZONTAL_FIXED : Constants.FOVMODE_VERTICAL_FIXED;
    }
    public get piSizeForWidth() {
        return this._fovMode == Camera.FOVMODE_HORIZONTAL_FIXED;
    }

    public get orthoRight() {
        return this.piSizeForWidth ? this._size : this._size * (this._aspect || 1);
    }
    public get orthoLeft() {
        return -1 * this.orthoRight;
    }
    public get orthoTop() {
        return !this.piSizeForWidth ? this._size : this._size / (this._aspect || 1);
    }
    public get orthoBottom() {
        return -1 * this.orthoTop;
    }
    
    private _mode: Constants.FREECAMERA_MODE = Constants.PERSPECTIVE_CAMERA;
    public get mode() {
        return this._mode;
    }
    public set mode(val: Constants.FREECAMERA_MODE) {
        this._mode = val;
        p3d.p3d_camera_mode(Engine.commands(), this.camera.entity, val == Camera.ORTHOGRAPHIC_CAMERA);
    }

    private _target: Vector3 = new Vector3(0, 0, 1);
    public get target() {
        this._scene.cmdsForceDirty.set(this.entity, this);
        return this._target;
    }
    public setTarget(val: Vector3) {
        this._target.copyFrom(val);
        this._scene.cmdsForceDirty.set(this.entity, this);
    }

    private _fovMode: Constants.CameraFovMode = Constants.FOVMODE_VERTICAL_FIXED;
    public set fovMode(val: Constants.CameraFovMode) {
        this._fovMode = val;
        p3d.p3d_camera_fixed_mode(Engine.commands(), this.camera.entity, this._fovMode == Camera.FOVMODE_HORIZONTAL_FIXED);
    }

    _onFrameUpdate(): void {
        if (this._isDisposed) {
            return;
        }
        super._onFrameUpdate();

        if (this._target._isDirty) {
            p3d.p3d_camera_target(Engine.commands(), this.camera.entity, this._target.x, this._target.y, this._target.z);
        }
        this._target._isDirty = false;
    }
}