
import { Engine } from "../engines/engine";
import { IsometryNode } from "../node"
import { Vector3 } from "../../maths/index";
import { Scene } from "../scene";
import { Constants, IIsometryNode } from "../../interface/index";
import { p3d } from "../types";

export class Light extends IsometryNode implements IIsometryNode {
    // diffuse: Color3;
    // specular: Color3;
    // intensity: number;

    private _color: Vector3 = new Vector3(1., 1., 1.);
    public color(r: number, g: number, b: number) {
        this._color.copyFromFloats(r, g, b);
        p3d.p3d_light_color(Engine.commands(), this.entity, this._color.x, this._color.y, this._color.z);
    }
    
    private _strength: number = 1.0;
    public strength(val: number) {
        this._strength = val;
        p3d.p3d_light_strength(Engine.commands(), this.entity, this._strength);
    }
}

export class DirectionLight extends Light {
    constructor(name: string, scene: Scene) {
        let entity = p3d.p3d_light(Engine.native(), Engine.commands(), scene.entity, Constants.LIGHT_DIRECTION);
        super(name, scene, entity);
    }
    private _direction: Vector3 = new Vector3(0, -1, 0);
    public direction(val: Vector3) {
        this._direction.copyFrom(val);
        p3d.p3d_light_direction(Engine.commands(), this.entity, this._direction.x, this._direction.y, this._direction.z);
    }
}

export class PointLight  extends Light {
    constructor(name: string, scene: Scene) {
        let entity = p3d.p3d_light(Engine.native(), Engine.commands(), scene.entity, Constants.LIGHT_POINT);
        super(name, scene, entity);
    }
    private _radius: number = 5.;
    public radius(val: number) {
        this._radius = val;
        p3d.p3d_light_radius(Engine.commands(), this.entity, this._radius);
    }
}

export class SpotLight  extends Light {
    constructor(name: string, scene: Scene) {
        let entity = p3d.p3d_light(Engine.native(), Engine.commands(), scene.entity, Constants.LIGHT_SPOT);
        super(name, scene, entity);
    }
    private _radius: number = 5.;
    public radius(val: number) {
        this._radius = val;
        p3d.p3d_light_radius(Engine.commands(), this.entity, this._radius);
    }
    private _inner: number = 0.2;
    private _outer: number = 0.5;
    public innerAngle(val: number) {
        this._inner = val;
        p3d.p3d_light_spot_angle(Engine.commands(), this.entity, this._inner, this._outer);
    }
    public outerAngle(val: number) {
        this._outer = val;
        p3d.p3d_light_spot_angle(Engine.commands(), this.entity, this._inner, this._outer);
    }
    private _direction: Vector3 = new Vector3(0, -1, 0);
    public direction(x: number, y: number, z: number) {
        this._direction.copyFromFloats(x, y, z);
        p3d.p3d_light_direction(Engine.commands(), this.entity, this._direction.x, this._direction.y, this._direction.z);
    }
}