import { p3d } from "../../types";
import { Engine } from "../../engines/engine";
import { Constants } from "../../../interface/index";
import { ExtensionTools } from "../../../base";

export interface ITextureInfo {
    url: string,
    srgb: boolean,
    compressed: boolean,
    isfile: boolean,
    cancombine: boolean,
    depth_or_array_layers: number,
    usages: number,
    width: number,
    height: number,
    size: number,
}

export interface ITextureLoader {
    resolve: (value: ITextureInfo | PromiseLike<ITextureInfo>) => void,
    reject: (reason?: any) => void,
    info: ITextureInfo,
    time: number,
}

export class BaseTexture {
    metadata: any = {};
    private _coordinatesIndex: number;
    public get coordinatesIndex() : number {
        return this._coordinatesIndex
    }
    
    public set coordinatesIndex(v : number) {
        this._coordinatesIndex = v;
        console.warn("TODO")
    }

    private static Counter: number = Number.MIN_SAFE_INTEGER;
    private static TempBuffer: Float32Array = new Float32Array(8);
    private static LastQueryTimeMS: number = 0;
    public static QueryTimeMS: number = 20;
    public static MaxWaitTimeMS: number = 120 * 1000;
    public static RecordFile: Map<number, ITextureLoader> = new Map();
    public static WaitingFile: [][] = [];
    public static CreateLoad(url: string, srgb: boolean, cancombine: boolean): Promise<ITextureInfo> {
        let pureName = ExtensionTools.GetPureResPath(url);
        let realurl = ExtensionTools.QueryRuntimeFullName(pureName);
        let compressed = realurl.endsWith(".ktx");
        return new Promise((resolve, reject) => {
            let index = BaseTexture.Counter++;
            let info = {
                resolve,
                reject,
                info: <ITextureInfo>{ url, srgb, width: 0, height: 0, size: 0, isfile: true, compressed, cancombine, depth_or_array_layers: 0, usages: Constants.TEXTURE_USEAGE_BINDING | Constants.TEXTURE_USEAGE_COPY_DST },
                time: Date.now()
            };
            BaseTexture.RecordFile.set(index, info);
        });
    }
    public static Query() {
        let time = Date.now();
        if (time < BaseTexture.LastQueryTimeMS + BaseTexture.QueryTimeMS) {
            return
        }
        BaseTexture.LastQueryTimeMS = time;
        BaseTexture.RecordFile.forEach((info, key) => {
            let isLoaded = p3d.p3d_query_texture(Engine.native(), Engine.param(), info.info.isfile, ExtensionTools.getAtom(info.info.url), info.info.cancombine, info.info.compressed, info.info.depth_or_array_layers, BaseTexture.TempBuffer);
            if (isLoaded) {
                // console.error("Loaded " + info.info.url);
                BaseTexture.RecordFile.delete(key);
                info.info.width     = BaseTexture.TempBuffer[0];
                info.info.height    = BaseTexture.TempBuffer[1];
                info.info.size      = BaseTexture.TempBuffer[2];
                info.resolve(info.info);
            } else if (info.time + BaseTexture.MaxWaitTimeMS < time) {
                BaseTexture.RecordFile.delete(key);
                info.info.width     = BaseTexture.TempBuffer[0];
                info.info.height    = BaseTexture.TempBuffer[1];
                info.info.size      = BaseTexture.TempBuffer[2];
                info.reject('Time Out');
            }
        });
    }
}
