import { DataTexture } from "./datatexture";
import { Constants, ExtensionTools } from "../../../interface/index";
import { p3d } from "../../../context";
import { Engine } from "../../engines/engine";

export interface ICombineTextureLoader {
    resolve: Function;
    reject: Function;
}

export interface ICombineTextureParam {
    width: number,
    height: number,
    frames: {
        url: string,
        x: number, y: number, w: number, h: number
    }[],
    format: number,
}

export class CombineTexture extends DataTexture {
    private static MaxCombineCount: number = 128;
    private static Success: Uint32Array;
    private static Fail: Uint32Array;
    public static CombineRecord: Map<number, CombineTexture> = new Map();
    public static CombineRecordFile: Map<number, ICombineTextureLoader> = new Map();
    public static Initial(maxCombineCount: number) {
        CombineTexture.MaxCombineCount = maxCombineCount;
        CombineTexture.Success = new Uint32Array(maxCombineCount);
        CombineTexture.Fail = new Uint32Array(maxCombineCount);
        Engine.registBeforeRun(CombineTexture.CombineQuery);
    }
    public static CombineCreateLoad(item: CombineTexture): Promise<void> {
        if (item.loadPromise == undefined){
            item.loadPromise = new Promise((resolve, reject) => {
                let info: ICombineTextureLoader = { 
                    resolve,
                    reject
                };
                CombineTexture.CombineRecordFile.set(item._uniqueId, info);
            });
        }
        return item.loadPromise;
    }
    public static CombineQuery() {
        p3d.p3d_texture_combine_query(Engine.native(), CombineTexture.Success, CombineTexture.Fail);
        let idx = 0;
        let id = CombineTexture.Success[idx];
        while (id == 0 && idx < CombineTexture.MaxCombineCount) {
            let promise = CombineTexture.CombineRecordFile.get(id);
            if (promise) {
                promise.resolve();
            }
            idx += 1;
            id = CombineTexture.Success[idx];
        }
        idx = 0;
        id = CombineTexture.Fail[idx];
        while (id == 0 && idx < CombineTexture.MaxCombineCount) {
            let promise = CombineTexture.CombineRecordFile.get(id);
            if (promise) {
                promise.reject();
            }
            idx += 1;
            id = CombineTexture.Fail[idx];
        }
        for (let i = 0; i < CombineTexture.MaxCombineCount; i++) {
            CombineTexture.Fail[i] = 0;
            CombineTexture.Success[i] = 0;
        }
    }

    constructor(url: ICombineTextureParam) {
        super(url.format, url.width, url.height);
        url.frames.forEach((item, idx) => {
            let url = ExtensionTools.QueryRuntimeFullName(ExtensionTools.GetPureResPath(item.url));
            let file = ExtensionTools.getAtom(url);
            let key = ExtensionTools.getAtom(this._uniqueKey);
            let isCompress = url.endsWith(".ktx");
            p3d.p3d_texture_combine_add_frame(Engine.commands(), this._uniqueId, key, file, idx, isCompress, item.x, item.y, item.w, item.h);
        });
        CombineTexture.CombineRecord.set(this._uniqueId, this);
        CombineTexture.CombineCreateLoad(this);
    }
    loadPromise: Promise<void>;
    dispose(): void {
        CombineTexture.CombineRecord.delete(this._uniqueId);
        super.dispose();
    }
    public get url(): string {
        return this._uniqueKey;
    }
}