
import { Scene } from "../../scene";
import { BaseTexture } from "./baseTexture";
import { Constants, Observable } from "../../../interface/index";
import { IFrameUpdate, IMaterialD3 } from "../../types";
import { ExtensionTools } from "../../../interface/index";
import { CustomRenderTarget } from "../../renderer";

export class Texture extends BaseTexture implements IFrameUpdate {
    public static TextureCombineCheck: RegExp[] = [];
    public static TextureCombineFiles: string[] = [];
    /** nearest is mag = nearest and min = nearest and mip = linear */
    public static NEAREST_SAMPLINGMODE = 1;
    /** Bilinear is mag = linear and min = linear and mip = nearest */
    public static BILINEAR_SAMPLINGMODE = 2;
    /** Trilinear is mag = linear and min = linear and mip = linear */
    public static TRILINEAR_SAMPLINGMODE = 3;
    /** nearest is mag = nearest and min = nearest and mip = linear */
    public static NEAREST_NEAREST_MIPLINEAR = 8;
    /** Bilinear is mag = linear and min = linear and mip = nearest */
    public static LINEAR_LINEAR_MIPNEAREST = 11;
    /** Trilinear is mag = linear and min = linear and mip = linear */
    public static LINEAR_LINEAR_MIPLINEAR = 3;
    /** mag = nearest and min = nearest and mip = nearest */
    public static NEAREST_NEAREST_MIPNEAREST = 4;
    /** mag = nearest and min = linear and mip = nearest */
    public static NEAREST_LINEAR_MIPNEAREST = 5;
    /** mag = nearest and min = linear and mip = linear */
    public static NEAREST_LINEAR_MIPLINEAR = 6;
    /** mag = nearest and min = linear and mip = none */
    public static NEAREST_LINEAR = 7;
    /** mag = nearest and min = nearest and mip = none */
    public static NEAREST_NEAREST = 1;
    /** mag = linear and min = nearest and mip = nearest */
    public static LINEAR_NEAREST_MIPNEAREST = 9;
    /** mag = linear and min = nearest and mip = linear */
    public static LINEAR_NEAREST_MIPLINEAR = 10;
    /** mag = linear and min = linear and mip = none */
    public static LINEAR_LINEAR = 2;
    /** mag = linear and min = nearest and mip = none */
    public static LINEAR_NEAREST = 12;

    public static CLAMP_ADDRESSMODE: Constants.AddressMode = Constants.CLAMP_TO_EDGE;
    public static REPEAT: Constants.AddressMode = Constants.REPEAT;
    public static MIRRORED_REPEAT: Constants.AddressMode = Constants.MIRRORED_REPEAT;
    
    url: string = null;
    srgb: boolean = false;
    filter: boolean = true;
    compressed: boolean = true;
    address_mode_u: Constants.AddressMode = Constants.CLAMP_TO_EDGE;
    address_mode_v: Constants.AddressMode = Constants.CLAMP_TO_EDGE;
    address_mode_w: Constants.AddressMode = Constants.CLAMP_TO_EDGE;
    mag_filter: Constants.FilterMode = Constants.NEAREST;
    min_filter: Constants.FilterMode = Constants.NEAREST;
    mipmap_filter: Constants.FilterMode = Constants.NEAREST;
    compare: Constants.CompareFunction = undefined;
    anisotropy_clamp: Constants.AnisotropyClamp = Constants.NONE;
    border_color: Constants.BorderColor = Constants.NONE;

    onDisposeObservable: Observable<void> = new Observable();

    public static isCompressed = function(url: string) {
        let pureName = ExtensionTools.GetPureResPath(url);
        let realurl = ExtensionTools.QueryRuntimeFullName(pureName);
        // debugger;
        return realurl?.endsWith(`.ktx`) || false;
    }

    isfile: boolean = true;
    cancombine: boolean = false;

    /**
     * 
     * @param url 
     * @param _scene 
     * @param noMipmap 
     * @param invertY 
     * @param samplingMode Texture.NEAREST_SAMPLINGMODE | Texture.BILINEAR_SAMPLINGMODE | Texture.TRILINEAR_SAMPLINGMODE | Texture.NEAREST_NEAREST_MIPLINEAR | Texture.LINEAR_LINEAR_MIPNEAREST | Texture.NEAREST_LINEAR_MIPNEAREST ...
     */
    constructor(url: string, private _scene: Scene, noMipmap?: boolean, invertY?: boolean, samplingMode?: number) {
        super();

        if (samplingMode == Texture.NEAREST_SAMPLINGMODE) {
            this.mag_filter = Constants.NEAREST;
            this.min_filter = Constants.NEAREST;
            this.mipmap_filter = Constants.NEAREST;
        }
        else if (samplingMode == Texture.BILINEAR_SAMPLINGMODE) {
            this.mag_filter = Constants.LINEAR;
            this.min_filter = Constants.LINEAR;
            this.mipmap_filter = Constants.NEAREST;
        }
        else if (samplingMode == Texture.TRILINEAR_SAMPLINGMODE) {
            this.mag_filter = Constants.LINEAR;
            this.min_filter = Constants.LINEAR;
            this.mipmap_filter = Constants.LINEAR;
        }
        else if (samplingMode == Texture.NEAREST_NEAREST_MIPLINEAR) {
            this.mag_filter = Constants.NEAREST;
            this.min_filter = Constants.NEAREST;
            this.mipmap_filter = Constants.LINEAR;
        }
        else if (samplingMode == Texture.LINEAR_LINEAR_MIPNEAREST) {
            this.mag_filter = Constants.LINEAR;
            this.min_filter = Constants.LINEAR;
            this.mipmap_filter = Constants.NEAREST;
        }
        else if (samplingMode == Texture.NEAREST_NEAREST_MIPNEAREST) {
            this.mag_filter = Constants.NEAREST;
            this.min_filter = Constants.NEAREST;
            this.mipmap_filter = Constants.NEAREST;
        }
        else if (samplingMode == Texture.NEAREST_LINEAR_MIPNEAREST) {
            this.mag_filter = Constants.NEAREST;
            this.min_filter = Constants.LINEAR;
            this.mipmap_filter = Constants.NEAREST;
        }
        else if (samplingMode == Texture.NEAREST_LINEAR_MIPLINEAR) {
            this.mag_filter = Constants.NEAREST;
            this.min_filter = Constants.NEAREST;
            this.mipmap_filter = Constants.LINEAR;
        }
        else if (samplingMode == Texture.NEAREST_LINEAR) {
            this.mag_filter = Constants.NEAREST;
            this.min_filter = Constants.LINEAR;
            this.mipmap_filter = Constants.NEAREST;
        }
    }
    public _material: IMaterialD3;
    public get scene() {
        return this._scene;
    }
    getScene() {
        return this._scene;
    }
    _onFrameUpdate(): void {
        this._levelDirty = false;
        this._tilloffDirty = false;
        this._offsetSpeedDirty = false;
    }
    onDisposeObserver: Observable<number> =  new Observable<number>();
    dispose(): void {
        // throw new Error("Method not implemented.");
        this.onDisposeObservable.notifyObservers();

        this._scene = undefined;
    }
    updateURL(url: string) {
        this.url = url;
        let pureName = ExtensionTools.GetPureResPath(url);
        let realurl = ExtensionTools.QueryRuntimeFullName(pureName);
        // debugger;
        this.compressed = realurl?.endsWith(`.ktx`) || false;
        // console.error(realurl, this.compressed);
        this.url = realurl;
        let len = Texture.TextureCombineCheck.length;
        for (let i =0; i < len; i++) {
            this.cancombine = Texture.TextureCombineCheck[i].test(url);
            if (this.cancombine) {
                return;
            }
        }
        this.cancombine = Texture.TextureCombineFiles.indexOf(url) >= 0;
    }

    _texture: CustomRenderTarget;

    public _offsetSpeedDirty: boolean = false;
    private _offsetSpeedU: number = 0;
    public set offsetSpeedU(val: number) {
        if (this._offsetSpeedU != val) {
            this._offsetSpeedDirty = true;
            this._offsetSpeedU = val;
            this._scene.textures.set(this, this);
            if (this._material) {
                this._scene.cmdsForceDirty.set(this._material.entity, this._material);
            }
        }
    }
    public get offsetSpeedU() {
        return this._offsetSpeedU;
    }
    private _offsetSpeedV: number = 0;
    public set offsetSpeedV(val: number) {
        val *= -1;
        if (this._offsetSpeedV != val) {
            this._offsetSpeedDirty = true;
            this._offsetSpeedV = val;
            this._scene.textures.set(this, this);
            if (this._material) {
                this._scene.cmdsForceDirty.set(this._material.entity, this._material);
            }
        }
    }
    public get offsetSpeedV() {
        return this._offsetSpeedV;
    }

    public _tilloffDirty: boolean = false;
    private _uOffset: number = 0;
    public set uOffset(val: number) {
        if (this._uOffset != val) {
            this._tilloffDirty = true;
            this._uOffset = val;
            this._scene.textures.set(this, this);
            if (this._material) {
                this._scene.cmdsForceDirty.set(this._material.entity, this._material);
            }
        }
    }
    public get uOffset() {
        return this._uOffset;
    }

    private _vOffset: number = 0;
    public set vOffset(val: number) {
        if (this._vOffset != val) {
            this._tilloffDirty = true;
            this._vOffset = val;
            this._scene.textures.set(this, this);
            if (this._material) {
                this._scene.cmdsForceDirty.set(this._material.entity, this._material);
            }
        }
    }
    public get vOffset() {
        return this._vOffset;
    }

    private _uScale: number = 1;
    public set uScale(val: number) {
        if (this._uScale != val) {
            this._tilloffDirty = true;
            this._uScale = val;
            this._scene.textures.set(this, this);
            if (this._material) {
                this._scene.cmdsForceDirty.set(this._material.entity, this._material);
            }
        }
    }
    public get uScale() {
        return this._uScale;
    }

    private _vScale: number = 1;
    public set vScale(val: number) {
        if (this._vScale != val) {
            this._tilloffDirty = true;
            this._vScale = val;
            this._scene.textures.set(this, this);
            if (this._material) {
                this._scene.cmdsForceDirty.set(this._material.entity, this._material);
            }
        }
    }
    public get vScale() {
        return this._vScale;
    }

    public _levelDirty: boolean = false;
    private _level: number = 1;
    public set level(val: number) {
        if (this._level != val) {
            this._levelDirty = true;
            this._level = val;
            this._scene.textures.set(this, this);
            if (this._material) {
                this._scene.cmdsForceDirty.set(this._material.entity, this._material);
            }
        }
    }
    public get level() {
        return this._level;
    }

    public set wrapU(val: Constants.AddressMode) {
        this.address_mode_u = val;
        if (this._material) {
            this._scene.cmdsForceDirty.set(this._material.entity, this._material);
        }
    }
    public set wrapV(val: Constants.AddressMode) {
        this.address_mode_v = val;
        if (this._material) {
            this._scene.cmdsForceDirty.set(this._material.entity, this._material);
        }
    }

    /**
     * @deprecated
     */
    hasAlpha: boolean;
}
