import { Engine } from "../engines/engine";
import { Scene } from "../scene";
import { EffectMaterial } from "./material";
import { Texture } from "./Textures/texture";
import { UniformMainTexture, UniformMainTextureUVOffsetSpeed } from "./node_material/mainTextureBlock";
import { Color3 } from "../../maths/math.color";
import { UniformEmissiveTexture } from "./node_material/emissiveBlock";
import { VARYING_COLOR, VARYING_NORMAL_WS, VARYING_POSITION_WS, VARYING_UV, VARYING_UV2, VARYING_UV3 } from "./node_material";
import { UniformOpacityTexture, UniformOpacityTextureUVOffsetSpeed } from "./node_material/opacityBlock";
import { Constants } from "../../interface/index";
import { p3d } from "../types";
import { ExtensionTools } from "../../base";

export class MainOpacityMaterial extends EffectMaterial {
    public static ShaderName = "MainOpacityMaterial";
    constructor(name: string, scene: Scene, shaderName = MainOpacityMaterial.ShaderName, _pass: Constants.PassTag = Constants.PASS_TAG_06) {
        super(name, scene, shaderName, _pass);
    }

    // protected _emissiveColorDirty: boolean = false;
    // protected _emissiveColor: Color3 = new Color3(0, 0, 0);
    // public set emissiveColor(val: Color3) {
    //     if (this._emissiveColor != val) {
    //         this._emissiveColor.copyFrom(val);
    //         this._emissiveColorDirty = true;
    //     }
    // }
    // public get emissiveColor() {
    //     return this._emissiveColor;
    // }

    // protected _emissiveTextureDirty: boolean = false;
    // protected _emissiveTexture: Texture;
    // public set emissiveTexture(val: Texture) {
    //     if (this._emissiveTexture != val) {
    //         this._emissiveTexture = val;
    //         this._emissiveTextureDirty = true;
    //     }
    // }

    protected _diffuseColorDirty: boolean = false;
    protected _diffuseColor: Color3 = new Color3(1, 1, 1);
    public set diffuseColor(val: Color3) {
        if (this._diffuseColor != val) {
            this._diffuseColor.copyFrom(val);
            this._diffuseColorDirty = true;
            this._scene.cmdsForceDirty.set(this.entity, this);
        }
    }
    public get diffuseColor() {
        this._scene.cmdsForceDirty.set(this.entity, this);
        return this._diffuseColor;
    }
    
    public set emissiveColor(val: Color3) {
        this.diffuseColor = val;
    }
    public get emissiveColor() {
        return this._diffuseColor;
    }


    protected _diffuseTextureDirty: boolean = false;
    protected _diffuseTexture: Texture;
    public set diffuseTexture(val: Texture) {
        if (this._diffuseTexture != val) {
            this._diffuseTexture = val;
            this._diffuseTextureDirty = true;
            if (val) val._material = this;
            this._scene.cmdsForceDirty.set(this.entity, this);
        }
    }
    public get diffuseTexture() {
        this._scene.cmdsForceDirty.set(this.entity, this);
        return this._diffuseTexture;
    }
    
    protected _opacityTextureDirty: boolean = false;
    protected _opacityTexture: Texture;
    public set opacityTexture(val: Texture) {
        if (this._opacityTexture != val) {
            this._opacityTexture = val;
            this._opacityTextureDirty = true;
            this._scene.cmdsForceDirty.set(this.entity, this);
            if (val) val._material = this;
        }
    }
    public get opacityTexture() {
        this._scene.cmdsForceDirty.set(this.entity, this);
        return this._opacityTexture;
    }

    public mainColor(r: number, g: number, b: number): void {
        this._diffuseColor.copyFromFloats(r, g, b);
        this._diffuseColorDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }

    _onFrameUpdate() {
        if (this._isDisposed) { return; }
        super._onFrameUpdate();
        if (this._diffuseColorDirty || (this._diffuseTextureDirty || (this._diffuseTexture && this._diffuseTexture._levelDirty))) {
            UniformMainTexture.color(this.entity, this._diffuseColor.r, this._diffuseColor.g, this._diffuseColor.b, this._diffuseTexture?.level || 1.);
        }
        if (this._diffuseTexture) {
            if (this._diffuseTextureDirty || this._diffuseTexture._tilloffDirty) {
                UniformMainTexture.tilloff(this.entity, this._diffuseTexture.uScale, this._diffuseTexture.vScale, this._diffuseTexture.uOffset, this._diffuseTexture.vOffset);
            }
            if (this._diffuseTextureDirty || this._diffuseTexture._offsetSpeedDirty) {
                UniformMainTextureUVOffsetSpeed.value(this.entity, this._diffuseTexture.offsetSpeedU, this._diffuseTexture.offsetSpeedV);
            }
            if (this._diffuseTextureDirty) {
                UniformMainTexture.texture(this.entity, this._diffuseTexture);
            }
        }
        
        // if (this._emissiveColorDirty || (this._emissiveTextureDirty || (this._emissiveTexture && this._emissiveTexture._levelDirty))) {
        //     EmissiveTextureBlock.emisisveColor(this.entity, this._emissiveColor.r, this._emissiveColor.g, this._emissiveColor.b, this._emissiveTexture?.level || 1.);
        // }
        // if (this._emissiveTexture) {
        //     if (this._emissiveTextureDirty || this._emissiveTexture._tilloffDirty) {
        //         EmissiveTextureBlock.tilloff(this.entity, this._emissiveTexture.uScale, this._emissiveTexture.vScale, this._emissiveTexture.uOffset, this._emissiveTexture.vOffset);
        //     }
        //     if (this._emissiveTextureDirty || this._emissiveTexture._offsetSpeedDirty) {
        //         EmissiveTextureUVOffsetSpeedBlock.value(this.entity, this._emissiveTexture.offsetSpeedU, this._emissiveTexture.offsetSpeedV);
        //     }
        //     if (this._emissiveTextureDirty) {
        //         EmissiveTextureBlock.texture(this.entity, this._emissiveTexture);
        //     }
        // }
        if (this._opacityTexture) {
            if (this._opacityTextureDirty || this._opacityTexture._levelDirty) {
                UniformOpacityTexture.level(this.entity, this._opacityTexture.level);
            }
            if (this._opacityTextureDirty || this._opacityTexture._tilloffDirty) {
                UniformOpacityTexture.tilloff(this.entity, this._opacityTexture.uScale, this._opacityTexture.vScale, this._opacityTexture.uOffset, this._opacityTexture.vOffset);
            }
            if (this._opacityTextureDirty || this._opacityTexture._offsetSpeedDirty) {
                UniformOpacityTextureUVOffsetSpeed.value(this.entity, this._opacityTexture.offsetSpeedU, this._opacityTexture.offsetSpeedV);
            }
            if (this._opacityTextureDirty) {
                UniformOpacityTexture.texture(this.entity, this._opacityTexture);
            }
        }


        this._diffuseTextureDirty = false;
        this._diffuseColorDirty = false;
        // this._emissiveTextureDirty = false;
        // this._emissiveColorDirty = false;
        this._opacityTextureDirty = false;
    }

    public static regist() {
        if (Engine.CheckShader(MainOpacityMaterial.ShaderName)) {
            return;
        }
        let uniforms = p3d.MaterialUniformDefines.create();
        let includes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformMainTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTextureUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformOpacityTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTextureUVOffsetSpeed);
        //         p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformEmissiveTexture.BLOCK_NAME));
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTextureUVOffsetSpeed);


        let varyings = p3d.P3DShaderVaryings.create();
        [
            // ["vec3", "v_pos"],
            // ["vec3", "v_normal"],
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
            // ["vec2", "v_uv2"],
            // ["vec2", "v_uv3"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let varying = VARYING_POSITION_WS | VARYING_NORMAL_WS | VARYING_UV | VARYING_COLOR | VARYING_UV2 | VARYING_UV3;

        let binds_defines_base = Constants.BIND_MODEL | Constants.BIND_VIEWER | Constants.BIND_EFFECT_VALUE | Constants.BIND_EFFECT_TEXTURE;

        Engine.RegistShader(
            MainOpacityMaterial.ShaderName,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            mat4 finalWorld = PI_ObjectToWorld;
        
            vec4 position =  vec4(A_POSITION, 1.);
            vec4 worldPos =  finalWorld * position;
            // vec4 worldPos =  position;
        
            gl_Position = PI_MATRIX_VP * worldPos;

            // gl_Position = position;
        
            // v_pos = worldPos.xyz;
        
            // mat3 normalWorld = mat3(finalWorld);
            // v_normal = A_NORMAL;
            
            v_uv    = A_UV;
            // v_uv2   = A_UV;
            // v_uv3   = A_UV;
            v_color = A_COLOR4;
            `,
            `
            vec4 baseColor = v_color;
            float alpha = opacity(matParam);
            alpha   *= v_color.a;
        
            float glossiness 			= 0.;
            vec3 diffuseColor 		    = vec3(0., 0., 0.);
            vec3 baseAmbientColor		= vec3(1., 1., 1.);
            vec4 refractionColor		= vec4(0., 0., 0., 1.);
            vec4 reflectionColor 		= vec4(0., 0., 0., 1.);
        
            // ----------------------------------------------------------
        
            vec4 mainTextureColor   = mainTexture(v_uv, applyUVOffsetSpeed(matParam.uMainUVOS), matParam);
            baseColor.rgb           *= mainTextureColor.rgb * mainStrength(matParam);
            alpha                   *= mainTextureColor.a;
        
            diffuseColor            = mainColor(matParam);
        
            vec4 opacityData        = opacityTexture(v_uv, applyUVOffsetSpeed(matParam.uOpacityUVOS), matParam);
            alpha                   *= opacityChannel(opacityData, matParam);
        
            if (cutoff(alpha, matParam)) {
                discard;
            }
        
            // vec3 emissiveColor = emissiveColor(matParam);
            // emissiveColor *= emissiveTexture(v_uv, applyUVOffsetSpeed(matParam.uEmissiveUVOS), matParam).rgb * emissiveStrength(matParam);
        
            vec3 finalSpecular 		= vec3(0., 0., 0.);
            vec3 finalDiffuse       = diffuseColor * baseColor.rgb;
            
            vec4 finalColor 		= vec4(
                finalDiffuse * baseAmbientColor
                + 
                finalSpecular
                +
                refractionColor.rgb
                + 
                reflectionColor.rgb
                ,
                alpha
            );
        
            finalColor				= max(finalColor, 0.0);
        
            // Premulty
            // finalColor.rgb          *= finalColor.a;
        
            gl_FragColor = finalColor;        
            `,
            includes,
            ``,
            varyings,
            binds_defines_base
        );
    }
}