import { Engine } from "../../engines/engine";
import { Texture } from "../Textures/texture";
import { INodeMaterialBlock, INodeMaterialTextureInfo, Mat4Arr, Vec2Arr, Vec3Arr, Vec4Arr, Constants } from "../../../interface/index";
import { CustomRenderTarget } from "../../../render/renderer";
import { p3d } from "../../types";
import { ExtensionTools } from "../../../base";
import { BABYLON } from "../../../render3d/babylon";

export class TextureAbout {
    public static KEY_TILLOFF: string = undefined;
    public static KEY_TEX: string     = undefined;
    public static KEY_LEVEL: string = undefined;
    public static TilloffBoundle: boolean = true;
    public static tilloff(mat: number, scaleU: number, scaleV: number, offsetU: number, offsetV: number) {
        p3d.p3d_material_uniform_vec4(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_TILLOFF), scaleU, scaleV, offsetU, offsetV);
    }
    public static level(mat: number, strength: number) {
        p3d.p3d_material_uniform_float(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_LEVEL), strength);
    }
    public static texture(mat: number, val: Texture | CustomRenderTarget) {
        // console.warn("Text Compressed: " + val.compressed);
        if (val instanceof Texture) {
            UniformTexture.texture(mat, this.KEY_TEX, val);
        } else {
            // UniformTexture.target(mat, this.KEY_TEX, this.KEY_TILLOFF, val, );
            p3d.p3d_material_uniform_tex_from_render_target(
                Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_TEX), ExtensionTools.getAtom(this.KEY_TILLOFF), val.key,
                true, BABYLON.Constants.CLAMP_ADDRESSMODE, BABYLON.Constants.CLAMP_ADDRESSMODE, BABYLON.Constants.CLAMP_ADDRESSMODE,
                BABYLON.Constants.LINEAR, BABYLON.Constants.LINEAR, BABYLON.Constants.LINEAR, BABYLON.Constants.NONE, BABYLON.Constants.NONE, BABYLON.Constants.NONE
            );
        }
        // console.warn("Load: " + val.url);
    }
}

export class TextureAboutAnimenable {
    public static KEY_USCALE: string = undefined;
    public static KEY_VSCALE: string = undefined;
    public static KEY_UOFFSET: string = undefined;
    public static KEY_VOFFSET: string = undefined;
    public static KEY_TEX: string     = undefined;
    public static KEY_LEVEL: string = undefined;
    public static TilloffBoundle: boolean = true;
    public static tilloff(mat: number, scaleU: number, scaleV: number, offsetU: number, offsetV: number) {
        p3d.p3d_material_uniform_float(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_USCALE), scaleU);
        p3d.p3d_material_uniform_float(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_VSCALE), scaleV);
        p3d.p3d_material_uniform_float(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_UOFFSET), offsetU);
        p3d.p3d_material_uniform_float(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_VOFFSET), offsetV);
    }
    public static level(mat: number, strength: number) {
        p3d.p3d_material_uniform_float(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_LEVEL), strength);
    }
    public static texture(mat: number, val: Texture) {
        // console.warn("Text Compressed: " + val.compressed);
        if (val._texture instanceof CustomRenderTarget) {
            UniformTexture.target(mat, this.KEY_TEX, this.KEY_LEVEL, val._texture, val);
        } else {
            UniformTexture.texture(mat, this.KEY_TEX, val);
        }
        // console.warn("Load: " + val.url);
    }
}

export class UniformTexture {
    public static texture(mat: number, key: string, val: Texture) {
        p3d.p3d_material_uniform_tex(Engine.commands(), mat, ExtensionTools.getAtom(key),
            ExtensionTools.getAtom(val.url), val.srgb, val.compressed, val.filter, val.address_mode_u, val.address_mode_v, val.address_mode_w,
            val.mag_filter, val.min_filter, val.mipmap_filter, val.anisotropy_clamp, val.border_color, val.isfile, val.cancombine, val.compare
        );
    }
    public static target(mat: number, key: string, key_tilloff: string, val: CustomRenderTarget, tex: Texture) {
        // console.warn("Text Compressed: " + val.compressed);
        p3d.p3d_material_uniform_tex_from_render_target(
            Engine.commands(), mat, ExtensionTools.getAtom(key), ExtensionTools.getAtom(key_tilloff), val.key,
            tex.filter, tex.address_mode_u, tex.address_mode_v, tex.address_mode_w,
            tex.mag_filter, tex.min_filter, tex.mipmap_filter, tex.anisotropy_clamp, tex.border_color, tex.compare
        );
    }
}

export class UVOffsetSpeed {
    public static KEY_PARAM:    string = undefined;
    public static value(mat: number, speedU: number, speedV: number) {
        // debugger;
        // console.warn(this.KEY_PARAM);
        p3d.p3d_material_uniform_vec2(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_PARAM), speedU, speedV);
    }
}

export class NodeMaterialBlock implements INodeMaterialBlock {
    key: string = "";
    bind_defines: Constants.BindingState[] = [];
    depends: string[] = [];
    vs_define: string = "";
    fs_define: string = "";
    mat4: Map<string, Mat4Arr> = new Map();
    vec4: Map<string, Vec4Arr> = new Map();
    vec3: Map<string, Vec3Arr> = new Map();
    vec2: Map<string, Vec2Arr> = new Map();
    float: Map<string, [number, boolean]> = new Map();
    uint: Map<string, [number, boolean]> = new Map();
    texture: Map<string, INodeMaterialTextureInfo> = new Map();
}