import { Engine } from "../../engines/engine";
import { TextureAbout, UVOffsetSpeed } from "./base";
import { Constants, INodeMaterialBlock, INodeMaterialTextureInfo, Mat4Arr, Vec2Arr, Vec3Arr, Vec4Arr } from "../../../interface/index";
import { p3d } from "../../types";
import { ExtensionTools } from "../../../base";

export class UniformEmissiveTexture extends TextureAbout implements INodeMaterialBlock {
    public static KEY_COLOR: string     = Constants.U_EMISSIVE_COLOR
    public static KEY_TILLOFF: string   = Constants.U_EMISSIVE_TILLOFF
    public static KEY_TEX: string       = Constants.U_EMISSIVE_TEX
    public static KEY_LEVEL: string     = Constants.U_EMISSIVE_TEX_LV
    public static emisisveColor(mat: number, r: number, g: number, b: number, level: number) {
        p3d.p3d_material_uniform_vec4(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_COLOR), r, g, b, level);
    }

    public static BLOCK_NAME = "EMISSIVE_TEXTURE";
    key: string;
    bind_defines: Constants.BindingState[] = [Constants.BIND_EFFECT_VALUE];
    depends: string[] = [];
    vs_define: string = "";
    fs_define: string = "";
    mat4: Map<string, Mat4Arr> = new Map();
    vec4: Map<string, Vec4Arr> = new Map();
    vec3: Map<string, Vec3Arr> = new Map();
    vec2: Map<string, Vec2Arr> = new Map();
    float: Map<string, [number, boolean]> = new Map();
    uint: Map<string, [number, boolean]> = new Map();
    texture: Map<string, INodeMaterialTextureInfo> = new Map();
    constructor() {
        super();

        UniformEmissiveTexture.BLOCK_NAME = "EMISSIVE_TEXTURE_UNBUNDLE";
        UniformEmissiveTexture.TilloffBoundle = false;

        this.key = UniformEmissiveTexture.BLOCK_NAME;
        this.vec3.set(UniformEmissiveTexture.KEY_COLOR, [0, 0, 0, false]);
        this.vec4.set(UniformEmissiveTexture.KEY_TILLOFF, [1, 1, 0, 0, false]);
        this.float.set(UniformEmissiveTexture.KEY_LEVEL, [1, false]);
        this.texture.set(UniformEmissiveTexture.KEY_TEX, <INodeMaterialTextureInfo>{
            url: "",
            filterable: true,
            stage: Constants.SHADER_STAGE_FRAGMENT,
            default: Constants.DEFAULT_TEXTURE_WHITE,
            dimension: Constants.TextureViewDimensionD2,
        });

        this.fs_define = `
vec4 emissiveTexture(vec2 vMainUV, vec2 vUVOS, MatParam maParam) {
    return Get${UniformEmissiveTexture.KEY_TEX}(vMainUV, vUVOS, maParam.${UniformEmissiveTexture.KEY_TILLOFF});
}
vec3 emissiveColor(MatParam maParam) {
    return maParam.uEmissiveInfo.xyz;
}
float emissiveStrength(MatParam maParam) {
    return maParam._EmissiveTexLevel;
}
        `;
    }

}
export class UniformEmissiveTextureUVOffsetSpeed extends UVOffsetSpeed {
    public static KEY_PARAM:    string = Constants.U_EMISSIVE_UVOS
}