import { Engine } from "../../engines/engine";
import { TextureAbout, TextureAboutAnimenable, UVOffsetSpeed } from "./base";
import { Constants, INodeMaterialBlock, INodeMaterialTextureInfo, Mat4Arr, Vec2Arr, Vec3Arr, Vec4Arr } from "../../../interface/index";
import { p3d } from "../../types";
import { ExtensionTools } from "../../../base";

export class UniformMainTextureSimple implements INodeMaterialBlock {
    public static KEY_TEX: string       = Constants.U_MAIN_TEX;

    public static BLOCK_NAME = "MAIN_TEXTURE_SIMPLE";
    key: string = UniformMainTextureSimple.BLOCK_NAME;
    bind_defines: Constants.BindingState[] = [];
    depends: string[] = [];
    vs_define: string = "";
    fs_define: string = "";
    mat4: Map<string, Mat4Arr> = new Map();
    vec4: Map<string, Vec4Arr> = new Map();
    vec3: Map<string, Vec3Arr> = new Map();
    vec2: Map<string, Vec2Arr> = new Map();
    float: Map<string, [number, boolean]> = new Map();
    uint: Map<string, [number, boolean]> = new Map();
    texture: Map<string, INodeMaterialTextureInfo> = new Map();
    constructor() {
        UniformMainTextureSimple.BLOCK_NAME = "MAIN_TEXTURE_SIMPLE";
        this.key = UniformMainTextureSimple.BLOCK_NAME;
        this.texture.set(UniformMainTextureSimple.KEY_TEX, <INodeMaterialTextureInfo>{
            url: "",
            filterable: true,
            stage: Constants.SHADER_STAGE_FRAGMENT,
            default: Constants.DEFAULT_TEXTURE_WHITE,
            dimension: Constants.TextureViewDimensionD2,
        });

        this.fs_define = `
vec4 mainTexture(vec2 vMainUV) {
    return Get${UniformMainTextureSimple.KEY_TEX}(vMainUV, vec2(0., 0.), vec4(1., 1., 0., 0.));
}
        `;
    }

}
export class UniformMainTexture extends TextureAboutAnimenable implements INodeMaterialBlock {
    public static KEY_COLOR: string     = Constants.U_MAIN_COLOR;
    // public static KEY_TILLOFF: string   = Constants.U_MAIN_TILLOFF;
    public static KEY_TEX: string       = Constants.U_MAIN_TEX;
    public static KEY_LEVEL: string     = Constants.U_MAIN_TEX_LV;
    public static KEY_USCALE: string    = "uMainUScale";
    public static KEY_VSCALE: string    = "uMainVScale";
    public static KEY_UOFFSET: string   = "uMainUOffset";
    public static KEY_VOFFSET: string   = "uMainVOffset";
    
    public static color(mat: number, r: number, g: number, b: number, a: number) {
        p3d.p3d_material_uniform_vec4(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_COLOR), r, g, b, a);
    }


    public static BLOCK_NAME = "MAIN_TEXTURE";
    key: string = UniformMainTexture.BLOCK_NAME;
    bind_defines: Constants.BindingState[] = [Constants.BIND_EFFECT_VALUE];
    depends: string[] = [];
    vs_define: string = "";
    fs_define: string = "";
    mat4: Map<string, Mat4Arr> = new Map();
    vec4: Map<string, Vec4Arr> = new Map();
    vec3: Map<string, Vec3Arr> = new Map();
    vec2: Map<string, Vec2Arr> = new Map();
    float: Map<string, [number, boolean]> = new Map();
    uint: Map<string, [number, boolean]> = new Map();
    texture: Map<string, INodeMaterialTextureInfo> = new Map();
    constructor() {
        super();
        UniformMainTexture.BLOCK_NAME = "MAIN_TEXTURE_UNBUNDLE";
        UniformMainTexture.TilloffBoundle = false;
        this.key = UniformMainTexture.BLOCK_NAME;

        this.vec3.set(UniformMainTexture.KEY_COLOR, [1, 1, 1, true]);
        // this.vec4.set(UniformMainTexture.KEY_TILLOFF, [1, 1, 0, 0, true]);
        this.float.set(UniformMainTexture.KEY_LEVEL, [1, false]);
        this.float.set(UniformMainTexture.KEY_USCALE, [1, false]);
        this.float.set(UniformMainTexture.KEY_VSCALE, [1, false]);
        this.float.set(UniformMainTexture.KEY_UOFFSET, [0, false]);
        this.float.set(UniformMainTexture.KEY_VOFFSET, [0, false]);
        this.texture.set(UniformMainTexture.KEY_TEX, <INodeMaterialTextureInfo>{
            url: "",
            filterable: true,
            stage: Constants.SHADER_STAGE_FRAGMENT,
            default: Constants.DEFAULT_TEXTURE_WHITE,
            dimension: Constants.TextureViewDimensionD2,
        });

        // let tilloff = `maParam.${UniformMainTexture.KEY_TILLOFF}`;
        let tilloff = `vec4(maParam.${UniformMainTexture.KEY_USCALE}, maParam.${UniformMainTexture.KEY_VSCALE}, maParam.${UniformMainTexture.KEY_UOFFSET}, maParam.${UniformMainTexture.KEY_VOFFSET})`;

        this.fs_define = `
vec4 mainTexture(vec2 vMainUV, vec2 vUVOS, MatParam maParam) {
    return Get${UniformMainTexture.KEY_TEX}(vMainUV, vUVOS, ${tilloff});
}
vec3 mainColor(MatParam maParam) {
    return maParam.${UniformMainTexture.KEY_COLOR}.xyz;
}
float mainStrength(MatParam maParam) {
    return maParam.${UniformMainTexture.KEY_LEVEL};
}
        `;
    }

}
export class UniformMainTextureUVOffsetSpeed extends UVOffsetSpeed {
    public static KEY_PARAM:    string = Constants.U_MAIN_UVOS;
}

export class UniformMainRenderTarget extends TextureAbout implements INodeMaterialBlock {
    public static KEY_COLOR: string     = Constants.U_MAIN_COLOR;
    public static KEY_TILLOFF: string   = Constants.U_MAIN_TILLOFF;
    public static KEY_TEX: string       = Constants.U_MAIN_TEX;
    public static KEY_LEVEL: string     = Constants.U_MAIN_TEX_LV;
    
    public static color(mat: number, r: number, g: number, b: number, a: number) {
        p3d.p3d_material_uniform_vec4(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_COLOR), r, g, b, a);
    }

    public static BLOCK_NAME = "MAIN_RENDERTARGET";
    key: string = UniformMainRenderTarget.BLOCK_NAME;
    bind_defines: Constants.BindingState[] = [Constants.BIND_EFFECT_VALUE];
    depends: string[] = [];
    vs_define: string = "";
    fs_define: string = "";
    mat4: Map<string, Mat4Arr> = new Map();
    vec4: Map<string, Vec4Arr> = new Map();
    vec3: Map<string, Vec3Arr> = new Map();
    vec2: Map<string, Vec2Arr> = new Map();
    float: Map<string, [number, boolean]> = new Map();
    uint: Map<string, [number, boolean]> = new Map();
    texture: Map<string, INodeMaterialTextureInfo> = new Map();
    constructor() {
        super();
        UniformMainRenderTarget.BLOCK_NAME = "MAIN_RENDERTARGET_BUNDLE";
        UniformMainRenderTarget.TilloffBoundle = false;
        this.key = UniformMainRenderTarget.BLOCK_NAME;

        this.vec3.set(UniformMainRenderTarget.KEY_COLOR, [1, 1, 1, true]);
        this.vec4.set(UniformMainRenderTarget.KEY_TILLOFF, [1, 1, 0, 0, true]);
        this.float.set(UniformMainRenderTarget.KEY_LEVEL, [1, false]);
        this.texture.set(UniformMainRenderTarget.KEY_TEX, <INodeMaterialTextureInfo>{
            url: "",
            filterable: true,
            stage: Constants.SHADER_STAGE_FRAGMENT,
            default: Constants.DEFAULT_TEXTURE_WHITE,
            dimension: Constants.TextureViewDimensionD2,
        });

        let tilloff = `maParam.${UniformMainRenderTarget.KEY_TILLOFF}`;
        this.fs_define = `
vec4 mainTexture(vec2 vMainUV, vec2 vUVOS, MatParam maParam) {
    return Get${UniformMainRenderTarget.KEY_TEX}(vMainUV, vUVOS, ${tilloff});
}
vec3 mainColor(MatParam maParam) {
    return maParam.${UniformMainRenderTarget.KEY_COLOR}.xyz;
}
float mainStrength(MatParam maParam) {
    return maParam.${UniformMainRenderTarget.KEY_LEVEL};
}
        `;
    }

}