import { INodeMaterialBlock, INodeMaterialBlockManager } from "../../../interface/index";
import { Engine } from "../../engines/engine";
import { p3d } from "../../types";
import { ExtensionTools } from "../../../base";

export class NodeMaterialBlockManager implements INodeMaterialBlockManager {
    pool: Set<string> = new Set();
    regist(block: INodeMaterialBlock): void {
        if (this.pool.has(block.key)) {
            throw new Error(`Regist NodeMaterialBlock [${block.key}] Again !!!`);
        } else {
            let bindDefines = 0;
            block.bind_defines.forEach(v => {
                bindDefines = bindDefines | v;
            });
            let blockres = p3d.NodeMaterialBlock.create(ExtensionTools.getAtom(block.key), block.vs_define, ``, block.fs_define, ``, bindDefines);
            block.mat4.forEach((v, k) => {
                p3d.p3d_node_material_block_mat4( blockres, ExtensionTools.getAtom(k), ...v );
            });
            block.vec4.forEach((v, k) => {
                p3d.p3d_node_material_block_vec4( blockres, ExtensionTools.getAtom(k), ...v );
            });
            block.vec3.forEach((v, k) => {
                p3d.p3d_node_material_block_vec3( blockres, ExtensionTools.getAtom(k), ...v );
            });
            block.vec2.forEach((v, k) => {
                p3d.p3d_node_material_block_vec2( blockres, ExtensionTools.getAtom(k), ...v );
            });
            block.float.forEach((v, k) => {
                p3d.p3d_node_material_block_float( blockres, ExtensionTools.getAtom(k), ...v );
            });
            block.uint.forEach((v, k) => {
                p3d.p3d_node_material_block_uint( blockres, ExtensionTools.getAtom(k), ...v );
            });
            block.depends.forEach((v) => {
                p3d.p3d_node_material_block_extend_depend( blockres, ExtensionTools.getAtom(v) );
            });
            block.texture.forEach((v, k) => {
                p3d.p3d_node_material_block_texture( blockres, ExtensionTools.getAtom(k), v.filterable, v.stage, v.default, v.dimension );
            });
            p3d.p3d_node_material_block_extend_bind_define( blockres, bindDefines );
            p3d.p3d_node_material_block_regist(
                Engine.native(), Engine.param(), blockres
            );
            blockres.free && blockres.free();
        }
    }
    public static Single: NodeMaterialBlockManager;
    public static Init() {
        NodeMaterialBlockManager.Single = new NodeMaterialBlockManager();
    }
}