import { TextureAbout, UVOffsetSpeed } from "./base";
import { Constants, INodeMaterialBlock, INodeMaterialTextureInfo, Mat4Arr, Vec2Arr, Vec3Arr, Vec4Arr } from "../../../interface/index";

export class UniformMaskTexture extends TextureAbout implements INodeMaterialBlock {
    public static KEY_TILLOFF: string = Constants.U_MASK_TILLOFF;
    public static KEY_TEX: string = Constants.U_MASK_TEX;

    public static BLOCK_NAME = "MASK_TEXTURE";
    key: string = UniformMaskTexture.BLOCK_NAME;
    bind_defines: Constants.BindingState[] = [Constants.BIND_EFFECT_VALUE];
    depends: string[] = [];
    vs_define: string = "";
    fs_define: string = "";
    mat4: Map<string, Mat4Arr> = new Map();
    vec4: Map<string, Vec4Arr> = new Map();
    vec3: Map<string, Vec3Arr> = new Map();
    vec2: Map<string, Vec2Arr> = new Map();
    float: Map<string, [number, boolean]> = new Map();
    uint: Map<string, [number, boolean]> = new Map();
    texture: Map<string, INodeMaterialTextureInfo> = new Map();
    constructor() {
        super();

        UniformMaskTexture.BLOCK_NAME = "MASK_TEXTURE_UNBUNDLE";
        UniformMaskTexture.TilloffBoundle = false;
        this.key = UniformMaskTexture.BLOCK_NAME;

        // this.vec3.set(UniformMaskTexture.KEY_COLOR, [1, 1, 1, true]);
        this.vec4.set(UniformMaskTexture.KEY_TILLOFF, [1, 1, 0, 0, false]);
        // this.float.set(UniformMaskTexture.KEY_LEVEL, [1, true]);
        this.texture.set(UniformMaskTexture.KEY_TEX, <INodeMaterialTextureInfo>{
            url: "",
            filterable: true,
            stage: Constants.SHADER_STAGE_FRAGMENT,
            default: Constants.DEFAULT_TEXTURE_WHITE,
            dimension: Constants.TextureViewDimensionD2,
        });

        this.fs_define = `
vec4 maskTexture(vec2 vMainUV, vec2 vUVOS, MatParam maParam) {
    return Get${UniformMaskTexture.KEY_TEX}(vMainUV, vUVOS, maParam.${UniformMaskTexture.KEY_TILLOFF});
}
        `;
    }

}

export class UniformMaskTextureUVOffsetSpeed extends UVOffsetSpeed {
    public static KEY_PARAM:    string = Constants.U_MASK_UVOS;
}