import { Engine } from "../../engines/engine";
import { BlockTextureChannel } from "./channel";
import { TextureAbout, UVOffsetSpeed } from "./base";
import { Constants, INodeMaterialBlock, INodeMaterialTextureInfo, Mat4Arr, Vec2Arr, Vec3Arr, Vec4Arr } from "../../../interface/index";
import { p3d } from "../../types";
import { ExtensionTools } from "../../../base";

export class UniformOpacity {
    public static KEY_ALPHA: string = "uOpacity";
    public static alpha(mat: number, val: number) {
        p3d.p3d_material_uniform_float(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_ALPHA), val);
    }
}

export class UniformOpacityTexture extends TextureAbout implements INodeMaterialBlock {
    public static KEY_LEVEL:    string = Constants.U_OPACITY_LEVEL;
    public static KEY_TILLOFF:  string = Constants.U_OPACITY_TILLOFF;
    public static KEY_CHANNEL:  string = Constants.U_OPACITY_CHANNEL;
    public static KEY_TEX:      string = Constants.U_OPACITY_TEX;
    public static level(mat: number, val: number) {
        p3d.p3d_material_uniform_float(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_LEVEL), val);
    }
    public static channel(mat: number, val: BlockTextureChannel) {
        p3d.p3d_material_uniform_uint(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_CHANNEL), val);
    }

    public static KEY_USCALE: string    = `uOpacityUScale`;
    public static KEY_UOFFSET: string   = `uOpacityUOffset`;
    public static KEY_VSCALE: string    = `uOpacityVScale`;
    public static KEY_VOFFSET: string   = `uOpacityVOffset`;

    public static BLOCK_NAME = "OPACITY_TEXTURE";
    key: string = UniformOpacityTexture.BLOCK_NAME;
    bind_defines: Constants.BindingState[] = [Constants.BIND_EFFECT_VALUE];
    depends: string[] = [];
    vs_define: string = "";
    fs_define: string = "";
    mat4: Map<string, Mat4Arr> = new Map();
    vec4: Map<string, Vec4Arr> = new Map();
    vec3: Map<string, Vec3Arr> = new Map();
    vec2: Map<string, Vec2Arr> = new Map();
    float: Map<string, [number, boolean]> = new Map();
    uint: Map<string, [number, boolean]> = new Map();
    texture: Map<string, INodeMaterialTextureInfo> = new Map();
    constructor() {
        super();

        UniformOpacityTexture.BLOCK_NAME = "OPACITY_TEXTURE_UNBUNDLE";
        UniformOpacityTexture.TilloffBoundle = false;
        this.key = UniformOpacityTexture.BLOCK_NAME;

        // this.vec3.set(UniformOpacityTexture.KEY_COLOR, [1, 1, 1, true]);
        this.vec4.set(UniformOpacityTexture.KEY_TILLOFF, [1, 1, 0, 0, false]);
        // this.float.set(UniformOpacityTexture.KEY_VSCALE, [1, false]);
        // this.float.set(UniformOpacityTexture.KEY_UOFFSET, [0, false]);
        // this.float.set(UniformOpacityTexture.KEY_VOFFSET, [0, false]);
        this.float.set(UniformOpacityTexture.KEY_LEVEL, [1, false]);
        this.uint.set(UniformOpacityTexture.KEY_CHANNEL, [BlockTextureChannel.CHANNEL_A, false]);
        this.texture.set(UniformOpacityTexture.KEY_TEX, <INodeMaterialTextureInfo>{
            url: "",
            filterable: true,
            stage: Constants.SHADER_STAGE_FRAGMENT,
            default: Constants.DEFAULT_TEXTURE_WHITE,
            dimension: Constants.TextureViewDimensionD2,
        });

        this.fs_define = `
vec4 opacityTexture(vec2 vMainUV, vec2 vUVOS, MatParam maParam) {
    return Get${UniformOpacityTexture.KEY_TEX}(vMainUV, vUVOS, maParam.${UniformOpacityTexture.KEY_TILLOFF});
}
float opacityChannel(vec4 data, MatParam maParam) {
    return valueByChannel(data, maParam.uOpacityChannel) * maParam.uOpacityLevel;
}
        `;
    }

}

export class UniformOpacityTextureUVOffsetSpeed extends UVOffsetSpeed {
    public static KEY_PARAM:    string = Constants.U_OPACITY_UVOS;
}

export class UniformOpacity2Texture extends TextureAbout {
    public static KEY_LEVEL:    string = Constants.U_OPACITY2_LEVEL;
    public static KEY_TILLOFF:  string = Constants.U_OPACITY2_TILLOFF;
    public static KEY_CHANNEL:  string = Constants.U_OPACITY2_CHANNEL;
    public static KEY_TEX:      string = Constants.U_OPACITY2_TEX;
    public static level(mat: number, val: number) {
        p3d.p3d_material_uniform_float(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_LEVEL), val);
    }
    public static channel(mat: number, val: BlockTextureChannel) {
        p3d.p3d_material_uniform_uint(Engine.commands(), mat, ExtensionTools.getAtom(this.KEY_CHANNEL), val);
    }
}

export class UniformOpacity2TextureUVOffsetSpeed extends UVOffsetSpeed {
    public static KEY_PARAM:    string = Constants.U_OPACITY2_UVOS;
}