import { Engine } from "../engines/engine";
import { MainOpacityOpacityFresnelMaterial } from "./main_opacity_opacity_fresnel";
import { p3d } from "../types";
import { ExtensionTools } from "../../base";
import { Texture } from "./Textures/texture";
import { UniformMainTexture, UniformMainTextureUVOffsetSpeed } from "./node_material/mainTextureBlock";
import { Color3 } from "../../maths/math.color";
import { VARYING_COLOR, VARYING_NORMAL_WS, VARYING_POSITION_WS, VARYING_UV, VARYING_UV2, VARYING_UV3, VARYING_UV4 } from "./node_material";
import { Vector4 } from "../../maths/math.vector";
import { Constants } from "../../interface/index";
import { UniformOpacityTexture } from "./node_material/opacityBlock";
import { UniformEmissiveTexture } from "./node_material/emissiveBlock";
import { FresnelParameters } from "./fresnel";
import { UniformEmissiveFresnel } from "./node_material/fresnelBlock";
import { Material } from "./material";
import { Scene } from "../scene";

class TwoOpacityMix extends MainOpacityOpacityFresnelMaterial {
    private _mixOpacityActive          : number = 1;
    public set mixOpacityActive            (val: boolean) {
        this._mixOpacityActive          = val ? 1 : 0;
        this._mixOpacityInfoDirty           = true;
    }

    private _mixOpacityDirty: boolean = false;
    private _mixOpacityUScale          : number = 1;
    public set mixOpacityUScale            (val: number) {
        this._mixOpacityUScale          = val;
        this._mixOpacityDirty           = true;
    }
    private _mixOpacityVScale          : number = 1;
    public set mixOpacityVScale            (val: number) {
        this._mixOpacityVScale          = val;
        this._mixOpacityDirty           = true;
    }
    private _mixOpacityUOffset         : number = 0;
    public set mixOpacityUOffset           (val: number) {
        this._mixOpacityUOffset         = val;
        this._mixOpacityDirty           = true;
    }
    private _mixOpacityVOffset         : number = 0;
    public set mixOpacityVOffset           (val: number) {
        this._mixOpacityVOffset         = val;
        this._mixOpacityDirty           = true;
    }
    
    private _mixOpacitySpeedDirty: boolean = false;
    private _mixOpacityUSpeed          : number = 0;
    public set mixOpacityUSpeed            (val: number) {
        this._mixOpacityUSpeed          = val;
        this._mixOpacitySpeedDirty      = true;
    }
    private _mixOpacityVSpeed          : number = 0;
    public set mixOpacityVSpeed            (val: number) {
        this._mixOpacityVSpeed          = val;
        this._mixOpacitySpeedDirty      = true;
    }
    
    private _mixOpacityInfoDirty: boolean = false;
    private _mixOpacityIntensity       : number = 0;
    public set mixOpacityIntensity         (val: number) {
        this._mixOpacityIntensity       = val;
        this._mixOpacityInfoDirty       = true;
    }
    
    private _mixOpacityControlDirty: boolean = false;
    private _mixOpacityControlUScale   : number = 1;
    public set mixOpacityControlUScale     (val: number) {
        this._mixOpacityControlUScale   = val;
        this._mixOpacityControlDirty    = true;
    }
    private _mixOpacityControlVScale   : number = 1;
    public set mixOpacityControlVScale     (val: number) {
        this._mixOpacityControlVScale   = val;
        this._mixOpacityControlDirty    = true;
    }
    private _mixOpacityControlUOffset  : number = 0;
    public set mixOpacityControlUOffset    (val: number) {
        this._mixOpacityControlUOffset  = val;
        this._mixOpacityControlDirty    = true;
    }
    private _mixOpacityControlVOffset  : number = 0;
    public set mixOpacityControlVOffset    (val: number) {
        this._mixOpacityControlVOffset  = val;
        this._mixOpacityControlDirty    = true;
    }
    private _mixOpacityControlIntensity: number = 1;
    public set mixOpacityControlIntensity  (val: number) {
        this._mixOpacityControlIntensity= val;
        this._mixOpacityInfoDirty       = true;
    }
    private _mixOpacityControlValue    : number = 0;
    public set mixOpacityControlValue      (val: number) {
        this._mixOpacityControlValue    = val;
        this._mixOpacityInfoDirty       = true;
    }

    _onFrameUpdate(): void {
        if (this._disposed) {
            return;
        }
        super._onFrameUpdate();
        if (this._mixOpacityDirty) {
            p3d.p3d_material_uniform_vec4(Engine.commands(), this.entity, ExtensionTools.getAtom(TwoOpacityMix.KEY_MixOpacityTilloff), this._mixOpacityUScale, this._mixOpacityVScale, this._mixOpacityUOffset, this._mixOpacityVOffset, );
        }
        if (this._mixOpacityInfoDirty) {
            p3d.p3d_material_uniform_vec4(Engine.commands(), this.entity, ExtensionTools.getAtom(TwoOpacityMix.KEY_MixOpacityInfo), this._mixOpacityActive, this._mixOpacityIntensity, this._mixOpacityControlValue, this._mixOpacityControlIntensity);
        }
        if (this._mixOpacityControlDirty) {
            p3d.p3d_material_uniform_vec4(Engine.commands(), this.entity, ExtensionTools.getAtom(TwoOpacityMix.KEY_MixOpacityControlTilloff), this._mixOpacityControlUScale, this._mixOpacityControlVScale, this._mixOpacityControlUOffset, this._mixOpacityControlVOffset, );
        }
        if (this._mixOpacitySpeedDirty) {
            p3d.p3d_material_uniform_vec2(Engine.commands(), this.entity, ExtensionTools.getAtom(TwoOpacityMix.KEY_MixOpacityUVSpeed), this._mixOpacityUSpeed, this._mixOpacityVSpeed);
        }

        this._mixOpacityDirty = false;
        this._mixOpacityInfoDirty = false;
        this._mixOpacityControlDirty = false;
        this._mixOpacitySpeedDirty = false;
    }

    public static KEY_MixOpacityInfo = "uMixOpacityInfo";
    public static KEY_MixOpacityTilloff = "uMixOpacityTilloff";
    public static KEY_MixOpacityControlTilloff = "uMixOpacityControlTilloff";
    public static KEY_MixOpacityUVSpeed = "uMixOpacityUVSpeed";
}

class DistortionUV extends TwoOpacityMix {
    public static KEY_distortionInfo = "distortionInfo";
    public static KEY_control = "control";
    public static KEY_MaskTilloff = "uMaskTilloff";
    public static KEY_MaskTexture = "uMaskTex";
    public static KEY_MaskUVSpeed = "uMaskUVOS";

    private _distortionInfo: Vector4 = new Vector4(0, 0, 0, 0);
    private _control: Vector4 = new Vector4(0, 0, 0, 0);
    public _emissiveColor: Color3 = new Color3(1, 1, 1);

    public get emissiveColor() {
        return this.diffuseColor;
    }
    public set emissiveColor(value: Color3) {
        this.diffuseColor = value;
    }

    private _infoDirty: boolean = true;
    public get distortionX() {
        return this._distortionInfo.x;
    }
    public set distortionX(value: number) {
        this._distortionInfo.x = value;
        this._infoDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }
    public get distortionY() {
        return this._distortionInfo.y;
    }
    public set distortionY(value: number) {
        this._distortionInfo.y = value;
        this._infoDirty = true;
    }
    public get distortionSpeedX() {
        return this._distortionInfo.z;
    }
    public set distortionSpeedX(value: number) {
        this._distortionInfo.z = value;
        this._infoDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }
    public get distortionSpeedY() {
        return this._distortionInfo.w;
    }
    public set distortionSpeedY(value: number) {
        this._distortionInfo.w = value;
        this._infoDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }

    private _controlDirty: boolean = true;
    public get maskFlowMode() {
        return this._control.y;
    }
    public set maskFlowMode(value: number) {
        this._control.y = value;
        this._controlDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }
    public get vertexColorFactor() {
        return this._control.z;
    }
    public set vertexColorFactor(value: number) {
        this._control.z = value;
        this._controlDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }
    public get opacityFromMaskBlue() {
        return this._control.w > 0;
    }
    public set opacityFromMaskBlue(value: boolean) {
        this._control.w = value ? 1 : 0;
        this._controlDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }

    private _maskTexture: Texture;
    private _maskTextureDirty: boolean = false;
    public set maskTexture(val: Texture) {
        if (this._maskTexture != val) {
            this._maskTexture = val;
            this._maskTextureDirty = true;
            if (val) val._material = this;
        }
    }
    public get maskTexture() {
        this._scene.cmdsForceDirty.set(this.entity, this);
        return this._maskTexture;
    }
    
    private _clipModeDirty: boolean = true;
    private _clipByOpacity: boolean = false;
    public set clipByOpacity(val: boolean) {
        this._clipByOpacity = val;
        this._clipModeDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }
    public get clipByOpacity() {
        return this._clipByOpacity;
    }


    // private _emissiveFresnelParameters: FresnelParameters = new FresnelParameters();
    // public set emissiveFresnelParameters(val: FresnelParameters) {
    //     this._emissiveFresnelParameters = val;
    //     (<any>val)._dirty = true;
    //     this._scene.cmdsForceDirty.set(this.entity, this);
    // }
    // public get emissiveFresnelParameters() {
    //     this._scene.cmdsForceDirty.set(this.entity, this);
    //     return this._emissiveFresnelParameters;
    // }

    _onFrameUpdate() {
        if (this._disposed) {
            return;
        }
        super._onFrameUpdate();
        
        if (this._maskTexture) {
            if (this._maskTextureDirty || this._maskTexture._tilloffDirty) {
                p3d.p3d_material_uniform_vec4(
                    Engine.commands(), this.entity, ExtensionTools.getAtom(DistortionUV.KEY_MaskTilloff),
                    this._maskTexture.uScale, this._maskTexture.vScale, this._maskTexture.uOffset, this._maskTexture.vOffset
                );
            }
            if (this._maskTextureDirty || this._maskTexture._offsetSpeedDirty) {
                p3d.p3d_material_uniform_vec2(
                    Engine.commands(), this.entity, ExtensionTools.getAtom(DistortionUV.KEY_MaskUVSpeed),
                    this._maskTexture.offsetSpeedU, this._maskTexture.offsetSpeedV
                );
            }
            if (this._maskTextureDirty) {
                let val = this._maskTexture;
                // console.warn("Text Compressed: " + val.compressed);
                p3d.p3d_material_uniform_tex(
                    Engine.commands(), this.entity, ExtensionTools.getAtom(DistortionUV.KEY_MaskTexture),
                    ExtensionTools.getAtom(val.url), val.srgb, val.compressed, val.filter, val.address_mode_u, val.address_mode_v, val.address_mode_w,
                    val.mag_filter, val.min_filter, val.mipmap_filter, val.anisotropy_clamp, val.border_color, val.isfile, val.cancombine, val.compare
                );
            }
        }

        if (this._controlDirty) {
            p3d.p3d_material_uniform_vec4(
                Engine.commands(), this.entity, ExtensionTools.getAtom(DistortionUV.KEY_control),
                this._control.x, this._control.y, this._control.z, this._control.w
            );
        }

        if (this._infoDirty) {
            p3d.p3d_material_uniform_vec4(
                Engine.commands(), this.entity, ExtensionTools.getAtom(DistortionUV.KEY_distortionInfo),
                this._distortionInfo.x, this._distortionInfo.y, this._distortionInfo.z, this._distortionInfo.w
            );
        }

        if (this._clipModeDirty) {
            p3d.p3d_material_uniform_float(
                Engine.commands(), this.entity, ExtensionTools.getAtom(StandardEffectMaterial.CLIP_BY_OPACITY),
                this._clipByOpacity ? 1. : 0.
            );
        }

        // if (this._emissiveFresnelParameters.dirty) {
        //     this._emissiveFresnelParameters._update();

        //     let leftColor = (<any>this._emissiveFresnelParameters)._leftColor;
        //     let rightColor = (<any>this._emissiveFresnelParameters)._rightColor;

        //     UniformEmissiveFresnel.left(this.entity, leftColor.r, leftColor.g, leftColor.b);
        //     UniformEmissiveFresnel.right(this.entity, rightColor.r, rightColor.g, rightColor.b);
        //     UniformEmissiveFresnel.biasAndPower(this.entity, this._emissiveFresnelParameters.bias, this._emissiveFresnelParameters.power);
        // }

        this._maskTextureDirty = false;
        this._diffuseTextureDirty = false;
        this._diffuseColorDirty = false;
        this._controlDirty = false;
        this._infoDirty = false;
        this._clipModeDirty = false;
    }
}

export class StandardEffectMaterial extends DistortionUV {
    constructor(name: string, scene: Scene, _pass: Constants.PassTag = Constants.PASS_TAG_06) {
        super(name, scene, StandardEffectMaterial.ShaderName, _pass);
    }
    public static ShaderName: string = "StandardEffectMaterial";
    public static CLIP_BY_OPACITY: string = "uClipByOpcity";

    public static regist() {
        if (Engine.CheckShader(StandardEffectMaterial.ShaderName)) {
            return;
        }
        let uniforms = p3d.MaterialUniformDefines.create();
        let includes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockFloat);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockFresnel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorSpace);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockViewDirection);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformMainTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTextureUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformOpacityTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTextureUVOffsetSpeed);
        // // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTexture);
        // p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformEmissiveTexture.BLOCK_NAME));
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTextureUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityFresnel);
        
        p3d.p3d_shader_uniform_vec4(uniforms, TwoOpacityMix.KEY_MixOpacityInfo, 0, 0, 0, 0, false);
        p3d.p3d_shader_uniform_vec4(uniforms, TwoOpacityMix.KEY_MixOpacityTilloff, 1, 1, 0, 0, false);
        p3d.p3d_shader_uniform_vec4(uniforms, TwoOpacityMix.KEY_MixOpacityControlTilloff, 1, 1, 0, 0, false);
        p3d.p3d_shader_uniform_vec2(uniforms, TwoOpacityMix.KEY_MixOpacityUVSpeed, 0, 0, false);
        
        p3d.p3d_shader_uniform_float(uniforms, StandardEffectMaterial.CLIP_BY_OPACITY, 0, false);
        
        p3d.p3d_shader_uniform_vec4(uniforms, DistortionUV.KEY_distortionInfo, 0, 0, 0, 0, false);
        p3d.p3d_shader_uniform_vec4(uniforms, DistortionUV.KEY_control, 0, 0, 0, 0, false);
        p3d.p3d_shader_uniform_tex(uniforms,  DistortionUV.KEY_MaskTexture, true, Constants.SHADER_STAGE_VERTEXFRAGMENT, Constants.DEFAULT_TEXTURE_BLACK, Constants.TextureViewDimensionD2);
        p3d.p3d_shader_uniform_vec4(uniforms, DistortionUV.KEY_MaskTilloff, 1, 1, 0, 0, false);
        p3d.p3d_shader_uniform_vec2(uniforms, DistortionUV.KEY_MaskUVSpeed, 0, 0, false);

        Material.ApplyVertexAbout(uniforms);

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec3", "v_pos"],
            ["vec3", "v_normal"],
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
            ["vec2", "v_uv2"],
            ["vec2", "v_uv3"],
            ["vec2", "v_uv4"],
            ["vec2", "vMixOpacityUV"],
            ["vec2", "vMixOpacityControlUV"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let varying = VARYING_POSITION_WS | VARYING_NORMAL_WS | VARYING_UV | VARYING_COLOR | VARYING_UV2 | VARYING_UV3;

        let binds_defines_base = Constants.BIND_MODEL | Constants.BIND_VIEWER | Constants.BIND_EFFECT_VALUE | Constants.BIND_EFFECT_TEXTURE;
        Engine.RegistShader(
            StandardEffectMaterial.ShaderName,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            mat4 finalWorld = PI_ObjectToWorld;
        
            vec4 position =  vec4(A_POSITION, 1.);
            vec4 worldPos =  finalWorld * position;
            // vec4 worldPos =  position;
        
            gl_Position = PI_MATRIX_VP * worldPos;

            // gl_Position = position;
        
            v_pos = worldPos.xyz;
        
            mat3 normalWorld = mat3(finalWorld);
            v_normal = normalWorld * A_NORMAL;

            float progress  = PI_Time.y; // matParam.${DistortionUV.KEY_control}.x;
            float maskValueFlow = matParam.${DistortionUV.KEY_control}.y;
            float maskUVFlow = 1.0 - maskValueFlow;
            vec2 diffValue = vec2(progress, progress) * matParam.${DistortionUV.KEY_distortionInfo}.zw;

            v_uv2 = (A_UV + vec2(maskUVFlow, maskUVFlow) * diffValue) * matParam.${DistortionUV.KEY_MaskTilloff}.xy + matParam.${DistortionUV.KEY_MaskTilloff}.zw;
            v_uv3 = vec2(maskValueFlow, maskValueFlow) * diffValue;
            v_uv4 = A_UV;

            v_uv    = A_UV;
            v_color = A_COLOR4;
            vMixOpacityUV = A_UV * matParam.${TwoOpacityMix.KEY_MixOpacityTilloff}.xy + matParam.${TwoOpacityMix.KEY_MixOpacityTilloff}.zw + matParam.${TwoOpacityMix.KEY_MixOpacityUVSpeed} * vec2(progress, progress);
            vMixOpacityControlUV = A_UV * matParam.${TwoOpacityMix.KEY_MixOpacityControlTilloff}.xy + matParam.${TwoOpacityMix.KEY_MixOpacityControlTilloff}.zw;
            `,
            `
            vec4 baseColor = v_color;
            float alpha = opacity(matParam);
            alpha   *= v_color.a;

            vec3 diffuseColor 		    = vec3(1., 1., 1.);
            vec3 baseAmbientColor		= vec3(1., 1., 1.);
            
            vec3 normalW 				= normalize(v_normal);
            vec3 viewDirectionW		    = WorldSpaceViewDir(v_pos);
            float NdotVunclamped 		= dot(normalW, viewDirectionW);
            float NdotV					= absEps(NdotVunclamped);
            vec3 V 					    = viewDirectionW;
            vec3 N 					    = normalW;
            float absNdV                = abs(NdotVunclamped);
        
            // ----------------------------------------------------------
            vec3 diffuseBase 	    = vec3(1., 1., 1.);

            vec4 textureMask        = Get${DistortionUV.KEY_MaskTexture}(v_uv2, vec2(0., 0.), vec4(1., 1., 0., 0.));
            vec2 uvDistortion       = (textureMask.rg * 2. - vec2(1., 1.) + v_uv3) * ${DistortionUV.KEY_distortionInfo}.xy;

            vec4 mainTextureColor   = mainTexture(v_uv, applyUVOffsetSpeed(uMainUVOS) + uvDistortion);
            baseColor.rgb           *= mainTextureColor.rgb * mainStrength() * mainColor();
            alpha                   *= mainTextureColor.a;
            
            if (matParam.${DistortionUV.KEY_control}.w > 0.0) {
                baseColor *= Get${DistortionUV.KEY_MaskTexture}(v_uv4, vec2(0., 0.), vec4(1., 1., 0., 0.)).b;
            }
        
            vec4 opacityData        = opacityTexture(v_uv, applyUVOffsetSpeed(matParam.uOpacityUVOS), matParam);
            float opacity           = opacityChannel(opacityData, matParam);
            if (matParam.${StandardEffectMaterial.CLIP_BY_OPACITY} > 0.5) {
                if (cutoff(opacity, matParam)) {
                    discard;
                }
            } else {
                alpha               *= opacity;
            }

            if (matParam.${TwoOpacityMix.KEY_MixOpacityInfo}.x > 0.5) {
                float mixOpacityAlpha           = opacityTexture(vMixOpacityUV, vec2(0., 0.), matParam).r * matParam.${TwoOpacityMix.KEY_MixOpacityInfo}.y;
                float mixOpacityControlValue    = opacityTexture(vMixOpacityControlUV, vec2(0., 0.), matParam).g * matParam.${TwoOpacityMix.KEY_MixOpacityInfo}.w;
                float mixOpacityControlFactor   = smoothstep(mixOpacityControlValue - matParam.${TwoOpacityMix.KEY_MixOpacityInfo}.z * 0.5, mixOpacityControlValue, matParam.${TwoOpacityMix.KEY_MixOpacityInfo}.z);
                alpha 						    = mix(1.0, mixOpacityControlValue * sqrt(alpha * mixOpacityAlpha), mixOpacityControlFactor);
            }
            alpha 					*= opacityFresnel(absNdV, matParam);
        

            vec3 finalDiffuse       = (diffuseBase * diffuseColor) * baseColor.rgb;
            
            vec4 finalColor 		= vec4(
                finalDiffuse
                ,
                alpha
            );
        
            finalColor				= max(finalColor, 0.0);
        
            // Premulty
            // finalColor.rgb          *= finalColor.a;
        
            gl_FragColor = finalColor;        
            `,
            includes,
            ``,
            varyings,
            binds_defines_base
        );
    }
}