import { Material } from "./material";
import { Scene } from "../scene";
import { Engine } from "../engines/engine";
import { VARYING_COLOR, VARYING_NORMAL_WS, VARYING_POSITION_WS, VARYING_UV } from "./node_material";
import { Constants } from "../../interface/index";
import { p3d } from "../types";
import { ExtensionTools } from "../../base";

export class StripesVirtualMaterial extends Material {
    public static regist() {
        let uniforms = p3d.MaterialUniformDefines.create();
        let includes = p3d.NodematerialIncludes.create();

        p3d.p3d_shader_uniform_float(uniforms, "uStep", 1, false);
        p3d.p3d_shader_uniform_float(uniforms, "uSpeed", 1, false);
        p3d.p3d_shader_uniform_float(uniforms, "uFadeStart", 0, false);
        p3d.p3d_shader_uniform_float(uniforms, "uFadeEnd", 1, false);
        
        p3d.p3d_shader_uniform_vec4(uniforms, "uDirection", 0.5, 0.5, 0.5, 1., false);
        p3d.p3d_shader_uniform_vec4(uniforms, "uColor0", 0.1, 0.5, 0.1, 1.0, false);
        p3d.p3d_shader_uniform_vec4(uniforms, "uColor1", 0.1, 1.0, 0.1, 1.0, false);

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec3", "v_pos"],
            ["vec3", "v_normal"],
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let varying = VARYING_POSITION_WS | VARYING_NORMAL_WS | VARYING_UV | VARYING_COLOR;
        let binds_defines_base = Constants.BIND_MODEL | Constants.BIND_VIEWER | Constants.BIND_EFFECT_VALUE | Constants.BIND_EFFECT_TEXTURE;

        Engine.RegistShader(
            "StripesVirtual",
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            mat4 finalWorld = PI_ObjectToWorld;
        
            vec4 position =  vec4(A_POSITION, 1.);
            vec4 worldPos =  finalWorld * position;
            // vec4 worldPos =  position;
        
            gl_Position = PI_MATRIX_VP * worldPos;

            // gl_Position = position;
        
            v_pos = worldPos.xyz;
        
            mat3 normalWorld = mat3(finalWorld);
            v_normal = A_NORMAL;
            
            v_uv = A_UV;
            v_color = A_COLOR4;
            `,
            `
            float temp          = dot(v_pos, matParam.uDirection.xyz);
            temp                = fract(temp / matParam.uStep + matParam.uSpeed * PI_Time.y);
            temp                = abs((temp - 0.5) * 2.);
            temp                = smoothstep(matParam.uFadeStart, matParam.uFadeEnd, temp);
            
            vec4 finalColor     = mix(matParam.uColor0, matParam.uColor1, temp);
            
            gl_FragColor = finalColor;
            `,
            includes,
            ``,
            varyings,
            binds_defines_base
        );
    }
    constructor(name: string, scene: Scene, _pass: Constants.PassTag = Constants.PASS_TAG_06) {
        super(name, scene, _pass);
        Material.modify(this, "StripesVirtual");
    }
}