import { Engine } from "../engines/engine";
import { Material } from "./material";
import { UniformMainTexture } from "./node_material/mainTextureBlock";
import { UniformEmissiveTexture } from "./node_material/emissiveBlock";
import { VARYING_COLOR, VARYING_NORMAL_WS, VARYING_POSITION_WS, VARYING_UV } from "./node_material";
import { UniformOpacityTexture } from "./node_material/opacityBlock";
import { MainOpacityMaterial } from "./main_opacity";
import { Constants } from "../../interface/index";
import { p3d } from "../types";
import { ExtensionTools } from "../../base";

export class TwoOpacityMixMaterial extends MainOpacityMaterial {
    private _mixOpacityDirty: boolean = false;
    private _mixOpacityUScale          : number = 1;
    public set mixOpacityUScale            (val: number) {
        this._mixOpacityUScale          = val;
        this._mixOpacityDirty           = true;
    }
    private _mixOpacityVScale          : number = 1;
    public set mixOpacityVScale            (val: number) {
        this._mixOpacityVScale          = val;
        this._mixOpacityDirty           = true;
    }
    private _mixOpacityUOffset         : number = 0;
    public set mixOpacityUOffset           (val: number) {
        this._mixOpacityUOffset         = val;
        this._mixOpacityDirty           = true;
    }
    private _mixOpacityVOffset         : number = 0;
    public set mixOpacityVOffset           (val: number) {
        this._mixOpacityVOffset         = val;
        this._mixOpacityDirty           = true;
    }
    
    private _mixOpacitySpeedDirty: boolean = false;
    private _mixOpacityUSpeed          : number = 0;
    public set mixOpacityUSpeed            (val: number) {
        this._mixOpacityUSpeed          = val;
        this._mixOpacitySpeedDirty      = true;
    }
    private _mixOpacityVSpeed          : number = 0;
    public set mixOpacityVSpeed            (val: number) {
        this._mixOpacityVSpeed          = val;
        this._mixOpacitySpeedDirty      = true;
    }
    
    private _mixOpacityInfoDirty: boolean = false;
    private _mixOpacityIntensity       : number = 0;
    public set mixOpacityIntensity         (val: number) {
        this._mixOpacityIntensity       = val;
        this._mixOpacityInfoDirty       = true;
    }
    
    private _mixOpacityControlDirty: boolean = false;
    private _mixOpacityControlUScale   : number = 1;
    public set mixOpacityControlUScale     (val: number) {
        this._mixOpacityControlUScale   = val;
        this._mixOpacityControlDirty    = true;
    }
    private _mixOpacityControlVScale   : number = 1;
    public set mixOpacityControlVScale     (val: number) {
        this._mixOpacityControlVScale   = val;
        this._mixOpacityControlDirty    = true;
    }
    private _mixOpacityControlUOffset  : number = 0;
    public set mixOpacityControlUOffset    (val: number) {
        this._mixOpacityControlUOffset  = val;
        this._mixOpacityControlDirty    = true;
    }
    private _mixOpacityControlVOffset  : number = 0;
    public set mixOpacityControlVOffset    (val: number) {
        this._mixOpacityControlVOffset  = val;
        this._mixOpacityControlDirty    = true;
    }
    private _mixOpacityControlIntensity: number = 1;
    public set mixOpacityControlIntensity  (val: number) {
        this._mixOpacityControlIntensity= val;
        this._mixOpacityInfoDirty       = true;
    }
    private _mixOpacityControlValue    : number = 0;
    public set mixOpacityControlValue      (val: number) {
        this._mixOpacityControlValue    = val;
        this._mixOpacityInfoDirty       = true;
    }

    _onFrameUpdate(): void {
        if (this._disposed) {
            return;
        }
        super._onFrameUpdate();
        if (this._mixOpacityDirty) {
            p3d.p3d_material_uniform_vec4(Engine.commands(), this.entity, ExtensionTools.getAtom(TwoOpacityMixMaterial.KEY_MixOpacityTilloff), this._mixOpacityUScale, this._mixOpacityVScale, this._mixOpacityUOffset, this._mixOpacityVOffset, );
        }
        if (this._mixOpacityInfoDirty) {
            p3d.p3d_material_uniform_vec4(Engine.commands(), this.entity, ExtensionTools.getAtom(TwoOpacityMixMaterial.KEY_MixOpacityInfo), 0., this._mixOpacityIntensity, this._mixOpacityControlValue, this._mixOpacityControlIntensity);
        }
        if (this._mixOpacityControlDirty) {
            p3d.p3d_material_uniform_vec4(Engine.commands(), this.entity, ExtensionTools.getAtom(TwoOpacityMixMaterial.KEY_MixOpacityControlTilloff), this._mixOpacityControlUScale, this._mixOpacityControlVScale, this._mixOpacityControlUOffset, this._mixOpacityControlVOffset, );
        }
        if (this._mixOpacitySpeedDirty) {
            p3d.p3d_material_uniform_vec2(Engine.commands(), this.entity, ExtensionTools.getAtom(TwoOpacityMixMaterial.KEY_MixOpacityUVSpeed), this._mixOpacityUSpeed, this._mixOpacityVSpeed);
        }

        this._mixOpacityDirty = false;
        this._mixOpacityInfoDirty = false;
        this._mixOpacityControlDirty = false;
        this._mixOpacitySpeedDirty = false;
    }

    public static KEY_MixOpacityInfo = "uMixOpacityInfo";
    public static KEY_MixOpacityTilloff = "uMixOpacityTilloff";
    public static KEY_MixOpacityControlTilloff = "uMixOpacityControlTilloff";
    public static KEY_MixOpacityUVSpeed = "uMixOpacityUVSpeed";
    public static regist() {
        if (Engine.CheckShader(MainOpacityMaterial.ShaderName)) {
            return;
        }
        let uniforms = p3d.MaterialUniformDefines.create();
        let includes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformMainTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTextureUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformOpacityTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTextureUVOffsetSpeed);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformEmissiveTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTextureUVOffsetSpeed);

        p3d.p3d_shader_uniform_vec4(uniforms, TwoOpacityMixMaterial.KEY_MixOpacityInfo, 0, 0, 0, 1, false);
        p3d.p3d_shader_uniform_vec4(uniforms, TwoOpacityMixMaterial.KEY_MixOpacityTilloff, 1, 1, 0, 0, false);
        p3d.p3d_shader_uniform_vec4(uniforms, TwoOpacityMixMaterial.KEY_MixOpacityControlTilloff, 1, 1, 0, 0, false);
        p3d.p3d_shader_uniform_vec2(uniforms, TwoOpacityMixMaterial.KEY_MixOpacityUVSpeed, 0, 0, false);

        Material.ApplyVertexAbout(uniforms);

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec3", "v_pos"],
            ["vec3", "v_normal"],
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
            ["vec2", "v_uv2"],
            ["vec2", "v_uv3"],
            ["vec2", "vMixOpacityUV"],
            ["vec2", "vMixOpacityControlUV"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });

        let varying = VARYING_POSITION_WS | VARYING_NORMAL_WS | VARYING_UV | VARYING_COLOR;
        let binds_defines_base = Constants.BIND_MODEL | Constants.BIND_VIEWER | Constants.BIND_EFFECT_VALUE | Constants.BIND_EFFECT_TEXTURE;

        Engine.RegistShader(
            MainOpacityMaterial.ShaderName,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            mat4 finalWorld = PI_ObjectToWorld;
        
            vec4 position =  vec4(A_POSITION, 1.);
            vec4 worldPos =  finalWorld * position;
            // vec4 worldPos =  position;
        
            gl_Position = PI_MATRIX_VP * worldPos;

            // gl_Position = position;
        
            v_pos = worldPos.xyz;
        
            mat3 normalWorld = mat3(finalWorld);
            v_normal = A_NORMAL;
            
            v_uv    = A_UV;
            v_uv2   = A_UV;
            v_uv3   = A_UV;
            v_color = A_COLOR4;
            vMixOpacityUV = A_UV * matParam.${TwoOpacityMixMaterial.KEY_MixOpacityTilloff}.xy + matParam.${TwoOpacityMixMaterial.KEY_MixOpacityTilloff}.zw + matParam.${TwoOpacityMixMaterial.KEY_MixOpacityUVSpeed} * PI_Time.y;
            vMixOpacityControlUV = A_UV * matParam.${TwoOpacityMixMaterial.KEY_MixOpacityControlTilloff}.xy + matParam.${TwoOpacityMixMaterial.KEY_MixOpacityControlTilloff}.zw;
            `,
            `
            vec4 baseColor = vec4(1., 1., 1., 1.);
            if (matParam.${Material.KEY_VERTEX_ABOUT} & matParam.${Material.MODE_VERTEX_COLOR} > 0) {
                baseColor.rgb = v_color.rgb;
            }
            float alpha = opacity(matParam);
            if (matParam.${Material.KEY_VERTEX_ABOUT} & matParam.${Material.MODE_VERTEX_ALPHA} > 0) {
                alpha *= v_color.a;
            }
        
            float glossiness 			= 0.;
            vec3 specularColor 		    = vec3(0., 0., 0.);
            vec3 diffuseColor 		    = vec3(0., 0., 0.);
            vec3 baseAmbientColor		= vec3(1., 1., 1.);
            vec4 refractionColor		= vec4(0., 0., 0., 1.);
            vec4 reflectionColor 		= vec4(0., 0., 0., 1.);
        
            // ----------------------------------------------------------
            vec3 diffuseBase 					= vec3(0., 0., 0.);
            vec3 specularBase 					= vec3(0., 0., 0.);
        
            vec4 mainTextureColor   = mainTexture(v_uv, applyUVOffsetSpeed(matParam.uMainUVOS), matParam);
            baseColor.rgb           *= mainTextureColor.rgb * mainStrength(matParam);
            alpha                   *= mainTextureColor.a;
        
            diffuseColor            = mainColor(matParam);
        
            vec4 opacityData        = opacityTexture(v_uv2, applyUVOffsetSpeed(matParam.uOpacityUVOS), matParam);
            alpha                   *= opacityChannel(opacityData, matParam);
            
            float mixOpacityAlpha   = opacityTexture(vMixOpacityUV, vec2(0., 0.), matParam).r * matParam.${TwoOpacityMixMaterial.KEY_MixOpacityInfo}.y;

            float mixOpacityControlValue = opacityTexture(vMixOpacityControlUV, vec2(0., 0.), matParam).g * matParam.${TwoOpacityMixMaterial.KEY_MixOpacityInfo}.w;
            float mixOpacityControlFactor = smoothstep(mixOpacityControlValue - matParam.${TwoOpacityMixMaterial.KEY_MixOpacityInfo}.z * 0.5, mixOpacityControlValue, matParam.${TwoOpacityMixMaterial.KEY_MixOpacityInfo}.z);
            alpha 						= mix(1.0, mixOpacityControlValue * sqrt(alpha * mixOpacityAlpha), mixOpacityControlFactor);
        
            if (cutoff(alpha, matParam)) {
                discard;
            }
        
            vec3 emissiveColor = emissiveColor(matParam);
            emissiveColor *= emissiveTexture(v_uv3, applyUVOffsetSpeed(matParam.uEmissiveUVOS), matParam).rgb * emissiveStrength(matParam);
        
            vec3 finalSpecular 		= specularBase * specularColor;
            vec3 finalDiffuse       = (diffuseBase * diffuseColor + emissiveColor) * baseColor.rgb;
            
            vec4 finalColor 		= vec4(
                finalDiffuse * baseAmbientColor
                + 
                finalSpecular
                +
                refractionColor.rgb
                + 
                reflectionColor.rgb
                ,
                alpha
            );
        
            finalColor				= max(finalColor, 0.0);
        
            // Premulty
            // finalColor.rgb          *= finalColor.a;
        
            gl_FragColor = finalColor;        
            `,
            includes,
            ``,
            varyings,
            binds_defines_base
        );
    }
}