
import { Engine } from "../engines/engine";
import { Scene } from "../scene";
import { Mesh } from "./mesh";
import * as GLTF2 from "../../gltf2/glTF2Interface";
import { VertexBuffer } from "./buffer";
import { Constants, IIndiceBuffer, IVertexBuffer, GeometryDesc } from "../../interface/index";
import { p3d } from "../types";
import { IGeometryMeta } from "../../context";

export class Geometry {
    private static VerticeBufferCounter: number = 0;
    public static VerticeBufferIDs: Map<string, number> = new Map();
    public static GetVerticeBufferID(name: string, force: boolean = false): number {
        let id = 0;
        if (force) {
            this.VerticeBufferCounter++;
            id = this.VerticeBufferCounter;
            this.VerticeBufferIDs.set(name, id);
            return id;
        } else {
            id = this.VerticeBufferIDs.get(name);
            if (id == undefined) {
                this.VerticeBufferCounter++;
                id = this.VerticeBufferCounter;
                this.VerticeBufferIDs.set(name, id);
            }
            return id;
        }
    }
    public static GetIndiceBufferID(name: string, force: boolean = false): number {
        return this.GetVerticeBufferID(`I#${name}`, force);
    }
    public static query_indices_buffer(key: string): boolean {
        // if (p3d.ve)
        return false;
    }
    public static query_vertex_buffer(key: string): boolean {
        // if (p3d.ve)
        return false;
    }
    public static create_indices_buffer(key: string, data: Uint16Array) {
        Engine.CreateIndiceBuffer(key, data);
    }
    public static create_vertex_buffer(key: string, data: Float32Array) {
        Engine.CreateVertexBuffer(key, data);
    }

    public static vertex_format(accessor: GLTF2.IAccessor): Constants.VertexFormat {
        
        if (accessor.componentType == GLTF2.AccessorComponentType.FLOAT) {
            if (accessor.type == GLTF2.AccessorType.VEC2) {
                return Constants.VERTEX_FORMAT_F32X2;
            }
            else if (accessor.type == GLTF2.AccessorType.VEC3) {
                return Constants.VERTEX_FORMAT_F32X3;
            }
            else if (accessor.type == GLTF2.AccessorType.VEC4) {
                return Constants.VERTEX_FORMAT_F32X4;
            }
            else {
                throw new Error("Not Supoort AccessorType: " + accessor.type);
            }
        }
        else if (accessor.componentType == GLTF2.AccessorComponentType.UNSIGNED_SHORT) {
            if (accessor.type == GLTF2.AccessorType.VEC2) {
                return Constants.VERTEX_FORMAT_U16X2;
            }
            else if (accessor.type == GLTF2.AccessorType.VEC4) {
                return Constants.VERTEX_FORMAT_U16X4;
            }
            else {
                throw new Error("Not Supoort AccessorType: " + accessor.type);
            }
        }
        else if (accessor.componentType == GLTF2.AccessorComponentType.UNSIGNED_INT) {
            if (accessor.type == GLTF2.AccessorType.VEC2) {
                return Constants.VERTEX_FORMAT_U32X2;
            }
            else if (accessor.type == GLTF2.AccessorType.VEC3) {
                return Constants.VERTEX_FORMAT_U32X3;
            }
            else if (accessor.type == GLTF2.AccessorType.VEC4) {
                return Constants.VERTEX_FORMAT_U32X4;
            }
            else {
                throw new Error("Not Supoort AccessorType: " + accessor.type);
            }
        }
        else if (accessor.componentType == GLTF2.AccessorComponentType.SHORT) {
            if (accessor.type == GLTF2.AccessorType.VEC2) {
                return Constants.VERTEX_FORMAT_S16X2;
            }
            else if (accessor.type == GLTF2.AccessorType.VEC3) {
                throw new Error("Not Supoort AccessorType: " + accessor.type);
            }
            else if (accessor.type == GLTF2.AccessorType.VEC4) {
                return Constants.VERTEX_FORMAT_S16X4;
            }
            else {
                throw new Error("Not Supoort AccessorType: " + accessor.type);
            }
        }
        else if (accessor.componentType == GLTF2.AccessorComponentType.BYTE) {
            if (accessor.type == GLTF2.AccessorType.VEC4) {
                return Constants.VERTEX_FORMAT_S8X4;
            }
            else {
                throw new Error("Not Supoort AccessorType: " + accessor.type);
            }
        }
        else if (accessor.componentType == GLTF2.AccessorComponentType.UNSIGNED_BYTE) {
            if (accessor.type == GLTF2.AccessorType.VEC4) {
                return Constants.VERTEX_FORMAT_U8X4;
            }
            else {
                throw new Error("Not Supoort AccessorType: " + accessor.type);
            }
        } else {
            throw new Error("Not Supoort AccessorType: " + accessor.type);
        }
    }
    
    public static GetAttributeElements(accessor: GLTF2.IAccessor) {
        if (accessor.type == GLTF2.AccessorType.VEC2) {
            return 2;
        }
        else if (accessor.type == GLTF2.AccessorType.VEC3) {
            return 3;
        }
        else if (accessor.type == GLTF2.AccessorType.VEC4) {
            return 4;
        }
        else if (accessor.type == GLTF2.AccessorType.SCALAR) {
            return 1;
        }
        else {
            throw new Error("Not Supoort AccessorType: " + accessor.type);
        }
    }

    public static vertex_kind(attr: string): Constants.VertexAttr {
        if (attr == VertexBuffer.POSITION) {
            return Constants.VERTEX_ATTR_POSITION;
        }
        else if (attr == VertexBuffer.COLOR_0 || attr == VertexBuffer.COLOR_00) {
            return Constants.VERTEX_ATTR_COLOR4;
        }
        else if (attr == VertexBuffer.NORMAL) {
            return Constants.VERTEX_ATTR_NORMAL;
        }
        else if (attr == VertexBuffer.TEXCOORD_0) {
            return Constants.VERTEX_ATTR_UV;
        }
        else if (attr == VertexBuffer.TEXCOORD_1) {
            return Constants.VERTEX_ATTR_UV2;
        }
        else if (attr == VertexBuffer.TANGENT) {
            return Constants.VERTEX_ATTR_TANGENT;
        }
        else if (attr == VertexBuffer.JOINTS_0) {
            return Constants.VERTEX_ATTR_MATRICESINDICES;
        }
        else if (attr == VertexBuffer.JOINTS_1) {
            return Constants.VERTEX_ATTR_MATRICESINDICESEXTRA;
        }
        else if (attr == VertexBuffer.WEIGHTS_0) {
            return Constants.VERTEX_ATTR_MATRICESWEIGHTS;
        }
        else if (attr == VertexBuffer.WEIGHTS_1) {
            return Constants.VERTEX_ATTR_MATRICESWEIGHTSEXTRA;
        }
        else {
            throw new Error("Not Supoort Attr: " + attr);
        }
    }
    
    public static morph_vertex_kind(attr: string, idx: number): Constants.VertexAttr {
        if (attr == VertexBuffer.POSITION) {
            if (idx == 0) {
                return Constants.VERTEX_ATTR_POSITION_M;
            }
            else if (idx == 1) {
                return Constants.VERTEX_ATTR_POSITION_M1;
            }
            else if (idx == 2) {
                return Constants.VERTEX_ATTR_POSITION_M2;
            }
            else if (idx == 3) {
                return Constants.VERTEX_ATTR_POSITION_M3;
            }
        }
        else if (attr == VertexBuffer.COLOR_0 || attr == VertexBuffer.COLOR_00) {
            if (idx == 0) {
                return Constants.VERTEX_ATTR_COLOR4_M;
            }
            else if (idx == 1) {
                return Constants.VERTEX_ATTR_COLOR4_M1;
            }
            else if (idx == 2) {
                return Constants.VERTEX_ATTR_COLOR4_M2;
            }
            else if (idx == 3) {
                return Constants.VERTEX_ATTR_COLOR4_M3;
            }
        }
        else if (attr == VertexBuffer.NORMAL) {
            if (idx == 0) {
                return Constants.VERTEX_ATTR_NORMAL_M;
            }
            else if (idx == 1) {
                return Constants.VERTEX_ATTR_NORMAL_M1;
            }
            else if (idx == 2) {
                return Constants.VERTEX_ATTR_NORMAL_M2;
            }
            else if (idx == 3) {
                return Constants.VERTEX_ATTR_NORMAL_M3;
            }
        }
        else if (attr == VertexBuffer.TEXCOORD_0) {
            if (idx == 0) {
                return Constants.VERTEX_ATTR_UV_M;
            }
            else if (idx == 1) {
                return Constants.VERTEX_ATTR_UV_M1;
            }
            else if (idx == 2) {
                return Constants.VERTEX_ATTR_UV_M2;
            }
            else if (idx == 3) {
                return Constants.VERTEX_ATTR_UV_M3;
            }
        }
        throw new Error("Not Supoort Attr: " + attr);
    }

    private _desc: GeometryDesc = { vertices: [], indices: null };
    constructor(protected _id: string, protected _scene: Scene) {
        
    }
    setVerticesBuffer(vb: IVertexBuffer) {
        this._desc.vertices.push(vb);
    }
    // setInstanceBuffer(insb: EInstanceKind) {
    //     this._desc.instance.push(insb);
    // }
    setIndices(ib: IIndiceBuffer) {
        this._desc.indices = ib;
    }
    applyToMesh(mesh: Mesh, meta?: IGeometryMeta) {
        if (!meta) {
            meta = p3d.GeometryMeta.create_empty();
        }

        this._desc.vertices.forEach((vb) => {
            // let name = getAtom(vb.name);
            let vbmeta = p3d.VBMeta.create(vb.name, vb.range ? vb.range[0] : null, vb.range ? vb.range[1] : null);
            vb.attrs.forEach((attr) => {
                p3d.p3d_vbmeta_attr(vbmeta, attr.kind, attr.format);
            });
            p3d.p3d_geo_set_vertex(meta, vbmeta);

            // name.free();
            vbmeta.free && vbmeta.free();
        });

        if (this._desc.indices) {
            let range = this._desc.indices.range;
            let name = this._desc.indices.name;
            p3d.p3d_geo_set_indice(
                meta,
                name,
                range ? range[0] : null,
                range ? range[1] : null,
                this._desc.indices.as_u16
            );
            // name.free();
        }

        (<any>mesh).__geometryID = p3d.p3d_mesh_geometry(Engine.native(), Engine.commands(), mesh.entity, meta);

        meta.free && meta.free();
    }
}
