

import { Engine } from "../engines/engine";
import { AbstractMesh } from "./abstractMesh";
import { Constants, ExtensionTools } from "../../interface/index";
import { Scene } from "../scene";
import { IInstanceSourceD3, IInstancedMeshD3, p3d } from "../types";
import { N4Pool } from "../engines/datapool";
import { IAtom } from "context";

export class InstancedMesh extends AbstractMesh implements IInstancedMeshD3 {
    public static INSTANCE_INDEX           = 0b0000_0000_0000_0000_0000_0000_0000_0001;
    public static INSTANCE_BASE            = 0b0000_0000_0000_0000_0000_0000_0000_0010;
    public static INSTANCE_COLOR           = 0b0000_0000_0000_0000_0000_0000_0000_0100;
    public static INSTANCE_TILL_OFF_1      = 0b0000_0000_0000_0000_0000_0000_0000_1000;
    public static INSTANCE_SKIN            = 0b0000_0000_0000_0000_0000_0000_0001_0000;
    public static INSTANCE_CUSTOM_VEC4_A   = 0b0000_0000_0000_0000_0000_0000_0010_0000;
    public static INSTANCE_CUSTOM_VEC4_B   = 0b0000_0000_0000_0000_0000_0000_0100_0000;
    public static INSTANCE_CUSTOM_UVEC4_A  = 0b0000_0000_0000_0000_0000_0000_1000_0000;
    public static INSTANCE_CUSTOM_IVEC4_B  = 0b0000_0000_0000_0000_0000_0001_0000_0000;
    public static INSTANCE_TILL_OFF_2      = 0b0000_0000_0000_0000_0000_0010_0000_0000;
    public static INSTANCE_VELOCITY        = 0b0000_0000_0000_0000_0000_0100_0000_0000;
    public static InstanceAttrAtoms: Map<string, IAtom> = new Map();

    protected _sourceMesh: IInstanceSourceD3;
    constructor(name: string, source: IInstanceSourceD3) {
        let entity = source._cacheInstances.pop();
        if (entity == undefined) {
            entity = p3d.p3d_instance_mesh(Engine.native(), Engine.commands(), (<any>source).entity);
        } else {
            p3d.p3d_node_enable(Engine.commands(), entity, true);
            p3d.p3d_local_scaling(Engine.commands(), entity, 1,1,1);
            p3d.p3d_local_position(Engine.commands(), entity, 0,0,0);
            p3d.p3d_local_quaternion(Engine.commands(), entity, 0,0,0,1);
        }
        super(name, <Scene>source._scene, entity);
        this._sourceMesh = source;
        this._sourceMesh._instances.set(this.entity, this);
        // this._instancedBuffer = new InstancedBuffer(this);
        // TODO
    }

    get sourceMesh() {
        return this._sourceMesh;
    }

    get boundingBox() {
        return this.sourceMesh.boundingBox;
    }

    private _instancedBuffer: Map<string, [number, number, number, number]> = new Map();
    public get instancedBuffers() {
        return this._instancedBuffer;
    }

    private _instanceValueDirty: Set<string> = new Set();

    public _applyInstanceAttribute() {
        if (this._isDisposed) return;
        if (this._instanceValueDirty.size > 0) {
            this._instanceValueDirty.forEach(key => {
                let state = this.sourceMesh._instancestates.get(key);
                let val = this._instancedBuffer.get(key);
                if (state) {
                    let atom = InstancedMesh.InstanceAttrAtoms.get(key);
                    if (state.vtype == Constants.INSTANCE_ATTR_FLOAT) {
                        p3d.p3d_instance_mesh_float(Engine.commands(), this._entity, val[0], atom);
                    }
                    else if (state.vtype == Constants.INSTANCE_ATTR_VEC2) {
                        p3d.p3d_instance_mesh_vec2(Engine.commands(), this._entity, val[0], val[1], atom);
                    }
                    else if (state.vtype == Constants.INSTANCE_ATTR_VEC3) {
                        p3d.p3d_instance_mesh_vec3(Engine.commands(), this._entity, val[0], val[1], val[2], atom);
                    }
                    else if (state.vtype == Constants.INSTANCE_ATTR_VEC4) {
                        p3d.p3d_instance_mesh_vec4(Engine.commands(), this._entity, val[0], val[1], val[2], val[3], atom);
                    }
                    else if (state.vtype == Constants.INSTANCE_ATTR_SINT) {
                        p3d.p3d_instance_mesh_sint(Engine.commands(), this._entity, val[0], atom);
                    }
                    else if (state.vtype == Constants.INSTANCE_ATTR_UINT) {
                        p3d.p3d_instance_mesh_uint(Engine.commands(), this._entity, val[0], atom);
                    }
                    else if (state.vtype == Constants.INSTANCE_ATTR_U16X2) {
                        p3d.p3d_instance_mesh_u16x2(Engine.commands(), this._entity, val[0], val[1], atom);
                    }
                    else if (state.vtype == Constants.INSTANCE_ATTR_U16X4) {
                        p3d.p3d_instance_mesh_u16x4(Engine.commands(), this._entity, val[0], val[1], val[2], val[3], atom);
                    }
                    else if (state.vtype == Constants.INSTANCE_ATTR_UNORM16X2) {
                        p3d.p3d_instance_mesh_u16x2(Engine.commands(), this._entity, val[0], val[1], atom);
                    }
                    else if (state.vtype == Constants.INSTANCE_ATTR_UNORM16X4) {
                        p3d.p3d_instance_mesh_u16x4(Engine.commands(), this._entity, val[0], val[1], val[2], val[3], atom);
                    }
                    else if (state.vtype == Constants.INSTANCE_ATTR_UNORM8X4) {
                        p3d.p3d_instance_mesh_u8x4(Engine.commands(), this._entity, val[0], val[1], val[2], val[3], atom);
                    }
                    else if (state.vtype == Constants.INSTANCE_ATTR_IVEC4) {
                        p3d.p3d_instance_mesh_ivec4(Engine.commands(), this._entity, val[0], val[1], val[2], val[3], atom);
                    }
                }
            });
        }
        this._instanceValueDirty.clear();
    }

    instanceAttribute(key: string, a: number, b?: number, c?: number, d?: number) {
        let state = this.sourceMesh._instancestates.get(key);
        if (state) {
            if (InstancedMesh.InstanceAttrAtoms.has(key) == false) {
                InstancedMesh.InstanceAttrAtoms.set(key, ExtensionTools.getAtom(key));
            }
            let val = this._instancedBuffer.get(key);
            if (val == undefined) {
                val = N4Pool.request(a, b, c, d);
            } else {
                val[0] = a; val[1] = b; val[2] = c; val[3] = d;
            }
            this._instancedBuffer.set(key, val);
            this._instanceValueDirty.add(key);
            this._scene.cmdsMeshInstanceAttr.set(this._entity, this);
            // this._applyInstanceAttribute();
        }
    }

    public _isUsing: boolean = true;
    dispose(): void {
        this._isUsing = false;
        this.parent = null;
        this.setEnabled(false);
        // this.setPosition(0, 0, 0);
        // this.setScaling(1, 1, 1);
        // this.setRotation(0, 0, 0);

        this._instancedBuffer.forEach(function(val) {
            N4Pool.recycle(val);
        });
        this._instancedBuffer.clear();
        this._sourceMesh._instances.delete(this.entity);
        this._sourceMesh._cacheInstances.push(this.entity);
        this._instanceValueDirty.clear();

        super.dispose(false);
    }
}