import { Engine } from "../engines/engine";
import { Scene } from "../scene";
import { Transform } from "../node";
import { Quaternion, Space, Vector3 } from "../../maths/index";
import { ITransformNodeD3, p3d } from "../types";

export class TransformNode extends Transform implements ITransformNodeD3 {
    public static _rotationAxisCache: Quaternion = new Quaternion();
    metadata: any = {};

    constructor(name: string, scene: Scene) {
        let entity = scene.cacheTransformNodes.pop();
        if (entity == undefined) {
            entity = p3d.p3d_transform_node(Engine.native(), Engine.commands(), scene.entity);
        } else {
            p3d.p3d_node_enable(Engine.commands(), entity, true);
            p3d.p3d_local_scaling(Engine.commands(), entity, 1,1,1);
            p3d.p3d_local_position(Engine.commands(), entity, 0,0,0);
            p3d.p3d_local_quaternion(Engine.commands(), entity, 0,0,0,1);
        }
        super(name, scene, entity);
    }

    forceDirty() {
        this._scene?.cmdsForceDirty.set(this._entity, this);
    }

    dispose(): void {
        if (this._disposed) {
            return;
        }
        let entity = this.entity;
        let scene = this._scene.entity;

        super.dispose(false);
        this._scene.cacheTransformNodes.push(entity);
        // p3d.p3d_local_scaling(Engine.commands(), entity, 1,1,1);
        // p3d.p3d_local_position(Engine.commands(), entity, 0,0,0);
        // p3d.p3d_local_quaternion(Engine.commands(), entity, 0,0,0,1);
        p3d.p3d_node_enable(Engine.commands(), entity, false);
        p3d.p3d_transform_node_parent(Engine.commands(), entity, scene);
    }
}
