import { Engine, Entity } from "../engines/engine";
import { Light } from "../lights";
import { Scene } from "../scene";
import { ICustomRenderTarget, IRenderer, IShadowGenerator, TAutoRenderTarget, Constants, PIWebGLFramebuffer } from "../../interface/index";
import { IRendererD3, p3d } from "../types";

export class ShadowGenerator extends Entity implements IShadowGenerator, IRendererD3  {
    private _atlasSize: number;
    private _minZ: number;
    private _maxZ: number;
    private _frustumSize: number;
    private _bias: number;
    private _normalBias: number;
    private _depthScale: number;
    public readonly pass: Constants.PassTag;
    constructor(
        public readonly _scene: Scene,
        public light: Light,
        pass: Constants.PassTag,
        bias: number = 0.005,
        normalBias: number = 0.005,
        depthScale: number = 10,
        minZ: number = 0.1,
        maxZ: number = 100,
        frustumSize: number = 10,
    ) {
        let entity = p3d.p3d_shadow_generator(Engine.native(), Engine.commands(), _scene.entity, light.entity, pass);
        super(entity);

        this.bias = bias;
        this.normalBias = normalBias;
        this.minZ = minZ;
        this.maxZ = maxZ;
        this.frustumSize = frustumSize;
        this.depthScale = depthScale;
        this.pass = pass;
    }
    name?: string;
    _onFrameUpdate(): void {
        throw new Error("Method not implemented.");
    }

    autoClearColor(flag: boolean): void {
        p3d.p3d_render_auto_clear_color(Engine.commands(), this.entity, flag);
    }
    autoClearDepth(flag: boolean): void {
        p3d.p3d_render_auto_clear_depth(Engine.commands(), this.entity, flag);
    }
    autoClearStencil(flag: boolean): void {
        p3d.p3d_render_auto_clear_stencil(Engine.commands(), this.entity, flag);
    }
    clearColor(r: number, g: number, b: number, a: number): void {
        p3d.p3d_render_clear_color(Engine.commands(), this.entity, r, g, b, a);
    }
    clearDepth(depth: number): void {
        p3d.p3d_render_clear_depth(Engine.commands(), this.entity, depth);
    }
    clearStencil(value: number): void {
        p3d.p3d_render_clear_stencil(Engine.commands(), this.entity, value);
    }
    viewport(x: number, y: number, width: number, height: number): void {
        //
    }
    setRenderTarget(target: ICustomRenderTarget | TAutoRenderTarget): void {
        if ((<ICustomRenderTarget>target).frameBuffer) {
            p3d.p3d_scene_shadowmap(Engine.commands(), this._scene.entity, (<PIWebGLFramebuffer>(<ICustomRenderTarget>target).frameBuffer)?.key);
            p3d.p3d_render_target_key(Engine.commands(), this.entity, (<PIWebGLFramebuffer>(<ICustomRenderTarget>target).frameBuffer)?.key);
        } else {
            p3d.p3d_render_target_auto(Engine.commands(), this.entity, (<TAutoRenderTarget>target).width, (<TAutoRenderTarget>target).height, (<TAutoRenderTarget>target).colorFormat, (<TAutoRenderTarget>target).depthStencilFormat);
        }
    }

    public set bias(val: number) {
        this._bias = val;
    }
    public get bias() {
        return this._bias;
    }

    public set normalBias(val: number) {
        this._normalBias = val;
    }
    public get normalBias() {
        return this._normalBias;
    }

    public set minZ(val: number) {
        this._minZ = val;
    }
    public get minZ() {
        return this._minZ;
    }

    public set maxZ(val: number) {
        this._maxZ = val;
    }
    public get maxZ() {
        return this._maxZ;
    }

    public set frustumSize(val: number) {
        this._frustumSize = val;
    }
    public get frustumSize() {
        return this._frustumSize;
    }

    public set depthScale(val: number) {
        this._depthScale = val;
    }
    public get depthScale() {
        return this._depthScale;
    }

    public update() {
        p3d.p3d_shadow_base_param(Engine.commands(), this.entity, this._bias, this._normalBias, this._depthScale);
        p3d.p3d_shadow_frustum(Engine.commands(), this.entity, this._frustumSize, this._minZ, this._maxZ);
    }
}