
import { EntityID, IAnimationGroup, ICamera, IEntity, ISprite, IVector3, Observable, Option } from "./entity";
import { IDisposable, ILayerMask, INode, INodeEuler, INodePostion, INodeRotation, INodeScaling, IScene
    , ITransformNode, INodeCollider, IMesh, IMeshRenderQueue, IFreeCamera, IOrthographicCamera, IPerspectiveCamera
    , ITargetCamera, IRenderPipeline, IRenderer, IMaterial, IParticleSystem, InstanceAttribute, IInstanceSource, IInstancedBuffer, Constants, IInstancedMesh
} from "../interface/index";
import { PointerEventTypes, PointerInfo } from "./events/pointerEvents";
import { EventData } from "../interface/event";
import { KeyboardInfo } from "./events/keyboardEvents";

export { p3d } from "../context";

export interface IEntityD3 extends IEntity {
    entity: EntityID,
}

export interface IRenderPipelineD3 extends IRenderPipeline {
    renderers: IRendererD3[];
} 

export interface IFrameUpdate {
    _onFrameUpdate(): void;
}

export interface INodeD3 extends IEntityD3, INode, IFrameUpdate, IDisposable, ITreeNode {
    enable: boolean;
    rayID: any;
    onPointerObservable: Observable<PointerInfo>;
}

export interface IRendererD3 extends IEntityD3, IRenderer, IFrameUpdate {

}
export interface ITransformNodeD3 extends IEntityD3, ITransformNode, IFrameUpdate, INodeD3 {

}
export interface INodePositionD3 extends IEntityD3, INodePostion {
    
}
export interface INodeRotationD3 extends IEntityD3, INodeRotation {
    
}
export interface INodeEulerD3 extends IEntityD3, INodeEuler {
    
}
export interface INodeScalingD3 extends IEntityD3, INodeScaling {
    
}
export interface ILayerMaskD3 extends IEntityD3, ILayerMask {
    
}
export interface IMeshRenderQueueD3 extends IEntityD3, IMeshRenderQueue {
}
export interface IMeshD3 extends IMesh, ITransformNodeD3, IMeshRenderQueue {
    dispose(doNotRecurse?: boolean,disposeMaterialAndTextures?: boolean): void;
}

export interface IMaterialD3 extends IEntityD3, IMaterial, IDisposable, IFrameUpdate {
    
}

export interface IOrthographicCameraD3 extends IEntityD3, IOrthographicCamera {
    
}
export interface ITargetCameraD3 extends IEntityD3, ITargetCamera {
    
}
export interface IPerspectiveCameraD3 extends IEntityD3, IPerspectiveCamera {
    
}
export interface IFreeCameraD3 extends IEntityD3, IFreeCamera {
    
}
export interface INodeColliderD3 extends IEntityD3, INodeCollider {
    
}
export interface IParticleSystemD3 extends IEntityD3, IParticleSystem {
    
}

export interface IAnimationEventDataStart {
    group: EntityID | string;
    target?: number;
}

export interface IAnimationEventDataEnd {
    group: EntityID | string;
    target?: number;
}

export interface IAnimationEventDataLoop {
    group: EntityID | string;
    target?: number;
    loop: number;
}

export interface IAnimationEventDataFrame {
    group: EntityID | string;
    target?: number;
    data: any;
}

export interface IAnimationGroup3D extends IAnimationGroup {
    _onStart(data: IAnimationEventDataStart): IAnimationGroup3D;
    _onEnd(data: IAnimationEventDataEnd): IAnimationGroup3D;
    _onLoop(data: IAnimationEventDataLoop): IAnimationGroup3D;
    _onFrame(data: IAnimationEventDataFrame): IAnimationGroup3D;
}
export interface ISprite3D {
    _applySpriteData(data: Uint16Array, offset: number): number;
}

export interface IInstancedBufferD3 extends IInstancedBuffer {
    // get(index: string): number[];
}
export interface IInstancedMeshD3 extends IInstancedMesh {
    sourceMesh: IInstanceSourceD3;
    instancedBuffers: IInstancedBufferD3;
    instanceAttribute(key: string, a: number, b?: number, c?: number, d?: number): void;
    _applyInstanceAttribute(): void;
}

export interface IInstanceSourceD3 extends IInstanceSource {
    _instancestates: Map<string, InstanceAttribute>;
    _cacheInstances: number[];
    _instances: Map<EntityID, IInstancedMeshD3>;
}

export interface ITreeNode {
    name?: string,
    /**
     * Get all direct children of this node
     * @param predicate — defines an optional predicate that will be called on every evaluated child, the predicate must return true for a given child to be part of the result, otherwise it will be ignored
     * @param directDescendantsOnly — defines if true only direct descendants of 'this' will be considered, if false direct and also indirect (children of children, an so on in a recursive manner) descendants of 'this' will be considered (Default: true)
     * @returns — an array of Node
     */
    getChildren(predicate?: (node: ITreeNode) => boolean, directDescendantsOnly?: boolean): ITreeNode[];
}

export interface ITreeContext {
    name: string,
    rootNodes: ITreeNode[];
}

export interface IPickingInfo {
    /** @hidden */
    _pickingUnavailable: boolean;

    /**
     * If the pick collided with an object
     */
    hit: boolean;
    /**
     * Distance away where the pick collided
     */
    distance: number;
    /**
     * The location of pick collision
     */
    pickedPoint: Option<IVector3>;
    /**
     * The mesh corresponding the the pick collision
     */
    pickedMesh: Option<INodeD3>;
    /** (See getTextureCoordinates) The barycentric U coordinate that is used when calculating the texture coordinates of the collision.*/
    bu: number;
    /** (See getTextureCoordinates) The barycentric V coordinate that is used when calculating the texture coordinates of the collision.*/
    bv: number;
    /** The index of the face on the mesh that was picked, or the index of the Line if the picked Mesh is a LinesMesh */
    faceId: number;
    /** The index of the face on the subMesh that was picked, or the index of the Line if the picked Mesh is a LinesMesh */
    subMeshFaceId: number;
    /** Id of the the submesh that was picked */
    subMeshId: number;
    /** If a sprite was picked, this will be the sprite the pick collided with */
    pickedSprite: Option<any>;

    /** If we are pikcing a mesh with thin instance, this will give you the picked thin instance */
    thinInstanceIndex: number;
    /**
     * If a mesh was used to do the picking (eg. 6dof controller) this will be populated.
     */
    originMesh: Option<INodeD3>;
    /**
     * The ray that was used to perform the picking.
     */
    ray: Option<any>;
}

export interface ISceneD3 extends IEntityD3, IScene, IFrameUpdate {
    materials: Map<EntityID, IMaterialD3>;
    textures: Map<IFrameUpdate, IFrameUpdate>;
    meshes: Map<EntityID, IFrameUpdate>;
    transforms: Map<EntityID, INodeD3>;

    cmdsForceDirty: Map<EntityID, IFrameUpdate>;

    // cmdsParent: Map<EntityID, IDirtyParent>;
    cmdsLayerMask: Map<EntityID, ILayerMaskD3>;
    cmdsCollider: Map<EntityID, INodeColliderD3>;

    cmdsMeshMaterial: Map<EntityID, IMeshD3>;
    cmdsMeshRenderQuene: Map<EntityID, IMeshRenderQueueD3>;
    cmdsMeshRenderAlign: Map<EntityID, IMeshD3>;
    cmdsMeshBlend: Map<EntityID, IMeshD3>;
    cmdsMeshCullMode: Map<EntityID, IMeshD3>;
    cmdsMeshDepthTest: Map<EntityID, IMeshD3>;
    cmdsMeshDepthWrite: Map<EntityID, IMeshD3>;
    
    run(): void;
    boundingBoxDebug(display: boolean, pass: number): void;
    isDisposed: boolean;
    pickWithColliderInfo(x: number, y: number, camera?: ICamera, isProjectionXY?: boolean): IPickingInfo;
    clientRect: { left: number, top: number, width: number, height: number };
    
    /** Callback called when a pointer move is detected */
    onPointerMove: (evt: EventData, pickInfo: IPickingInfo, type: PointerEventTypes) => void;
    /** Callback called when a pointer down is detected  */
    onPointerDown: (evt: EventData, pickInfo: IPickingInfo, type: PointerEventTypes) => void;
    /** Callback called when a pointer up is detected  */
    onPointerUp: (evt: EventData, pickInfo: Option<IPickingInfo>, type: PointerEventTypes) => void;
    /** Callback called when a pointer pick is detected */
    onPointerPick: (evt: EventData, pickInfo: IPickingInfo) => void;
    onKeyboardObservable: Observable<KeyboardInfo>;
}

export interface IDeviceLimits {
    max_bind_groups : number,
    max_bindings_per_bind_group : number,
    max_non_sampler_bindings : number,
    max_sampled_textures_per_shader_stage : number,
    max_samplers_per_shader_stage : number,
    max_storage_buffer_binding_size : number,
    max_texture_array_layers : number,
    max_texture_dimension_1d : number,
    max_texture_dimension_2d : number,
    max_texture_dimension_3d : number,
    max_uniform_buffer_binding_size : number,
    max_uniform_buffers_per_shader_stage : number,
    max_vertex_attributes : number,
    max_vertex_buffer_array_stride : number,
    max_vertex_buffers : number,
    min_storage_buffer_offset_alignment : number,
    min_uniform_buffer_offset_alignment : number,
}

export interface IDeviceFeatures {
    ADDRESS_MODE_CLAMP_TO_ZERO : boolean,
    BGRA8UNORM_STORAGE : boolean,
    BUFFER_BINDING_ARRAY : boolean,
    CLEAR_TEXTURE : boolean,
    CONSERVATIVE_RASTERIZATION : boolean,
    DEPTH32FLOAT_STENCIL8 : boolean,
    DEPTH_CLIP_CONTROL : boolean,
    DUAL_SOURCE_BLENDING : boolean,
    FLOAT32_FILTERABLE : boolean,
    INDIRECT_FIRST_INSTANCE : boolean,
    MAPPABLE_PRIMARY_BUFFERS : boolean,
    MULTIVIEW : boolean,
    MULTI_DRAW_INDIRECT : boolean,
    MULTI_DRAW_INDIRECT_COUNT : boolean,
    PARTIALLY_BOUND_BINDING_ARRAY : boolean,
    PIPELINE_STATISTICS_QUERY : boolean,
    POLYGON_MODE_LINE : boolean,
    POLYGON_MODE_POINT : boolean,
    PUSH_CONSTANTS : boolean,
    RAY_QUERY : boolean,
    RAY_TRACING_ACCELERATION_STRUCTURE : boolean,
    RG11B10UFLOAT_RENDERABLE : boolean,
    SAMPLED_TEXTURE_AND_STORAGE_BUFFER_ARRAY_NON_UNIFORM_INDEXING : boolean,
    SHADER_EARLY_DEPTH_TEST : boolean,
    SHADER_F16 : boolean,
    SHADER_F64 : boolean,
    SHADER_I16 : boolean,
    SHADER_PRIMITIVE_INDEX : boolean,
    SHADER_UNUSED_VERTEX_OUTPUT : boolean,
    SPIRV_SHADER_PASSTHROUGH : boolean,
    STORAGE_RESOURCE_BINDING_ARRAY : boolean,
    TEXTURE_ADAPTER_SPECIFIC_FORMAT_FEATURES : boolean,
    TEXTURE_BINDING_ARRAY : boolean,
    TEXTURE_COMPRESSION_ASTC : boolean,
    TEXTURE_COMPRESSION_ASTC_HDR : boolean,
    TEXTURE_COMPRESSION_BC : boolean,
    TEXTURE_COMPRESSION_ETC2 : boolean,
    TEXTURE_FORMAT_16BIT_NORM : boolean,
    TEXTURE_FORMAT_NV12 : boolean,
    TIMESTAMP_QUERY : boolean,
    TIMESTAMP_QUERY_INSIDE_PASSES : boolean,
    UNIFORM_BUFFER_AND_STORAGE_TEXTURE_ARRAY_NON_UNIFORM_INDEXING : boolean,
    VERTEX_ATTRIBUTE_64BIT : boolean,
    VERTEX_WRITABLE_STORAGE : boolean,
}

export interface IDeviceDownLevelFlags {
    ANISOTROPIC_FILTERING : boolean,
    BASE_VERTEX : boolean,
    BUFFER_BINDINGS_NOT_16_BYTE_ALIGNED: boolean,
    COMPARISON_SAMPLERS : boolean,
    COMPUTE_SHADERS : boolean,
    CUBE_ARRAY_TEXTURES : boolean,
    DEPTH_BIAS_CLAMP : boolean,
    DEPTH_TEXTURE_AND_BUFFER_COPIES : boolean,
    FRAGMENT_STORAGE : boolean,
    FRAGMENT_WRITABLE_STORAGE : boolean,
    FULL_DRAW_INDEX_UINT32 : boolean,
    INDEPENDENT_BLEND : boolean,
    INDIRECT_EXECUTION : boolean,
    MULTISAMPLED_SHADING : boolean,
    NON_POWER_OF_TWO_MIPMAPPED_TEXTURES: boolean,
    READ_ONLY_DEPTH_STENCIL : boolean,
    SURFACE_VIEW_FORMATS : boolean,
    UNRESTRICTED_EXTERNAL_TEXTURE_COPIES: boolean,
    UNRESTRICTED_INDEX_BUFFER : boolean,
    VERTEX_STORAGE : boolean,
    VIEW_FORMATS : boolean,
    WEBGPU_TEXTURE_FORMAT_SUPPORT : boolean,

}