/**
 * 
 */
import { BABYLON } from '../babylon';
import { Engine } from '../engine';
// import { BABYLON_TEXTURE_TYPE, ITextureArgCfg, PiBABYLONTexture } from '../../util/res_mgr';
import { ExtensionTools, Res, ResTab } from '../../interface/index';

declare let module;
let logger;
ExtensionTools.RegistStartUp(() => { logger = ExtensionTools.Logger(module.name, "BABYLON"); });

export function queryImageRuntimePath(srcPath: string): string {
    let purePath = srcPath;
    let runtimePath = srcPath;
    if (ExtensionTools.GetPureResPath) {
        purePath = ExtensionTools.GetPureResPath(srcPath);
        runtimePath = ExtensionTools.QueryRuntimeFullName(purePath);
    }
    if (!runtimePath) {
        runtimePath = srcPath;
    }
    return runtimePath;
}

/**
 * @example
 * ```typescript
 * let imagefiles: Map<string, string> = new Map();
 * TexturePathFilter.filter = function() {
 *     let temp = url.match(/\/([^\/]*)$/);
 *     if (url.startsWith("Unity导出资源的在项目内的保存路径") && temp.length > 1) {
 *         let key = temp[1];
 *         let unique = imagefiles.get(key);
 *         if (!unique) {
 *             unique = url;
 *             imagefiles.set(key, unique);
 *         }
 *         url = unique;
 *     }
 *     return url;
 * }
 * ```
 */
export class TexturePathFilter {
    public static filter(url: string): string {
        return url;
    }
}

// // tslint:disable-next-line:max-func-body-length
// export const initTexture = () => {
//     BABYLON.Texture.CreateFromBase64String = function(this: BABYLON.Texture, data, name, scene, noMipmap, invertY, samplingMode, onLoad, onError, format) {
//         if (samplingMode === void 0) { samplingMode = BABYLON.Texture.TRILINEAR_SAMPLINGMODE; }
//         if (onLoad === void 0) { onLoad = null; }
//         if (onError === void 0) { onError = null; }
//         if (format === void 0) { format = 5; }
//         let texture = new BABYLON.Texture(null, scene, noMipmap, invertY, samplingMode, onLoad, onError, data, false, format);
//         texture.updateURL("data:" + name, data, onLoad);

//         return texture;
//     };
//     BABYLON.Texture.prototype.delayLoad = function(res?: Res<ITextureArgCfg, PiBABYLONTexture>) {

//         if (this.delayLoadState !== BABYLON.Constants.DELAYLOADSTATE_NOTLOADED) {
//             return;
//         }

//         let scene = this.getScene();

//         if (!scene) {
//             return;
//         }

//         if (!res) {
//             return;
//         }

//         if (this._pi_re) {
//             // debugger;
//             logger.error('Texture.prototype.delayLoad');
//         }
//         this._pi_res = res;

//         this.delayLoadState = BABYLON.Constants.DELAYLOADSTATE_LOADED;
//         this._texture = res.link.internalTexture;

//         if (this._delayedOnLoad) {
//             if (this._texture.isReady) {
//                 BABYLON.Tools.SetImmediate(this._delayedOnLoad);
//             } else {
//                 this._texture.onLoadedObservable.add(this._delayedOnLoad);
//             }
//         }

//         this._delayedOnLoad = null;
//         this._delayedOnError = null;
//     };

//     BABYLON.Texture.prototype.updateURL = function(this: BABYLON.Texture, url: string, buffer: BABYLON.Nullable<string | ArrayBuffer | HTMLImageElement | Blob> = null, onLoad?: () => void, internalOnload?: () => void): void {
//         if (this.url) {
//             this.releaseInternalTexture();
//             this.getScene()!.markAllMaterialsAsDirty(BABYLON.Constants.MATERIAL_TextureDirtyFlag);
//         }

//         if (!PiBABYLONTexture.getTextureModifiyInfos(url)) {
//             url = TexturePathFilter.filter(url);
//             url = queryImageRuntimePath(url);
//         }

//         this.url = url;
//         this._buffer = buffer;
//         this.delayLoadState = BABYLON.Constants.DELAYLOADSTATE_NOTLOADED;

//         const noMipmap  = this.noMipmap;
//         const invertY   = this.invertY;

//         if (onLoad) {
//             (<any>this)._delayedOnLoad = onLoad;
//         }

//         const engine    = (<Engine>this._scene.getEngine());
//         const scene     = this._scene;
//         const resTab    = Engine.ofl.resTab;

//         resTab.load(BABYLON_TEXTURE_TYPE, url, [<ITextureArgCfg>{engine, invertY, noMipmap, scene, samplingMode: this.samplingMode, internalTexture: undefined, buffer}]).then((res: Res<ITextureArgCfg, PiBABYLONTexture>) => {
//             (<any>this).delayLoad(res);
//             internalOnload && internalOnload();
//         }).catch((e) => {
//             logger.error(e);
//         });
//     };

//     BABYLON.Texture.prototype.computeTextureOffsetWithSpeedUV = function(this: BABYLON.Texture) {
//         if (this._scene?.animationsEnabled) {
//             if (Math.abs(this.offsetSpeedU) > 0.0001) {
//                 this.uOffset = Math.round(this._scene._animationTime) / this.offsetSpeedU;
//             }
//             if (Math.abs(this.offsetSpeedV) > 0.0001) {
//                 this.vOffset = Math.round(this._scene._animationTime) / this.offsetSpeedV;
//                 if (!this.invertY) {
//                     this.vOffset = 1.0 - this.vOffset;
//                 }
//             }
//         }
//     };
// };