import { BABYLON } from "../render3d/babylon";
import { IPath, IRibbonPoint } from "./interface";

export class RibbonBuilder {
    public static TmpVectors = {
        Vector3: [BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero()]
    }

    public static FormatPath(path: BABYLON.Vector3[], size: number, resultPath: IPath, closePath: boolean = false, minimumDistance: number = 0.0001, baseNormal: BABYLON.Vector3 = new BABYLON.Vector3(0, 0, 1), baseDirection: BABYLON.Vector3 = new BABYLON.Vector3(1, 0, 0), baseTangent: BABYLON.Vector3 = new BABYLON.Vector3(0, 1, 0)) {
        let distance = 0;
        let pointCount = 0;
        resultPath.points.length = 0;
        {
            const len = path.length;
            let lastPoint = path[0];
            let point: IRibbonPoint = {
                position: lastPoint.clone(),
                direction: BABYLON.Vector3.Zero(),
                normal: BABYLON.Vector3.Zero(),
                tanget: BABYLON.Vector3.Zero(),
                distance: 0,
                size: size,
            };
            resultPath.points.push(point);
            pointCount += 1;

            for (let i = 1; i < len; ++i) {
                const pathPoint = path[i];
                let direction = pathPoint.subtract(lastPoint);
                let tempDistance = direction.length();
                if (tempDistance >= minimumDistance) {
                    distance += tempDistance;
                    direction.scaleInPlace(1 / tempDistance);

                    let point: IRibbonPoint = {
                        position: pathPoint.clone(),
                        direction: direction,
                        normal: BABYLON.Vector3.Zero(),
                        tanget: BABYLON.Vector3.Zero(),
                        distance: distance,
                        size: size,
                    };
                    resultPath.points.push(point);

                    if (pointCount == 1) {
                        resultPath.points[pointCount - 1].direction.copyFrom(direction);
                    } else {
                        resultPath.points[pointCount - 1].direction.addInPlace(direction).normalize();
                    }

                    lastPoint = pathPoint;
                    pointCount += 1;
                }
            }
            if (closePath) {
                const pathPoint = path[0];
                let direction = pathPoint.subtract(lastPoint);
                let tempDistance = direction.length();
                distance += tempDistance;
                if (tempDistance < minimumDistance) {
                    direction.copyFrom(resultPath.points[pointCount - 1].direction);
                } else {
                    direction.scaleInPlace(1 / tempDistance);
                }
                direction.addInPlace(resultPath.points[0].direction).normalize();

                let point: IRibbonPoint = {
                    position: pathPoint.clone(),
                    direction: direction,
                    normal: BABYLON.Vector3.Zero(),
                    tanget: BABYLON.Vector3.Zero(),
                    distance: distance,
                    size: size,
                };
                resultPath.points.push(point);
                resultPath.points[0].direction.copyFrom(direction);
                pointCount += 1;
            }
            let lastNormal = baseNormal;
            let inNormal: BABYLON.Vector3 = BABYLON.Vector3.Zero();
            let outNormal: BABYLON.Vector3 = BABYLON.Vector3.Zero();
            let preDirection: BABYLON.Vector3;
            let nextDirection: BABYLON.Vector3;
            for (let i = 0; i < pointCount; i++) {
                let point = resultPath.points[i];
                let prePoint: IRibbonPoint;
                let nextPoint: IRibbonPoint;
                
                if (closePath) {

                } else {
                    if (i == 0) {
                        point.normal.copyFrom(baseNormal);
                    } else if (i == pointCount - 1) {
                        prePoint = resultPath.points[i - 1];
                        BABYLON.Vector3.CrossToRef(point.direction, prePoint.direction, point.normal);
                        if (point.normal.normalize().lengthSquared() <= 0.000001) {
                            point.normal.copyFrom(prePoint.normal);
                        }
                    } else {
                        prePoint = resultPath.points[i - 1];
                        nextPoint = resultPath.points[i + 1];
                        BABYLON.Vector3.CrossToRef(point.direction, prePoint.direction, inNormal);
                        if (inNormal.normalize().lengthSquared() <= 0.000001) {
                            inNormal.copyFrom(prePoint.normal);
                        }
                        BABYLON.Vector3.CrossToRef(nextPoint.direction, point.direction, outNormal);
                        if (outNormal.normalize().lengthSquared() <= 0.000001) {
                            outNormal.copyFrom(nextPoint.normal);
                        }
                        point.normal.copyFrom(inNormal).addInPlace(outNormal).scaleInPlace(0.5).normalize();
                        if (point.normal.normalize().lengthSquared() <= 0.000001) {
                            point.normal.copyFrom(baseNormal);
                        }
                    }
                    BABYLON.Vector3.CrossToRef(point.normal, point.direction, point.tanget);
                    if (point.tanget.normalize().lengthSquared() <= 0.000001) {
                        point.tanget.copyFrom(baseTangent);
                    }
                }
            }
        }
    }

    public static FormatPathForceNormal(path: BABYLON.Vector3[], size: number, resultPath: IPath, closePath: boolean = false, minimumDistance: number = 0.0001, forceNormal: BABYLON.Vector3 = new BABYLON.Vector3(0, 0, 1), baseDirection: BABYLON.Vector3 = new BABYLON.Vector3(1, 0, 0), baseTangent: BABYLON.Vector3 = new BABYLON.Vector3(0, 1, 0)) {
        let distance = 0;
        let pointCount = 0;
        resultPath.points.length = 0;
        {
            const len = path.length;
            let lastPoint = path[0];
            let point: IRibbonPoint = {
                position: lastPoint.clone(),
                direction: BABYLON.Vector3.Zero(),
                normal: forceNormal.clone(),
                tanget: BABYLON.Vector3.Zero(),
                distance: 0,
                size: size,
            };
            resultPath.points.push(point);
            pointCount += 1;

            for (let i = 1; i < len; ++i) {
                const pathPoint = path[i];
                let direction = pathPoint.subtract(lastPoint);
                let tempDistance = direction.length();
                if (tempDistance >= minimumDistance) {
                    distance += tempDistance;
                    direction.scaleInPlace(1 / tempDistance);

                    let point: IRibbonPoint = {
                        position: pathPoint.clone(),
                        direction: direction,
                        normal: forceNormal.clone(),
                        tanget: BABYLON.Vector3.Zero(),
                        distance: distance,
                        size: size,
                    };
                    resultPath.points.push(point);

                    if (pointCount == 1) {
                        resultPath.points[pointCount - 1].direction.copyFrom(direction);
                    } else {
                        resultPath.points[pointCount - 1].direction.addInPlace(direction).normalize();
                    }

                    lastPoint = pathPoint;
                    pointCount += 1;
                }
            }
            if (closePath) {
                const pathPoint = path[0];
                let direction = pathPoint.subtract(lastPoint);
                let tempDistance = direction.length();
                distance += tempDistance;
                if (tempDistance < minimumDistance) {
                    direction.copyFrom(resultPath.points[pointCount - 1].direction);
                } else {
                    direction.scaleInPlace(1 / tempDistance);
                }
                direction.addInPlace(resultPath.points[0].direction).normalize();

                let point: IRibbonPoint = {
                    position: pathPoint.clone(),
                    direction: direction,
                    normal: forceNormal.clone(),
                    tanget: BABYLON.Vector3.Zero(),
                    distance: distance,
                    size: size,
                };
                resultPath.points.push(point);
                resultPath.points[0].direction.copyFrom(direction);
                pointCount += 1;
            }
            let lastNormal = forceNormal.clone();
            let inNormal: BABYLON.Vector3 = BABYLON.Vector3.Zero();
            let outNormal: BABYLON.Vector3 = BABYLON.Vector3.Zero();
            let preDirection: BABYLON.Vector3;
            let nextDirection: BABYLON.Vector3;
            for (let i = 0; i < pointCount; i++) {
                let point = resultPath.points[i];
                let prePoint: IRibbonPoint;
                let nextPoint: IRibbonPoint;
                
                // if (closePath) {

                // } else {
                    BABYLON.Vector3.CrossToRef(point.normal, point.direction, point.tanget);
                    if (point.tanget.normalize().lengthSquared() <= 0.000001) {
                        point.tanget.copyFrom(baseTangent);
                    }
                // }
            }
        }
    }

    public static Build(resultPath: IPath, pointOffset: number, positions: Float32Array, colors: Float32Array, uvs: Float32Array, indices: Uint16Array, strip: number, posOffset: number, colorOffset: number, uvOffset: number): number {
        let pointCount = resultPath.points.length;
        let idx = pointOffset;
        let vertexCountPerPoint = 3;
        let indiceIndexPerPoint = 6;
        let indiceCountPerPoint = 12;
        let maxDistance = resultPath.points[pointCount - 1].distance;
        for (let i = 0; i < pointCount; i++) {
            let point = resultPath.points[i];
            let x0 = point.position.x;
            let y0 = point.position.y;
            let z0 = point.position.z;
            let xs = point.tanget.x * point.size;
            let ys = point.tanget.y * point.size;
            let zs = point.tanget.z * point.size;
            let index = idx * vertexCountPerPoint;
            {
                positions[(index + 0) * strip + posOffset + 0] = x0 + xs;
                positions[(index + 0) * strip + posOffset + 1] = y0 + ys;
                positions[(index + 0) * strip + posOffset + 2] = z0 + zs;
                positions[(index + 1) * strip + posOffset + 0] = x0 - 0;
                positions[(index + 1) * strip + posOffset + 1] = y0 - 0;
                positions[(index + 1) * strip + posOffset + 2] = z0 - 0;
                positions[(index + 2) * strip + posOffset + 0] = x0 - xs;
                positions[(index + 2) * strip + posOffset + 1] = y0 - ys;
                positions[(index + 2) * strip + posOffset + 2] = z0 - zs;
            }
            {
                // let pp = Math.abs(point.distance / maxDistance - 0.5);
                // let edge = 0.48;
                // pp = 0.85 * (pp < edge ? 1.0 : pp / (0.5 - pp));
                let pp = 0.85;
                colors[(index + 0) * strip + colorOffset + 0] = 1;
                colors[(index + 0) * strip + colorOffset + 1] = 1;
                colors[(index + 0) * strip + colorOffset + 2] = 1;
                colors[(index + 0) * strip + colorOffset + 3] = 0.2;
                colors[(index + 1) * strip + colorOffset + 0] = 1;
                colors[(index + 1) * strip + colorOffset + 1] = 1;
                colors[(index + 1) * strip + colorOffset + 2] = 1;
                colors[(index + 1) * strip + colorOffset + 3] = pp;
                colors[(index + 2) * strip + colorOffset + 0] = 1;
                colors[(index + 2) * strip + colorOffset + 1] = 1;
                colors[(index + 2) * strip + colorOffset + 2] = 1;
                colors[(index + 2) * strip + colorOffset + 3] = 0.2;
            }
            {
                uvs[(index + 0) * strip + uvOffset + 0] = point.distance;
                uvs[(index + 0) * strip + uvOffset + 1] = 0;
                uvs[(index + 1) * strip + uvOffset + 0] = point.distance;
                uvs[(index + 1) * strip + uvOffset + 1] = 0.5;
                uvs[(index + 2) * strip + uvOffset + 0] = point.distance;
                uvs[(index + 2) * strip + uvOffset + 1] = 1;
            }
            if (i > 0)
            {
                let preIndex = (idx - 1) * vertexCountPerPoint;
                let indiceIndex = (idx - 1) * indiceCountPerPoint;
                indices[indiceIndex + 0] = preIndex + 0;
                indices[indiceIndex + 1] = preIndex + 1;
                indices[indiceIndex + 2] = preIndex + 3;

                indices[indiceIndex + 3] = preIndex + 3;
                indices[indiceIndex + 4] = preIndex + 1;
                indices[indiceIndex + 5] = preIndex + 4;
                
                indices[indiceIndex + 6] = preIndex + 1;
                indices[indiceIndex + 7] = preIndex + 2;
                indices[indiceIndex + 8] = preIndex + 4;
                
                indices[indiceIndex + 9] = preIndex + 4;
                indices[indiceIndex +10] = preIndex + 2;
                indices[indiceIndex +11] = preIndex + 5;
            }
            idx += 1;
        }

        return (idx - 1) * 12;
    }

    public static FromMaskImage(imageData: Uint8Array, positions: Float32Array, uvs: Float32Array, colors: Float32Array) {

    }
}
