
import { Scene } from "./render3d/scene";
import { BABYLON } from "./render3d/babylon";
import { SceneRenderFlags, IModelObj, IRayCastInfo, Scene3DEventInfo, Nullable, EventListener } from "./interface/index";
import { InputManager } from "./render/inputs/inputManager";

export class SceneStruct {
    public static SceneMap: Map<string, SceneStruct> = new Map();
    /**
     * 场景内渲染光表
     */
    public readonly lightMap: Map<string, BABYLON.Light> = new Map();
    _modelMap: Map<any, IModelObj> = new Map();
    /**
     * 底层场景对象
     */
    declare public impl: Scene;
    
    private _prefabID: number = 0;
    public prefabID() {
        return this._prefabID++;
    }
    constructor(sceneName: string, engine: BABYLON.Engine) {
        this.name = sceneName;
        this.engine = engine;
        this.impl = new Scene(engine);
        SceneStruct.SceneMap.set(sceneName, this);
        
        // this.onPointerDown = (e: PointerEvent) => {
        //     (<any>this.impl)._inputManager._onPointerDown(e);
        // };
        // this.onPointerMove = (e: PointerEvent) => {
        //     (<any>this.impl)._inputManager._onPointerMove(e);
        // };
        // this.onPointerUp = (e: PointerEvent) => {
        //     (<any>this.impl)._inputManager._onPointerUp(e);
        // };
        // this.onPointerClick = (e: PointerEvent) => {
        //     const rayInfo = this.impl.pick3D(e.x, e.y);
        //     this.pointerClickListenMap.forEach((func) => {
        //         func({ rayInfo, e, s: null });
        //     });
        // };
        this.inputManager = new InputManager(this.impl);
    }
    get frameStep(): number {
        throw new Error("Method not implemented.");
    }
    _frameCount: number;

    public readonly pointerDownListenMap: Map<Function, (info: Scene3DEventInfo) => any> = new Map();
    public readonly pointerMoveListenMap: Map<Function, (info: Scene3DEventInfo) => any> = new Map();
    public readonly pointerUpListenMap: Map<Function, (info: Scene3DEventInfo) => any> = new Map();
    public readonly pointerClickListenMap: Map<Function, (info: Scene3DEventInfo) => any> = new Map();
    
    public readonly onBeforeDisposeObserver: BABYLON.Observable<any> = new BABYLON.Observable();
    public onBeforeDispose(func: () => void) {
        this.onBeforeDisposeObserver.addOnce(func);
    }
    applyEnvironmentBRDFTexture(path: string): void {
        throw new Error("Method not implemented.");
    }
    // insertMesh(targetPath: string, targetFile: string, modelName?: string | string[]): IModelObj {
    //     let loader = new BABYLON.GLTF2.Loader(this.impl, targetPath, targetFile + ".gltf", targetFile);
    //     if (modelName instanceof Array) {
    //         loader.includeNodeNames = modelName;
    //     } else if (modelName) {
    //         loader.includeNodeNames = [modelName];
    //     }
    //     // loader.imageSolts = model.gltfModifyConfig?.imageSolts;
    //     // loader.materialSolts = model.gltfModifyConfig?.materialSolts;
    //     // loader.materialReplace = model.gltfModifyConfig?.materialReplace;
    //     let result = <IModelObj> {
    //         loadPromise: undefined,
    //         rootImpl: undefined,
    //         meshes: new Map(),
    //         particleSysList: new Map(),
    //         animationMap: new Map(),
    //         dispose: function(this: IModelObj) {
    //             this.isDisposed = true;
    //             this.rootImpl?.dispose();
    //             this.animationMap = undefined;
    //             this.loadPromise = undefined;
    //             this.materials = undefined;
    //             this.meshes = undefined;
    //             this.particleSysList = undefined;
    //             this.transformNodes = undefined;
    //         } 
    //     };
    //     result.loadPromise = loader.loadAsync(LoaderTool.AnalyByJS).then((res) => {
    //         console.log("Loaded " + targetPath, targetFile);
    //         result.rootImpl = res._root_;
    //         res.meshes?.forEach((item) => {
    //             result.meshes.set(item.id, item);
    //         });

    //         res.particles?.forEach((item) => {
    //             result.particleSysList.set(item.id, item);
    //         });
            
    //         result.transformNodes = res.nodes;

    //         res.animationGroups?.forEach((item) => {
    //             result.animationMap.set(item.name, item);
    //         });

    //         result.animations = res.animationGroups;

    //         return result;
    //     });

    //     return result;
    // }
    isDirty: boolean;
    engine: BABYLON.Engine;
    autoOptimize: boolean;
    name: string;
    protected _flag: SceneRenderFlags = SceneRenderFlags.pause;
    public get renderFlag() {
        return this._flag;
    }
    protected _isActive: boolean = false;
    camera: BABYLON.Camera;
    cameraMap: Map<string, BABYLON.Camera> = new Map();
    beforeRenderCallList: Map<Function, (scene: SceneStruct) => void> = new Map();
    onBeforeDrawPhaseCallList: Map<Function, (scene: SceneStruct) => void> = new Map();
    afterRenderCallList: Map<Function, (scene: SceneStruct) => void> = new Map();

    onPointerUp = (e: PointerEvent) => {

    }
    onPointerClick = (e: PointerEvent) => {

    }
    onPointerMove = (e: PointerEvent) => {

    }
    onPointerDown = (e: PointerEvent) => {

    }
    // render = (ctx?: GlContext) => {
    //     this.impl.ctx = ctx;
    //     if (this._isActive) {
    //         this.beforeRenderCallList.forEach((call) => {
    //             try {
    //                 call(this);
    //             } catch (error) {
                    
    //             }
    //         });
    //         this.impl.render();
    //         this.afterRenderCallList.forEach((call) => {
    //             try {
    //                 call(this);
    //             } catch (error) {
                    
    //             }
    //         });
    //     }
    //     return this._isActive;
    // }
    activeObserver = (camera: BABYLON.Camera) => {
        
    };
    changeCameraObserver = (camera: BABYLON.Camera) => {

    };
    removeCameraObserver = (camera: BABYLON.Camera) => {

    };
    setCurrCamera(cameraName: string): void {
        this.camera = this.cameraMap.get(cameraName);
        (<any>this.impl)._activeCamera = this.camera;
    }
    removeCamera(cameraName: string): void {
        let camera = this.cameraMap.get(cameraName);
        if (camera != undefined) {
            camera.dispose();
        }
        this.cameraMap.delete(cameraName);
    }
    addCamera(camera: BABYLON.Camera): void {
        this.cameraMap.set(camera.name, camera);
    }
    pickCall = (e: BABYLON.PointerInfo, s: BABYLON.EventState) => {
        let rayInfo: IRayCastInfo;
        switch (e.type) {
            case BABYLON.PointerEventTypes.POINTERDOWN: {
                if (this.pointerDownListenMap.size > 0) {
                    // rayInfo = this.impl.pick3D();
                    rayInfo = <any>e.pickInfo;
                    rayInfo.target = rayInfo.pickedMesh;
                    if ((<any>rayInfo.pickedMesh)) {
                        rayInfo.rayID = (<any>rayInfo.pickedMesh).rayID;
                    }

                    this.pointerDownListenMap.forEach((func) => {
                        func({ rayInfo, e, s });
                    });
                }
                break;
            }
            case BABYLON.PointerEventTypes.POINTERMOVE: {
                rayInfo = this.impl.pick3D(e.event.x, e.event.y);
                this.pointerMoveListenMap.forEach((func) => {
                    func({ rayInfo, e, s });
                });
                break;
            }
            case BABYLON.PointerEventTypes.POINTERUP: {
                // rayInfo = this.impl.pick3D();

                rayInfo = <any>e.pickInfo;
                rayInfo.target = rayInfo.pickedMesh;
                if ((<any>rayInfo.pickedMesh)) {
                    rayInfo.rayID = (<any>rayInfo.pickedMesh).rayID;
                }

                this.pointerUpListenMap.forEach((func) => {
                    func({ rayInfo, e, s });
                });
                break;
            }
            // tslint:disable-next-line:no-empty
            default: { }
        }
    };
    dispose(): void {
        this.onBeforeDisposeObserver.notifyObservers(undefined);
        this.impl.dispose();
        this.impl = undefined;
        this._flag = SceneRenderFlags.dispose;
        SceneStruct.SceneMap.delete(this.name);
    }
    pause = () => {
        // throw new Error("Method not implemented.");
        this._isActive = false;
        this._flag = SceneRenderFlags.pause;
    }
    active = () => {
        // throw new Error("Method not implemented.");
        this._isActive = true;
        this._flag = SceneRenderFlags.active;
    }
    registerBeforeRenderCall(f: (scene: SceneStruct) => void): void {
        this.beforeRenderCallList.set(f, f);
    }
    unregisterBeforeRenderCall(f: (scene: SceneStruct) => void): void {
        this.beforeRenderCallList.delete(f);
    }
    registerAfterRenderCall(f: (scene: SceneStruct) => void): void {
        this.afterRenderCallList.set(f, f);
    }
    unregisterAfterRenderCall(f: (scene: SceneStruct) => void): void {
        this.afterRenderCallList.delete(f);
    }
    registerBeforeDrawPhaseCall(f: (scene: SceneStruct) => void): void {
        this.beforeRenderCallList.set(f, f);
    }
    unregisterBeforeDrawPhaseCall(f: (scene: SceneStruct) => void): void {
        this.beforeRenderCallList.delete(f);
    }

    protected inputManager: InputManager;
    attachControl(attachUp = true, attachDown = true, attachMove = true, elementToAttachTo: Nullable<{
        addEventListener(type: string, listener: EventListener, options?: boolean | AddEventListenerOptions): void;
        removeEventListener(type: string, listener: EventListener, options?: boolean | AddEventListenerOptions): void;
    }> = null, elementAttachedToForKeyBoard?: BABYLON.Nullable<{
        addEventListener(type: string, listener: BABYLON.KeyboardEventListener, options?: boolean | AddEventListenerOptions): void;
        removeEventListener(type: string, listener: BABYLON.KeyboardEventListener, options?: boolean | AddEventListenerOptions): void;
    }>) {
        this.inputManager.attachControl(attachUp, attachDown, attachMove, elementToAttachTo, elementAttachedToForKeyBoard);
    }
}