/**
 *
 */
import { BABYLON } from './render3d/babylon';
import { ITransformCfg } from './interface/index';
import { ModelObj } from './model_obj';
import { ExtensionTools } from './base';

declare let module;
let logger;
ExtensionTools.RegistStartUp(() => { logger = ExtensionTools.Logger(module.name, "BABYLON"); });

// tslint:disable-next-line:no-unnecessary-class
export class NodeTools {
    public static sceneAsRightHand: boolean = false;
    /**
     * 旋转 Mesh - 自转
     * @param mesh 目标mesh
     * @param rotate 旋转参数： [ x, y, z ]
     */
    public static rotateNode(node: BABYLON.Transform, rotate: number[]) {
        // if (!node.rotationQuaternion) {
        //     node.rotationQuaternion = new BABYLON.Quaternion();
        // }

        NodeTools.rotateQuaternion(node, rotate);
    }
    /**
     * 旋转 Mesh - 自转
     * @param node 目标
     * @param rotate 旋转参数： [ x, y, z ]
     */
    public static rotateQuaternion(node: BABYLON.Transform, rotate: number[]) {
        // YXZ
        let quaternion = BABYLON.Quaternion.Identity();
        if (NodeTools.sceneAsRightHand) {
            BABYLON.Quaternion.RotationYawPitchRollToRef(rotate[1] - Math.PI, rotate[0], -rotate[2], quaternion);
        } else {
            BABYLON.Quaternion.RotationYawPitchRollToRef(rotate[1], rotate[0], rotate[2], quaternion);
        }

        node.setRotationQuaternion(quaternion.x, quaternion.y, quaternion.z, quaternion.w);
    }
    public static positNode(node: BABYLON.Transform, data: number[]) {
        if (NodeTools.sceneAsRightHand) {
            node.setPosition(-data[0], data[1], -data[2]);
        } else {
            node.setPosition(data[0], data[1], data[2]);
        }
    }
    public static scaleNode(node: BABYLON.Transform, data: number[]) {
        if (NodeTools.sceneAsRightHand) {
            node.setScaling(data[0], data[1], -data[2]);
        } else {
            node.setScaling(data[0], data[1], data[2]);
        }
    }
    public static nodeLookAt(node: BABYLON.Transform, data: number[]) {
        if (NodeTools.sceneAsRightHand) {
            node.lookAt(new BABYLON.Vector3(data[0], data[1], -data[2]));
        } else {
            node.lookAt(new BABYLON.Vector3(data[0], data[1], data[2]));
        }
    }
    public static nodeTransform(node: BABYLON.Transform, transform: ITransformCfg) {
        if (transform !== undefined) {
            if (transform.position) {
                NodeTools.positNode(node, transform.position);
            }
            if (transform.scale) {
                NodeTools.scaleNode(node, transform.scale);
            }
            if (transform.rotate) {
                NodeTools.rotateNode(node, transform.rotate);
            }
        }
    }
}

// tslint:disable-next-line:no-unnecessary-class
export class CameraTool {
    /**
     * 修改 正交相机视角
     * @param size 正交相机视角大小
     */
    public static changeCameraOrth(camera: BABYLON.FreeCamera, size: number, width: number, height: number, sizeForWidth: boolean = false) {
        const yDistance = sizeForWidth ? Math.round(size / width / height * 100) / 100 : size;
        const xDistance = sizeForWidth ? size : Math.round(size * width / height * 100) / 100;
        // camera.orthoLeft = -xDistance;
        // camera.orthoRight = xDistance;
        // camera.orthoTop = yDistance;
        // camera.orthoBottom = -yDistance;
        camera.size = size;
        camera.piSizeForWidth = sizeForWidth;
    }

    public static getAccurateSize(xDistance: number, w: number, h: number) {
        return xDistance / (w / h);
    }
}

// tslint:disable-next-line:no-unnecessary-class
export class LoaderTool {
    /**
     * 新框架该值不影响
     */
    public static UseNewGLTFLoader = true;
    public static AnalyByJS = false;
    private static _ResPath        = 'app/scene_res/res/';
    private static _SceneResPath   = 'scenes/';
    private static _ModelResPath   = 'models/';
    private static _NodeResPath    = 'models/';
    private static _EffectResPath  = 'effects/';

    public static getResPath() {
        return this._ResPath;
    }
    public static getSceneResPath() {
        return this._SceneResPath;
    }
    public static getModelResPath() {
        return this._ModelResPath;
    }
    public static getNodeResPath() {
        return this._NodeResPath;
    }
    public static getEffectResPath() {
        return this._EffectResPath;
    }
    // /**
    //  *
    //  * @param model 目标模型
    //  * @param targetPath 如果赋值(完整路径)将忽略 model 配置的路径
    //  * @param targetFile 如果赋值(完整路径)将忽略 model 配置的路径
    //  */
    // public static loadScene(model: AppendModelObj, targetPath?: string, targetFile?: string) {
    //     const name = model.name;
    //     const path = model.path;
    //     const fileName = model.fileName;
    //     const modelName = model.modelName;
    //     const sceneImpl = model.scene.impl;

    //     targetPath = targetPath ? targetPath : `${this.getResPath()}${this.getSceneResPath()}${path}`;
    //     targetFile = targetFile ? targetFile : `${fileName}.gltf`;

    //     let loader = new BABYLON.GLTF2.Loader(sceneImpl, targetPath, targetFile, model.name);
    //     if (modelName instanceof Array) {
    //         loader.includeNodeNames = modelName;
    //     } else if (modelName) {
    //         loader.includeNodeNames = [modelName];
    //     }
    //     loader.loadAsync(LoaderTool.AnalyByJS).then((e) => {
    //         logger.info('%o', e);
    //     });
    //     // loader.
    //     // BABYLON.SceneLoader.Append(targetPath, targetFile, sceneImpl, model.appened, (e) => {
    //     //     logger.info('%o', e);
    //     // });
    // }
    // /**
    //  *
    //  * @param model 目标模型
    //  * @param targetPath 如果赋值(完整路径)将忽略 model 配置的路径
    //  * @param targetFile 如果赋值(完整路径)将忽略 model 配置的路径
    //  */
    // public static loadSceneAsync(model: AppendModelObj, targetPath?: string, targetFile?: string) {
    //     const name = model.name;
    //     const path = model.path;
    //     const fileName = model.fileName;
    //     const modelName = model.modelName;
    //     const sceneImpl = model.scene.impl;

    //     targetPath = targetPath ? targetPath : `${this.getResPath()}${this.getSceneResPath()}${path}`;
    //     targetFile = targetFile ? targetFile : `${fileName}.gltf`;

    //     let loader = new BABYLON.GLTF2.Loader(sceneImpl, targetPath, targetFile, model.name);
    //     if (modelName instanceof Array) {
    //         loader.includeNodeNames = modelName;
    //     } else if (modelName) {
    //         loader.includeNodeNames = [modelName];
    //     }
    //     return loader.loadAsync(LoaderTool.AnalyByJS).then((res) => {
    //         // model.appened(res);
    //     });

    //     // return BABYLON.SceneLoader.AppendAsync(targetPath, targetFile, sceneImpl, (e) => {
    //     //     logger.info('%o', e);
    //     // }).then(
    //     //         (res) => {
    //     //             model.appened(res);
    //     //         }
    //     //     );
    // }
    /**
     * @deprecated
     * @param model 目标模型
     * @param targetPath 如果赋值(完整路径)将忽略 model 配置的路径
     * @param targetFile 如果赋值(完整路径)将忽略 model 配置的路径
     */
    public static insertMesh(model: ModelObj, targetPath?: string, targetFile?: string) {
        // const name = model.name;
        // const path = model.path;
        // const fileName = model.fileName;
        // const modelName = model.modelName;
        // const sceneImpl = model.scene.impl;

        // targetPath = targetPath ? targetPath : `${this.getResPath()}${this.getModelResPath()}${path}`;
        // targetFile = targetFile ? targetFile : `${fileName}.gltf`;

        // let loader = new BABYLON.GLTF2.Loader(sceneImpl, targetPath, targetFile, model.name);
        // if (modelName instanceof Array) {
        //     loader.includeNodeNames = modelName;
        // } else if (modelName) {
        //     loader.includeNodeNames = [modelName];
        // }
        // loader.imageSolts = model.gltfModifyConfig?.imageSolts;
        // loader.materialSolts = model.gltfModifyConfig?.materialSolts;
        // loader.materialReplace = model.gltfModifyConfig?.materialReplace;
        // loader.loadAsync(LoaderTool.AnalyByJS).then((res) => {
        //     model.loaded(res._root_, res.meshes, res.particles, [], res.animationGroups);
        // });

        // BABYLON.SceneLoader.ImportMesh(modelName, targetPath, targetFile, sceneImpl, model.loaded, (e) => {
        //     logger.info('%o', e);
        // }, undefined, undefined, model.gltfModifyConfig);

    }
    /**
     *
     * @param model 目标模型
     * @param targetPath 如果赋值(完整路径)将忽略 model 配置的路径
     * @param targetFile 如果赋值(完整路径)将忽略 model 配置的路径
     */
    public static insertMeshAsync(model: ModelObj, targetPath?: string, targetFile?: string) {
        const name = model.name;
        const path = model.path;
        const fileName = model.fileName;
        const modelName = model.modelName;
        const sceneImpl = model.scene.impl;

        targetPath = targetPath ? targetPath : `${this.getResPath()}${this.getModelResPath()}${path}`;
        targetFile = targetFile ? targetFile : `${fileName}.gltf`;

        let loader = new BABYLON.GLTF2.Loader(sceneImpl, targetPath, targetFile, model.name);
        if (modelName instanceof Array) {
            loader.includeNodeNames = modelName;
        } else if (modelName) {
            loader.includeNodeNames = [modelName];
        }
        loader.imageSolts = model.gltfModifyConfig?.imageSolts;
        loader.materialSolts = model.gltfModifyConfig?.materialSolts;
        loader.materialReplace = model.gltfModifyConfig?.materialReplace;
        return loader.loadAsync(LoaderTool.AnalyByJS);
    }
    /**
     * @deprecated
     * @param model 目标模型
     * @param targetPath 如果赋值(完整路径)将忽略 model 配置的路径
     * @param targetFile 如果赋值(完整路径)将忽略 model 配置的路径
     */
    public static insertEffect(model: ModelObj, targetPath?: string, targetFile?: string) {
        // const name = model.name;
        // const path = model.path;
        // const fileName = model.fileName;
        // const modelName = model.modelName;
        // const sceneImpl = model.scene.impl;

        // targetPath = targetPath ? targetPath : `${this.getResPath()}${this.getEffectResPath()}${path}`;
        // targetFile = targetFile ? targetFile : `${fileName}.gltf`;

        // let loader = new BABYLON.GLTF2.Loader(sceneImpl, targetPath, targetFile, model.name);
        // if (modelName instanceof Array) {
        //     loader.includeNodeNames = modelName;
        // } else if (modelName) {
        //     loader.includeNodeNames = [modelName];
        // }
        // loader.imageSolts = model.gltfModifyConfig?.imageSolts;
        // loader.materialSolts = model.gltfModifyConfig?.materialSolts;
        // loader.materialReplace = model.gltfModifyConfig?.materialReplace;
        // return loader.loadAsync(LoaderTool.AnalyByJS).then((res) => {
        //     model.loaded(res._root_, res.meshes, res.particles, [], res.animationGroups);
        // });
    }
    /**
     * @deprecated
     * @param model 目标模型
     * @param targetPath 如果赋值(完整路径)将忽略 model 配置的路径
     * @param targetFile 如果赋值(完整路径)将忽略 model 配置的路径
     */
    public static insertEffectAsync(model: ModelObj, targetPath?: string, targetFile?: string) {
        // const name = model.name;
        // const path = model.path;
        // const fileName = model.fileName;
        // const modelName = model.modelName;
        // const sceneImpl = model.scene.impl;

        // targetPath = targetPath ? targetPath : `${this.getResPath()}${this.getEffectResPath()}${path}`;
        // targetFile = targetFile ? targetFile : `${fileName}.gltf`;

        // // return BABYLON.SceneLoader.ImportMeshAsync(modelName, targetPath, targetFile, sceneImpl, undefined, undefined, model.gltfModifyConfig)
        // //     .then(
        // //         (res) => {
        // //             return model.loaded(res.meshes, res.particleSystems, res.skeletons, res.animationGroups);
        // //         }
        // //     );
        
        // let loader = new BABYLON.GLTF2.Loader(sceneImpl, targetPath, targetFile, model.name);
        // if (modelName instanceof Array) {
        //     loader.includeNodeNames = modelName;
        // } else if (modelName) {
        //     loader.includeNodeNames = [modelName];
        // }
        // loader.imageSolts = model.gltfModifyConfig?.imageSolts;
        // loader.materialSolts = model.gltfModifyConfig?.materialSolts;
        // loader.materialReplace = model.gltfModifyConfig?.materialReplace;
        // return loader.loadAsync(LoaderTool.AnalyByJS).then((res) => {
        //     model.loaded(res._root_, res.meshes, res.particles, [], res.animationGroups);
        // });
    }
    // public static loadSklai(model: ModelObj, targetPath?: string, targetFile?: string): Promise<[string, Res<any, ArrayBuffer>]> {
    //     const name = model.name;
    //     const path = model.path;
    //     const fileName = model.fileName;
    //     const modelName = model.modelName;
    //     const sceneImpl = model.scene.impl;

    //     targetPath = targetPath ? targetPath : `${this.getResPath()}${this.getModelResPath()}${path}`;
    //     targetFile = targetFile ? targetFile : `${fileName}.sklai`;

    //     let url = targetPath + targetFile;
    //     return sceneImpl.resTab.load(RES_FILE, url, ["Uint8Array"])
    //         .then(
    //             (r) => {
    //                 return [url, r];
    //             }
    //         );
    // }
}