import { CustomScene, IModelOpt, IPrefabType, ModelObj, ModelPrefab, PrefabModelObjPool, SceneStruct, Sprite } from "../index";
import { BABYLON } from "../render3d/babylon";

export interface IModelByReplaceSpriteCfg extends IModelOpt {
    spriteFrame: string;
    spriteAtlas: string;
}

export class ModelByReplaceSprite extends ModelPrefab {
    public static registPrefabCfg(prefabPool: PrefabModelObjPool, gltfPath: string, gltfName: string, modelName: string, spriteFrame: string, spriteAtlas: string) {
        prefabPool.registPrefabCfg(modelName, { path: gltfPath, fileName: gltfName, single: false, caches: 4, spriteFrame, spriteAtlas });
        PrefabModelObjPool.registPrefabFactory(modelName, ModelByReplaceSprite.Creation);
        PrefabModelObjPool.registPrefabResFactory(modelName, ModelObj.Creation);
    }
    public static Creation(scene: CustomScene, info: IPrefabType, pool: PrefabModelObjPool, model: ModelObj): ModelPrefab {
        let cfg: IModelByReplaceSpriteCfg = <IModelByReplaceSpriteCfg>pool.cfgmap.get(info);
        return new ModelByReplaceSprite(scene, info, pool, model, cfg.spriteFrame, cfg.spriteAtlas);
    }

    private sprite: Sprite;
    constructor(scene: CustomScene, info: IPrefabType, pool: PrefabModelObjPool, model: ModelObj, spriteFrame: string, spriteAtlas: string) {
        super(scene, info, pool, model);
        this.sprite = new Sprite(spriteFrame, spriteAtlas, scene.impl);
        this.sprite.setEnabled(false);
        this._loadPromise = new Promise(
            (resolve, reject) => {
                if (this._isDisposed) {
                    reject("Disposed!!");
                } else {
                    if (this._data.isLoaded) {
                        // 取正确的节点作为父节点
                        this.setSpriteParent(this._data.rootImpl, this.sprite);
                        this._loaded();
                        resolve(this);
                    } else {
                        this._data.loadPromise.then(() => {
                            // 取正确的节点作为父节点
                            this.setSpriteParent(this._data.rootImpl, this.sprite);
                            this._loaded();
                            resolve(this);
                        })
                            .catch(e => reject(e));
                    }
                }
            }
        );
    }
    /**
     * 默认实现为 rootimpl 为 父节点
     * @param rootImpl 
     * @param sprite 
     */
    public setSpriteParent = (rootImpl: BABYLON.ITransformNodeD3, sprite: Sprite) => {
        sprite.parent = rootImpl;
    }
    public setAlpha(param: number): ModelByReplaceSprite {
        super.setAlpha(param);
        if (this.sprite) {
            this.sprite.alpha = param;
        }
        return this;
    }
    public setColor(r: number, g: number, b: number): ModelByReplaceSprite {
        super.setColor(r, g, b);
        if (this.sprite) {
            this.sprite.color = new BABYLON.Color3(r, g, b);
        }
        return this;
    }
    public set alphaIndex(val: number) {
        this._data.alphaIndex = val;
        this.sprite.alphaIndex = val;
    }
    public get alphaIndex(): number {
        return this._data.alphaIndex;
    }
    public set renderingGroupId(val: number) {
        this._data.renderingGroupId = val;
        this.sprite.renderingGroupId = val;
    }
    public get renderingGroupId(): number {
        return this._data.renderingGroupId;
    }
    dispose(): void {
        super.dispose();
        if (this.sprite) {
            this.sprite.parent = undefined;
            this.sprite.dispose();
            this.sprite = undefined;
        }
    }
    protected _getColliderNode() {
        return this.sprite;
    }
}