import { EAtlasRenderBase, EAtlasRenderHS, EAtlasRenderHSB, EAtlasRenderMode, ITexturePackAtlasCompact, ITexturePackFrameCompact } from "../interface/texturepacker";
import { Constants, ExtensionTools } from "../interface/index";
import { ISpriteAtlas, ISpriteAtlasCells, ISpriteAtlasRects, ITexturePackAtlas, ITexturePackFrame } from "../interface/sprite";
import { BABYLON } from "../render3d/babylon";
import { IOption, MaxRectsPacker } from "../maxrects_packer/maxrects-packer";
import { IRectangle } from "../maxrects_packer/geom/Rectangle";
import { CombineTexture, ICombineTextureParam } from "../render/materials/Textures/combine_texture";

export class TextureFrame implements ITexturePackFrame {
    // frame: { x: number; y: number; w: number; h: number; };
    get rotated() {
        return this.data[0] > 0;
    }
    get trimmed() {
        return this.data[1] > 0;
    }
    get spriteSourceSize() {
        return { x: this.data[4], y: this.data[5], w: this.data[6], h: this.data[7] };
    }
    get sourceSize() {
        return { w: this.data[2], h: this.data[3] };
    }
    get frame() {
        return { x: this.data[8], y: this.data[9], w: this.data[10], h: this.data[11] };
    }
    data: ITexturePackFrameCompact;
    constructor(data: ITexturePackFrameCompact) {
        this.data = data;
        // this.rotated = data[0] > 0;
        // this.trimmed = data[1] > 0;
        // this.sourceSize = { w: data[2], h: data[3] };
        // this.spriteSourceSize = { x: data[4], y: data[5], w: data[6], h: data[7] };
        // this.frame = { x: data[8], y: data[9], w: data[10], h: data[11] };
    }
    toU16(data: Uint16Array, offset: number) {
        for (let i = 0; i < 12; i++) {
            data[i + offset] = this.data[i];
        }
    }
}

export class TexturePackAtlas implements ITexturePackAtlas {
    frames: { [index: string]: ITexturePackFrame; };
    animations?: { [index: string]: string[]; };
    meta: { app: "https://www.codeandweb.com/texturepacker" | string; version: string; image: string; format: string; size: { w: number; h: number; }; smartupdate: string; };
    image: string;
    samplerMode?: number;
    alphaMode?: number;
    noMipmap?: boolean;
    renderMode?: EAtlasRenderMode;
    public static CompressBlockInfo: [RegExp, number, number][] = [];
    public static fromJson(input: ITexturePackAtlas[] | ITexturePackAtlasCompact[]) {
        if (input instanceof Array) {
            let items = [];
            input.forEach((item) => {
                let data: ITexturePackAtlas | ITexturePackAtlasCompact = <ITexturePackAtlas | ITexturePackAtlasCompact>item;
                
                let w = (<ITexturePackAtlas>data).meta.size.w;
                let h = (<ITexturePackAtlas>data).meta.size.h;
                let realfilename = ExtensionTools.GetPureResPath(data.image);
                realfilename    = ExtensionTools.QueryRuntimeFullName(realfilename);
                TexturePackAtlas.CompressBlockInfo.forEach(([reg, cw, ch]) => {
                    if (reg.test(realfilename)) {
                        w = Math.ceil(w / cw) * cw;
                        h = Math.ceil(h / ch) * ch;
                    }
                });
                (<ITexturePackAtlas>data).meta.size.w = w;
                (<ITexturePackAtlas>data).meta.size.h = h;
                if ((<ITexturePackAtlasCompact>data).compact) {
                    
                    let result = new TexturePackAtlas();
                    result.frames = {};
                    result.animations = data.animations;
                    result.meta = data.meta;
                    result.image = data.image;
                    for (let key in data.frames) {
                        result.frames[key] = new TextureFrame((<ITexturePackAtlasCompact>data).frames[key]);
                    }
                    result.samplerMode = data.samplerMode;
                    result.alphaMode = data.alphaMode;
                    result.noMipmap = data.noMipmap;
                    result.renderMode = data.renderMode;
                    items.push(result)
                } else {
                    let result = new TexturePackAtlas();
                    result.frames = (<ITexturePackAtlas>data).frames;
                    result.animations = data.animations;
                    result.meta = data.meta;
                    result.image = data.image;
                    result.samplerMode = data.samplerMode;
                    result.alphaMode = data.alphaMode;
                    result.noMipmap = data.noMipmap;
                    result.renderMode = data.renderMode;
                    items.push(result)
                }
            })
            return items;
        } else {
            let data: ITexturePackAtlas | ITexturePackAtlasCompact = <ITexturePackAtlas | ITexturePackAtlasCompact>input;
            let w = (<ITexturePackAtlas>data).meta.size.w;
            let h = (<ITexturePackAtlas>data).meta.size.h;
            let realfilename = ExtensionTools.GetPureResPath(data.image);
            realfilename    = ExtensionTools.QueryRuntimeFullName(realfilename);
            TexturePackAtlas.CompressBlockInfo.forEach(([reg, cw, ch]) => {
                if (reg.test(realfilename)) {
                    w = Math.ceil(w / cw) * cw;
                    h = Math.ceil(h / ch) * ch;
                }
            });
            (<ITexturePackAtlas>data).meta.size.w = w;
            (<ITexturePackAtlas>data).meta.size.h = h;
            if ((<ITexturePackAtlasCompact>data).compact) {
                let result = new TexturePackAtlas();
                result.frames = {};
                result.animations = data.animations;
                result.meta = data.meta;
                result.image = data.image;
                for (let key in data.frames) {
                    result.frames[key] = new TextureFrame((<ITexturePackAtlasCompact>data).frames[key]);
                }

                result.samplerMode = data.samplerMode;
                result.alphaMode = data.alphaMode;
                result.noMipmap = data.noMipmap;
                result.renderMode = data.renderMode;
                return result;
            } else {
                let result = new TexturePackAtlas();
                result.frames = (<ITexturePackAtlas>data).frames;
                result.animations = data.animations;
                result.meta = data.meta;
                result.image = data.image;
                result.samplerMode = data.samplerMode;
                result.alphaMode = data.alphaMode;
                result.noMipmap = data.noMipmap;
                result.renderMode = data.renderMode;
                return result;
            }
        }
    }
}

export interface IAtlasParam {
    spriteLimit?: number,
    samplerMode?: number,
    alphaMode?: number,
    noMipmap?: boolean,
    renderMode?: EAtlasRenderMode;
}

export class GlobalAtlasManager {
    public static combineAtlasLink: Map<string, string> = new Map();
    public static combineAtlasRect: Map<string, string> = new Map();

    public static atlasMap: Map<string, Atlas> = new Map();
    public static atlasSpriteLimit: Map<string, IAtlasParam> = new Map();
    private static atlasLoadMap: Map<string, Promise<null>> = new Map();
    private static readonly Default: [number, number, number, number] = [1, 1, 0, 0]
    public static load(atlasConfigPath: string, spriteLimit?: number, samplerMode?: number, alphaMode?: number, noMipmap: boolean = true, renderMode: EAtlasRenderMode = 0) {
        let result = GlobalAtlasManager.atlasLoadMap.get(atlasConfigPath);
        if (result) {
            return result;
        } else {
            let res = ExtensionTools.CreateResTab();
            result = ExtensionTools.LoadFile(res, atlasConfigPath).then((data) => {
                let json = ExtensionTools.UTF8Decode((<ArrayBuffer>data));
                let atlas = TexturePackAtlas.fromJson(JSON.parse(json));
                GlobalAtlasManager.regist(atlas, spriteLimit, samplerMode, alphaMode, noMipmap, renderMode);
                res.clear();
                return null;
            });
            GlobalAtlasManager.atlasLoadMap.set(atlasConfigPath, result);
            return result;
        }
    }
    public static regist(atlas: ISpriteAtlas | ISpriteAtlas[], spriteLimit?: number, samplerMode?: number, alphaMode?: number, noMipmap: boolean = true, renderMode: EAtlasRenderMode = 0) {
        if (atlas instanceof Array) {
            atlas.forEach((item) => {
                spriteLimit = GlobalAtlasManager.atlasSpriteLimit.get(item.image)?.spriteLimit || spriteLimit;
                samplerMode = GlobalAtlasManager.atlasSpriteLimit.get(item.image)?.samplerMode || samplerMode || item.samplerMode;
                alphaMode = GlobalAtlasManager.atlasSpriteLimit.get(item.image)?.alphaMode || alphaMode || item.alphaMode;
                noMipmap = GlobalAtlasManager.atlasSpriteLimit.get(item.image)?.noMipmap || noMipmap || item.noMipmap;
                renderMode = GlobalAtlasManager.atlasSpriteLimit.get(item.image)?.renderMode || renderMode || item.renderMode;
                GlobalAtlasManager.atlasMap.set(item.image, new Atlas(item, samplerMode, alphaMode, noMipmap, renderMode));
                GlobalAtlasManager.atlasSpriteLimit.set(item.image, { spriteLimit, samplerMode, alphaMode, noMipmap, renderMode });
            });
        } else {
            spriteLimit = GlobalAtlasManager.atlasSpriteLimit.get(atlas.image)?.spriteLimit || spriteLimit;
            samplerMode = GlobalAtlasManager.atlasSpriteLimit.get(atlas.image)?.samplerMode || samplerMode || atlas.samplerMode;
            alphaMode = GlobalAtlasManager.atlasSpriteLimit.get(atlas.image)?.alphaMode || alphaMode || atlas.alphaMode;
            noMipmap = GlobalAtlasManager.atlasSpriteLimit.get(atlas.image)?.noMipmap || noMipmap || atlas.noMipmap;
            renderMode = GlobalAtlasManager.atlasSpriteLimit.get(atlas.image)?.renderMode || renderMode || atlas.renderMode;
            GlobalAtlasManager.atlasMap.set(atlas.image, new Atlas(atlas, samplerMode, alphaMode, noMipmap, renderMode));
            GlobalAtlasManager.atlasSpriteLimit.set(atlas.image, { spriteLimit, samplerMode, alphaMode, noMipmap, renderMode });
        }
    }
    public static getAtlas(atlas: ISpriteAtlas | string): Atlas {
        if ((<ISpriteAtlas>atlas).image) {
            let result = GlobalAtlasManager.atlasMap.get(<string>(<ISpriteAtlas>atlas).image);
            if (result == undefined) {
                GlobalAtlasManager.regist(<ISpriteAtlas>atlas);
            }

            return GlobalAtlasManager.atlasMap.get(<string>(<ISpriteAtlas>atlas).image);
        } else {
            return GlobalAtlasManager.atlasMap.get(<string>atlas);
        }
    }
    public static queryRealAtlasName(atlas: string): string {
        return GlobalAtlasManager.combineAtlasLink.get(atlas) || atlas;
    }
    /**
     * 
     * @param spriteFrame 
     * @param atlasKey 如果知道散图属于哪个图集可以加快运行速度
     * @returns 
     */
    public static getFrame(spriteFrame: string | number, atlasKey?: string): [number, number, number, number] {
        if (atlasKey) {
            let realAtlasKey = GlobalAtlasManager.queryRealAtlasName(atlasKey);
            let atlas = this.atlasMap.get(realAtlasKey);
            if (realAtlasKey != atlasKey) {
                return atlas.getFrame(`${atlasKey}#${spriteFrame}`) || GlobalAtlasManager.Default;
            } else {
                return atlas.getFrame(spriteFrame) || GlobalAtlasManager.Default;
            }
        } else {
            let keys = this.atlasMap.keys();
            let key = keys.next().value;
            let result = undefined;
            while (key) {
                let atlas = this.atlasMap.get(key);

                if (atlas) {
                    result = atlas.getFrame(spriteFrame);
                    if (result) {
                        break;
                    }
                }
                key = keys.next().value;
            }

            return result || GlobalAtlasManager.Default;
        }
    }
    /**
     * 
     * @param spriteFrame 
     * @returns 
     */
    public static getTexturePackFrame(spriteFrame: string | number, atlasKey: string): ITexturePackFrame {
        let realAtlasKey = GlobalAtlasManager.queryRealAtlasName(atlasKey);
        let atlas = this.atlasMap.get(realAtlasKey);
        if ((<ITexturePackAtlas>atlas.atlas)?.frames) {
            if (realAtlasKey != atlasKey) {
                return (<ITexturePackAtlas>atlas.atlas)?.frames[`${atlasKey}#${spriteFrame}`];
            } else {
                return (<ITexturePackAtlas>atlas.atlas)?.frames[spriteFrame];
            }
        } else {
            return undefined;
        }
    }
    public static getTexture(atlasKey: string, scene: BABYLON.Scene): BABYLON.Texture | undefined {
        let realAtlasKey = GlobalAtlasManager.queryRealAtlasName(atlasKey);
        let atlas = this.atlasMap.get(realAtlasKey);
        return atlas?.getTexture(scene);
    }
    public static extendAtlasFrame(atlasKey: string, key: string | number, rect: [number, number, number, number]): void {
        let atlas = this.atlasMap.get(atlasKey);
        if (atlas) {
            atlas.extendAtlasFrame(key, rect);
        }
    }
}

export class Atlas {
    protected _atlasFrames: { [index: string | number]: [number, number, number, number] };
    public readonly atlasImage: string;
    constructor(
        public readonly atlas: ISpriteAtlas,
        public readonly samplerMode: number = BABYLON.Texture.LINEAR_LINEAR,
        public readonly alphaMode: number = Constants.ALPHA_COMBINE,
        public readonly noMipmap: boolean = true,
        public readonly renderMode: EAtlasRenderMode = 0
    ) {
        if ((<ITexturePackAtlas>atlas).meta) {
            this.atlasImage = (<ITexturePackAtlas>atlas).image;
        } else {
            this.atlasImage = (<ISpriteAtlasRects>atlas).image;
        }

        if ((<ISpriteAtlasRects>this.atlas).rects) {
            this._atlasFrames = (<ISpriteAtlasRects>this.atlas).rects;
        } else if ((<ITexturePackAtlas>atlas).meta) {
            let atlasFrames: { [index: string]: [number, number, number, number] } = {};

            let w = (<ITexturePackAtlas>atlas).meta.size.w;
            let h = (<ITexturePackAtlas>atlas).meta.size.h;
            for (let key in (<ITexturePackAtlas>atlas).frames) {
                let frame = (<ITexturePackAtlas>atlas).frames[key];
                if (frame.rotated) {
                    let uoff = frame.frame.x / w;
                    let voff = frame.frame.y / h;
                    let hsize = frame.frame.h / w;
                    let vsize = frame.frame.w / h;
                    atlasFrames[key] = [hsize, vsize, uoff, voff];
                } else {
                    let uoff = frame.frame.x / w;
                    let voff = frame.frame.y / h;
                    let hsize = frame.frame.w / w;
                    let vsize = frame.frame.h / h;
                    atlasFrames[key] = [hsize, vsize, uoff, voff];
                }
            }
            this._atlasFrames = <{ [index: string]: [number, number, number, number] }>atlasFrames;
        } else {
            let atlasFrames: [number, number, number, number][] = [];

            let hCount = (<ISpriteAtlasCells>this.atlas).hCount;
            let vCount = (<ISpriteAtlasCells>this.atlas).vCount;
            let hsize = (<ISpriteAtlasCells>this.atlas).hsize;
            let vsize = (<ISpriteAtlasCells>this.atlas).vsize;
            let frames = hCount * vCount;
            for (let i = 0; i < frames; i++) {
                let uoff = i % hCount * hsize;
                let voff = Math.floor(i / hCount) * vsize;
                atlasFrames[i] = [hsize, vsize, uoff, voff];
            }

            this._atlasFrames = <{ [index: number]: [number, number, number, number] }>atlasFrames;
        }
    }
    public getTexture(scene: BABYLON.Scene): BABYLON.Texture | undefined {
        return;
    }
    extendAtlasFrame(key: string | number, rect: [number, number, number, number]): void {
        this._atlasFrames[key] = rect;
    }
    getFrame(spriteFrame: string | number): [number, number, number, number] {
        return this._atlasFrames[spriteFrame];
    }
}

export interface ICustomRect extends IRectangle {
    key: string,
    url: string,
    rot?: boolean,
    /**
     * all input with same will be packed in the same bin.
     */
    tag?: string,
    spriteSourceSize: { x: number, y: number, w: number, h: number },
    sourceSize: { w: number, h: number }
}


export class CombinedAtlas extends Atlas {
    constructor(
        private _texture: CombineTexture,
        atlas: ITexturePackAtlas,
        samplerMode?: number, alphaMode?: number, noMipmap?: boolean, renderMode?: EAtlasRenderMode
    ) {
        super(atlas, samplerMode, alphaMode, noMipmap, renderMode);
    }
    public getTexture(scene: BABYLON.Scene): BABYLON.Texture | undefined {
        let atlas = <ITexturePackAtlas>this.atlas;
        let texture = new BABYLON.Texture(null, scene, false, false, BABYLON.Texture.LINEAR_LINEAR);
        texture.isfile = false;
        texture.updateURL(atlas.image);
        texture.hasAlpha = true;
        texture.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
        texture.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
        texture.srgb = false;
        texture.compressed = false;

        return texture;
    }
    dispose() {
        let atlas = <ITexturePackAtlas>this.atlas;
        this._texture.dispose();
        GlobalAtlasManager.atlasMap.delete(atlas.image);
    }
}
export class AtlasCombine {
    public loadPromise: Promise<void[]>;
    private _combineAtlasList: CombinedAtlas[] = [];
    private _isDisposed: boolean = false;
    constructor(
        protected atlasList: string[],
        spriteLimit?: number, samplerMode?: number, alphaMode?: number, noMipmap?: boolean, renderMode?: EAtlasRenderMode
    ) {
        let maxWidth = 4096;
        let maxHeight = 4096;
        const options: IOption = {
            smart: true,
            pot: false,
            square: false,
            allowRotation: false,
            tag: false,
            border: 0,
        }; // Set packing options
        let packer = new MaxRectsPacker<ICustomRect>(maxWidth, maxHeight, 0, options);
        let inputs: ICustomRect[] = [];
        atlasList.forEach((atlasName) => {
            let atlas = <ITexturePackAtlas | ITexturePackAtlasCompact>GlobalAtlasManager.atlasMap.get(atlasName).atlas;
            let w = atlas.meta.size.w;
            let h = atlas.meta.size.h;
            inputs.push(<ICustomRect>{
                key: atlas.image,
                url: atlas.image,
                x: 0,
                y: 0,
                width:  w,
                height: h,
                spriteSourceSize: { x: 0, y: 0, w: w, h: h },
                sourceSize: { w: w, h: h }
            })
        });

        packer.addArray(inputs); // Start packing with input array
        packer.next(); // Start a new packer bin
        packer.bins.forEach((bin) => {
            let result: ITexturePackAtlas = {
                frames: {},
                meta: {
                    app: "pi_texturepacker",
                    version: "",
                    image: ``,
                    format: "",
                    size: { w: maxWidth, h: maxHeight },
                    smartupdate: "",
                },
                animations: {},
                image: ``,
            };
            let combineinfo: ICombineTextureParam = {
                width: bin.maxWidth,
                height: bin.maxHeight,
                format: BABYLON.Constants.RGBA8 ,
                frames: []
            }
            bin.rects.forEach((rect, idx) => {
                let sourceAtlas = <ITexturePackAtlas>GlobalAtlasManager.atlasMap.get(rect.key).atlas;
                for (let key in sourceAtlas.frames) {
                    let sourceFrame = sourceAtlas.frames[key];
                    let sourceFrameData = sourceAtlas.frames[key].frame;
                    let newKey = `${sourceAtlas.image}#${key}`;
                    result.frames[newKey] = {
                        frame: { x: sourceFrameData.x + rect.x, y: sourceFrameData.y + rect.y, w: sourceFrameData.w, h: sourceFrameData.h },
                        rotated: sourceFrame.rotated,
                        trimmed: sourceFrame.trimmed,
                        spriteSourceSize: sourceFrame.spriteSourceSize,
                        sourceSize: sourceFrame.sourceSize
                    };
                }
                combineinfo.frames.push({
                    url: sourceAtlas.image,
                    x: rect.x,
                    y: rect.y,
                    w: sourceAtlas.meta.size.w,
                    h: sourceAtlas.meta.size.h,
                });
            });

            let combineTexture = new CombineTexture(combineinfo);
            result.image = combineTexture.url;
            result.meta.image = combineTexture.url;
            let _atlas = new CombinedAtlas(combineTexture, result, samplerMode, alphaMode, noMipmap, renderMode);
            this._combineAtlasList.push(_atlas);
            GlobalAtlasManager.atlasMap.set(result.image, _atlas);
            GlobalAtlasManager.atlasSpriteLimit.set(result.image, { spriteLimit, samplerMode, alphaMode, noMipmap, renderMode });
            
            bin.rects.forEach((rect, idx) => {
                GlobalAtlasManager.combineAtlasLink.set(rect.key, result.image);
            });
        });
    }
    destroy() {
        this.atlasList.forEach((key) => {
            GlobalAtlasManager.combineAtlasLink.delete(key);
        });
        this._isDisposed = true;
        this._combineAtlasList.forEach((item) => {
            item.dispose();
        });
        this._combineAtlasList.length = 0;
    }
}