import { CustomScene, IModelOpt, IPrefabType, Model, ModelObj, ModelPrefab, NodeObj, PrefabModelObjPool } from "../index";
import { BABYLON } from "../render3d/babylon";
import { ExtensionTools, InstanceAttribute } from "../interface/index";
import { p3d } from "../render/types";

interface ISceneWithModelRectTextureProgress extends CustomScene {
    __RectTextureProgressMeshes: Map<string, BABYLON.Mesh>;
}

interface ICreation extends IModelOpt {
    texPath: string,
}

export class ModelRectTextureProgress extends ModelObj {
    public static regist(scene: CustomScene, texPath: string) {
        TextureProgressMaterial.regist();
        scene.prefabPool.registPrefabCfg(texPath, { path: "null", fileName: "", single: false, cacheCount: 8, texPath });
        PrefabModelObjPool.registPrefabFactory(texPath, ModelPrefab.Creation);
        PrefabModelObjPool.registPrefabResFactory(texPath, ModelRectTextureProgress.Creation);

        let tempscene = <ISceneWithModelRectTextureProgress>scene;
        if (!tempscene.__RectTextureProgressMeshes) {
            tempscene.__RectTextureProgressMeshes = new Map();
        }
        scene.impl.onDisposeObservable.add(() => {
            tempscene.__RectTextureProgressMeshes.forEach(source => {
                source.dispose(false, true);
                (<any>source).__instance_keep = undefined;
            });
            tempscene.__RectTextureProgressMeshes.clear();
        });
        if (!tempscene.__RectTextureProgressMeshes.get(texPath)) {
            let source = createSourceMesh(scene.impl, texPath);
            let ins = source.createInstance(`Keep_ModelRectTextureProgress`);
            ins.setEnabled(false);
            (<any>source).__instance_keep = ins;
            tempscene.__RectTextureProgressMeshes.set(texPath, source);
        }
    }
    // public static Creation(scene: CustomScene, info: IPrefabType, pool: PrefabModelObjPool, model: ModelRectTextureProgress) {
    //     let result = new NodeObj(scene, info, pool, model);
    //     return result;
    // }
    public static Creation(name: string, scene: CustomScene, opt?: ICreation) {
        let model = new ModelRectTextureProgress(name, scene);
        let tempscene = <ISceneWithModelRectTextureProgress>scene;
        let source = tempscene.__RectTextureProgressMeshes.get(opt.texPath);
        let instance = source.createInstance(name);

        let rootImpl = instance;
        model.rootImpl = rootImpl;
        model.instance = instance;
        model._loadPromise = Promise.resolve().then(() => {
            model.loaded(rootImpl, [rootImpl], [], [], [], [], []);
            return model;
        });

        return model;
    }
    protected instance: BABYLON.InstancedMesh;
    public set progress(val: number) {
        this.instance.instanceAttribute(TextureProgressMaterial.KEY_PROGRESS, val, 0);
    }
    public setAlpha(val: number) {
        super.setAlpha(val);
        this.progress = val;
    }
    public set alphaIndex(val: number) {
        super.alphaIndex = val;
        this.instance.alphaIndex = val;
    }
    constructor(name: string, scene: CustomScene) {
        super(name, scene);
    }

    _dispose(disposeMaterialAndTextures: boolean = true) {
        this.instance.parent = null;
        this.instance.dispose();
        super._dispose();
    }
}

export class TextureProgressMaterial extends BABYLON.MainOpacityMaterial {

    public static KEY_PROGRESS = "A_PROGRESS";
    public static ShaderName = "TextureProgress";
    
    constructor(name: string, scene: BABYLON.Scene, shaderName = TextureProgressMaterial.ShaderName) {
        super(name, scene, shaderName);
    }
    
    public static regist() {
        if (p3d.p3d_check_shader(BABYLON.Engine.native(), BABYLON.Engine.param(), TextureProgressMaterial.ShaderName)) {
            return;
        }
        let uniforms = p3d.MaterialUniformDefines.create();
        let includes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
                // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(BABYLON.UniformMainTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTextureUVOffsetSpeed);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTexture);
        // p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(BABYLON.UniformOpacityTexture.BLOCK_NAME));
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTextureUVOffsetSpeed);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTexture);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTextureUVOffsetSpeed);

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec3", "v_pos"],
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
            ["vec2", "v_progress"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let binds_defines_base = BABYLON.Constants.BIND_MODEL | BABYLON.Constants.BIND_VIEWER | BABYLON.Constants.BIND_EFFECT_VALUE;

        p3d.p3d_regist_material(
            BABYLON.Engine.native(), BABYLON.Engine.param(),
            TextureProgressMaterial.ShaderName,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            mat4 finalWorld = PI_ObjectToWorld;
        
            vec4 position =  vec4(A_POSITION, 1.);
            vec4 worldPos =  finalWorld * position;
            // vec4 worldPos =  position;
        
            gl_Position = PI_MATRIX_VP * worldPos;
        
            v_pos = worldPos.xyz;
            
            v_uv    = A_UV;
            v_progress   = ${TextureProgressMaterial.KEY_PROGRESS};
            v_color = A_COLOR4;
            `,
            `
            vec4 baseColor = v_color;
            float alpha = 1.0;
            alpha   *= v_color.a;
        
            float glossiness 			= 0.;
            vec3 emissionColor 		    = vec3(1., 1., 1.);
            vec3 baseAmbientColor		= vec3(1., 1., 1.);
            vec4 refractionColor		= vec4(0., 0., 0., 1.);
            vec4 reflectionColor 		= vec4(0., 0., 0., 1.);
        
            // ----------------------------------------------------------
            vec2 bgUV           = vec2(v_uv.x, v_uv.y * 0.5);
        
            vec4 mainTextureColor   = mainTexture(bgUV, applyUVOffsetSpeed(uMainUVOS));
            baseColor.rgb           *= mainTextureColor.rgb * mainStrength();

            vec2 gressUV            = vec2(v_uv.x, v_uv.y * 0.5 + 0.5);
            vec4 gressColor         = mainTexture(gressUV, applyUVOffsetSpeed(uMainUVOS)) * mainColor();
            
            float gressAlpha        = step(v_uv.x, v_progress.x) * gressColor.a;
            baseColor               = baseColor * (1.0 - gressAlpha) + gressColor * gressAlpha;
            alpha                   *= baseColor.a;
        
            vec3 finalSpecular 		= vec3(0., 0., 0.);
            vec3 finalDiffuse       = emissionColor * baseColor.rgb;
            
            vec4 finalColor 		= vec4(
                finalDiffuse * baseAmbientColor
                + 
                finalSpecular
                +
                refractionColor.rgb
                + 
                reflectionColor.rgb
                ,
                alpha
            );
        
            finalColor				= max(finalColor, 0.0);
        
            // Premulty
            // finalColor.rgb          *= finalColor.a;
        
            gl_FragColor = finalColor;        
            `,
            includes,
            "",
            varyings,
            binds_defines_base
        );

        varyings.free && varyings.free();
        uniforms.free && uniforms.free();
        includes.free && includes.free();
    }
}

function createSourceMesh(scene: BABYLON.Scene, texPath: string): BABYLON.Mesh {

    // 创建自定义材质
    let custom_material = new TextureProgressMaterial("TxtureProgress", scene);
    let texture = new BABYLON.Texture(texPath, scene, false, false);
    texture.updateURL(texPath);
    custom_material.diffuseTexture = texture;
    custom_material.diffuseColor = BABYLON.Color3.White();
    custom_material.diffuseTexture.hasAlpha = true;
    custom_material.diffuseTexture.wrapU = BABYLON.Constants.CLAMP_ADDRESSMODE;
    custom_material.diffuseTexture.wrapV = BABYLON.Constants.CLAMP_ADDRESSMODE;

    let _mesh = BABYLON.PlaneBuilder.CreatePlane(
        `Progress`, { size: 1 }, scene,
        [{ key: "A_PROGRESS", vtype: BABYLON.Constants.INSTANCE_ATTR_VEC2, code: "" }],
        true
    );
    _mesh.material = custom_material;
    // _mesh.billboardMode = BABYLON.Mesh.BILLBOARDMODE_ALL;
    // _mesh.registerInstancedBuffer(`a_progress`, 2);
    // _mesh.instancedBuffers.a_progress = new BABYLON.Vector2(0.0, 0.0);
    // _mesh.setEnabled(false);

    // custom_material.needDepthPrePass = true;

    _mesh.alphaIndex = 3010;

    return _mesh;
}

function progressMaterial() {

}
