import { ITileAsset } from "./interface";

export interface ITileAssetExtend extends ITileAsset {
    imageUSize: number;
    imageVSize: number;
    imagePath: string;
    properties: { name: string; value: number; }[]
}

export class TileMapAsset {
    public assets: Map<string, ITileAssetExtend> = new Map();
    record(source: string, asset: ITileAsset) {
        if (this.assets.get(source)) {
            console.warn(`Record ${source} Again !!!`);
        } else {
            let data: ITileAssetExtend = <any>asset;
            let imageRow = asset.tilecount / asset.columns;
            let imageRealWidth = asset.columns * asset.tilewidth;
            let imageRealHeight = imageRow * asset.tileheight;
            imageRealWidth = nextPowerOfTwo(imageRealWidth);
            imageRealHeight = nextPowerOfTwo(imageRealHeight);

            data.imageUSize = asset.tilewidth / imageRealWidth;
            data.imageVSize = asset.tileheight / imageRealHeight;
            this.assets.set(source, data);
        }
    }
}

/**
 * @description 下一个2的冥的数
 */
function nextPowerOfTwo(value: number) {
    --value;
    value |= value >> 1;
    value |= value >> 2;
    value |= value >> 4;
    value |= value >> 8;
    value |= value >> 16;

    return ++value;
}